package net.shrine.adapter


import net.shrine.adapter.audit.AdapterAuditDb

import scala.Option.option2Iterable
import scala.util.Try
import scala.xml.{NodeSeq, XML}
import net.shrine.adapter.dao.AdapterDao
import net.shrine.adapter.dao.model.{ShrineQuery, ShrineQueryResult}
import net.shrine.protocol.{AuthenticationInfo, BaseShrineRequest, BroadcastMessage, DoneButHasNoResultsForQueuedQueryException, ErrorResponse, HasQueryResults, HiveCredentials, I2b2ResultEnvelope, NoResultsForQueuedQueryException, QueryInstance, QueryResult, ReadInstanceResultsRequest, ReadInstanceResultsResponse, ReadQueryInstancesRequest, ReadQueryInstancesResponse, ReadResultRequest, ReadResultResponse, ResultOutputType, ShrineRequest, ShrineResponse}
import net.shrine.protocol.query.QueryDefinition

import scala.concurrent.duration.Duration
import net.shrine.problem.{AbstractProblem, ProblemSources}

import scala.util.control.NonFatal
import cats.syntax.either._
import net.shrine.http4s.client.legacy.{HttpResponse, Poster}

/**
 * @author clint
 * @since Nov 2, 2012
 *
 */

//todo cleanup with SHRINE-3627
abstract class AbstractReadQueryResultAdapter[Req <: BaseShrineRequest, Rsp <: ShrineResponse with HasQueryResults](
  poster: Poster,
  override val hiveCredentials: HiveCredentials,
  dao: AdapterDao,
  doObfuscation: Boolean,
  getQueryId: Req => Long,
  getProjectId: Req => String,
  toResponse: (Long, QueryResult) => Rsp,
  breakdownTypes: Set[ResultOutputType],
  collectAdapterAudit:Boolean,
  obfuscator:Obfuscator,
  queuedQueryTimeToLive:Duration
) extends WithHiveCredentialsAdapter(hiveCredentials) {


  override protected[adapter] def processRequest(message: BroadcastMessage): ShrineResponse = {
    //Req is either  a ReadQueryResultRequest or a ReadInstanceResultsRequest. Both have the project id
    val req = message.request.asInstanceOf[Req]
    val queryId: Long = getQueryId(req)
    val researcherAuthn = req.authn
    val waitTime = req.waitTime

    askAdapterDatabaseThenCrc(queryId,waitTime,researcherAuthn).fold(e => e, r => r)
  }

  def askCrc(
              queryId: Long,  //network query id
              waitTime: Duration,
              researcherAuthn: AuthenticationInfo
            ): Either[ErrorResponse, ShrineResponse ] = {

    try {
      dao.findQueryByNetworkId(queryId).map { shrineQueryRow =>
        dao.findResultsFor(queryId).map { shrineQueryResult: ShrineQueryResult =>
          debug(s"Query $queryId is incomplete in the adapter's database - will ask CRC for results")
          //if this times out (or hits some other CRC web api trouble) just send back the best answer from the SHRINE adapter database
          requestQueuedResultFromCrc(
            queryId = queryId,
            i2b2MasterQueryId = shrineQueryRow.i2b2MasterQueryId.toLong,
            waitTime = waitTime,
            researcherAuthn = researcherAuthn,
            shrineQueryResult = shrineQueryResult
          )
        }.getOrElse {
          debug(s"Query $queryId found but not the results of the original call to the CRC. That call may not have returned from the initial CRC call yet")
          //don't record this in the adapter database
          Left(ErrorResponse(QueryResultNotAvailable(queryId)))
        }

      }.getOrElse {
        //don't record this in the adapter database
        debug(s"Query $queryId not found in the Shrine DB")
        Left(ErrorResponse(QueryNotFound(queryId)))
      }
    } catch {
      case NonFatal(x) => Left(ErrorResponse(CouldNotRetrieveQueryFromCrc(queryId, x)))
    }
  }

  //preserves the old web api path through the code with a cut/paste
  private def askAdapterDatabaseThenCrc(
                              queryId: Long,  //network query id
                              waitTime: Duration,
                              researcherAuthn: AuthenticationInfo
                            ): Either[ErrorResponse, ShrineResponse] = {

    try {
      dao.findQueryByNetworkId(queryId).map { shrineQueryRow =>
        dao.findResultsFor(queryId).map { shrineQueryResult: ShrineQueryResult =>
          if (shrineQueryResult.isDone) {
            debug(s"Query $queryId is done and already stored, returning stored results")
            makeResponseFrom(queryId, shrineQueryResult)
          } else {
            debug(s"Query $queryId is incomplete in the adapter's database - will ask CRC for results")
            //if this times out (or hits some other CRC web api trouble) just send back the best answer from the SHRINE adapter database
            requestQueuedResultFromCrc(
              queryId,
              shrineQueryRow.i2b2MasterQueryId.toLong,
              waitTime,
              researcherAuthn,
              shrineQueryResult)
          }
        }.getOrElse {
          debug(s"Query $queryId found but not the results of the original call to the CRC. That call may not have returned from the initial CRC call yet")
          //don't record this in the adapter database
          Left(ErrorResponse(QueryResultNotAvailable(queryId)))
        }

      }.getOrElse {
        //don't record this in the adapter database
        debug(s"Query $queryId not found in the Shrine DB")
        Left(ErrorResponse(QueryNotFound(queryId)))
      }
    } catch {
      case NonFatal(x) => Left(ErrorResponse(CouldNotRetrieveQueryFromCrc(queryId, x)))
    }
  }

  private def requestQueuedResultFromCrc(
                                          queryId: Long,  //network query id
                                          i2b2MasterQueryId: Long,
                                          waitTime: Duration,
                                          researcherAuthn: AuthenticationInfo,
                                          shrineQueryResult: ShrineQueryResult
                                        ): Either[ErrorResponse, ShrineResponse ] = {
    debug(s"Query $queryId is incomplete in the adapter's database - will ask CRC for results")

    val errorOrResponse = queryInstanceInCrc(queryId,i2b2MasterQueryId,waitTime).flatMap{ rqiResponse: ReadQueryInstancesResponse =>
      if (rqiResponse.queryInstances.forall(_.queryStatus.isDone) &&
        !rqiResponse.queryInstances.exists(_.queryStatus.isError)) {
        //get that queryInstanceId carefully
        rqiResponse.queryInstances.headOption.map{ queryInstance: QueryInstance =>
          retrieveIfComplete(
            queryId = queryId,
            i2b2QueryInstanceId = queryInstance.queryInstanceId.toLong,
            waitTime = waitTime,
            researcherAuthn = researcherAuthn,
            shrineQueryResult = shrineQueryResult
          )
        }.getOrElse(Left(ErrorResponse(NoQueryInstanceFromCrc(queryId,rqiResponse))))
      } else {
        //the CRC told us the query was incomplete or has an error.
        //todo depending on the error, it might be best to leave the QUEUED state in place and just log for some SHRINE-3327
        debug(s"Query is in an incomplete state or has an error. Replying with $rqiResponse ")
        Right(rqiResponse)
      }
    }
    //record select responses to the adpater's database
    errorOrResponse.fold({ errorResponse: ErrorResponse =>
      //if an erroring query has aged out, record it in the adapters database as an error (and give up asking the CRC for it)
      if(System.currentTimeMillis() - shrineQueryResult.dateCreated.toGregorianCalendar.getTimeInMillis > queuedQueryTimeToLive.toMillis) {
        val errorResult = QueryResult.errorResult(Some(errorResponse.errorMessage), s"Adapter gave up on query after $queuedQueryTimeToLive", errorResponse.problem)
        val response: Rsp = toResponse(queryId, errorResult)
        storeResultIfNecessary(shrineQueryResult, response, researcherAuthn, queryId, Set.empty)
      }
    },{ _ => //shrineQueryResult already saved
    })

    errorOrResponse
  }

  private def queryInstanceInCrc(
                                  queryId: Long,
                                  i2b2MasterQueryId: Long,
                                  waitTime: Duration
                                ): Either[ErrorResponse, ReadQueryInstancesResponse] = {

    //Use a ReadQueryInstancesRequest to find out if it is safe to ask for the results
    val readQueryInstancesRequest: ReadQueryInstancesRequest = ReadQueryInstancesRequest(
      projectId = hiveCredentials.projectId,
      waitTime = waitTime,
      authn = hiveCredentials.toAuthenticationInfo,
      i2b2MasterQueryId = i2b2MasterQueryId
    )

    val readQueryInstanceHttpResponse: HttpResponse = poster.post(readQueryInstancesRequest.toI2b2String)
    debug(s"ReadQueryInstancesResponse httpResponse for $queryId is $readQueryInstanceHttpResponse")

    if (readQueryInstanceHttpResponse.statusCode == 200) {
      try {
        val rqiResponse = ReadQueryInstancesResponse.fromI2b2(readQueryInstanceHttpResponse.body)
        debug(s"rqiResponse is  $rqiResponse")
        Right(rqiResponse)
      } catch {
        case noResults: NoResultsForQueuedQueryException =>
          Left(ErrorResponse(NoResultsForQueuedQuery(noResults)))
        case doneNoResults: DoneButHasNoResultsForQueuedQueryException =>
          Left(ErrorResponse(DoneButHasNoResultsForQueuedQuery(doneNoResults)))
        case NonFatal(x) =>
          info(s"Caught a new exception while parsing ReadQueryInstancesResponse ${readQueryInstanceHttpResponse.body}",x)
          try {
            Left(ErrorResponse.fromI2b2(readQueryInstanceHttpResponse.body))
          } catch { //not an error response from the CRC
            case NonFatal(y) => Left(ErrorResponse(CouldNotRetrieveQueryFromCrc(queryId,y)))
          }
      }
    } else {
      val x = new IllegalStateException(s"Got status code ${readQueryInstanceHttpResponse.statusCode} from the CRC, expected 200 OK. $readQueryInstanceHttpResponse")
      x.fillInStackTrace()
      CouldNotRetrieveQueryFromCrc(queryId,x)
      throw x
    }

  }

  /*
Send a ReadInstanceResultsRequest to see if the query is actually completed before risking asking for a result
   */
  private def retrieveIfComplete(queryId: Long,
                                 i2b2QueryInstanceId: Long,
                                 waitTime: Duration,
                                 researcherAuthn: AuthenticationInfo,
                                 shrineQueryResult: ShrineQueryResult
                                ): Either[ErrorResponse, ShrineResponse with HasQueryResults] = {

    val queryCompletedAttempt: Either[ErrorResponse, ReadInstanceResultsResponse] = readInstanceResults(queryId,i2b2QueryInstanceId,waitTime)

    queryCompletedAttempt.fold(Left(_),{ queryCompletedResponse: ReadInstanceResultsResponse =>
      if(queryCompletedResponse.results.forall(_.statusType.isDone)) {
        //Finally ask for the results.
        debug(s"OK to ask the CRC for the results for $queryId")
        retrieveQueryResults(queryId, waitTime, researcherAuthn, shrineQueryResult, queryCompletedResponse)
      }else {
        debug(s"Not safe to ask the CRC for the results for $queryId queryCompletedAttempt is $queryCompletedAttempt")
        queryCompletedAttempt
      }
    })
  }

  //See if the query is actually completed.
  private def readInstanceResults( queryId: Long,
                                   localQueryId: Long,
                                   waitTime: Duration
                                   ): Either[ErrorResponse, ReadInstanceResultsResponse] = {

    val readInstanceResultsRequest = ReadInstanceResultsRequest(
      projectId = hiveCredentials.projectId,
      waitTime = waitTime,
      authn = hiveCredentials.toAuthenticationInfo,
      queryInstanceId = localQueryId
    )
    val readInstanceResultsHttpResponse: HttpResponse = poster.post(readInstanceResultsRequest.toI2b2String)
    debug(s"readInstanceResultsHttpResponse for $queryId is $readInstanceResultsHttpResponse")

    if (readInstanceResultsHttpResponse.statusCode == 200) {
      try {
        val rirr: ReadInstanceResultsResponse =
          ReadInstanceResultsResponse.
            fromI2b2(breakdownTypes)(XML.loadString(readInstanceResultsHttpResponse.body)).
            copy(shrineNetworkQueryId = queryId)
        debug(s"ReadInstanceResultsResponse is $rirr")

        Right(rirr)
      } catch {
        case noResults: NoResultsForQueuedQueryException =>
          Left(ErrorResponse(NoResultsForQueuedQuery(noResults)))
        case doneNoResults: DoneButHasNoResultsForQueuedQueryException =>
          Left(ErrorResponse(DoneButHasNoResultsForQueuedQuery(doneNoResults)))
      }
    } else {
      val x = new IllegalStateException(s"Got status code ${readInstanceResultsHttpResponse.statusCode} from the CRC, expected 200 OK. $readInstanceResultsHttpResponse")
      x.fillInStackTrace()
      CouldNotRetrieveQueryFromCrc(queryId,x)
      throw x
    }
  }

  private def makeResponseFrom(queryId: Long, shrineQueryResult: ShrineQueryResult): Either[ErrorResponse, Rsp] = {
    shrineQueryResult.toQueryResults(doObfuscation).
      map(x => Right(toResponse(queryId, x))).
      getOrElse(Left(ErrorResponse(QueryNotFound(queryId))))
  }

  private def retrieveQueryResults(
                                    queryId: Long,
                                    waitTime:Duration,
                                    researcherAuthn:AuthenticationInfo,
                                    shrineQueryResult: ShrineQueryResult,
                                    queryCompletedResponse: ReadInstanceResultsResponse
                                  ): Either[ErrorResponse, Rsp] = {

    //todo use the queryCompletedResponse to get the count
    val countResponse: Either[ErrorResponse, ReadResultResponse] = retrieveCountResults(waitTime, researcherAuthn, shrineQueryResult)

    countResponse.flatMap{ goodCountResponse: ReadResultResponse =>

      val breakdownResults = retrieveBreakdownResults(waitTime, researcherAuthn, queryCompletedResponse)

      breakdownResults.fold({ breakdownErrorResponses =>
        //Someday store errors in breakdowns after they get redesigned. For now report the first error
        //storeResultIfNecessary(shrineQueryResult, countResponse, researcherAuthn, queryId, getFailedBreakdownTypes(breakdownResponseAttempts))

        Left(breakdownErrorResponses.head)
      }, { breakdownsByType =>
        val queryResultWithBreakdowns: QueryResult = goodCountResponse.metadata.withBreakdowns(breakdownsByType)
        val queryResultToReturn: QueryResult = if (doObfuscation) obfuscator.obfuscate(queryResultWithBreakdowns) else queryResultWithBreakdowns

        val response: Rsp = toResponse(queryId, queryResultToReturn)
        //NB: Only store the result if needed, that is, if all results are done
        storeResultIfNecessary(shrineQueryResult, response, researcherAuthn, queryId, Set.empty)

        Right(response)
      })
    }
  }

  private def retrieveCountResults(
                                    waitTime:Duration,
                                    researcherAuthn:AuthenticationInfo,
                                    shrineQueryResult: ShrineQueryResult
                                  ): Either[ErrorResponse, ReadResultResponse] = {
    val countRequest = ReadResultRequest(hiveCredentials.projectId, waitTime, hiveCredentials.toAuthenticationInfo, shrineQueryResult.count.localId.toString)
    val countAttempt: ShrineResponse = delegateResultRetrievingAdapter.process(researcherAuthn,countRequest)

    countAttempt match {
      case response: ReadResultResponse => Right(response)
      case e:ErrorResponse =>
        error(s"Error requesting count result from the CRC: '$e'")
        Left(e)
    }
  }

  private def retrieveBreakdownResults(
                                        waitTime:Duration,
                                        researcherAuthn:AuthenticationInfo,
                                        queryCompletedResponse: ReadInstanceResultsResponse
                                      ): Either[Seq[ErrorResponse], Map[ResultOutputType, I2b2ResultEnvelope]] = {

    val breakdownIds: Seq[String] = queryCompletedResponse.results.filter(_.resultType.exists(_.isBreakdown)).map(_.resultId.toString)

    val breakdownAttempts: Seq[ShrineResponse] = breakdownIds.map{ breakdownId =>
      val request = ReadResultRequest(hiveCredentials.projectId, waitTime, hiveCredentials.toAuthenticationInfo, breakdownId)
      delegateResultRetrievingAdapter.process(researcherAuthn, request)
    }

    //Log errors retrieving breakdown
    val breakdownErrorResponses: Seq[ErrorResponse] = breakdownAttempts.collect{
      case e:ErrorResponse =>
        error(s"Error requesting breakdown result from the CRC: '$e'")
        e
    }

    if(breakdownErrorResponses.isEmpty) {
      val breakdownResponses: Seq[ReadResultResponse] = breakdownAttempts.collect {
        case response: ReadResultResponse => response
      }

      val breakdownsByType: Map[ResultOutputType, I2b2ResultEnvelope] = (for {
        breakdownResponse <- breakdownResponses
        resultType <- breakdownResponse.metadata.resultType
      } yield resultType -> breakdownResponse.data).toMap

      Right(breakdownsByType)
    } else Left(breakdownErrorResponses)
  }

  private def storeResultIfNecessary(shrineQueryResult: ShrineQueryResult,
                                     response: Rsp,
                                     authn: AuthenticationInfo,
                                     queryId: Long, //network query id
                                     failedBreakdownTypes: Set[ResultOutputType]): Unit = {
    val responseIsDone = response.results.forall(_.statusType.isDone)

    if (responseIsDone) {
      storeResult(shrineQueryResult, response, authn, queryId, failedBreakdownTypes)
    }
    if (collectAdapterAudit) AdapterAuditDb.db.insertResultSent(queryId, response)
  }

  private def storeResult(
                           shrineQueryResult: ShrineQueryResult,
                           response: Rsp,
                           authn: AuthenticationInfo,
                           queryId: Long,
                           failedBreakdownTypes: Set[ResultOutputType]) {
    val rawResults = response.results
    val obfuscatedResults = obfuscator.obfuscateResults(doObfuscation)(response.results)

    for {
      shrineQuery: ShrineQuery <- dao.findQueryByNetworkId(queryId)
      queryResult: QueryResult <- rawResults.headOption
      obfuscatedQueryResult: QueryResult <- obfuscatedResults.headOption
    } {
      val queryDefinition = QueryDefinition(shrineQuery.name, shrineQuery.queryDefinition.expr)

      dao.inTransaction {
        dao.deleteQuery(queryId)

        dao.storeResults(authn, shrineQueryResult.localId, queryId, queryDefinition, rawResults, obfuscatedResults, failedBreakdownTypes.toSeq, queryResult.breakdowns, obfuscatedQueryResult.breakdowns)
      }
    }
  }

  private type Unmarshaller[R] = Set[ResultOutputType] => NodeSeq => Try[R]
  
  private final class DelegateAdapter[Rqst <: ShrineRequest, Rspns <: ShrineResponse](unmarshaller: Unmarshaller[Rspns])
    extends CrcAdapter[Rqst, Rspns](poster, hiveCredentials) {

    def process(authn: AuthenticationInfo, req: Rqst): Rspns = processRequest(BroadcastMessage(authn, req)).asInstanceOf[Rspns]

    override protected def parseShrineResponse(xml: NodeSeq): ShrineResponse = unmarshaller(breakdownTypes)(xml).get
  }

  private lazy val delegateResultRetrievingAdapter = new DelegateAdapter[ReadResultRequest, ReadResultResponse](ReadResultResponse.fromI2b2)
}

case class QueryNotFound(queryId:Long) extends AbstractProblem(ProblemSources.Adapter) {
  override def summary: String = s"Query not found"
  override def description:String = s"No query with id $queryId found on ${stamp.host.getHostName}"
}

case class QueryResultNotAvailable(queryId:Long) extends AbstractProblem(ProblemSources.Adapter) {
  override def summary: String = s"Query $queryId found but its results are not available yet."
  override def description:String = s"Query $queryId found but its results are not available yet on ${stamp.host.getHostName} The call to the CRC may not have completed."
}

case class CouldNotRetrieveQueryFromCrc(queryId:Long,x: Throwable) extends AbstractProblem(ProblemSources.Adapter) {
  override def summary: String = s"Could not retrieve query $queryId from the CRC"
  override def description:String = s"Unhandled exception while retrieving query $queryId while retrieving it from the CRC on ${stamp.host.getHostName}"
  override def throwable = Some(x)
}

case class NoResultsForQueuedQuery(nrc: NoResultsForQueuedQueryException) extends AbstractProblem(ProblemSources.Adapter) {
  override val throwable = Some(nrc)
  override val summary: String = "No results returned from CRC for a queued query."
  override val description = "The CRC returned a response with status but no results when asked for an update on a QUEUED query."
  override def detailsText = Some(s"Response from CRC was ${nrc.xmlResponseFromCrc}")
}

case class DoneButHasNoResultsForQueuedQuery(dbhnr: DoneButHasNoResultsForQueuedQueryException) extends AbstractProblem(ProblemSources.Adapter) {
  override val throwable = Some(dbhnr)
  override val summary: String = "The CRC reports a DONE status but has no results for a queued query"
  override val description = "The CRC returned a response with status DONE but no results when asked for an update on a QUEUED query."
  override def detailsText = Some(s"Response from CRC was ${dbhnr.xmlResponseFromCrc}")
}

case class NoQueryInstanceFromCrc(queryId:Long,rqiResponse: ReadQueryInstancesResponse) extends AbstractProblem(ProblemSources.Adapter) {
  override def summary: String = s"Query not found"
  override def description:String = s"No queryInstance with for $queryId found on ${stamp.host.getHostName}"
  override def detailsText = Some(s"Response from CRC was $rqiResponse")
}