package net.shrine.adapter

import java.sql.SQLException
import java.util.Date

import net.shrine.adapter.dao.BotDetectedException
import net.shrine.log.Loggable
import net.shrine.problem.{AbstractProblem, RawProblem, ProblemNotYetEncoded, ProblemSources}
import net.shrine.protocol.{BaseShrineResponse, BroadcastMessage, ErrorResponse, NoResultsForQueuedQueryException, ShrineRequest}

import scala.util.control.NonFatal

/**
 * @author Bill Simons
 * @since 4/8/11
 * @see http://cbmi.med.harvard.edu
 * @see http://chip.org
 *       <p/>
 *       NOTICE: This software comes with NO guarantees whatsoever and is
 *       licensed as Lgpl Open Source
 * @see http://www.gnu.org/licenses/lgpl.html
 */

//todo cleanup with SHRINE-3627
abstract class Adapter extends Loggable {
  
  //noinspection RedundantBlock
  final def perform(message: BroadcastMessage): BaseShrineResponse = {
    def problemToErrorResponse(problem:RawProblem):ErrorResponse = {
      //TODO: Is this needed? LoggingProblemHandler.handleProblem(problem)
      ErrorResponse(problem)
    }

    val shrineResponse = try {
      processRequest(message)
    } catch {
      case e: BotDetectedException => problemToErrorResponse(BotDetected(e))

      case e @ CrcInvocationException(invokedCrcUrl, request, cause) => problemToErrorResponse(CrcCouldNotBeInvoked(invokedCrcUrl,request,e))

      case e: AdapterMappingException => problemToErrorResponse(AdapterMappingProblem(e))

      case e: SQLException => problemToErrorResponse(AdapterDatabaseProblem(e))

      case e: NoResultsForQueuedQueryException => problemToErrorResponse(NoResultsForQueuedQuery(e))

      case NonFatal(e) => {
        val summary = if(message == null) "Unknown problem in Adapter.perform with null BroadcastMessage"
                      else s"Unexpected exception in Adapter"
        problemToErrorResponse(ProblemNotYetEncoded(summary,e))
      }
    }

    shrineResponse
  }

  protected[adapter] def processRequest(message: BroadcastMessage): BaseShrineResponse
  
  //NOOP, may be overridden by subclasses
  //todo only used by AbstractReadQueryResultAdapter . Probably OK to get rid of this.
  def shutdown(): Unit = ()
}

case class CrcCouldNotBeInvoked(crcUrl:String,request:ShrineRequest,x:CrcInvocationException) extends AbstractProblem(ProblemSources.Adapter) {
  override val throwable = Some(x)
  override val summary: String = s"Error communicating with I2B2 CRC."
  override val description: String = s"Error invoking the CRC at '$crcUrl' with a ${request.getClass.getSimpleName} due to ${throwable.get}."
  override def detailsText: Option[String] = Some(s"Request is $request")
}

case class AdapterMappingProblem(x:AdapterMappingException) extends AbstractProblem(ProblemSources.Adapter) {

  override val throwable = Some(x)
  override val summary: String = x.message
  override val description = s"The Shrine Adapter on ${stamp.host.getHostName} cannot map this query to its local terms."
  override def detailsText: Option[String] = Some(
    s"""
       |Query Definition is ${x.runQueryRequest.queryDefinition}
       |RunQueryRequest is ${x.runQueryRequest.elideAuthenticationInfo}
     """.stripMargin)
}

case class AdapterDatabaseProblem(x:SQLException) extends AbstractProblem(ProblemSources.Adapter) {

  override val throwable = Some(x)
  override val summary: String = "Problem using the Adapter database."
  override val description = "The Shrine Adapter encountered a problem using a database."
}

case class BotDetected(bdx:BotDetectedException) extends AbstractProblem(ProblemSources.Adapter) {

  override val throwable = Some(bdx)
  override val summary: String = s"A user has run so many queries in a period of time that the adapter suspects a bot."
  override val description: String = s"${bdx.domain}:${bdx.username} has run ${bdx.detectedCount} queries since ${new Date(bdx.sinceMs)}, more than the limit of ${bdx.limit} allowed in this time frame."
}
