package net.shrine.adapter

import com.typesafe.config.Config
import net.shrine.adapter.dao.squeryl.tables.Tables
import net.shrine.adapter.dao.squeryl.{SquerylAdapterDao, SquerylI2b2AdminDao}
import net.shrine.adapter.dao.{AdapterDao, I2b2AdminDao}
import net.shrine.adapter.mappings.AdapterMappings
import net.shrine.adapter.service.I2b2AdminService
import net.shrine.adapter.translators.{ExpressionTranslator, QueryDefinitionTranslator}
import net.shrine.config.ConfigExtensions
import net.shrine.crypto.BouncyKeyStoreCollection
import net.shrine.dao.squeryl.SquerylInitializer
import net.shrine.http4s.client.legacy.{EndpointConfig, Poster}
import net.shrine.log.Log
import net.shrine.protocol.{HiveCredentials, ResultOutputType}

import scala.collection.JavaConverters._
import scala.concurrent.duration.Duration

/**
  * All the parts required for an adapter.
  *
  * @author david 
  * @since 1.22
  */
//todo cleanup with SHRINE-3627 and moving the AIM 3 APIs to http4s
case class AdapterComponents(
                              i2b2AdminService: I2b2AdminService,
                              adapterDao: AdapterDao,
                              adapterMappings: AdapterMappings,
                              lastModified: Long
                            )

object AdapterComponents {
  def apply(adapterConfig: Config,
            certCollection: BouncyKeyStoreCollection,
            squerylInitializer: SquerylInitializer,
            breakdownTypes: Set[ResultOutputType],
            crcHiveCredentials: HiveCredentials,
            pmPoster: Poster):
  AdapterComponents = {
    val crcEndpoint: EndpointConfig = adapterConfig.getConfigured("crcEndpoint",EndpointConfig(_))

    val crcPoster: Poster = Poster(certCollection,crcEndpoint)

    val squerylAdapterTables: Tables = new Tables

    val adapterDao: AdapterDao = new SquerylAdapterDao(squerylInitializer, squerylAdapterTables)(breakdownTypes)

    //NB: Is i2b2HiveCredentials.projectId the right project id to use?
    val i2b2AdminDao: I2b2AdminDao = new SquerylI2b2AdminDao(crcHiveCredentials.projectId, squerylInitializer, squerylAdapterTables)

    val adapterMappingsFile = adapterConfig.getString("adapterMappingsFileName")
    val adapterMappings: AdapterMappings = AdapterMappings(adapterMappingsFile)

    val expressionTranslator: ExpressionTranslator = ExpressionTranslator(adapterMappings)

    val queryDefinitionTranslator: QueryDefinitionTranslator = new QueryDefinitionTranslator(expressionTranslator)

    val doObfuscation = adapterConfig.getBoolean("setSizeObfuscation")
    val collectAdapterAudit = adapterConfig.getBoolean("audit.collectAdapterAudit")

    val botCountTimeThresholds: Seq[(Long, Duration)] = {
      import scala.concurrent.duration.DurationLong
      val countsAndMilliseconds: Seq[Config] = adapterConfig.getConfig("botDefense").getConfigList("countsAndMilliseconds").asScala
      countsAndMilliseconds.map(pairConfig => (pairConfig.getLong("count"),pairConfig.getLong("milliseconds").milliseconds))
    }

    val obfuscator:Obfuscator = adapterConfig.getConfigured("obfuscation",Obfuscator(_))

    Log.info(s"obfuscator is $obfuscator")

    //todo only used by i2b2AdminService . Can it be removed from there for SHRINE-3540 ?
    val runQueryAdapter = RunQueryAdapter(
      poster = crcPoster,
      dao = adapterDao,
      hiveCredentials = crcHiveCredentials,
      conceptTranslator = queryDefinitionTranslator,
      doObfuscation = doObfuscation,
      runQueriesImmediately = adapterConfig.getBoolean("immediatelyRunIncomingQueries"), 
      breakdownTypes = breakdownTypes,
      collectAdapterAudit = collectAdapterAudit,
      botCountTimeThresholds = botCountTimeThresholds,
      obfuscator = obfuscator,
      adapterConfig.get("crcRunQueryTimeLimit",Duration(_)),
      certCollection
    )

    AdapterComponents(
      i2b2AdminService = new I2b2AdminService(
        dao = adapterDao,
        i2b2AdminDao = i2b2AdminDao,
        pmPoster = pmPoster,
        runQueryAdapter = runQueryAdapter
      ),
      adapterDao = adapterDao,
      adapterMappings = adapterMappings,
      lastModified = adapterMappings.lastModified)
  }
}