package net.shrine.adapter

import cats.effect.{IO, Timer}
import fs2.Stream
import net.shrine.adapter.dao.QueryResultStatus
import net.shrine.config.ConfigExtensions
import net.shrine.http4s.catsio.{ExecutionContexts, IOScheduler}
import net.shrine.hub.mom.ShrineMomClient
import net.shrine.log.Log
import net.shrine.problem.{AbstractProblem, JsonProblemDigest, ProblemSources, RawProblem}
import net.shrine.protocol.version.v1.{UpdateResult, UpdateResultWithError, UpdateResultWithProgress}
import net.shrine.protocol.version.{NodeKey, QueryId, ResultStatuses}
import net.shrine.protocol.{AuthenticationInfo, ErrorResponse, QueryResult, ReadInstanceResultsResponse, ReadQueryInstancesResponse, ReadQueryResultResponse, ResultOutputType, ShrineResponse}
import net.shrine.source.ConfigSource

import scala.concurrent.duration.FiniteDuration
import scala.util.control.NonFatal

/**
  * Poll for results from QUEUED queries and send those results to the hub when found.
  *
  * @since 1.26.5.1
 */

object QueuedQueriesPoller {

  val queuedQueryTimeToLive:FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.adapter.queuedQueryTimeToLive")

  val queuedQueryRestTimeBeforePolling:FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.adapter.queuedQueryRestTimeBeforePolling")

  val queuedQueryCrcPollHttpCallTimeout:FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.adapter.queuedQueryCrcPollHttpCallTimeout")

  val crcRunQueryTimeLimit:FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.adapter.crcRunQueryTimeLimit")

  implicit val timer: Timer[IO] = IO.timer(ExecutionContexts.crcExecutionContext)

  private val scheduler: IOScheduler = IOScheduler.scheduleWithFixedDelay(
    initialDelay = queuedQueryRestTimeBeforePolling,
    interval = queuedQueryRestTimeBeforePolling,
    task = () => checkForQueuedQueries(),
    name = "QueuedQueriesPoller"
  )

  def restart(): IO[Unit] = scheduler.restart()

  def stop(): IO[Unit] = scheduler.stop()

  lazy val readQueryResultAdapter:ReadQueryResultAdapter = ReadQueryResultAdapter(ConfigSource.config)

  /**
    * @return IO[true] if any queued queries are found, false if not
    */
  def checkForQueuedQueries():IO[Boolean] = {

    //if any potential queued queries are found, return true
    Log.debug(s"Starting checkForQueuedQueries")

//Check the squyrl database for not completed, not running queries
    val queuedQueriesIO: IO[Seq[QueryResultStatus]] = IO(readQueryResultAdapter.selectQueuedQueryIds)

    // Filter out queries that were very recently QUEUED
    val oldEnoughIO: IO[(Seq[QueryResultStatus],Boolean)] = queuedQueriesIO.map{ queuedQueries =>
      Log.debug(s"Found ${queuedQueries.size} queries. ${queuedQueries.mkString(", ")}")
      val now = System.currentTimeMillis()
      (queuedQueries.filter{ queuedQuery =>
        //Wait at least polling time plus the CRC time. queuedQuery.timestamp.getTime can be the start time
        (now - queuedQuery.timestamp.getTime) > queuedQueryRestTimeBeforePolling.toMillis + crcRunQueryTimeLimit.toMillis
      },queuedQueries.nonEmpty)
    }
//Ask the CRC for an update
    oldEnoughIO.flatMap{ oldEnoughAndCheckAgain =>
      val oldEnough: Seq[QueryResultStatus] = oldEnoughAndCheckAgain._1
      val oldQueryIds = oldEnough.map(_.networkQueryId).distinct

      val oldEnoughAndDistinct: Seq[QueryResultStatus] = oldQueryIds.flatMap { id =>
        oldEnough.collectFirst { case q: QueryResultStatus if q.networkQueryId == id => q } }

      Log.debug(s"Found ${oldEnoughAndDistinct.size} queries old enough to check. ${oldEnoughAndDistinct.mkString(", ")}")

      val askCrcIOs = oldEnoughAndDistinct.map{ queuedQuery:QueryResultStatus =>
        Log.debug(s"Will ask the CRC about query ${queuedQuery.networkQueryId}")
        val auth = AuthenticationInfo.noPassword(queuedQuery.username,queuedQuery.domain)
        IO(readQueryResultAdapter.askCrc(
          queuedQuery.networkQueryId,
          queuedQueryCrcPollHttpCallTimeout,
          auth
        )).flatMap{ interpretAndMaybeSendAnUpdate(queuedQuery,_) }.handleErrorWith{
          case NonFatal(x) => {
            //trap NonFatal exceptions, log, and move on to the next query
            //this will handle bad behavior from the CRC, just stumbling on the problematic query and continuing
            IO(Log.error(s"Caught exception in QueuedQueriesPoller while checking on query ${queuedQuery.networkQueryId}",x))
          }
        }
      }
      val itOfStreams = askCrcIOs.map(Stream.eval)
      val seqResultStatus = itOfStreams.foldLeft[Stream[IO,Unit]](Stream.empty){ (soFar,next) =>
        soFar.append(next)
      }.compile.drain

//If there are no queued queries then let the polling stop (to be restarted on the next queued query)
      seqResultStatus.map(_ => oldEnoughAndCheckAgain._2)
    }.handleErrorWith{
      case NonFatal(x) =>
        Log.error(s"Caught exception in QueuedQueriesPoller.checkForQueuedQueries",x)
        IO.raiseError(x)
    }
  }

  def interpretAndMaybeSendAnUpdate(
                                     oldQueryResult:QueryResultStatus,
                                     crcResponse:Either[ErrorResponse, ShrineResponse]
                                   ):IO[Unit] = {
    val queryId = new QueryId(oldQueryResult.networkQueryId)

    Log.debug(s"$queryId's response is $crcResponse")
    val maybeUpdateFromCrc: Option[UpdateResult] = crcResponse.fold({ errorResponse: ErrorResponse =>
      readQueryResultAdapter.storeErrorFromShrine(oldQueryResult,errorResponse.problem) //Give the CRC one chance - See SHRINE-3327
      errorResponse.problem match {
        case qrna:QueryResultNotAvailable =>
          Log.error(s"Not reporting $qrna to the hub. If it is recurring please set the status for this query to FINISHED in the shrine_query_history.QUERY_RESULT table")
          None       //don't report QueryResultNotAvailable SHRINE-3585
        case rp:RawProblem => Some(UpdateResultWithError(
          queryId = queryId,
          adapterNodeKey = NodeKey.localNodeKey,
          problem = JsonProblemDigest(rp),
          status = ResultStatuses.ErrorFromCrc,
          statusMessage = Option(errorResponse.errorMessage)
        ))
        case _ => throw new IllegalStateException("Error response problems must be subclasses of RawProblem")
      }
    },{ shrineResponse: ShrineResponse  =>

      Log.debug(s"interpretAndMaybeSendAnUpdate shrineResponse is $shrineResponse")

      val maybeUpdateResult: Option[UpdateResult] = shrineResponse match {
        case rqrr:ReadQueryResultResponse =>
          val updateResult = rqrr.singleNodeResult.createUpdateResult(
            queryId,
            rqrr.queryId,
            rqrr.results.head.resultType.getOrElse(ResultOutputType.PATIENT_COUNT_XML)
          )
          if(updateResult.status.isFinal) Some(updateResult)
          else None //only send an update if there's finally a result
        case rirr:ReadInstanceResultsResponse => // this is an intermediate state unless there's an error. If so store it.
          val updateResult = rirr.results.head.createUpdateResult(
            queryId,
            rirr.results.head.resultId,
            rirr.results.head.resultType.getOrElse(ResultOutputType.PATIENT_COUNT_XML)
          )
          updateResult match {
            case updateResultWithError:UpdateResultWithError =>
              //If this update is an error then store it.
              // A ReadInstanceResultsResponse - if in error - will likely not be stored
              readQueryResultAdapter.storeErrorFromShrine(oldQueryResult, updateResultWithError.problem)
              Some(updateResultWithError)
            case _ => None
          }
        case rqir:ReadQueryInstancesResponse => // this is an intermediate state unless there's an error. If so store it.
          val maybeUpdateResult = rqir.createUpdateResult
          maybeUpdateResult.foreach {
            case updateResultWithError:UpdateResultWithError => readQueryResultAdapter.storeErrorFromShrine(oldQueryResult, updateResultWithError.problem)
            case _ =>
          }
          maybeUpdateResult
        case _ => throw new IllegalArgumentException(s"Code does not handle a ${shrineResponse.getClass.getSimpleName} $shrineResponse")
      }
      maybeUpdateResult
    })

    val maybeUpdate: Option[UpdateResult] = maybeUpdateFromCrc.fold{
      maybeQueryTooOld(oldQueryResult)  //no update from the CRC, so check timeout
    } {
      //still in some flavor of QUEUED
      case progress: UpdateResultWithProgress => maybeQueryTooOld(oldQueryResult).orElse(Some(progress))
      case update => Some(update)
    }
    maybeUpdate.fold(IO.unit)(sendUpdateResultMessageIO)
  }

  private def maybeQueryTooOld(oldQueryResult:QueryResultStatus):Option[UpdateResult] = {
    val now = System.currentTimeMillis()
    if ((now - oldQueryResult.timestamp.getTime) > queuedQueryTimeToLive.toMillis ) {
      val queryId = new QueryId(oldQueryResult.networkQueryId)
      val problem = QueryQueuedTimeToLiveExceeded(queryId,queuedQueryTimeToLive)

      //store error to the database
      readQueryResultAdapter.storeErrorFromShrine(oldQueryResult,problem)

      Some(UpdateResultWithError(
        queryId = queryId,
        adapterNodeKey = NodeKey.localNodeKey,
        problem = JsonProblemDigest(problem),
        status = ResultStatuses.ErrorFromCrc,
        statusMessage = Option(problem.description)
      ))
    }
    else None
  }

  private def sendUpdateResultMessageIO(updateResult:UpdateResult): IO[Unit] = {
    ShrineMomClient.sendToHubIO(updateResult.queryId,updateResult,UpdateResult,s" updateResult for ${updateResult.queryId} to the hub")
  }
}

case class QueryQueuedTimeToLiveExceeded(queryId: QueryId,queryTimeToLive:FiniteDuration) extends AbstractProblem(ProblemSources.Adapter) {
  override val throwable: Option[Throwable] = None
  override val summary: String = s"Shrine's adapter has given up on this query after $queryTimeToLive"
  override val description = s"Shrine's adapter has polled the CRC for a result from query $queryId for $queryTimeToLive before giving up."
  override def detailsText: None.type = None
}
