package net.shrine.adapter

import com.typesafe.config.Config
import javax.sql.DataSource
import net.shrine.adapter.dao.model.{ShrineQuery, ShrineQueryResult}
import net.shrine.adapter.dao.squeryl.SquerylAdapterDao
import net.shrine.adapter.dao.squeryl.tables.Tables
import net.shrine.adapter.dao.{AdapterDao, QueryResultStatus}
import net.shrine.config.ConfigExtensions
import net.shrine.crypto.BouncyKeyStoreCollection
import net.shrine.dao.squeryl.{DataSourceSquerylInitializer, SquerylDbAdapterSelecter, SquerylInitializer}
import net.shrine.http4s.client.legacy.{EndpointConfig, Poster}
import net.shrine.problem.Problem
import net.shrine.protocol.query.QueryDefinition
import net.shrine.protocol.{AuthenticationInfo, HiveCredentials, QueryResult, ReadQueryResultRequest, ReadQueryResultResponse, ResultOutputType, ResultOutputTypes}
import net.shrine.slick.TestableDataSourceCreator
import net.shrine.source.ConfigSource
import org.squeryl.internals.DatabaseAdapter

import scala.concurrent.duration.Duration


/**
	* For GetQueryResult , which is i2b2's GetQueryResult .
	*
 * @author clint
 * @since Nov 2, 2012
 * 
 */
//todo cleanup with SHRINE-3627
final class ReadQueryResultAdapter(
	poster: Poster,
	override val hiveCredentials: HiveCredentials,
	dao: AdapterDao,
	doObfuscation: Boolean,
	breakdownTypes: Set[ResultOutputType],
	collectAdapterAudit:Boolean,
	obfuscator: Obfuscator,
	queuedQueryTimeToLive:Duration
) extends AbstractReadQueryResultAdapter[ReadQueryResultRequest, ReadQueryResultResponse](
	poster = poster,
	hiveCredentials = hiveCredentials,
	dao = dao,
	doObfuscation = doObfuscation,
	getQueryId = _.queryId,
	getProjectId = _.projectId,
	toResponse = ReadQueryResultResponse(_, _),
	breakdownTypes = breakdownTypes,
	collectAdapterAudit = collectAdapterAudit,
	obfuscator = obfuscator,
	queuedQueryTimeToLive = queuedQueryTimeToLive
) {
	def selectQueuedQueryIds: Seq[QueryResultStatus] = dao.findQueuedQueryIds

	def storeErrorFromShrine(oldQueryResult:QueryResultStatus,problem:Problem): Unit = {
		val authn = AuthenticationInfo.noPassword(oldQueryResult.username,oldQueryResult.domain)
		val newErrorQueryResult = QueryResult.errorResult(Some(problem.description),problem.summary,problem)

		dao.findQueryByNetworkId(oldQueryResult.networkQueryId).fold{
			error(s"storeErrorFromShrine No ShrineQuery in the database for ${oldQueryResult.networkQueryId}. This should not happen.")
		}{ shrineQuery: ShrineQuery =>
			val queryDefinition = QueryDefinition(shrineQuery.name, shrineQuery.queryDefinition.expr)
			val maybeOldResults: Option[ShrineQueryResult] = dao.findResultsFor(oldQueryResult.networkQueryId)
			//maybeOldResults is only used to find errors in breakdowns
			val failedBreakdownTypes: Seq[ResultOutputType] = maybeOldResults.map(_.breakdowns.map(_.resultType)).getOrElse(Seq.empty)

			dao.inTransaction {
				dao.deleteQuery(oldQueryResult.networkQueryId)
				dao.storeResults(authn, shrineQuery.i2b2MasterQueryId, oldQueryResult.networkQueryId, queryDefinition, Seq(newErrorQueryResult), Seq(newErrorQueryResult), failedBreakdownTypes, Map.empty, Map.empty)
			}
		}
	}
}

object ReadQueryResultAdapter {
	def apply(config:Config): ReadQueryResultAdapter = {
		val config: Config = ConfigSource.config
		val shrineConfig = config.getConfig("shrine")

		val certCollection: BouncyKeyStoreCollection = BouncyKeyStoreCollection.fromConfig

		val adapterConfig = shrineConfig.getConfig("adapter")

		val crcEndpoint: EndpointConfig = adapterConfig.getConfigured("crcEndpoint",EndpointConfig(_))
		val crcPoster: Poster = Poster(certCollection,crcEndpoint)

		val squerylAdapter: DatabaseAdapter = SquerylDbAdapterSelecter.determineAdapter(shrineConfig.getString("shrineDatabaseType"))
		val dataSource: DataSource = TestableDataSourceCreator.dataSource(shrineConfig.getConfig("squerylDataSource.database"))

		val squerylInitializer: SquerylInitializer = new DataSourceSquerylInitializer(dataSource, squerylAdapter)

		val squerylAdapterTables: Tables = new Tables
		val breakdownTypes: Set[ResultOutputType] = shrineConfig.getOptionConfigured("breakdownResultOutputTypes", ResultOutputTypes.fromConfig).getOrElse(Set.empty)

		val adapterDao: AdapterDao = new SquerylAdapterDao(squerylInitializer, squerylAdapterTables)(breakdownTypes)

		val crcHiveCredentials: HiveCredentials = shrineConfig.getConfigured("hiveCredentials", HiveCredentials(_, HiveCredentials.CRC))

		val doObfuscation = adapterConfig.getBoolean("setSizeObfuscation")
		val collectAdapterAudit = adapterConfig.getBoolean("audit.collectAdapterAudit")

		val obfuscator:Obfuscator = adapterConfig.getConfigured("obfuscation",Obfuscator(_))

		val queuedQueryTimeToLive:Duration = adapterConfig.get("queuedQueryTimeToLive",Duration(_))

		new ReadQueryResultAdapter(
			crcPoster,
			crcHiveCredentials,
			adapterDao,
			doObfuscation,
			breakdownTypes,
			collectAdapterAudit,
			obfuscator,
			queuedQueryTimeToLive
		)
	}
}