package net.shrine.adapter

import java.sql.SQLException

import cats.effect.{Fiber, IO}
import com.typesafe.config.Config
import javax.sql.DataSource
import net.shrine.adapter.audit.AdapterAuditDb
import net.shrine.adapter.dao.AdapterDao
import net.shrine.adapter.dao.squeryl.SquerylAdapterDao
import net.shrine.adapter.dao.squeryl.tables.Tables
import net.shrine.adapter.mappings.AdapterMappings
import net.shrine.adapter.translators.{ExpressionTranslator, QueryDefinitionTranslator}
import net.shrine.config.ConfigExtensions
import net.shrine.crypto.{BouncyKeyStoreCollection, SignatureNotVerifiedException}
import net.shrine.dao.squeryl.{DataSourceSquerylInitializer, SquerylDbAdapterSelecter, SquerylInitializer}
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.http4s.client.legacy.{EndpointConfig, Poster}
import net.shrine.hub.mom.ShrineMomClient
import net.shrine.log.Log
import net.shrine.problem._
import net.shrine.protocol.query.{CouldNotMapAllTermsException, MappingException, QueryDefinition}
import net.shrine.protocol.version.v1.{RunQueryForResult, UpdateResult, UpdateResultWithError, UpdateResultWithProgress}
import net.shrine.protocol.version.{CouldNotVerifySignature, NodeKey, QueryId, ResultStatuses}
import net.shrine.protocol._
import net.shrine.slick.TestableDataSourceCreator
import net.shrine.source.ConfigSource
import net.shrine.util.XmlDateHelper
import org.squeryl.internals.DatabaseAdapter

import scala.concurrent.blocking
import scala.concurrent.duration.{Duration, DurationInt}
import scala.language.postfixOps
import scala.util.control.NonFatal
import scala.util.{Failure, Success, Try}
import scala.xml.{NodeSeq, XML}

/**
 * @author Bill Simons
 * @author clint
 * @since 4/15/11
 * @see http://cbmi.med.harvard.edu
 * @see http://chip.org
 *       <p/>
 *       NOTICE: This software comes with NO guarantees whatsoever and is
 *       licensed as Lgpl Open Source
 * @see http://www.gnu.org/licenses/lgpl.html
 */
//todo cleanup with SHRINE-3627
final case class RunQueryAdapter(
  poster: Poster,
  dao: AdapterDao,
  override val hiveCredentials: HiveCredentials,
  conceptTranslator: QueryDefinitionTranslator,
  doObfuscation: Boolean,
  runQueriesImmediately: Boolean,
  breakdownTypes: Set[ResultOutputType],
  collectAdapterAudit:Boolean,
  botCountTimeThresholds:Seq[(Long,Duration)],
  obfuscator: Obfuscator,
  crcRunQueryTimeLimit:Duration = 30 seconds,
  certCollection:BouncyKeyStoreCollection
) extends CrcAdapter[RunQueryRequest, RunQueryResponse](poster, hiveCredentials) {

  logStartup()

  def startRunQueryForExpectedResult(runQueryForResult: RunQueryForResult): IO[Fiber[IO, Unit]] = {
    IO(runQueryForExpectedResult(runQueryForResult)).start(IO.contextShift(ExecutionContexts.crcExecutionContext))
  }

  def runQueryForExpectedResult(runQueryForResult: RunQueryForResult):Unit = {
    val thisNodeKey = ConfigSource.config.get("shrine.nodeKey",new NodeKey(_))

    if (collectAdapterAudit) AdapterAuditDb.db.insertQueryReceived(runQueryForResult)

    val queryId:QueryId = runQueryForResult.resultProgress.queryId

    val updateResultProgressReceived = UpdateResultWithProgress(queryId,thisNodeKey,ResultStatuses.ReceivedByAdapter)
    sendUpdateResult(updateResultProgressReceived)

    import cats.syntax.either._
    val runQueryRequestTry: Try[RunQueryRequest] = runQueryForResult.query.verifySignature(certCollection).toTry.map{ _ =>
        runQueryForResultToRunQueryRequest(runQueryForResult)
    }

    val resultFromRun: Try[(UpdateResult, Option[ShrineResponse])] = runQueryRequestTry.map { runQueryRequest: RunQueryRequest =>

      if (runQueriesImmediately) {
        debug(s"Performing query from user ${runQueryRequest.authn.domain}:${runQueryRequest.authn.username}")

        blocking(dao.checkIfBot(runQueryRequest.authn,botCountTimeThresholds))

        val translatedRequest: ShrineRequest = translateRequest(runQueryRequest)
        if (collectAdapterAudit) AdapterAuditDb.db.insertExecutionStarted(runQueryRequest)

        runQueryInCrcAndStoreResults(runQueryRequest,queryId,translatedRequest,thisNodeKey)
      } else {
        debug(s"Storing query from user ${runQueryRequest.authn.domain}:${runQueryRequest.authn.username} to run manually")

        storeQuery(runQueryRequest.authn, runQueryRequest)

        val updateProgress = UpdateResultWithProgress(queryId,thisNodeKey,ResultStatuses.QueuedForManualSubmission,Some("Adapter configured to submit queries manually"))
        (updateProgress, None)
      }
    }.transform(s => Try(s),{ throwable:Throwable =>
        val problem = throwable match {
          case x:SignatureNotVerifiedException => CouldNotVerifySignature(x,runQueryForResult.resultProgress.id)
          case e: AdapterMappingException => AdapterMappingProblem(e)
          case e: SQLException => AdapterDatabaseProblem(e)
          case e: CrcInvocationException => CrcCouldNotBeInvoked(e.invokedUrl,e.request,e)

          case NonFatal(x) => ExceptionWhileAttemptingToRunQuery(x)
        }
        Try{
          val errorUpdate = UpdateResultWithError.create(queryId,thisNodeKey,problem,ResultStatuses.ErrorInShrine)
          debug(s"errorUpdate is $errorUpdate",throwable)
          (errorUpdate,None)
        }
      }
    )

    resultFromRun.map { r =>
      sendUpdateResult(r._1) //todo use an IO someday

      if (collectAdapterAudit) r._2.foreach(AdapterAuditDb.db.insertResultSent(queryId.underlying, _))
    }
  }

  private def runQueryInCrcAndStoreResults (runQueryRequest: RunQueryRequest,queryId: QueryId,translatedRequest: ShrineRequest, thisNodeKey:NodeKey): (UpdateResult, Option[ShrineResponse]) = {

    val updateResultProgressToCrc = UpdateResultWithProgress(queryId,thisNodeKey,ResultStatuses.SubmittedToCRC)

    sendUpdateResult(updateResultProgressToCrc)
    try {
      val i2b2Response: String = callCrc(translatedRequest)

      //NB: Pass through ErrorResponses received from the CRC.
      //See: https://open.med.harvard.edu/jira/browse/SHRINE-794
      val responseFromCrc: ShrineResponse = parseShrineErrorResponseWithFallback(i2b2Response) match {
        case e: ErrorResponse => e
        case rawRunQueryResponse: RawCrcRunQueryResponse => processRawCrcRunQueryResponse(runQueryRequest.authn, runQueryRequest, rawRunQueryResponse)
      }
      if (collectAdapterAudit) AdapterAuditDb.db.insertExecutionCompletedShrineResponse(runQueryRequest, responseFromCrc)

      val updateResult: UpdateResult = responseToResult(queryId, responseFromCrc, thisNodeKey)
      debug(s"updateResult is $updateResult")
      //if the responseToResult is some flavor of QUEUED, start polling.
      QueuedQueriesPoller.restart().unsafeRunSync()

      (updateResult, Option(responseFromCrc))
    } catch {
      case NonFatal(x) =>

        val errorUpdate = UpdateResultWithError.create(queryId,thisNodeKey,ExceptionWhileAttemptingToRunQuery(x),ResultStatuses.ErrorFromCrc)
        (errorUpdate,None)
    }
  }

  private def runQueryForResultToRunQueryRequest(runQueryForResult: RunQueryForResult):RunQueryRequest = {

    val authToUse = AuthenticationInfo.noPasswordFromResearcher(runQueryForResult.researcher)

    val outputTypes: Set[ResultOutputType] = try {
      RunQueryRequest.determineI2b2OutputTypes(breakdownTypes)(XML.loadString(runQueryForResult.query.outputTypesXml))
    } catch {
      case NonFatal(x) => throw CouldNotInterpretI2b2OutputTypesException(runQueryForResult.query.outputTypesXml,x)
    }

    val queryDefinition: QueryDefinition = try {
      QueryDefinition.fromI2b2(runQueryForResult.query.queryDefinitionXml).get
    } catch {
      case NonFatal(x) => throw CouldNotInterpretQueryDefinitionException(runQueryForResult.query.queryDefinitionXml,x)
    }

    RunQueryRequest(
      projectId = runQueryForResult.query.projectName,
      waitTime = crcRunQueryTimeLimit,
      authn = authToUse,
      networkQueryId = runQueryForResult.query.id.underlying,
      topicIdAndName = Some((runQueryForResult.query.topicId.underlying.toString,runQueryForResult.topic.name)),
      outputTypes = outputTypes,
      queryDefinition = queryDefinition,
      nodeId = Some(XmlNodeName.fromProtocolNode(runQueryForResult.node))
    )
  }

  override protected[adapter] def parseShrineResponse(xml: NodeSeq): RawCrcRunQueryResponse = RawCrcRunQueryResponse.fromI2b2(breakdownTypes)(xml).get //TODO: Avoid .get call

  override protected[adapter] def translateNetworkToLocal(request: RunQueryRequest): RunQueryRequest = {
    try { request.mapQueryDefinition(conceptTranslator.translate) }
    catch {
      case c @ CouldNotMapAllTermsException(unmappable) =>
        throw AdapterMappingException(request, s"Could not map query term(s) ${unmappable.mkString(", ")}}.", c)
      case m @ MappingException(message) => throw AdapterMappingException(request, message, m)
    }
  }

  override protected[adapter] def processRequest(message: BroadcastMessage): ShrineResponse = {

    if (collectAdapterAudit) AdapterAuditDb.db.insertQueryReceived(message)

    dao.checkIfBot(message.networkAuthn,botCountTimeThresholds)

    val runQueryReq = message.request.asInstanceOf[RunQueryRequest]

    //todo remove this behavior with the special HMS-only identity code SHRINE-2723
    //We need to use the network identity from the BroadcastMessage, since that will have the network username 
    //(ie, ecommons) of the querying user. Using the AuthenticationInfo from the incoming request breaks the fetching
    //of previous queries on deployed systems where the credentials in the identity param to this method and the authn
    //field of the incoming request are different, like the HMS Shrine deployment.
    //NB: Credential field is wiped out to preserve old behavior -Clint 14 Nov, 2013
    val authnToUse = message.networkAuthn.copy(credential = Credential("", isToken = false))

    //use the local adapter's crcRunQueryTimeLimit for the query
    val runQueryReqWithTimeLimit = runQueryReq.copy(waitTime = crcRunQueryTimeLimit)

    if (!runQueriesImmediately) {
      debug(s"Queueing query from user ${message.networkAuthn.domain}:${message.networkAuthn.username}")

      storeQuery(authnToUse, runQueryReqWithTimeLimit)

    } else {
      debug(s"Performing query from user ${message.networkAuthn.domain}:${message.networkAuthn.username}")

      val shrineResponse: ShrineResponse = runQuery(authnToUse, message.copy(request = runQueryReqWithTimeLimit.withAuthn(authnToUse)), runQueryReqWithTimeLimit.withAuthn(authnToUse))
      val thisNodeKey = ConfigSource.config.get("shrine.nodeKey",new NodeKey(_))

      //if the responseToResult is some flavor of QUEUED, start polling.
      QueuedQueriesPoller.restart().unsafeRunSync()
      sendUpdateResult(responseToResult(new QueryId(runQueryReqWithTimeLimit.networkQueryId),shrineResponse,thisNodeKey))

      shrineResponse
    }
  }

  private def responseToResult(queryId: QueryId, shrineResponse: ShrineResponse,thisNodeKey:NodeKey):UpdateResult = {

    debug(s"start responseToResult for query $queryId")

    shrineResponse match {
      case runQueryResponse:RunQueryResponse => runQueryResponse.singleNodeResult.createUpdateResult(
        queryId,
        runQueryResponse.queryInstanceId,
        runQueryResponse.results.headOption.flatMap(_.resultType).getOrElse(ResultOutputType.PATIENT_COUNT_XML)
      )
      case er: ErrorResponse => UpdateResultWithError.create(
        queryId = queryId,
        adapterNodeKey = thisNodeKey,
        problem = er.problem,
        status = ResultStatuses.ErrorFromCrc,
        statusMessage = Some(er.errorMessage))
    }
  }

  private def sendUpdateResult(updateResult:UpdateResult):Unit = {
    ShrineMomClient.sendToHubIO(updateResult.queryId,updateResult,UpdateResult,s"$updateResult to hub").
      handleErrorWith(t => IO(Log.error("Exception while running RunQueryAdapter.sendUpdateResult", t))).unsafeRunAsyncAndForget()
  }

  private def storeQuery(authnToUse: AuthenticationInfo, request: RunQueryRequest): RunQueryResponse = {
    //Use dummy ids for what we would have received from the CRC
    val masterId: Long = -1L
    val queryInstanceId: Long = -1L
    val resultId: Long = -1L

    //TODO: is this right?? Or maybe it's project id?
    val groupId = authnToUse.domain

    val invalidSetSize = -1L
    val now = XmlDateHelper.now
    val queryResult = QueryResult(
      resultId = resultId,
      instanceId = queryInstanceId,
      resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
      setSize = invalidSetSize,
      startDate = Some(now),
      endDate = Some(now),
      description = Some("Query enqueued for later processing"),
      statusType = QueryResult.StatusType.Held,
      statusMessage = Some("Query enqueued for later processing")
    )

    dao.inTransaction {
      val insertedQueryId = dao.insertQuery(masterId.toString, request.networkQueryId, authnToUse, request.queryDefinition, isFlagged = false, hasBeenRun = false, flagMessage = None)

      val insertedQueryResultIds = dao.insertQueryResults(insertedQueryId, Seq(queryResult))

      //NB: We need to insert dummy QueryResult and Count records so that calls to StoredQueries.retrieve() in 
      //AbstractReadQueryResultAdapter, called when retrieving results for previously-queued-or-incomplete 
      //queries, will work.
      val countQueryResultId = insertedQueryResultIds(ResultOutputType.PATIENT_COUNT_XML).head

      dao.insertCountResult(countQueryResultId, -1L, -1L)
    }

    RunQueryResponse(masterId, XmlDateHelper.now, authnToUse.username, groupId, request.queryDefinition, queryInstanceId, queryResult)
  }

  private def runQuery(authnToUse: AuthenticationInfo, message: BroadcastMessage, request: RunQueryRequest): ShrineResponse = {
    if (collectAdapterAudit) AdapterAuditDb.db.insertExecutionStarted(request)

    //NB: Pass through ErrorResponses received from the CRC.
    //See: https://open.med.harvard.edu/jira/browse/SHRINE-794
    val result: ShrineResponse = super.processRequest(message) match {
      case e: ErrorResponse => e
      case rawRunQueryResponse: RawCrcRunQueryResponse => processRawCrcRunQueryResponse(authnToUse, request, rawRunQueryResponse)
    }
    if (collectAdapterAudit) AdapterAuditDb.db.insertExecutionCompletedShrineResponse(request,result)

    result
  }

  private[adapter] def processRawCrcRunQueryResponse(
                                                      authnToUse: AuthenticationInfo,
                                                      request: RunQueryRequest,
                                                      rawRunQueryResponse: RawCrcRunQueryResponse): RunQueryResponse = {
    def isBreakdown(result: QueryResult) = result.resultType.exists(_.isBreakdown)

    //here in the adapter only ever expect one count QueryResult
    val originalResults: Seq[QueryResult] = rawRunQueryResponse.results

    val (originalBreakdownResults, originalNonBreakDownResults): (Seq[QueryResult],Seq[QueryResult]) = originalResults.partition(isBreakdown)

    val resultsWithLeastCompleteState = resultsWithCountStateOfLeastCompleteState(request,originalResults,originalBreakdownResults)
    debug(s"resultsWithLeastCompleteState is $resultsWithLeastCompleteState")

    //there's always at least one count result, but it's not always the head
    val countResultWithLeastCompleteState = resultsWithLeastCompleteState.filter(_.resultType.exists(!_.isBreakdown)).head

    //if all the results are done without error it's OK to call the CRC and ask about breakdowns
    val originalBreakdownCountAttempts: Seq[(QueryResult, Try[QueryResult])] =
      if(countResultWithLeastCompleteState.statusType.isDone && !countResultWithLeastCompleteState.statusType.isError) {
        attemptToRetrieveBreakdowns(request, originalBreakdownResults)
      }
      //if any breakdowns are in some error or not-done state, don't attempt to get the breakdown counts from the CRC
      else Seq.empty

    val (successfulBreakdownCountAttempts, failedBreakdownCountAttempts) = originalBreakdownCountAttempts.partition { case (_, t) => t.isSuccess }

    val failedBreakdownCountAttemptsWithProblems = failedBreakdownCountAttempts.map { attempt: (QueryResult, Try[QueryResult]) =>
      val originalResult: QueryResult = attempt._1
      val queryResult:QueryResult = if (originalResult.problem.isDefined) originalResult
      else {
        attempt._2 match {
          case Success(_) => originalResult
          case Failure(x) => //noinspection RedundantBlock
          {
            val problem:RawProblem = x match {
              case e: ErrorFromCrcException => ErrorFromCrcBreakdown(e)
              case e: MissingCrCXmlResultException => CannotInterpretCrcBreakdownXml(e)
              case NonFatal(e) => {
                val summary = s"Unexpected exception while interpreting breakdown response"
                ProblemNotYetEncoded(summary, e)
              }
            }
            originalResult.copy(
              statusType = QueryResult.StatusType.Error,
              statusMessage = Some(problem.summary),
              problem = Some(problem)
            )
          }
        }
      }
      (queryResult,attempt._2)
    }

    logBreakdownFailures(rawRunQueryResponse, failedBreakdownCountAttemptsWithProblems)

    val successAndFailedBreakdowns: Seq[QueryResult] =
      successfulBreakdownCountAttempts.map(_._2.get) ++ failedBreakdownCountAttemptsWithProblems.map(_._1)

    val originalMergedBreakdowns: Map[ResultOutputType, I2b2ResultEnvelope] = {
      val withBreakdownCounts = successfulBreakdownCountAttempts.collect { case (_, Success(queryResultWithBreakdowns)) => queryResultWithBreakdowns }

      withBreakdownCounts.map(_.breakdowns).fold(Map.empty)(_ ++ _)
    }

    //Now we've got a new set of breakdown results from the CRC
    //some of which could conceivably be in different states than the first query
    //so check it a second time
    val resultWithBreakdownState = resultsWithCountStateOfLeastCompleteState(request,resultsWithLeastCompleteState,successAndFailedBreakdowns)

    val obfuscatedQueryResults: Seq[QueryResult] = resultWithBreakdownState.map(obfuscator.obfuscate)

    debug(s"obfuscatedQueryResults is $obfuscatedQueryResults")

    val obfuscatedNonBreakdownQueryResults: Seq[QueryResult] = obfuscatedQueryResults.filterNot(isBreakdown)

    val obfuscatedMergedBreakdowns: Map[ResultOutputType, I2b2ResultEnvelope] = originalMergedBreakdowns.mapValues(_.mapValues(obfuscator.obfuscate))

    val failedBreakdownTypes: Seq[ResultOutputType] = failedBreakdownCountAttemptsWithProblems.flatMap { case (qr, _) => qr.resultType }

    dao.storeResults(
      authn = authnToUse,
      masterId = rawRunQueryResponse.queryId.toString,
      networkQueryId = request.networkQueryId,
      queryDefinition = request.queryDefinition,
      rawQueryResults = resultWithBreakdownState,
      obfuscatedQueryResults = obfuscatedQueryResults,
      failedBreakdownTypes = failedBreakdownTypes,
      mergedBreakdowns = originalMergedBreakdowns,
      obfuscatedBreakdowns = obfuscatedMergedBreakdowns)

    // at this point the queryResult could be a mix of successes and failures.
    // SHRINE reports only the successes. See SHRINE-1567 for details
    val queryResults: Seq[QueryResult] = if (doObfuscation) obfuscatedNonBreakdownQueryResults else originalNonBreakDownResults
    val breakdownsToReturn: Map[ResultOutputType, I2b2ResultEnvelope] = if (doObfuscation) obfuscatedMergedBreakdowns else originalMergedBreakdowns

    //can failedBreakdownCountAttempts be mixed back in here?
    //there's always at least one count result, but it's not always the head
    val obfuscatedCountResult: QueryResult = queryResults.filter(_.resultType.exists(!_.isBreakdown)).head

    val resultWithBreakdowns: QueryResult = obfuscatedCountResult.withBreakdowns(breakdownsToReturn)

    debug(s"resultWithBreakdowns is $resultWithBreakdowns")

    if(debugEnabled) {
      def justBreakdowns(breakdowns: Map[ResultOutputType, I2b2ResultEnvelope]) = breakdowns.mapValues(_.data)

      debug(s"Returning QueryResult with count ${resultWithBreakdowns.setSize} ")
      debug(s"Returning QueryResult with breakdowns ${justBreakdowns(resultWithBreakdowns.breakdowns)} ")
    }

    //if any results are queued the resultWithBreakdowns should already be queued

    //todo for results with errors - the error should be in by now as well
    //for the first run of a query with breakdowns that has any errors - this is the right behavior.
    val problem: Option[Problem] = failedBreakdownCountAttemptsWithProblems.headOption.flatMap(x => x._1.problem)
    val queryResult: QueryResult = problem.fold(resultWithBreakdowns){pd =>
      if(resultWithBreakdowns.problem.isEmpty )
        QueryResult.errorResult(Some(pd.description),"Error in breakdown from CRC",pd)
      else resultWithBreakdowns
    }

    debug(s"Final queryResult: $queryResult")

    rawRunQueryResponse.toRunQueryResponse.withResult(queryResult)
  }

  private def resultsWithCountStateOfLeastCompleteState(
                                                         request: RunQueryRequest,
                                                         originalResults: Seq[QueryResult],
                                                         originalBreakdownResults: Seq[QueryResult]
                                                       ): Seq[QueryResult] = {
    //there's always at least one count result, but it's not always the head
    val countResult = originalResults.filter(_.resultType.exists(!_.isBreakdown)).head

    val countQueryState: QueryResult.StatusType = countResult.statusType
    val firstBreakdownError: Option[QueryResult] = originalBreakdownResults.find(_.statusType.isError)
    val firstBreakdownQueued: Option[QueryResult] = originalBreakdownResults.find(!_.statusType.isDone)
    val expectedBreakdownsAreAbsentFromRaw:Boolean = request.outputTypes.exists(_.isBreakdown) && originalBreakdownResults.isEmpty

    //if the query is in error, pass through
    if(countQueryState.isError) originalResults
    //if any breakdowns are in error change the query's state to error
    else if(firstBreakdownError.isDefined) Seq(countResult.consolidateWithError(firstBreakdownError.get))
    //if the query is queued, pass through
    else if(countQueryState.crcPromisedToFinishAfterReply) originalResults
    //else if any breakdowns are queued change the query's state to that queued state
    else if(firstBreakdownQueued.isDefined) Seq(countResult.withStatus(
      statusType = firstBreakdownQueued.get.statusType,
      statusMessage = firstBreakdownQueued.get.statusMessage.orElse(Some("No status message for the firstBreakdownQueued"))
    ))
    //if breakdowns are expected, but don't appear at all yet, send back some pending state
    else if(expectedBreakdownsAreAbsentFromRaw) Seq(countResult.copy(
      statusType = QueryResult.StatusType.Processing,
      statusMessage = Some("The CRC does not know about the breakdowns yet, but should soon")
    ))
    //else everything is finished. Pass through
    else originalResults
  }

  private def getResultFromCrc(parentRequest: RunQueryRequest, networkResultId: Long): Try[ReadResultResponse] = {
    def readResultRequest(runQueryReq: RunQueryRequest, networkResultId: Long) = ReadResultRequest(hiveCredentials.projectId, runQueryReq.waitTime, hiveCredentials.toAuthenticationInfo, networkResultId.toString)

    Try(XML.loadString(callCrc(readResultRequest(parentRequest, networkResultId)))).flatMap(ReadResultResponse.fromI2b2(breakdownTypes))
  }

  private[adapter] def attemptToRetrieveCount(runQueryReq: RunQueryRequest, originalCountQueryResult: QueryResult): (QueryResult, Try[QueryResult]) = {
    originalCountQueryResult -> (for {
      countData <- getResultFromCrc(runQueryReq, originalCountQueryResult.resultId)
    } yield originalCountQueryResult.withSetSize(countData.metadata.setSize))
  }

  private[adapter] def attemptToRetrieveBreakdowns(runQueryReq: RunQueryRequest, breakdownResults: Seq[QueryResult]): Seq[(QueryResult, Try[QueryResult])] = {
    breakdownResults.map { origBreakdownResult =>
      origBreakdownResult -> (for {
        breakdownData <- getResultFromCrc(runQueryReq, origBreakdownResult.resultId).map(_.data)
      } yield origBreakdownResult.withBreakdown(breakdownData))
    }
  }

  private[adapter] def logBreakdownFailures(response: RawCrcRunQueryResponse,
    failures: Seq[(QueryResult, Try[QueryResult])]) {
    for {
      (origQueryResult, Failure(e)) <- failures
    } {
      error(s"Couldn't load breakdown for QueryResult with masterId: ${response.queryId}, instanceId: ${origQueryResult.instanceId}, resultId: ${origQueryResult.resultId}. Asked for result type: ${origQueryResult.resultType}", e)
    }
  }

  private def logStartup(): Unit = {
    val message = {
      if (runQueriesImmediately) { s"${getClass.getSimpleName} will run queries immediately" }
      else { s"${getClass.getSimpleName} will queue queries for later execution" }
    }

    info(message)
  }
}

object RunQueryAdapter {
  def apply(config:Config): RunQueryAdapter = {
    val config: Config = ConfigSource.config
    val shrineConfig = config.getConfig("shrine")
    val certCollection: BouncyKeyStoreCollection = BouncyKeyStoreCollection.fromConfig

    val adapterConfig = shrineConfig.getConfig("adapter")

    val crcEndpoint: EndpointConfig = adapterConfig.getConfigured("crcEndpoint",EndpointConfig(_))
    val crcPoster: Poster = Poster(certCollection,crcEndpoint)

    val squerylAdapter: DatabaseAdapter = SquerylDbAdapterSelecter.determineAdapter(shrineConfig.getString("shrineDatabaseType"))
    val dataSource: DataSource = TestableDataSourceCreator.dataSource(shrineConfig.getConfig("squerylDataSource.database"))

    val squerylInitializer: SquerylInitializer = new DataSourceSquerylInitializer(dataSource, squerylAdapter)

    val squerylAdapterTables: Tables = new Tables
    val breakdownTypes: Set[ResultOutputType] = shrineConfig.getOptionConfigured("breakdownResultOutputTypes", ResultOutputTypes.fromConfig).getOrElse(Set.empty)

    val adapterDao: AdapterDao = new SquerylAdapterDao(squerylInitializer, squerylAdapterTables)(breakdownTypes)

    val crcHiveCredentials: HiveCredentials = shrineConfig.getConfigured("hiveCredentials", HiveCredentials(_, HiveCredentials.CRC))

    val adapterMappingsFile = adapterConfig.getString("adapterMappingsFileName")
    val adapterMappings: AdapterMappings = AdapterMappings(adapterMappingsFile)
    val expressionTranslator: ExpressionTranslator = ExpressionTranslator(adapterMappings)

    val queryDefinitionTranslator: QueryDefinitionTranslator = new QueryDefinitionTranslator(expressionTranslator)

    val doObfuscation = adapterConfig.getBoolean("setSizeObfuscation")
    val collectAdapterAudit = adapterConfig.getBoolean("audit.collectAdapterAudit")
    val runQueriesImmediately = adapterConfig.getOption("immediatelyRunIncomingQueries", _.getBoolean).getOrElse(true) //todo use reference.conf SHRINE-3090

    val botCountTimeThresholds: Seq[(Long, Duration)] = {
      import scala.collection.JavaConverters._
      import scala.concurrent.duration.DurationLong

      val countsAndMilliseconds: Seq[Config] = adapterConfig.getConfig("botDefense").
                                                getConfigList("countsAndMilliseconds").asScala
      countsAndMilliseconds.map(pairConfig =>
        (pairConfig.getLong("count"),pairConfig.getLong("milliseconds").milliseconds))
    }

    val obfuscator:Obfuscator = adapterConfig.getConfigured("obfuscation",Obfuscator(_))

    val crcRunQueryTimeLimit = adapterConfig.get("crcRunQueryTimeLimit",Duration(_))

    new RunQueryAdapter(
      crcPoster,
      adapterDao,
      crcHiveCredentials,
      queryDefinitionTranslator,
      doObfuscation,
      runQueriesImmediately,
      breakdownTypes,
      collectAdapterAudit,
      botCountTimeThresholds,
      obfuscator,
      crcRunQueryTimeLimit,
      certCollection
    )
  }
}


case class ErrorFromCrcBreakdown(x:ErrorFromCrcException) extends AbstractProblem(ProblemSources.Adapter) {

  override val throwable = Some(x)
  override val summary: String = "The CRC reported an error."
  override val description = "The CRC reported an internal error."
}

case class CannotInterpretCrcBreakdownXml(x:MissingCrCXmlResultException) extends AbstractProblem(ProblemSources.Adapter) {

  override val throwable = Some(x)
  override val summary: String = "SHRINE cannot interpret the CRC response."
  override val description = "The CRC responded, but SHRINE could not interpret that response."
}

case class ExceptionWhileAttemptingToRunQuery(x:Throwable) extends AbstractProblem(ProblemSources.Adapter) {
  override val throwable = Some(x)
  override val summary: String = "The adapter could not run the query."
  override val description = "The adapter could not run the query due to an exception."
}

case class CouldNotInterpretI2b2OutputTypesException(outputTypesText:String,cause:Throwable)
  extends Exception(s"Could not interpret output types text of '$outputTypesText' ",cause)

case class CouldNotInterpretQueryDefinitionException(queryDefinitionText:String, cause:Throwable)
  extends Exception(s"Could not interpret output types text of '$queryDefinitionText' ",cause)
