package net.shrine.adapter.mappings

import cats.effect.{Async, ContextShift, IO}
import javax.sql.DataSource
import slick.jdbc.JdbcProfile
import slick.lifted.ProvenShape
import fs2.Stream
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.log.Log
import net.shrine.slick.TestableDataSourceCreator
import net.shrine.source.ConfigSource

import scala.concurrent.{ExecutionContext, Future, blocking}

case class AdapterMappingsDb(dataSource: DataSource, schema: AdapterMappingsSchema)
                            (implicit ex: ExecutionContext, cs: ContextShift[IO]) {
  import schema.jdbcProfile.api._
  import schema._

  val db = Database.forDataSource(dataSource, None)

  def checksum: IO[Long] = metainformation.map(_.map(_.checksum).getOrElse(-1))
  def fileLastModified: IO[Long] = metainformation.map(_.map(_.fileLastModified).getOrElse(-1))
  def filename: IO[String] = metainformation.map(_.map(_.filename).getOrElse(""))
  def loadedFromFile: IO[Long] = metainformation.map(_.map(_.loadedFromFile).getOrElse(-1))

  private def metainformation: IO[Option[AdapterMappingsMetadataRow]] = runIO {
    adapterMappingsMetadataTableQuery.sortBy(_.loadedFromFile.desc).result
  }.map(_.headOption)

  def createTables: IO[Unit] = runIO {
    DBIO.seq(
      adapterMappingsTableQuery.schema.create,
      adapterMappingsMetadataTableQuery.schema.create
    )
  }

  def dropTables: IO[Unit] = runIO {
    DBIO.seq(
      adapterMappingsTableQuery.schema.drop,
      adapterMappingsMetadataTableQuery.schema.drop
    )
  }

  def truncateMappings(): IO[Int] = runIO(adapterMappingsTableQuery.delete)

  def getAll: IO[Seq[AdapterMappingsRow]] = runIO(adapterMappingsTableQuery.result)

  def localTermsFor(shrineKey: String): IO[Set[String]] = getAllByShrineKey(shrineKey)

  def localTermsFor(shrineKeys: Seq[String]): IO[Map[String, Set[String]]] = getAllByShrineKeys(shrineKeys)

  def updateMetainformation(filename: String, fileLastModified: Long, checksum: Long): IO[Int] = runIO {
    adapterMappingsMetadataTableQuery += AdapterMappingsMetadataRow(filename, fileLastModified, System.currentTimeMillis(), checksum)
  }

  def upsert(mappings: Iterable[(String, String)]): IO[Int] = upsert(mappings.iterator)

  def upsert(mappings: Iterator[(String, String)]): IO[Int] =
    upsertChunked(mappings.map(mapping => AdapterMappingsRow(mapping._1, mapping._2)), 1000, 15)

  def upsertChunked(rows: Iterator[AdapterMappingsRow], chunkSize: Int, connections: Int): IO[Int] = {
    Stream.fromIterator(rows)(Async[IO])
      .chunkN(chunkSize)
      .parEvalMap(connections)(chunkedRows => IO.fromFuture(upsertRowsJdbc(chunkedRows.iterator)))
      .compile
      .fold(0)(_ + _)
  }

  private def upsertRowsJdbc(rows: Iterator[AdapterMappingsRow]): IO[Future[Int]] = IO {
    val sql = ConfigSource.config.getConfig("shrine.adapter").getString("adapterMappingsInsertStatement")

    val inserts = SimpleDBIO[Int] { session =>
      val statement = session.connection.prepareStatement(sql)
      rows.foreach { row =>
        statement.setString(1, row.shrineKey)
        statement.setString(2, row.adapterKey)
        statement.addBatch()
      }

      statement.executeBatch().sum
    }

    db.run(inserts.transactionally)
  }

  private def runIO[R](dbio: DBIOAction[R, NoStream, _]): IO[R] =
    cs.evalOn(ex)(IO.fromFuture(IO(blocking(db.run(dbio)))))

  private def getAllByShrineKey(shrineKey: String): IO[Set[String]] = runIO(lookupByShrineKey(shrineKey))

  private def getAllByShrineKeys(shrineKeys: Seq[String]): IO[Map[String, Set[String]]] = runIO {
    DBIO.sequence(shrineKeys.map(lookupByShrineKey))
  }.map(resultSets => shrineKeys.zip(resultSets).toMap)

  private def lookupByShrineKey(shrineKey: String):DBIO[Set[String]] = {
    adapterMappingsTableQuery
      .filter(_.shrineKey === shrineKey)
      .map(_.adapterKey)
      .result
      .map(_.toSet)
  }
}

object AdapterMappingsDb {

  implicit val ex: ExecutionContext = ExecutionContexts.databaseExecutionContext
  implicit val cs: ContextShift[IO] = IO.contextShift(ex)

  lazy val db = {
    val shrineConfig = ConfigSource.config.getConfig("shrine")
    val dataSource: DataSource = TestableDataSourceCreator.dataSource(shrineConfig.getConfig("adapter.audit.database"))
    val profile = TestableDataSourceCreator.slickDriver(shrineConfig)
    AdapterMappingsDb(profile, dataSource)
  }

  def apply(jdbcProfile: JdbcProfile, dataSource: DataSource)
           (implicit ex: ExecutionContext, cs: ContextShift[IO]): AdapterMappingsDb =
    AdapterMappingsDb(dataSource, AdapterMappingsSchema(jdbcProfile))

  def reloadMappings(iter: Iterator[(String, String)], filename: String, lastModified: Long, checksum: Long): IO[Long] = for {
    _ <- AdapterMappingsDb.db.truncateMappings()
    _ <- IO(Log.debug(s"Old adapter mappings have been deleted. Inserting new mappings..."))
    upserted <- IO(Log.logDuration("Reload mappings")(Log.debug(_)) {
      AdapterMappingsDb.db.upsert(iter).unsafeRunSync()
    })
    _ <- AdapterMappingsDb.db.updateMetainformation(filename, lastModified, checksum)
  } yield upserted
}

case class AdapterMappingsSchema(jdbcProfile: JdbcProfile) {
  import jdbcProfile.api._

  val adapterMappingsTableQuery = TableQuery[AdapterMappingsTable]
  val adapterMappingsMetadataTableQuery = TableQuery[AdapterMappingsMetadataTable]

  case class AdapterMappingsTable(tag: Tag) extends Table[AdapterMappingsRow](tag, "ADAPTER_MAPPING") {
    def shrineKey:Rep[String] = column[String]("SHRINE_KEY")
    def adapterKey:Rep[String] = column[String]("ADAPTER_KEY")

    override def * :ProvenShape[AdapterMappingsRow] = (shrineKey, adapterKey) <> (AdapterMappingsRow.tupled, AdapterMappingsRow.unapply)
  }

  case class AdapterMappingsMetadataTable(tag: Tag) extends Table[AdapterMappingsMetadataRow](tag, "ADAPTER_MAPPING_META") {
    def filename:Rep[String] = column[String]("FILENAME")
    def fileLastModified:Rep[Long] = column[Long]("FILE_LAST_MODIFIED")
    def loadedFromFile:Rep[Long] = column[Long]("LOADED_FROM_FILE")
    def checksum:Rep[Long] = column[Long]("CHECKSUM")

    override def * :ProvenShape[AdapterMappingsMetadataRow] = (filename, fileLastModified, loadedFromFile, checksum) <> (AdapterMappingsMetadataRow.tupled, AdapterMappingsMetadataRow.unapply)
  }

  case class AdapterMappingsRow(shrineKey: String, adapterKey: String)
  case class AdapterMappingsMetadataRow(filename: String, fileLastModified: Long, loadedFromFile: Long, checksum: Long)
}