package net.shrine.adapter.service

import net.shrine.adapter.RunQueryAdapter
import net.shrine.adapter.components.{HeldQueries, I2b2AdminPreviousQueries, I2b2AdminQueryingUsers, QueryDefinitions}
import net.shrine.adapter.dao.{AdapterDao, I2b2AdminDao}
import net.shrine.authorization.{PmAuthorizerComponent, PmHttpClientComponent}
import net.shrine.http4s.client.legacy.Poster
import net.shrine.i2b2.protocol.pm.User
import net.shrine.log.Loggable
import net.shrine.protocol.{I2b2AdminReadQueryDefinitionRequest, I2b2AdminRequestHandler, ReadI2b2AdminPreviousQueriesRequest, ReadI2b2AdminQueryingUsersRequest, RunHeldQueryRequest, ShrineRequest, ShrineResponse}



/**
 * @author clint
 * @since Apr 4, 2013
 * 
 * Exposes an http API dictated to us by the i2b2 team for use by the i2b2 workbench app. 
 * Details of the implementation of the API were mandated by people at the PI level. Sigh.
 */
final class I2b2AdminService(
    dao: AdapterDao,
	i2b2AdminDao: I2b2AdminDao,
	override val pmPoster: Poster,
	runQueryAdapter: RunQueryAdapter) extends 
		I2b2AdminRequestHandler with 
		PmAuthorizerComponent with 
		PmHttpClientComponent with
		Loggable {
  
  require(dao != null)
  require(i2b2AdminDao != null)
  require(pmPoster != null)
  
  info("I2b2 admin service initialized")
  
  private lazy val heldQueries = HeldQueries(dao, runQueryAdapter)
  
  private lazy val i2b2AdminPreviousQueries = I2b2AdminPreviousQueries(i2b2AdminDao)
  
  private lazy val i2b2AdminQueryingUsers = I2b2AdminQueryingUsers(i2b2AdminDao)
  
  private lazy val queryDefinitions = QueryDefinitions[I2b2AdminReadQueryDefinitionRequest](dao)

  def crcUrl = runQueryAdapter.poster.url

  def obfuscate = runQueryAdapter.doObfuscation

  //NB: shouldBroadcast is ignored; we never broadcast
  override def readQueryDefinition(request: I2b2AdminReadQueryDefinitionRequest, shouldBroadcast: Boolean): ShrineResponse = {
    checkWithPmAndThen(request) {
      queryDefinitions.get
    }
  }

  //NB: shouldBroadcast is ignored; we never broadcast
  override def readI2b2AdminPreviousQueries(request: ReadI2b2AdminPreviousQueriesRequest, shouldBroadcast: Boolean): ShrineResponse = {
    checkWithPmAndThen(request) {
      i2b2AdminPreviousQueries.get
    }
  }
  
  //NB: shouldBroadcast is ignored; we never broadcast
  override def readI2b2AdminQueryingUsers(request: ReadI2b2AdminQueryingUsersRequest, shouldBroadcast: Boolean): ShrineResponse = {
    checkWithPmAndThen(request) {
      i2b2AdminQueryingUsers.get
    }
  }
  
  //NB: shouldBroadcast is ignored; we never broadcast
  override def runHeldQuery(request: RunHeldQueryRequest, shouldBroadcast: Boolean): ShrineResponse = {
    checkWithPmAndThen(request) {
      heldQueries.run
    }
  }

  def checkWithPmAndThen[Req <: ShrineRequest](request: Req)(f: Req => ShrineResponse): ShrineResponse = {
    import PmAuthorizerComponent._

    val authorized = Pm.authorize(request.projectId, Set(User.Roles.Manager), request.authn) 
    
    authorized match {
      case Authorized(user) => f(request) 
      case na: NotAuthorized => na.toErrorResponse
    }
  }
}
