package net.shrine.adapter.mappings

import cats.effect.{ContextShift, IO}
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.util.ShouldMatchersForJUnit
import org.junit.Assert.{assertEquals, assertNotEquals}
import org.junit.{After, Before, Ignore, Test}

class AdapterMappingsDbTest extends ShouldMatchersForJUnit {
  implicit val cs: ContextShift[IO] = IO.contextShift(ExecutionContexts.databaseExecutionContext)
  val dao = AdapterMappingsDb.db

  @Test
  def testNoMappings(): Unit = {
    assertEquals(Set.empty, dao.localTermsFor("hello").unsafeRunSync())
  }

  @Test
  def testOneMapping(): Unit = {
    dao.upsert(Seq("foo" -> "bar")).unsafeRunSync()
    assertEquals(Set("bar"), dao.localTermsFor("foo").unsafeRunSync())
  }

  @Test
  def testTwoMappings(): Unit = {
    dao.upsert(Seq(
      "foo" -> "bar",
      "foo" -> "more")).unsafeRunSync()

    assertEquals(Set("bar", "more"), dao.localTermsFor("foo").unsafeRunSync())
  }

  @Test
  def testTwoMappingsRedundant(): Unit = {
    dao.upsert(Seq(
      "foo" -> "bar",
      "foo" -> "bar",
      "foo" -> "more")).unsafeRunSync()

    assertEquals(Set("bar", "more"), dao.localTermsFor("foo").unsafeRunSync())
  }

  @Test
  def testTwoMappingsIgnored(): Unit = {
    dao.upsert(Seq(
      "foo" -> "bar",
      "baz" -> "egg",
      "foo" -> "more")).unsafeRunSync()

    assertEquals(Set("bar", "more"), dao.localTermsFor("foo").unsafeRunSync())
  }

  @Test
  def testMultipleInsert(): Unit = {
    dao.upsert(Seq(
      "foo" -> "bar",
      "baz" -> "egg",
      "foo" -> "more")).unsafeRunSync()

    assertEquals(Set("bar", "more"), dao.localTermsFor("foo").unsafeRunSync())

    dao.upsert(Seq(
      "foo" -> "even more",
      "spam" -> "eggs",
      "foo" -> "more")).unsafeRunSync()

    assertEquals(Set("bar", "more", "even more"), dao.localTermsFor("foo").unsafeRunSync())
  }

  @Test
  def testMany(): Unit = {
    dao.upsert(Seq(
      "foo" -> "1",
      "foo" -> "2",
      "foo" -> "3",
      "foo" -> "4",
      "foo" -> "5",
      "foo" -> "6",
      "foo" -> "7",
      "foo" -> "8",
      "foo" -> "9",
      "foo" -> "10",
      "baz" -> "11",
      "foo" -> "12")).unsafeRunSync()

    assertEquals(Set("1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "12"), dao.localTermsFor("foo").unsafeRunSync())
  }

  @Test
  def testSelectMultipleTerms(): Unit = {
    dao.upsert(Seq(
      "foo" -> "bar",
      "baz" -> "egg",
      "foo" -> "more")).unsafeRunSync()

    // results are in the same order as initial
    assertEquals(Map("foo" -> Set("bar", "more"), "baz" -> Set("egg")), dao.localTermsFor(Seq("foo", "baz")).unsafeRunSync())
  }

  @Ignore // H2 database driver does not behave like mysql, even in mysql compatibility mode. See SHRINE-3606.
  @Test
  def testWithBackslash(): Unit = {
    dao.upsert(Seq(
      """\\SHRINE\SHRINE\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I am the same\""" -> """\\PCORI_DIAG\PCORI\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I am the same\""",
      """\\SHRINE\SHRINE\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I am the same\""" -> """\\PCORI_DIAG\PCORI\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I am the same\""",
      """\\SHRINE\SHRINE\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I am the same\""" -> """\\PCORI_DIAG\PCORI\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I'm actually different\""",
      """\\SHRINE\SHRINE\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I am really different\""" -> """\\PCORI_DIAG\PCORI\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I am really different\""",
    )).unsafeRunSync()

    assertEquals(Set(
      """\\PCORI_DIAG\PCORI\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I am the same\""",
      """\\PCORI_DIAG\PCORI\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I'm actually different\""",
    ), dao.localTermsFor("""\\SHRINE\SHRINE\DIAGNOSIS\10\(I00-I99) Dise~3w8h\(I80-I89) Dise~cgvv\(I87) I am the same\""").unsafeRunSync())
  }

  @Test
  def testMetaInformation(): Unit = {
    def assertMetadataMatches(filename: String, fileLastModified: Long, checksum: Long): Unit = {
      assertEquals(dao.fileLastModified.unsafeRunSync(), fileLastModified)
      assertEquals(dao.checksum.unsafeRunSync(), checksum)
      assertEquals(dao.filename.unsafeRunSync(), filename)
    }
    assertMetadataMatches("", -1, -1)
    assertEquals(dao.loadedFromFile.unsafeRunSync(), -1)

    dao.updateMetainformation("a",1L, 2L).unsafeRunSync()
    val loaded = dao.loadedFromFile.unsafeRunSync()

    assertMetadataMatches("a", 1L, 2L)
    assertNotEquals(-1, loaded)

    dao.updateMetainformation("b",3L, 4L).unsafeRunSync()
    assertMetadataMatches("b", 3L, 4L)
    assertNotEquals(loaded, dao.loadedFromFile.unsafeRunSync())
    assertNotEquals(-1, dao.loadedFromFile.unsafeRunSync())
  }

  @Test
  def testTruncate(): Unit = {
    dao.upsert(Seq("foo" -> "bar")).unsafeRunSync()
    assertNotEquals(0, dao.getAll.unsafeRunSync().size)
    dao.truncateMappings().unsafeRunSync()
    assertEquals(0, dao.getAll.unsafeRunSync().size)
  }

  @Before
  def beforeEach(): Unit = {
    dao.createTables.unsafeRunSync()
  }

  @After
  def afterAll(): Unit = {
    dao.dropTables.unsafeRunSync()
  }
}
