package net.shrine.adapter.mappings

import cats.effect.IO
import org.junit.Assert.{assertEquals, assertNotEquals}
import net.shrine.util.ShouldMatchersForJUnit
import org.junit.{After, Before, Test}

class AdapterMappingsTest extends ShouldMatchersForJUnit {
  val dao = AdapterMappingsDb.db
  val csv = "simple-mappings.csv"

  @Test
  def testReloadMappings(): Unit = {
    type MappingState = (String, Long, Long, Int)
    def assertMappingStateAllEqual(currentState: MappingState, expectedState: MappingState): Unit =
      assertEquals(expectedState, currentState)

    def readMappingState(): MappingState =
      (dao.filename.unsafeRunSync(),
       dao.fileLastModified.unsafeRunSync(),
       dao.checksum.unsafeRunSync(),
       dao.getAll.unsafeRunSync().size)

    def assertMappingStateNoneEqual(mappingState: MappingState, noneShouldEqual: MappingState): Unit = {
      assertNotEquals(mappingState._1, noneShouldEqual._1)
      assertNotEquals(mappingState._2, noneShouldEqual._2)
      assertNotEquals(mappingState._3, noneShouldEqual._3)
      assertNotEquals(mappingState._4, noneShouldEqual._4)
    }

    val initial: MappingState = readMappingState()
    assertMappingStateAllEqual(initial, ("", -1, -1, 0))

    // Should insert for the first time
    AdapterMappings.compareAndReloadMappings(csv).unsafeRunSync()
    val updated: MappingState = readMappingState()
    assertMappingStateNoneEqual(initial, updated)

    // When any of the three are the only change, make sure the mappings are reloaded
    val (sameFilename, sameLastModified, sameChecksum, _) = updated
    val differentFilename = "differentFilename"
    val differentLastModified = -2L
    val differentChecksum = -2L

    def assertMappingsUpdated(updatefn: IO[_]): Unit = {
      updatefn.unsafeRunSync()
      val insertionTime = dao.loadedFromFile.unsafeRunSync()
      AdapterMappings.compareAndReloadMappings(csv).unsafeRunSync()
      assertMappingStateAllEqual(readMappingState(), updated)
      assertNotEquals(dao.loadedFromFile.unsafeRunSync(), insertionTime)
    }

    assertMappingsUpdated(dao.updateMetainformation(differentFilename, sameLastModified, sameChecksum))
    assertMappingsUpdated(dao.updateMetainformation(sameFilename, differentLastModified, sameChecksum))
    assertMappingsUpdated(dao.updateMetainformation(sameFilename, sameLastModified, differentChecksum))
    // When nothing changes the mappings should not be reloaded.
    assertThrows[AssertionError]{
      assertMappingsUpdated(dao.updateMetainformation(sameFilename, sameLastModified, sameChecksum))
    }
  }

  @Before
  def beforeEach(): Unit = {
    dao.createTables.unsafeRunSync()
  }

  @After
  def afterEach(): Unit = {
    dao.dropTables.unsafeRunSync()
  }
}
