package net.shrine

/**
  * Created by marc-danie on 3/22/18.
  */

import java.util.Date

import cats.effect.{IO, Timer}
import org.http4s.{AuthedService, BasicCredentials, Header, HttpRoutes, Request, Response, ServerSentEvent, Status}
import fs2.Stream
import net.shrine.i2b2.protocol.pm.User
import net.shrine.log.Log
import net.shrine.protocol.Credential
import org.http4s.ServerSentEvent.EventId
import org.http4s.dsl.impl.{->, /, Auth}
import org.http4s.dsl.io.{GET, NotFound, Unauthorized, Ok, Root}
import org.http4s.dsl.io.{http4sNotFoundSyntax, http4sOkSyntax, http4sUnauthorizedSyntax}
import org.http4s.server.middleware.authentication.BasicAuth
import org.http4s.server.{AuthMiddleware, Router}
import org.http4s.util.CaseInsensitiveString

import scala.language.higherKinds
import scala.collection.mutable.Queue
import scala.concurrent.duration.DurationInt
import scala.concurrent.ExecutionContext.Implicits.global

object Http4sServerSentEventApp extends Auth {

  var dataFlow: Queue[String] = new Queue[String]
  dataFlow += "event: update\n"
  dataFlow += "id: 4373150586012137406\n"
  dataFlow += "data: this is an update\n"
  dataFlow += "event: message\n"
  dataFlow += "data: This is a test 1\n"
  dataFlow += "data: This is a test 2\n"
  dataFlow += "data: This is a test 3\n"
  dataFlow += "data: This is a test 4\n"
  dataFlow += "data: This is a test 5\n"
  dataFlow += "data: This is a test 6\n"
  dataFlow += "data: This is a test 7\n"
  dataFlow += "data: This is a test 8\n"
  dataFlow += "data: This is a test 9\n"
  dataFlow += "data: This is a test 10\n"

  def getData: String = {
    val result: String = dataFlow.dequeue()
    dataFlow.enqueue(result)
    result
  }

  val service: HttpRoutes[IO] = HttpRoutes.of[IO] {

     case req @ GET -> Root / "serverSentEvent" => {
       implicit val timer: Timer[IO] = IO.timer(global)

      //Create a Stream and wake it up every second to send data
      val eventStream: Stream[IO, ServerSentEvent] =
        Stream.awakeEvery[IO](1.second).map(i => ServerSentEvent(data = getData))
      Ok(eventStream)
    }

    case x => NotFound(s"The server sent events app does not respond to $x")
  }

  def authRouter: HttpRoutes[IO] = {
    Router[IO](
      "/auth" -> privateSSERoute,
      "/event" -> toAuthService
    )
  }

  val basicAuth: AuthedService[User, IO] = AuthedService[User, IO] {
    case GET -> Root  as user => createSession(Some(user), Some(true))

    case x => NotFound(s"The QEP service does not respond to $x")
  }

  var queryStatusFlow: Queue[String] = new Queue[String]
  queryStatusFlow += """{"status": "processing"}""".stripMargin
  queryStatusFlow += """{"status": "queued"}"""".stripMargin
  queryStatusFlow += """{"status": "finished"}""".stripMargin

  def getStatus : String  ={
    val result: String = queryStatusFlow.dequeue()
    queryStatusFlow.enqueue(result)
    result
  }

  def processSSE (req: Request[IO], tempToken: String, queryId: String): IO[Response[IO]] = {
    implicit val timer: Timer[IO] = IO.timer(global)

    //If there is a last-event-id header then the user was already validated using the tempToken
    if(validateEventId(req.headers.get(CaseInsensitiveString("LAST-EVENT-ID")))) {

      val lastEventId: String = req.headers.get(CaseInsensitiveString("LAST-EVENT-ID")).get.value

      val eventStream: Stream[IO, ServerSentEvent] =
        Stream.awakeEvery[IO](1.second).map(i => ServerSentEvent(data = getStatus, eventType = Some(s"queryResult-$lastEventId"), id = Some(EventId(lastEventId))))
      Ok(eventStream)
    }
    else if(validateToken(tempToken)){

      //todo this looks a bit spurious. createSession.flatmap() would be better
      val timeToken: String = createSession(isShort = Some(false)).unsafeRunSync().as[String].unsafeRunSync()

      val validSession: String =   """{"session": "valid"}""".stripMargin
      val authData: ServerSentEvent = ServerSentEvent(data = validSession, eventType = Some("auth"), id = Some(EventId(timeToken)))
      val vDataStream:  Stream[IO, ServerSentEvent] = Stream.emit(authData)
      Ok(vDataStream)
    }
    else{
      IO(Response(status = Unauthorized).withEntity("Token is not valid"))
    }
  }

  val toAuthService: HttpRoutes[IO] = HttpRoutes.of[IO] {

    case req @ GET -> Root / "ssEvent" / tempToken / queryId => {

      processSSE(req, tempToken, queryId)
    }

    case x => NotFound(s"The server sent events app does not respond to $x")
  }

  val privateSSERoute = ShrineBasicCredentials.basicAuthMock(basicAuth)

  def createSession(user: Option[User] = None, isShort: Option[Boolean] ): IO[Response[IO]] = {

    val expiration: Long = isShort match {
      case Some(true) => System.currentTimeMillis() + 30000// 30 seconds
      case _ => System.currentTimeMillis() + 1800 * 1000 //30 mins
    }

    IO(Response(status = Status.Ok).withEntity(expiration.toString))
  }

  def validateToken(token: String): Boolean ={

    val date = new Date(token.toLong)
    val now = new Date()

    now.before(date)
  }

  def validateEventId(header: Option[Header]): Boolean ={
    header match {
      case Some(h) => validateToken(h.value)
      case None => false
    }
  }
}

object ShrineBasicCredentials {

  def authenticator(userCred: BasicCredentials):IO[Option[User]] ={

    //This method does not really do any validation
    val username = "some-user"
    val domain = "some-domain"
    val credential = Credential("foo", isToken = true)

    val user = User("some-full-name", username, domain, credential, Map.empty, Map.empty)
    IO(Some(user))
  }

  val basicAuthMock: AuthMiddleware[IO, User] = BasicAuth("foo-realm", authenticator _)
}
