package net.shrine.api

import cats.effect.{ContextShift, IO, Timer}
import javax.servlet.annotation.WebListener
import javax.servlet.{ServletContextEvent, ServletContextListener}
import net.shrine.api.dashboard.DashboardService
import net.shrine.http4s.servlet.ShrineServletContextSyntax
import net.shrine.api.i2b2.I2B2Service
import net.shrine.api.qep.QepService
import net.shrine.api.static.data.StaticDataService
import net.shrine.hub.data.service.HubDataService
import net.shrine.hub.mom.{OkToRetry}
import net.shrine.hub.{HubDispatcher, HubLifecycle, OldResultsPoller}
import net.shrine.log.Log
import net.shrine.messagequeuemiddleware.MessageQueueWebApi
import net.shrine.receiver.Receiver
import net.shrine.source.ConfigSource
import net.shrine.adapter.QueuedQueriesPoller
import net.shrine.adapter.mappings.AdapterMappings
import net.shrine.api.steward.StewardService
import net.shrine.api.steward.email.AuditEmailer
import net.shrine.http4s.catsio.{ExecutionContexts, IOScheduler, RetryIO}
import net.shrine.http4s.client.Http4sHttpClient
import net.shrine.networkhealth.{NetworkHealthScheduler, NetworkHealthService}

import scala.concurrent.duration.DurationInt
import scala.language.postfixOps
import scala.concurrent.ExecutionContext.global
import cats.implicits.catsSyntaxApplicative
import cats.implicits.catsSyntaxApply

/**
  * This class is needed to load the http4s services in a servlet container
  * Created by marc-danie on 7/24/18.
  */
@WebListener
class Bootstrap extends ServletContextListener {
  implicit val cs: ContextShift[IO] = IO.contextShift(global)
  implicit val timer: Timer[IO] = IO.timer(global)

  def afterInitialized(sce: ServletContextEvent): IO[Unit] =
    ifIO(ConfigSource.config.getBoolean("shrine.hub.create")) {
      //todo find some way to determine that the hub is ready to receive. Wait for that.
      RetryIO
        .delayedExponentialBackoff(HubLifecycle.queuesFromDatabase(), 5 seconds, 8, {case OkToRetry(x) => true})
        .redeemWith((t: Throwable) => IO(Log.error("Something went wrong while starting the hub's queues", t)), //todo see SHRINE-3663 . This is a stop-shrine problem
          _ => IO(Log.debug("Hub's queues started successfully")) *> NetworkHealthScheduler.restart())
    }


  override def contextInitialized(sce: ServletContextEvent): Unit = {
    //todo why is servlet context name null ?  SHRINE-3044  Log.debug(s"contextInitialized for ${sce.getServletContext.getServletContextName}")
    Log.debug(s"contextInitialized for ${sce.getServletContext.getContextPath}")

    IOScheduler.startInParallelAndWaitForAll(
      ifIO(ConfigSource.config.getBoolean("shrine.hub.create")) {initHub(sce)}, // receiver starts after hub is set up
      ifIO(ConfigSource.config.getBoolean("shrine.adapter.create")) {initAdapter(sce)},
      Receiver.restart(),
      AuditEmailer.restart()
    ).flatMap(_ => mountServices(sce))
     .attempt
     .flatMap {
       case Left(t)=> IO(abortStartup(t))
       case Right(_) => IO(Log.debug("Shrine startup completed."))
     }
     .unsafeRunSync()

    afterInitialized(sce).unsafeRunAsyncAndForget() // Must be async so tomcat can set up the endpoints
  }

  def abortStartup(cause: Throwable): Unit = {
    Log.error("Shrine startup aborted", cause)
    throw new RuntimeException("Shrine startup aborted", cause) // This stops the servlet
  }

  def initHub(sce: ServletContextEvent): IO[Unit] =
    IOScheduler.startInParallelAndWaitForAll(
      IO(Log.debug(s"This node is a shrine hub.")),
      mountHubServices(sce),
      HubDispatcher.restart(),
      OldResultsPoller.restart()
    )

  def initAdapter(sce: ServletContextEvent): IO[Unit] = for {
    _ <- AdapterMappings.compareAndReloadMappings(ConfigSource.config.getString("shrine.adapter.adapterMappingsFileName"))
    _ <- QueuedQueriesPoller.restart()
  } yield Unit

  def mountServices(sce: ServletContextEvent):IO[Unit] = IO {
    //TODO: find out why AutoSlash middleware does not work - SHRINE-2581
    ShrineServletContextSyntax.mount(
      sce = sce,
      service = new NetworkHealthService().router,
      name = "SHRINE Network Health",
      mapping = "/networkhealth/*"
    )
    ShrineServletContextSyntax.mount(
      sce = sce,
      service = DashboardService().router,
      name = "SHRINE Dashboard Service",
      mapping = "/dashboard/*"
    )

    ShrineServletContextSyntax.mount(
      sce = sce,
      service = StaticDataService().service,
      name = "SHRINE Static Data Service",
      mapping = "/staticData/*"
    )

    ShrineServletContextSyntax.mount(
      sce = sce,
      service = QepService().router,
      name = "SHRINE QEP Service",
      mapping = "/qep/*"
    )

    ShrineServletContextSyntax.mount(
      sce = sce,
      service = I2B2Service().service,
      name = "SHRINE I2B2 Service",
      mapping = "/i2b2/*"
    )

    ShrineServletContextSyntax.mount(
      sce = sce,
      service = StewardService().service,
      name = "SHRINE Steward Service",
      mapping = "/steward/*"
    )
  }

  def mountHubServices(sce: ServletContextEvent):IO[Unit] = IO {
    HubLifecycle.initHubFromConfigIfEmptyIO().unsafeRunSync

    //todo do not bother if using AWS SQS
    ShrineServletContextSyntax.mount(
      sce = sce,
      service = MessageQueueWebApi().service,
      name = "SHRINE MOM",
      mapping = "/mom/*"
    )
    ShrineServletContextSyntax.mount(
      sce = sce,
      service = HubDataService.service,
      name = "SHRINE Hub Data Service",
      mapping = "/hub/*"
    )

    Log.debug(s"services mounted for ${sce.getServletContext.getServletContextName}")
  }

  override def contextDestroyed(sce: ServletContextEvent): Unit =
    IOScheduler.startInParallelAndWaitForAll(
      ifIO(ConfigSource.config.getBoolean("shrine.adapter.create")) {destroyAdapter},
      ifIO(ConfigSource.config.getBoolean("shrine.hub.create")) {destroyHub},
      Receiver.stop(),
      AuditEmailer.stop(),
      Http4sHttpClient.shutdown(),
      NetworkHealthScheduler.stop(),
    ).flatMap(_ => ExecutionContexts.shutdown())
      .unsafeRunSync()

  def destroyAdapter: IO[Unit] = QueuedQueriesPoller.stop()

  def destroyHub: IO[Unit] = OldResultsPoller.stop() *> HubDispatcher.stop()

  def ifIO(condition: => Boolean)(io: IO[Unit]): IO[Unit] = io.whenA(condition)
}
