package net.shrine.wiring

import com.typesafe.config.Config
import javax.sql.DataSource
import net.shrine.adapter.AdapterComponents
import net.shrine.adapter.service.{I2b2AdminResource, I2b2AdminService}
import net.shrine.config.ConfigExtensions
import net.shrine.crypto.{BouncyKeyStoreCollection, KeyStoreDescriptorParser}
import net.shrine.dao.squeryl.{DataSourceSquerylInitializer, SquerylDbAdapterSelecter, SquerylInitializer}
import net.shrine.http4s.client.legacy.{EndpointConfig, Poster}
import net.shrine.log.Loggable
import net.shrine.protocol.{HiveCredentials, ResultOutputType, ResultOutputTypes, XmlNodeName}
import net.shrine.qep.QueryEntryPointComponents
import net.shrine.slick.TestableDataSourceCreator
import net.shrine.source.ConfigSource
import org.squeryl.internals.DatabaseAdapter

/**
 * @author clint
 * @since Jan 14, 2014
 *
 * Application wiring for Shrine.
 */

//todo has to stay until after the Aim 3 APIs are moved in SHRINE-3540
object ShrineOrchestrator extends ShrineJaxrsResources with Loggable {

  override def resources: Iterable[AnyRef] = {
      i2b2AdminResource
  }

  //todo another pass to put things only used in one place into that place's apply(Config)

  //Load config from file on the classpath called "shrine.conf"
  val config: Config = ConfigSource.config
  val shrineConfig: Config = config.getConfig("shrine")


  //TODO: Don't assume keystore lives on the filesystem, could come from classpath, etc
  val keyStoreDescriptor = KeyStoreDescriptorParser(shrineConfig.getConfig("keystore"), shrineConfig.getConfigOrEmpty("hub"), shrineConfig.getConfigOrEmpty("queryEntryPoint"))
  val certCollection: BouncyKeyStoreCollection = BouncyKeyStoreCollection.fromConfig

  //todo used by the adapterService but not by the QEP. maybe each can have its own signerVerifier
  protected val squerylAdapter: DatabaseAdapter = SquerylDbAdapterSelecter.determineAdapter(shrineConfig.getString("shrineDatabaseType"))
  protected val dataSource: DataSource = TestableDataSourceCreator.dataSource(shrineConfig.getConfig("squerylDataSource.database"))

  //todo why does the qep need a HubDao ?
  protected val squerylInitializer: SquerylInitializer = new DataSourceSquerylInitializer(dataSource, squerylAdapter)

  //todo it'd be better for the adapter and qep to each have its own connection to the pm cell.
  private val pmEndpoint: EndpointConfig = shrineConfig.getConfigured("pmEndpoint", EndpointConfig(_))

  protected val breakdownTypes: Set[ResultOutputType] = shrineConfig.getOptionConfigured("breakdownResultOutputTypes", ResultOutputTypes.fromConfig).getOrElse(Set.empty)

  //todo really should be part of the adapter config, but is out in shrine's part of the name space
  val crcHiveCredentials: HiveCredentials = shrineConfig.getConfigured("hiveCredentials", HiveCredentials(_, HiveCredentials.CRC))
  val pmPoster: Poster = Poster(certCollection,pmEndpoint)
  val adapterComponents:Option[AdapterComponents] = shrineConfig.getOptionConfiguredIf("adapter", AdapterComponents(
    _,
    certCollection,
    squerylInitializer,
    breakdownTypes,
    crcHiveCredentials,
    pmPoster
  ))
  if (adapterComponents.isEmpty)
    warn("Adapter Components is improperly configured, please check the adapter section in shrine.conf")

  //todo maybe just break demeter too use this
  val i2b2AdminService: Option[I2b2AdminService] = adapterComponents.map(_.i2b2AdminService)

  protected val xmlNodeName: XmlNodeName = XmlNodeName(shrineConfig.getString("nodeKey"))

  val queryEntryPointComponents:Option[QueryEntryPointComponents] = shrineConfig.getOptionConfiguredIf("queryEntryPoint", QueryEntryPointComponents(_,
    certCollection,
    breakdownTypes,
    pmPoster, //todo could really have its own
    xmlNodeName
  ))

  protected val pmUrlString: String = pmEndpoint.url.toString

  private[shrine] val ontEndpoint: EndpointConfig = shrineConfig.getConfigured("ontEndpoint", EndpointConfig(_))
  protected val ontPoster: Poster = Poster(certCollection,ontEndpoint)

  //todo SHRINE-3540 really the only thing the shrine.war should have in 2.0
  protected val i2b2AdminResource: Option[I2b2AdminResource] = i2b2AdminService.map(I2b2AdminResource(_, breakdownTypes))

}

