package net.shrine.authorization

import com.typesafe.config.Config
import net.shrine.api.steward.{InboundShrineQuery, TopicIdAndName, TopicState}
import net.shrine.api.steward.db.{QueryParameters, StewardDatabase}
import net.shrine.authorization.AuthorizationResult.{Authorized, NotAuthorized}
import net.shrine.i2b2.protocol.pm.User
import net.shrine.log.Loggable
import net.shrine.protocol.query.QueryDefinition
import net.shrine.protocol.{ApprovedTopic, ErrorResponse, ReadApprovedQueryTopicsRequest, ReadApprovedQueryTopicsResponse, RunQueryRequest}
import net.shrine.source.ConfigSource

import scala.language.postfixOps

/**
 * A QueryAuthorizationService that talks to the standard data steward application to learn about topics (intents) and check that a
 * shrine query can be run
 *
 * @author david 
 * @since 4/2/15
 */

//todo move to the DSA service subproject for SHRINE-3053
final case class StewardQueryAuthorizationService() extends QueryAuthorizationService with Loggable {

  //Contact a data steward and either return an Authorized or a NotAuthorized or throw an exception
  override def authorizeRunQueryRequest(runQueryRequest: RunQueryRequest): AuthorizationResult = {
    debug(s"authorizeRunQueryRequest started for ${runQueryRequest.queryDefinition.name}")

    val interpreted = runQueryRequest.topicId.fold(
      authorizeRunQueryRequestNoTopic(runQueryRequest)
    )(
      authorizeRunQueryRequestForTopic(runQueryRequest,_)
    )

    debug(s"authorizeRunQueryRequest completed with $interpreted) for ${runQueryRequest.queryDefinition.name}")

    interpreted
  }

  def authorizeRunQueryRequestNoTopic(runQueryRequest: RunQueryRequest): AuthorizationResult = {
    val userName = runQueryRequest.authn.username
    val queryId = runQueryRequest.queryDefinition.name

    val inboundShrineQuery: InboundShrineQuery = InboundShrineQuery(runQueryRequest.networkQueryId,queryId,runQueryRequest.queryDefinition.toXml.toString())
    val result: (TopicState, Option[TopicIdAndName]) = StewardDatabase.db.logAndCheckQuery(userName, None, inboundShrineQuery)

    if(ConfigSource.config.getString("shrine.steward.createTopicsMode")=="TopicsIgnoredJustLog") Authorized(None)
    else NotAuthorized(runQueryRequest.authn.domain,runQueryRequest.authn.username,s"Topics required but not supplied by ${runQueryRequest.topicId}")
  }

  def authorizeRunQueryRequestForTopic(runQueryRequest: RunQueryRequest,topicIdString:String): AuthorizationResult = {
    val userName: String = runQueryRequest.authn.username
    val queryId = runQueryRequest.queryDefinition.name

    val inboundShrineQuery: InboundShrineQuery = InboundShrineQuery(runQueryRequest.networkQueryId,queryId,runQueryRequest.queryDefinition.toXml.toString())
    val result: (TopicState, Option[TopicIdAndName]) = StewardDatabase.db.logAndCheckQuery(userName, Some(topicIdString.toInt), inboundShrineQuery)

    result._2.fold(throw new AuthorizationException(runQueryRequest.authn.domain,runQueryRequest.authn.username,s"Error finding topic with id $topicIdString")){topicIdAndName =>

      if(result._1 == TopicState.approved) Authorized(Some((topicIdAndName.id, topicIdAndName.name)))
      else NotAuthorized(runQueryRequest.authn.domain,runQueryRequest.authn.username,s"Topic ${topicIdAndName.id} ${topicIdAndName.name} is ${result._1}, but must be ${TopicState.approved} to use")
    }
  }

  override def authorizeRunQuery(user: User, networkQueryId: Long, queryDefinition: QueryDefinition, topicId: Option[String], topicName: Option[String]): AuthorizationResult = {
    debug(s"authorizeRunQueryRequest started for ${queryDefinition.name}")

    val interpreted = topicId.fold(
      authorizeRunQueryNoTopic(user, queryDefinition, networkQueryId)
    )(
      authorizeRunQueryForTopic(user, queryDefinition, networkQueryId, _)
    )

    debug(s"authorizeRunQuery completed with $interpreted) for ${queryDefinition.name}")

    interpreted
  }

  def authorizeRunQueryNoTopic(user: User, queryDefinition: QueryDefinition, networkQueryId: Long): AuthorizationResult = {
    val userName = user.username
    val queryId = queryDefinition.name

    val inboundShrineQuery: InboundShrineQuery = InboundShrineQuery(networkQueryId,queryId,queryDefinition.toXml.toString())
    val result: (TopicState, Option[TopicIdAndName]) = StewardDatabase.db.logAndCheckQuery(userName, None, inboundShrineQuery)

    if(ConfigSource.config.getString("shrine.steward.createTopicsMode")=="TopicsIgnoredJustLog") Authorized(None)
    else NotAuthorized(user.domain,user.username,s"Topics required but not supplied}")
  }

  def authorizeRunQueryForTopic(user: User, queryDefinition: QueryDefinition, networkQueryId: Long, topicId: String): AuthorizationResult = {
    val userName: String = user.username
    val queryId = queryDefinition.name

    val inboundShrineQuery: InboundShrineQuery = InboundShrineQuery(networkQueryId,queryId, queryDefinition.toXml.toString())
    val result: (TopicState, Option[TopicIdAndName]) = StewardDatabase.db.logAndCheckQuery(userName, Some(topicId.toInt), inboundShrineQuery)

    result._2.map(topicIdAndName => {
      if(result._1 == TopicState.approved) Authorized(Some((topicIdAndName.id, topicIdAndName.name)))
      else NotAuthorized(user.domain,user.username,s"Topic ${topicIdAndName.id} ${topicIdAndName.name} is ${result._1}, but must be ${TopicState.approved} to use")
    }).getOrElse(NotAuthorized(user.domain,user.username,s"Error finding topic with id $topicId"))
  }

  //Either read the approved topics from a data steward or have an error response.
  override def readApprovedTopics(readTopicsRequest: ReadApprovedQueryTopicsRequest): Either[ErrorResponse, ReadApprovedQueryTopicsResponse] = {
    val userName = readTopicsRequest.authn.username

    val queryParameters = QueryParameters(researcherIdOption = Some(userName), stateOption = Some(TopicState.approved))
    val topicsForResearcher = StewardDatabase.db.selectTopicsForResearcher(queryParameters)

    val topics: Seq[ApprovedTopic] = topicsForResearcher.topics.map(topic => ApprovedTopic(topic.id, topic.name))
    Right(ReadApprovedQueryTopicsResponse(topics))
  }
}

object StewardQueryAuthorizationService {

  def apply(config:Config):StewardQueryAuthorizationService = {
    StewardQueryAuthorizationService()
  }
}
