package net.shrine.http4s.auth

import cats.effect.IO
import com.typesafe.config.Config
import net.shrine.i2b2.protocol.pm.{BadUsernameOrPasswordException, GetUserConfigurationRequest, PmUserWithoutProjectException, User}
import net.shrine.log.Loggable
import net.shrine.protocol.{AuthenticationInfo, Credential}
import net.shrine.source.ConfigSource
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.http4s.client.legacy.{EndpointConfig, Http4sLegacyClient, Poster}
import org.http4s.{AuthedRoutes, BasicCredentials, HttpRoutes}
import org.http4s.server.AuthMiddleware
import org.http4s.server.middleware.authentication.BasicAuth

import scala.collection.mutable
import scala.concurrent.blocking

object BasicAuthentication {

  val unauthorizedMsg = "Username or password does not match credentials."

  val basicAuthMiddleware: AuthMiddleware[IO, User] = BasicAuth(UserAuthentication.domain, authenticator) apply _ andThen UserAuthentication.addUnauthorizedMsg(unauthorizedMsg)

  def authenticator(userCred: BasicCredentials):IO[Option[User]] ={

    authenticator(userCred, isToken = false)
  }

  def authenticator(userCred: BasicCredentials, isToken: Boolean):IO[Option[User]] ={

    val config: Config = ConfigSource.config
    val userSource:UserSource = config.getString("shrine.authenticate.usersource.type") match {
      case PmUserSource.configName => PmUserSource(config, isToken)
      case ConfigUserSource.configName => ConfigUserSource(config)
      case x => throw new IllegalStateException(s"The config value for 'shrine.authenticate.usersource.type' must be either ${PmUserSource.configName} for actual use or ${ConfigUserSource.configName} for testing. '$x' cannot be used.")
    }

    userSource.authenticateUser(Some(userCred), isToken)
  }

  case class PmUserSource(config:Config, isToken: Boolean) extends UserSource with Loggable {
    def authenticateUser(userPassOption: Option[BasicCredentials], isToken: Boolean): IO[Option[User]] = {
      val httpClientContextShift = IO.contextShift(ExecutionContexts.httpClientExecutionContext)
      httpClientContextShift.evalOn(ExecutionContexts.httpClientExecutionContext)(IO {

        val noUser: Option[User] = None
        userPassOption.fold(noUser)(userPass => {
          val requestString = GetUserConfigurationRequest(AuthenticationInfo(UserAuthentication.domain,
            username = userPass.username,
            credential = Credential(userPass.password, isToken = isToken))).toI2b2String
          val httpResponse = blocking {
            pmPoster.post(requestString)
          }

          if (httpResponse.statusCode >= 400) {
            val message = s"HttpResponse status is ${httpResponse.statusCode} from PM via ${pmPoster.url} for ${userPass.username}. Response is ${httpResponse}"
            warn(message)
            throw new IllegalStateException(message)
          }

          try {
            // user object contains the token in the password field
            val user = User.fromI2b2(httpResponse.body).get
            Some(user)
          } catch {
            case x: BadUsernameOrPasswordException => {
              error(s"PM at ${pmPoster.url} found no (user,password) combination for ${userPass.username}.", x)
              None
            }
            case x: PmUserWithoutProjectException => {
              warn(s"PM at ${pmPoster.url} found no projects for ${userPass.username}.", x)
              None
            }
          }
        }
        )
      })
    }

    lazy val pmPoster: Poster = {
      val pmEndpoint: EndpointConfig = EndpointConfig(config.getConfig("shrine.pmEndpoint"))

      //todo for SHRINE-3661 - replace this - maybe need to take the bouncy castle certs and enforce them
      val httpClient = Http4sLegacyClient(pmEndpoint.timeout)
      Poster(pmEndpoint.url.toString, httpClient)
    }
  }

  object PmUserSource {
    val configName: String = getClass.getSimpleName.dropRight(1)
  }

  case class ConfigUserSource(config:Config) extends UserSource {

    val prefix = "shrine.authenticate.usersource"

    val subconfig = config.getConfig(prefix)

    import scala.collection.JavaConverters._
    val roles: mutable.Set[String] = subconfig.entrySet().asScala.
      filterNot(x => x.getKey == "type").
      filterNot(x => x.getKey == "domain").
      map(x => x.getKey.take(x.getKey.indexOf(".")))

    val rolesToUserNames: Map[String, String] = roles.map(x => (
      x,
      subconfig.getConfig(x).getString("username")
    )).toMap

    val userNamesToPasswords: Map[String, String] = rolesToUserNames.map(x => (
      x._2,
      subconfig.getConfig(x._1).getString("password")
    ))

    lazy val qepUserName = rolesToUserNames("qep")
    lazy val researcherUserName = rolesToUserNames("researcher")
    lazy val stewardUserName = rolesToUserNames("steward")

    def authenticateUser(userPass: Option[BasicCredentials], isToken: Boolean): IO[Option[User]] = IO {

      val noUser:Option[User] = None
      userPass.fold(noUser)(up =>{

        if (userNamesToPasswords(up.username) == up.password) {
          val params:Map[String,String] = if (up.username == "qep") Map(UserAuthentication.qepRole -> "true")
          else if (up.username == stewardUserName) Map(UserAuthentication.stewardRole -> "true")
          else Map.empty
          val user: User = User(fullName = up.username,
            username = up.username,
            domain = "domain",
            credential = Credential(up.password,isToken = isToken),
            params = params,
            rolesByProject = Map())
          Some(user)
        }
        else None
      })
    }
  }

  object ConfigUserSource {
    val configName = getClass.getSimpleName.dropRight(1)
  }

  trait UserSource {

    def authenticateUser(userPass: Option[BasicCredentials], isToken: Boolean): IO[Option[User]]

  }

}

object BasicAuthMiddleware {

  def apply(service: AuthedRoutes[User, IO]): HttpRoutes[IO] = {

    BasicAuthentication.basicAuthMiddleware(service)
  }
}

