package net.shrine.http4s.catsio

import java.util.concurrent.{ExecutorService, Executors, ForkJoinPool, ForkJoinWorkerThread, SynchronousQueue, ThreadFactory, ThreadPoolExecutor, TimeUnit}
import java.util.concurrent.atomic.AtomicInteger

import cats.effect.IO
import net.shrine.log.Log
import net.shrine.source.ConfigSource

import scala.concurrent.{BlockContext, CanAwait, ExecutionContext, ExecutionContextExecutorService}



/**
  * Execution contexts for shrine
  *
  * @author dwalend
  * @since 1.26
  */
object ExecutionContexts {

  /**
    * For the 30-seconds scale calls into the CRC. The CRC gets its own thread pool due to its 15-per-core limit
    */
  lazy val crcExecutionContext: ExecutionContextExecutorService = {
    val crcClientThreadPoolSize = ConfigSource.config.getInt("shrine.adapter.crcClientThreadPoolSize")
    forkJoinThreadPoolExecutionContext(
      prefix = "crcCall-",
      maxSize = crcClientThreadPoolSize,
      daemonic = true //OK to just drop communication with the CRC at shutdown
    )
  }

  /**
    * For the http4s server request-in/response-out work
    */
  lazy val http4sExecutionContext: ExecutionContextExecutorService = forkJoinThreadPoolExecutionContext(
    prefix = "http4sServer-",
    maxSize = Runtime.getRuntime.availableProcessors * 8
  )

  /**
    * For chattering with the database at 10-500 millisecond-timescales
    */
  lazy val databaseExecutionContext: ExecutionContextExecutorService = forkJoinThreadPoolExecutionContext(
    prefix = "database-",
    maxSize = 20 //Hard-coded in slick.util.AsyncExecutor
  )

  /**
    * For http client-side calls, at the 500 millisecond to 5 second time scales
    */
  lazy val httpClientExecutionContext: ExecutionContextExecutorService = cachedThreadPoolExecutionContext(
    prefix = "httpClient-"
  )

  /**
    * For long-polling calls, blocking queues, and similar work on the scale of 5 seconds or more
    */
  lazy val longActionExecutionContext: ExecutionContextExecutorService = cachedThreadPoolExecutionContext(
      prefix = "longAction-",
      daemonic = true //Fine to slam these down when tomcat stops
  )

  def shutdown(): IO[Unit] = IO {
    crcExecutionContext.shutdown() //OK if these long calls only make it part-way.
    longActionExecutionContext.shutdown() //Safe to stop these long actions, too
    http4sExecutionContext.shutdown() //stop taking new work.
    httpClientExecutionContext.shutdown() //shutdown http clients almost last. Give them a chance to finish writes.
    databaseExecutionContext.shutdown() //shutdown databases last. Give them a chance to finish writes.
  }

  private def fixedThreadPoolExecutionContext(prefix:String,size:Int,daemonic:Boolean = false): ExecutionContextExecutorService = {
    val threadFactory = ShrineThreadFactory(prefix,daemonic)
    val executorService: ExecutorService = Executors.newFixedThreadPool(size,threadFactory)

    val executionContext: ExecutionContextExecutorService = ExecutionContext.fromExecutorService(executorService)

    executionContext
  }

  private def forkJoinThreadPoolExecutionContext(prefix:String,maxSize:Int,daemonic:Boolean = false): ExecutionContextExecutorService = {
    val threadFactory = ShrineThreadFactory(prefix,daemonic)
    val executorService: ExecutorService = forkJoinExecutorService(maxSize,threadFactory)

    val executionContext: ExecutionContextExecutorService = ExecutionContext.fromExecutorService(executorService)

    executionContext
  }

  private def cachedThreadPoolExecutionContext(prefix:String,daemonic:Boolean = false): ExecutionContextExecutorService = {
    val threadFactory = ShrineThreadFactory(prefix,daemonic)
    val executorService: ExecutorService = Executors.newCachedThreadPool(threadFactory)

    val executionContext: ExecutionContextExecutorService = ExecutionContext.fromExecutorService(executorService)

    executionContext
  }

  private def boundedExecutorService(maxPoolSize:Int,threadFactory: ThreadFactory): ExecutorService = {
    new ThreadPoolExecutor(
      0,
      maxPoolSize,
      60L,
      TimeUnit.SECONDS,
      new SynchronousQueue[Runnable],
      threadFactory
    )
  }

  private def forkJoinExecutorService(maxSize:Int,threadFactory:ShrineThreadFactory):ExecutorService = {
    new ForkJoinPool(maxSize, threadFactory, threadFactory.uncaughtExceptionHandler, true)
  }

}

case class ShrineThreadFactory(
                                threadNamePrefix:String,
                                daemonic:Boolean
                              ) extends ThreadFactory with ForkJoinPool.ForkJoinWorkerThreadFactory {

  private val counter = new AtomicInteger(0)

  val uncaughtExceptionHandler: Thread.UncaughtExceptionHandler = new Thread.UncaughtExceptionHandler {
    def uncaughtException(thread: Thread, cause: Throwable): Unit = Log.error(s"${cause.getClass.getSimpleName} never caught in ${thread.getName}",cause)
  }

  def wire[T <: Thread](thread: T): T = {
    thread.setDaemon(daemonic)
    thread.setUncaughtExceptionHandler(uncaughtExceptionHandler)
    thread.setName(s"$threadNamePrefix${counter.getAndIncrement()}")
    thread
  }

  def newThread(runnable: Runnable): Thread = wire(new Thread(runnable))

  //mostly cut-paste from scala.concurrent.impl.ExecutionContextImpl.DefaultThreadFactory
  def newThread(fjp: ForkJoinPool): ForkJoinWorkerThread = wire(
    new ForkJoinWorkerThread(fjp) with BlockContext {
      override def blockOn[T](thunk: =>T)(implicit permission: CanAwait): T = {
        var result: T = null.asInstanceOf[T]
        ForkJoinPool.managedBlock(new ForkJoinPool.ManagedBlocker {
          @volatile var isDone: Boolean = false
          override def block(): Boolean = {
            result = try thunk finally { isDone = true }
            true
          }
          override def isReleasable: Boolean = isDone
        })
        result
      }
    }
  )
}
