package net.shrine.http4s.catsio

import java.util.concurrent.atomic.AtomicReference

import cats.data.NonEmptyList
import cats.implicits.catsSyntaxApply
import cats.implicits.catsSyntaxParallelSequence
import cats.effect.{CancelToken, ContextShift, IO, Timer}
import net.shrine.log.Log

import scala.concurrent.duration.{Duration, FiniteDuration}
import scala.util.control.NonFatal

/**
 * Execute IO's periodically. Interface loosely based off Monix's Scheduler.
 */
//todo rename - I don't think this is a Scheduler. this is a scheduled Task like the method comments say
trait IOScheduler {

  /**
   * Stop all executions of the currently scheduled task immediately.
   */
  def stop(): IO[Unit]

  /**
   * Stop all executions of the currently scheduled task immediately, and then rerun
   * the task with the specified parameters.
   */
  def restart(): IO[Unit]
}

object IOScheduler {

  /**
   * Execute all IO's in tasks concurrently using the given ContextShift, and wait for all results, returning Unit.
   */
  def startInParallelAndWaitForAll(tasks: IO[_]*)(implicit cs: ContextShift[IO]): IO[Unit] = {
    NonEmptyList.fromList(tasks.toList).fold(IO.unit){_.parSequence *> IO.unit}
  }

  /**
   * Execute task once after the initial delay, and subsequently after interval,
   * until task returns false or the scheduler is stopped by the stop() method.
   */
  def scheduleWithFixedDelay(initialDelay: FiniteDuration,
                             interval: FiniteDuration,
                             task: () => IO[Boolean],
                             name: String)
                            (implicit timer: Timer[IO]): IOScheduler =
    new IOSchedulerImpl(initialDelay, interval, task, name)

  /**
   * Execute task once only after the initial delay.
   */
  def scheduleOnce(initialDelay: FiniteDuration,
                   task: () => IO[_],
                   name: String)
                  (implicit timer: Timer[IO]): IOScheduler =
    scheduleWithFixedDelay(initialDelay, Duration.Zero, () => task().flatMap(_ => IO(false)), name)

  /**
   *  Execute task indefinitely, ignoring output of task, until stop or restart methods are called.
   */
  def scheduleIndefinitely(initialDelay: FiniteDuration,
                           interval: FiniteDuration,
                           task: () => IO[_],
                           name: String)
                          (implicit timer: Timer[IO]): IOScheduler =
    scheduleWithFixedDelay(initialDelay, interval, () => task().flatMap(_ => IO(true)), name)


  /**
   * Scheduler to perform a task repeatedly until the stop condition(s).
   * Will stop when either the stop() method is called, or the task returns false.
   *
   * @param name The name of the scheduled service.
   * @param initialDelay The time to wait before performing the first execution of task.
   * @param interval The time to wait between executions of task
   * @param task The task to be performed. If task should be performed again, return true, else return false.
   *
   * @param timer The timer to use to schedule the task. Cats effect's beefed up Execution Context
   */
  private class IOSchedulerImpl(initialDelay: FiniteDuration,
                                interval: FiniteDuration,
                                task: () => IO[Boolean],
                                name: String)
                               (implicit timer: Timer[IO])
      extends IOScheduler {

    private val atomicCancelToken = new AtomicReference[Option[CancelToken[IO]]](None)

    override def restart(): IO[Unit] = for {
        newToken <- pollUntilStop()
        _ <- IO(Log.debug(s"Started $name $newToken"))
        oldToken <- IO(atomicCancelToken.getAndSet(Option(newToken)))
      } yield oldToken.foreach(cancelAndLog)

    override def stop(): IO[Unit] =
      IO(atomicCancelToken.getAndSet(None).foreach(cancelAndLog))

    private def cancelAndLog(ct: CancelToken[IO]): Unit = {
      Log.debug(s"Stopping $name $ct")
      ct.unsafeRunSync()
      Log.debug(s"Stopped $name $ct")
    }

    private def pollUntilStop(): IO[CancelToken[IO]] =
      sleepAndCheck(initialDelay)
        .handleErrorWith {
          case NonFatal(x) =>
            Log.error(s"Caught exception in $name",x)
            IO.raiseError(x)
        }
        .runCancelable {
          case Right(_) => IO(Log.debug(s"End of $name.pollUntilStop()"))
          case Left(t) => IO(Log.error(s"Exception while running $name.pollUntilStop ", t))
        }
        .toIO

    private def sleepAndCheck(sleepTime:FiniteDuration):IO[Unit] =
      IO.sleep(sleepTime)
        .flatMap(_ => task())
        .flatMap{checkAgain:Boolean =>
          if(checkAgain) sleepAndCheck(interval)
          else IO.unit
        }
  }
}
