package net.shrine.http4s.catsio

import cats.effect.{IO, Timer}
import net.shrine.log.Log

import scala.concurrent.duration.{DurationInt, FiniteDuration}
import scala.language.postfixOps

/**
  * Enumerate the _actually_ useful ways to retry IOs. 
  */
object RetryIO {

  def delayedExponentialBackoff[A](task: IO[A],
                                   firstDelay:FiniteDuration,
                                   maxRetries: Int,
                                   retryIfTaskThrows: Throwable => Boolean)
                                  (implicit timer: Timer[IO]): IO[A] =
    retry(task)(firstDelay)(Some(maxRetries))(firstDelay * Math.pow(2, _).toInt)(retryIfTaskThrows)(timer)

  def keepTrying[A](task: IO[A],
                    delay:FiniteDuration,
                    retryIfTaskThrows: Throwable => Boolean)
                   (implicit timer: Timer[IO]): IO[A] =
    retry(task = task)(delay = 0.seconds)(maxRetries = None)(backoff = _ => delay)(retryIfTaskThrows = retryIfTaskThrows)(timer = timer)

  def keepTryingBounded[A](task: IO[A],
                           delay:FiniteDuration,
                           maxRetries: Int,
                           retryIfTaskThrows: Throwable => Boolean)
                          (implicit timer: Timer[IO]): IO[A] =
    retry(task)()(Some(maxRetries))(_ => delay)(retryIfTaskThrows)(timer)

  private[catsio] def retry[A](
                                task: IO[A])
              (delay: FiniteDuration = 0 seconds)
              (maxRetries: Option[Int] = None)
              (backoff: Int => FiniteDuration = _ => delay)
              (retryIfTaskThrows: Throwable => Boolean = _ => false)
              (timer: Timer[IO]): IO[A] = {
    var tried = 0
    def increment(): Unit = tried += 1

    def build(nextDelay: FiniteDuration): IO[A] = {
      IO.cancelBoundary
        .flatMap(_ => IO.sleep(nextDelay)(timer))
        .flatMap{_ =>
          if(maxRetries.fold(true)(tried < _)) { //for all but a last attempt
            task.handleErrorWith{
              case x if retryIfTaskThrows(x) => //try again despite survivable exceptions
                Log.info(s"Caught ${x.getClass.getSimpleName}. Will make ${tried+1} of ${maxRetries.fold("unlimited")(_.toString)} attempts of $task after $nextDelay",x)
                continue()
              case z => //raise this unanticipated exception and give up
                Log.info(s"Caught ${z.getClass.getSimpleName}. Will not attempt $task again.",z)
                IO.raiseError(z)
            }
          } else { //On a last attempt 
            Log.info(s"Last attempt of $task")
            task.handleErrorWith {
              x => // raise any exception and give up
                Log.info(s"Caught during last attempt of $task", x)
                IO.raiseError(x)
            }
          } 
        }
    }

    def continue(): IO[A] = {increment(); build(backoff(tried))}

    build(delay)
  }
}
