package net.shrine.http4s.client

import cats.effect.{ContextShift, IO}
import io.netty.handler.ssl.util.InsecureTrustManagerFactory
import io.netty.handler.ssl.{SslContext, SslContextBuilder}
import net.shrine.http4s.catsio.{ExecutionContexts, ShrineThreadFactory}
import org.asynchttpclient.proxy.ProxyServer
import org.asynchttpclient.{DefaultAsyncHttpClientConfig, Dsl, Realm}
import org.http4s.client.Client
import org.http4s.client.asynchttpclient.AsyncHttpClient
import org.http4s.{EntityDecoder, Request, Response, Status}

import scala.concurrent.blocking

/**
  * @author david
  * @since 1.25.3.1
  */
object Http4sHttpClient {

  private val httpClientBuilder: DefaultAsyncHttpClientConfig.Builder = createBuilder

  // The client and the IO representing all shutdown executions
  private val clientAndShutdown: (Client[IO], IO[Unit]) = {
    implicit val concurrentEffect: ContextShift[IO] = IO.contextShift(ExecutionContexts.httpClientExecutionContext)
    AsyncHttpClient.allocate[IO](httpClientBuilder.build).unsafeRunSync()
  }

  private val httpClient: Client[IO] = clientAndShutdown._1

  def shutdown(): IO[Unit] = clientAndShutdown._2

  def webFetchAndDecodeIO[A](request:Request[IO])(toA: (Status,String) => IO[A]): IO[A] = {
    blocking{
      httpClient.fetch(request) { response: Response[IO] =>
        EntityDecoder.decodeString(response).flatMap(toA(response.status, _))
      }
    }
  }

  def webFetchResponse(request: Request[IO], postProcess: Option[Response[IO] => IO[Response[IO]]] = None): IO[Response[IO]] = {
    httpClient.fetch(request) { response: Response[IO] =>
      EntityDecoder.decodeString(response).map {
        Response[IO](status = response.status).withEntity(_)
      }
    }
  }

  private def createBuilder: DefaultAsyncHttpClientConfig.Builder = {

    //Build up proxy parameters from standard locations
    val proxyHost = Option(System.getProperty("http.proxyHost"))
    val proxyPort = Option(System.getProperty("http.proxyPort")).map(_.toInt)
    val proxyUser = Option(System.getProperty("http.proxyUser"))
    val proxyPassword = Option(System.getProperty("http.proxyPassword"))

    val proxyParameters = (proxyHost,proxyPort,proxyUser,proxyPassword) match {
      case (None, None, None, None) => None
      case (Some(host), Some(port), None, None) => Some((host, port, None))
      case (Some(host), Some(port), Some(user), Some(password)) => Some((host, port, Some((user, password))))
      case _ => throw new IllegalArgumentException(
        s"""If either of http.proxyHost and http.proxyPort are set then both must be set (host: $proxyHost,port: $proxyPort).
           |If either of http.proxyUser and http.proxyPassword are set then both must be set (user: $proxyUser,password provided : ${proxyPassword.isDefined}).""".stripMargin)
    }

    //todo use a more particular trust manager - SHRINE-3661
    val sslContext: SslContext = SslContextBuilder.forClient().trustManager(InsecureTrustManagerFactory.INSTANCE).build()
    val sslConfigBuilder: DefaultAsyncHttpClientConfig.Builder = Dsl.config().setSslContext(sslContext)

    val proxyConfigBuilder: DefaultAsyncHttpClientConfig.Builder = proxyParameters.fold {
      sslConfigBuilder
    } { parameters =>
      val host = parameters._1
      val port = parameters._2
      val proxyServer: ProxyServer = parameters._3.fold {
        new ProxyServer.Builder(host, port).build()
      } { credentials =>
        val realm = new Realm.Builder(credentials._1, credentials._2)
        new ProxyServer.Builder(host, port).setRealm(realm).build()
      }
      sslConfigBuilder.setProxyServer(proxyServer)
    }
    proxyConfigBuilder.setThreadFactory(ShrineThreadFactory(
      threadNamePrefix = "http4s-async-http-client-worker-",
      daemonic = true
    ))
  }
}
