package net.shrine.http4s.client.legacy

import cats.effect.{ContextShift, IO, Timer}
import fs2.Chunk
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.http4s.client.Http4sHttpClient
import net.shrine.log.Loggable
import net.shrine.util.{StringEnrichments, XmlUtil}
import org.http4s.EntityEncoder.simple
import org.http4s.headers.`Content-Type`
import org.http4s.{Charset, DefaultCharset, EntityEncoder, MediaType, Method, Request, Status, Uri}

import scala.concurrent.duration.FiniteDuration

/**
 * An Http4s-based replacement for the old JerseyHttpClient. Do not use this in new code.
 */
final case class Http4sLegacyClient(timeout: FiniteDuration)
  extends HttpClient with Loggable {

  override def post(input: String, url: String): HttpResponse = {
    val contextShift: ContextShift[IO] = IO.contextShift(ExecutionContexts.crcExecutionContext)
    val timer: Timer[IO] = IO.timer(ExecutionContexts.crcExecutionContext)
    postIO(input,url).timeout(timeout)(timer,contextShift).unsafeRunSync()
  }

  def postIO(body: String, url: String): IO[HttpResponse] = {
    import scala.language.higherKinds

    def xmlStringEncoder[F[_]](implicit charset: Charset = DefaultCharset): EntityEncoder[F, String] = {
      val hdr = `Content-Type`(MediaType.text.xml).withCharset(charset)
      simple(hdr)(s => Chunk.bytes(s.getBytes(charset.nioCharset)))
    }

    val request: Request[IO] = Request(
      method = Method.POST, //always POST with i2b2
      uri = Uri.unsafeFromString(url)
    ).withEntity(body)(xmlStringEncoder)

    def prettyPrintIfXml(s: String): String = {
      import StringEnrichments._

      s.tryToXml.map(_.head).map(XmlUtil.prettyPrint).getOrElse(s)
    }

    def toHttpResponse(status:Status,contents:String):IO[HttpResponse] = IO (HttpResponse(status.code,contents))

    def safeLogResponse(httpResponse: HttpResponse):IO[Unit] = IO{
      if(httpResponse.statusCode < 400) {
        debug(s"Got response from '$url' of '${prettyPrintIfXml(httpResponse.body)}'")
      } else {
        debug(s"Got error code ${httpResponse.statusCode} from '$url' of '${httpResponse.body}'")
      }
    }

    for {
      _ <- IO(debug(s"Invoking '$url' with ${prettyPrintIfXml(body).length} characters of '${prettyPrintIfXml(body)}'"))
      httpResponse <- Http4sHttpClient.webFetchAndDecodeIO(request)(toHttpResponse)
      _ <- safeLogResponse(httpResponse)
    } yield httpResponse
  }
}