package net.shrine.http4s.auth


import cats.data.OptionT
import cats.effect.{ConcurrentEffect, ContextShift, IO}
import cats.implicits.{catsSyntaxEither => _}
import net.shrine.i2b2.protocol.pm.User
import org.http4s.{AuthedRoutes, BasicCredentials, Headers, Method, Request, Response, Uri}
import org.http4s.dsl.impl.{->, /, Auth}
import org.http4s.dsl.io.{GET, Ok, Root, http4sOkSyntax, Unauthorized}
import org.http4s.headers.Authorization
import org.junit.runner.RunWith
import org.scalatest.FlatSpec
import org.scalatestplus.junit.JUnitRunner

import scala.concurrent.ExecutionContext.Implicits.global
/**
  * Test web api of BasicAuthentication
  */
@RunWith(classOf[JUnitRunner])
class BasicAuthenticationTest extends FlatSpec with Auth  {

  //Auth is needed for 'as' to work in 'Root / "private" as user'
  private val authService  =  AuthedRoutes.of[User, IO] {
    case GET -> Root / "private" as user =>
      Ok(s"This page is protected using HTTP authentication; logged in as ${user.username}")
  }
  val wrappedAuthService = BasicAuthMiddleware(authService)

  def extractResponse(request: Request[IO]): Response[IO] = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val responseOptionIo: OptionT[IO, Response[IO]] = wrappedAuthService.run(request)

    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  "An authorized user" should "be allowed access to the service" in {
    val username = "admin"
    val password = "pwd"
    val pingRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/private"),
      headers = Headers.of(Authorization(BasicCredentials(username, password)))
    )

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("This page is protected using HTTP authentication; logged in as admin")(entityText)
  }

  "A user with no username or password" should "not be allowed access to the service" in {
    val pingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString("/private"))

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Unauthorized)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult(BasicAuthentication.unauthorizedMsg)(entityText)
  }

  "A user with a username and no password" should "not be allowed access to the service" in {
    val username = "admin"
    val password = ""
    val pingRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/private"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Unauthorized)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult(BasicAuthentication.unauthorizedMsg)(entityText)
  }
}
