package net.shrine.http4s.auth

import java.util.Base64

import cats.data.OptionT
import cats.effect.{ConcurrentEffect, ContextShift, IO}
import cats.implicits.{catsSyntaxEither => _}
import net.shrine.i2b2.protocol.pm.User
import org.http4s.{AuthScheme, AuthedRoutes, Credentials, Headers, Method, Request, Response, Uri}
import org.http4s.dsl.impl.{->, /, Auth}
import org.http4s.dsl.io.{GET, Ok, Root, Unauthorized, http4sOkSyntax}
import org.http4s.headers.Authorization
import org.junit.runner.RunWith
import org.scalatest.FlatSpec
import org.scalatestplus.junit.JUnitRunner

import scala.concurrent.ExecutionContext.Implicits.global
/**
  * Test web api of BearerAuthentication
  */
@RunWith(classOf[JUnitRunner])
class BearerAuthenticationTest extends FlatSpec with Auth  {

  //Auth is needed for 'as' to work in 'Root / "private" as user'
  private val authService  =  AuthedRoutes.of[User, IO] {
    case GET -> Root / "private" as user =>
      Ok(s"This page is protected using Bearer authentication; logged in as ${user.username}")
  }
  val wrappedAuthService = BearerAuthentication.bearerAuthMiddleware(authService)

  def extractResponse(request: Request[IO]): Response[IO] = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val responseOptionIo: OptionT[IO, Response[IO]] = wrappedAuthService.run(request)

    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  "An authorized user" should "be allowed access to the service" in {
    val bearerToken: String =
      """{"username": "adminSession", "sessionId": "1234567"}
      """.stripMargin

    val encodedBearerToken: String = Base64.getEncoder.encodeToString(bearerToken.getBytes)
    val request: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/private"),
      headers = Headers.of(Authorization(Credentials.Token(AuthScheme.Bearer, encodedBearerToken)))
    )

    val response: Response[IO]= extractResponse(request)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("This page is protected using Bearer authentication; logged in as adminSession")(entityText)
  }

  "A user with no username or session Id" should "not be allowed access to the service" in {
    val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString("/private"))

    val response: Response[IO]= extractResponse(request)

    assertResult(Unauthorized)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult(BearerAuthentication.unauthorizedMsg)(entityText)
  }

  "A user with a username and no session Id" should "not be allowed access to the service" in {
    val bearerToken: String =
      """{"username": "adminSession", "sessionId": ""}
      """.stripMargin
    val encodedBearerToken: String = Base64.getEncoder.encodeToString(bearerToken.getBytes)

    val pingRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/private"),
      headers = Headers.of(Authorization(Credentials.Token(AuthScheme.Bearer, encodedBearerToken))))

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Unauthorized)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult(BearerAuthentication.unauthorizedMsg)(entityText)
  }

  "A user with a username and bad session Id" should "not be allowed access to the service" in {
    val bearerToken: String =
      """{"username": "adminSession", "sessionId": "S1234567"}
      """.stripMargin
    val encodedBearerToken: String = Base64.getEncoder.encodeToString(bearerToken.getBytes)

    val pingRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/private"),
      headers = Headers.of(Authorization(Credentials.Token(AuthScheme.Bearer, encodedBearerToken))))

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Unauthorized)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult(BearerAuthentication.unauthorizedMsg)(entityText)
  }
}

