package net.shrine.http4s.catsio

import cats.effect.{IO, Timer}
import net.shrine.util.ShouldMatchersForJUnit
import org.junit.Test
import org.scalatest.Assertions

import scala.concurrent.duration.{DurationInt, FiniteDuration}
import scala.reflect.ClassTag
import scala.language.postfixOps

class RetryIOTest extends ShouldMatchersForJUnit {
  import RetryIOTest._

  @Test
  def testBase(): Unit = RetryIOTest.withTimer { implicit t: Timer[IO] =>
    assertSucceeds(RetryIO.retry(finalResult)()()()()(t))
  }

  @Test
  def testDelay(): Unit = RetryIOTest.withTimer { implicit t: Timer[IO] =>
    assertTakesAtLeast(delay, () => assertSucceeds(RetryIO.retry(finalResult)(delay)()()()(t)))
  }

  @Test
  def testExponentialBackoff(): Unit = RetryIOTest.withTimer { implicit t: Timer[IO] =>
    // first execution   -  1 delay  =  initialDelay (* 2^0)
    // second execution  -  2 delay  =  initialDelay * 2^1
    // third execution   -  4 delay  =  initialDelay * 2^2
    // all executions    -  7 delay
    assertTakesAtLeast(delay * 7, () =>
      assertTriesNTimes(3, () =>
        assertSucceeds(
          RetryIO.delayedExponentialBackoff(RetryIOTest.throwNTimes(2, finalResult), delay, 2, retryIfTaskThrows)
      ))
    )
  }

  @Test
  def testRetries(): Unit = RetryIOTest.withTimer { implicit t: Timer[IO] =>

    assertSucceeds(throwNTimes(0, finalResult))

    // Always tries once, then retries up to maxRetries times
    assertTriesNTimes(1, () => assertSucceeds(throwAndRetry(0, 0, 0 seconds)))
    assertTriesNTimes(2, () => assertSucceeds(throwAndRetry(1, 1, 0 seconds)))
    assertTriesNTimes(2, () => assertSucceeds(throwAndRetry(1, 2, 0 seconds)))
    assertTriesNTimes(4, () => assertSucceeds(throwAndRetry(3, 5, 0 seconds)))

    assertTriesNTimes(1, () =>
      assertThrowsException[RuntimeException] {
        throwAndRetry(1, 0, 0 seconds)
    })

    assertTriesNTimes(10, () =>
      assertThrowsException[RuntimeException] {
        throwAndRetry(10, 9, 0 seconds)
    })
  }

  @Test
  def testRetriesWithDelay(): Unit = RetryIOTest.withTimer { implicit t: Timer[IO] =>
    assertTakesAtLeast(delay, () => assertSucceeds(throwAndRetry(0, 0, delay)))
    assertTakesAtLeast(delay * 2, () => assertSucceeds(throwAndRetry(1, 1, delay)))
    assertTakesAtLeast(delay * 3, () => assertSucceeds(throwAndRetry(3, 3, delay)))
    assertTakesAtLeast(delay * 2, () => assertSucceeds(throwAndRetry(1, 3, delay)))

    assertTakesAtLeast(delay * 3, () =>
      assertTriesNTimes(3, () =>
        assertThrowsException[RuntimeException] {
          throwAndRetry(3, 2, delay)
        }
      )
    )
  }

  @Test
  def testUnretriableException(): Unit = RetryIOTest.withTimer { implicit t: Timer[IO] =>
    assertThrowsException[IllegalArgumentException] {
      RetryIO.retry(IO.raiseError(new IllegalArgumentException()))()()()()(t)
    }

    assertThrowsException[IllegalArgumentException] {
      RetryIO.retry(throwNTimes(2, finalResult).flatMap(_ => IO.raiseError(new IllegalArgumentException())))()(Some(2))()(retryIfTaskThrows)(t)
    }
  }

  @Test
  def testKeepTryingBounded():Unit = RetryIOTest.withTimer { implicit t: Timer[IO] =>

    assertSucceeds(throwNTimes(0, finalResult))

    assertThrowsException[RuntimeException] {
      RetryIO.keepTryingBounded(
        task = throwNTimes(3, finalResult),
        delay = 0 seconds,
        maxRetries = 2,
        retryIfTaskThrows = retryIfTaskThrows
      )
    }
  }

}

object RetryIOTest extends LoggingTest {
  import org.junit.Assert

  val delay: FiniteDuration = 100 milliseconds
  val retryIfTaskThrows: Throwable => Boolean = {case _: RuntimeException => true}
  val printedStatement: String = "ran"
  val finalResult: IO[Boolean] = IO{println(printedStatement); true}

  def throwAndRetry(throws: Int, retries: Int, delay: FiniteDuration)(implicit timer: Timer[IO]): IO[Boolean] = {
    RetryIO.retry(throwNTimes(throws, finalResult))(delay)(Some(retries))()(retryIfTaskThrows)(timer)
  }

  def assertThrowsException[I <: Throwable](io: IO[Boolean])(implicit tag: ClassTag[I]): Unit = {
    Assertions.assertThrows{io.unsafeRunSync()}
  }

  def throwNTimes(n: Int, result: IO[Boolean]): IO[Boolean] = {
    var times = 0
    result.map(a => if (times < n) {times += 1; throw new RuntimeException()} else a)
  }

  def withTimer(fn: Timer[IO] => Unit): Unit = {
    implicit val t: Timer[IO] = IO.timer(ExecutionContexts.longActionExecutionContext)
    fn(t)
  }

  def assertTakesAtLeast(duration: FiniteDuration, thunk: () => Unit): Unit = {
    val start = System.currentTimeMillis
    thunk()
    val timeTaken = System.currentTimeMillis - start
    Assert.assertTrue(timeTaken >= duration.toMillis)
  }

  def assertTriesNTimes(n: Int, thunk: () => Unit): Unit = {
    out.reset()
    thunk()
    Assert.assertEquals(n, out.toString().split("\n").count(_.equals(printedStatement)))
    out.reset()
  }

  def assertSucceeds(io: IO[Boolean]): Unit = io.map(a => Assert.assertEquals(true, a)).unsafeRunSync()
}