package net.shrine.protocol

import javax.xml.datatype.XMLGregorianCalendar
import net.shrine.log.Loggable
import net.shrine.problem.{AbstractProblem, Problem, ProblemNotYetEncoded, ProblemSources, RawProblem, XmlProblemDigest}
import net.shrine.protocol.QueryResult.StatusType
import net.shrine.protocol.version.ResultStatuses.QueuedByCRC
import net.shrine.protocol.version.{NodeKey, QueryId, ResultStatuses}
import net.shrine.protocol.version.v1.{UpdateResult, UpdateResultWithCrcResult, UpdateResultWithError, UpdateResultWithProgress}

import scala.xml.NodeSeq
import net.shrine.util.{NodeSeqEnrichments, OptionEnrichments, SEnum, Tries, XmlDateHelper, XmlUtil}
import net.shrine.serialization.{I2b2Marshaller, XmlMarshaller}

import scala.util.Try
import scala.collection.immutable.Seq

/**
 * @author Bill Simons
 * @since 4/15/11
 * @see http://cbmi.med.harvard.edu
 * @see http://chip.org
 *       <p/>
 *       NOTICE: This software comes with NO guarantees whatsoever and is
 *       licensed as Lgpl Open Source
 * @see http://www.gnu.org/licenses/lgpl.html
 *
 * NB: this is a case class to get a structural equality contract in hashCode and equals, mostly for testing
 */
final case class QueryResult (
                               resultId: Long,
                               instanceId: Long,
                               resultType: Option[ResultOutputType],
                               setSize: Long,
                               startDate: Option[XMLGregorianCalendar],
                               endDate: Option[XMLGregorianCalendar],
                               description: Option[String], //todo this is used for the adapter name
                               statusType: StatusType,
                               statusMessage: Option[String],
                               problem: Option[Problem] = None,
                               breakdowns: Map[ResultOutputType,I2b2ResultEnvelope] = Map.empty
) extends XmlMarshaller with I2b2Marshaller with Loggable {

  //only used in tests
  def this(
            resultId: Long,
            instanceId: Long,
            resultType: ResultOutputType,
            setSize: Long,
            startDate: XMLGregorianCalendar,
            endDate: XMLGregorianCalendar,
            statusType: QueryResult.StatusType) = {
    this(
      resultId,
      instanceId,
      Option(resultType),
      setSize,
      Option(startDate),
      Option(endDate),
      None, //description
      statusType,
      None) //statusMessage
  }

  def this(
            resultId: Long,
            instanceId: Long,
            resultType: ResultOutputType,
            setSize: Long,
            startDate: XMLGregorianCalendar,
            endDate: XMLGregorianCalendar,
            description: String,
            statusType: QueryResult.StatusType) = {
    this(
      resultId,
      instanceId,
      Option(resultType),
      setSize,
      Option(startDate),
      Option(endDate),
      Option(description),
      statusType,
      None) //statusMessage
  }

  def resultTypeIs(testedResultType: ResultOutputType): Boolean = resultType match {
    case Some(rt) => rt == testedResultType
    case _ => false
  }

  import QueryResult._

  //NB: Fragile, non-type-safe ==
  def isError: Boolean = statusType == StatusType.Error

  def elapsed: Option[Long] = {
    def inMillis(xmlGc: XMLGregorianCalendar) = xmlGc.toGregorianCalendar.getTimeInMillis

    for {
      start <- startDate
      end <- endDate
    } yield inMillis(end) - inMillis(start)
  }

  //Sorting isn't strictly necessary, but makes deterministic unit testing easier.
  //The number of breakdowns will be at most 4, so performance should not be an issue.
  private def sortedBreakdowns: Seq[I2b2ResultEnvelope] = {
    breakdowns.values.to[Seq].sortBy(_.resultType.name)
  }

  override def toI2b2: NodeSeq = {
    import OptionEnrichments._

    XmlUtil.stripWhitespace {
      <query_result_instance>
        <result_instance_id>{ resultId }</result_instance_id>
        <query_instance_id>{ instanceId }</query_instance_id>
        { description.toXml(<description/>) }
        {
          resultType.fold( ResultOutputType.ERROR.toI2b2NameOnly("") ){ rt =>
            if(rt.isBreakdown) rt.toI2b2NameOnly()
            else if (rt.isError) rt.toI2b2NameOnly()  //The result type can be an error
            else if (statusType.isError) rt.toI2b2NameOnly() //Or the status type can be an error
            else rt.toI2b2
          }
        }
        <set_size>{ setSize }</set_size>
        { startDate.toXml(<start_date/>) }
        { endDate.toXml(<end_date/>) }
        <query_status_type>
          <name>{ statusType }</name>
          { statusType.toI2b2(this) }
        </query_status_type>
        {
          //NB: Deliberately use Shrine XML format instead of the i2b2 one.  Adding breakdowns to i2b2-format XML here is deviating from the i2b2 XSD schema in any case,
          //so if we're going to do that, let's produce saner XML.
          sortedBreakdowns.map(_.toXml.head).map(XmlUtil.renameRootTag("breakdown_data"))
        }
      </query_result_instance>
    }
  }

  override def toXml: NodeSeq = XmlUtil.stripWhitespace {
    import OptionEnrichments._

    <queryResult>
      <resultId>{ resultId }</resultId>
      <instanceId>{ instanceId }</instanceId>
      { resultType.toXml(_.toXml) }
      <setSize>{ setSize }</setSize>
      { startDate.toXml(<startDate/>) }
      { endDate.toXml(<endDate/>) }
      { description.toXml(<description/>) }
      <status>{ statusType }</status>
      { statusMessage.toXml(<statusMessage/>) }
      {
        //Sorting isn't strictly necessary, but makes deterministic unit testing easier.
        //The number of breakdowns will be at most 4, so performance should not be an issue.
        sortedBreakdowns.map(_.toXml)
      }
      { problem.map(_.toXml).getOrElse("") }
    </queryResult>
  }

  def withId(id: Long): QueryResult = copy(resultId = id)

  def withInstanceId(id: Long): QueryResult = copy(instanceId = id)

  def modifySetSize(f: Long => Long): QueryResult = withSetSize(f(setSize))

  def withSetSize(size: Long): QueryResult = copy(setSize = size)

  def withDescription(desc: String): QueryResult = copy(description = Option(desc))

  def withResultType(resType: ResultOutputType): QueryResult = copy(resultType = Option(resType))

  def withBreakdown(breakdownData: I2b2ResultEnvelope): QueryResult = copy(breakdowns = breakdowns + (breakdownData.resultType -> breakdownData))

  def withBreakdowns(newBreakdowns: Map[ResultOutputType, I2b2ResultEnvelope]): QueryResult = copy(breakdowns = newBreakdowns)

  def withStatus(
                  statusType: StatusType,
                  statusMessage: Option[String]): QueryResult = copy(
                                                                      statusType = statusType,
                                                                      statusMessage = statusMessage
                                                                    )

  def createUpdateResult(
                          queryId:QueryId,
                          crcQueryInstanceId:Long,
                          resultOutputType: ResultOutputType
                        ):UpdateResult = statusType match {
    case finished:StatusType if finished == StatusType.Finished => //Only successful done state

      UpdateResultWithCrcResult(
        queryId = queryId,
        adapterNodeKey = NodeKey.localNodeKey,
        count = setSize.toInt,
        crcQueryInstanceId = crcQueryInstanceId,
        resultType = resultOutputType,
        statusMessage = statusMessage,
        breakdowns = I2b2ResultEnvelope.toV26Breakdowns(breakdowns)
      )
    case error:StatusType if error.isDone => //All other done states are errors
      UpdateResultWithError.create(
        queryId = queryId,
        adapterNodeKey = NodeKey.localNodeKey,
        problem = problem.getOrElse(ProblemNotYetEncoded(s"QueryResult with ${error.name} status but no problem.")),
        status = ResultStatuses.ErrorFromCrc,
        statusMessage = statusMessage
      )
    case _ => //Everything else is some flavor of "QUEUED - Ask the CRC Again Later."
      UpdateResultWithProgress(
        queryId = queryId,
        adapterNodeKey = NodeKey.localNodeKey,
        status = QueuedByCRC,
        statusMessage = statusMessage
      )
  }

  def consolidateWithError(queryResult: QueryResult): QueryResult ={
    copy(
      description = queryResult.description,
      statusMessage = queryResult.statusMessage,
      problem = queryResult.problem
    )
  }
}

object QueryResult {

  val defaultI2b2Id = Some(-1)

  final case class StatusType(
                               name: String,
                               isDone: Boolean,
                               i2b2Id: Option[Int] = defaultI2b2Id, //todo probably should not be an Option, just a default
                               isError: Boolean = false,
                               private val doToI2b2: QueryResult => NodeSeq = StatusType.defaultToI2b2,
                               isCrcCallCompleted:Boolean = true //todo OK to remove. Always true
    ) extends StatusType.Value {

    def crcPromisedToFinishAfterReply: Boolean = isCrcCallCompleted && !isDone

    def toI2b2(queryResult: QueryResult): NodeSeq = doToI2b2(queryResult)
  }

  object StatusType extends SEnum[StatusType] {
    private val defaultToI2b2: QueryResult => NodeSeq = { queryResult =>
      val i2b2Id: Int = queryResult.statusType.i2b2Id.getOrElse{
        throw new IllegalStateException(s"queryResult.statusType ${queryResult.statusType} has no i2b2Id")
      }
      <status_type_id>{ i2b2Id }</status_type_id><description>{ queryResult.statusType.name }</description>
    }

    val noMessage:NodeSeq = null
    val Error = StatusType("ERROR", isDone = true,  None, isError = true, { queryResult =>
      (queryResult.statusMessage, queryResult.problem) match {
        case (Some(msg),Some(pd)) => <description>{ if(msg != "ERROR") msg else pd.summary }</description> ++ pd.toXml
        case (Some(msg),None) => <description>{ msg }</description>
        case (None,Some(pd)) => <description>{ pd.summary }</description> ++ pd.toXml
        case (None, None) => noMessage
      }
    })

    //at least all of the states from https://github.com/i2b2/i2b2-core-server/blob/release-1709c-backup/edu.harvard.i2b2.crc/src/server/edu/harvard/i2b2/crc/ejb/QueryManagerBeanUtil.java
    val Finished = StatusType("FINISHED", isDone = true, Some(3))

    //All of these are forms of "QUEUEUD - Ask the CRC again later"
    val Processing = StatusType("PROCESSING", isDone = false, Some(2))
    val Running = StatusType("RUNNING", isDone = false, Some(2))
    val Queued = StatusType("QUEUED", isDone = false, Some(2))
    //TODO: What <status_type_id>s should these have?  Does anyone care?
    val Held = StatusType("HELD", isDone = false)
    val SmallQueue = StatusType("SMALL_QUEUE", isDone = false)
    val TimedOut = StatusType("TIMEDOUT", isDone = false)
    val MediumQueue = StatusType("MEDIUM_QUEUE", isDone = false)
    val LargeQueue = StatusType("LARGE_QUEUE", isDone = false)
    //three new states from i2b2 1.07.8.b
    val MediumQueueRunning = StatusType("MEDIUM_QUEUE_RUNNING", isDone = false)
    val LargeQueueRunning = StatusType("LARGE_QUEUE_RUNNING", isDone = false)

    val HubWillSubmit = StatusType("HUB_WILL_SUBMIT",isDone = true,isCrcCallCompleted = true)

    //other final error states from a conversation with the i2b2 team October 17, 2018
    val NoMoreQueue = StatusType("NO_MORE_QUEUE", isDone = true, isError = true, doToI2b2 = { queryResult =>
      queryResult.statusMessage.fold(noMessage){msg => <description>{ msg }</description>}
    })

    //from https://github.com/i2b2/i2b2-core-server/blob/38ba20a5b2cffa4bcfc397f19fa1a424c294264a/edu.harvard.i2b2.crc/src/server/edu/harvard/i2b2/crc/dao/setfinder/QueryStatusTypeId.java
    val Incomplete = StatusType("INCOMPLETE", isDone = true, Some(5), isError = true, { queryResult =>
      queryResult.statusMessage.fold(noMessage){msg => <description>{ msg }</description>}
    })

    //from https://github.com/i2b2/i2b2-core-server/search?q=NEVER_FINISHED&unscoped_q=NEVER_FINISHED
    val NeverFinished = StatusType("NEVER_FINISHED", isDone = true, isError = true, doToI2b2 = { queryResult =>
      queryResult.statusMessage.fold(noMessage){msg => <description>{ msg }</description>}
    })
    //from https://github.com/i2b2/i2b2-core-server/blob/38ba20a5b2cffa4bcfc397f19fa1a424c294264a/edu.harvard.i2b2.crc/src/server/edu/harvard/i2b2/crc/ejb/QueryInfoBean.java
    val Cancelled = StatusType("CANCELLED", isDone = true, Some(9), isError = true, { queryResult =>
      queryResult.statusMessage.fold(noMessage){msg => <description>{ msg }</description>}
    })

  }

  def extractLong(nodeSeq: NodeSeq)(elemName: String): Long = (nodeSeq \ elemName).text.toLong

  private def parseDate(lexicalRep: String): Option[XMLGregorianCalendar] = XmlDateHelper.parseXmlTime(lexicalRep).toOption

  def elemAt(path: String*)(xml: NodeSeq): NodeSeq = path.foldLeft(xml)(_ \ _)

  def asText(path: String*)(xml: NodeSeq): String = elemAt(path: _*)(xml).text.trim

  def asResultOutputTypeOption(elemNames: String*)(breakdownTypes: Set[ResultOutputType], xml: NodeSeq): Option[ResultOutputType] = {
    import ResultOutputType.valueOf

    val typeName = asText(elemNames: _*)(xml)

    valueOf(typeName) orElse valueOf(breakdownTypes)(typeName)
  }

  def extractResultOutputType(xml: NodeSeq)(parse: NodeSeq => Try[ResultOutputType]): Option[ResultOutputType] = {
    val attempt = parse(xml)

    attempt.toOption
  }

  def extractProblemDigest(xml: NodeSeq):Option[XmlProblemDigest] = {

    val subXml = xml \ "problem"
    if(subXml.nonEmpty) Some(XmlProblemDigest.fromXml(xml))
    else None
  }

  def fromXml(breakdownTypes: Set[ResultOutputType])(xml: NodeSeq): QueryResult = {
    def extract(elemName: String): Option[String] = {
      Option((xml \ elemName).text.trim).filter(!_.isEmpty)
    }

    def extractDate(elemName: String): Option[XMLGregorianCalendar] = extract(elemName).flatMap(parseDate)

    val asLong = extractLong(xml) _

    import NodeSeqEnrichments.Strictness._
    import Tries.sequence

    def extractBreakdowns(elemName: String): Map[ResultOutputType, I2b2ResultEnvelope] = {
      //noinspection ScalaUnnecessaryParentheses
      val mapAttempt = for {
        subXml <- xml.withChild(elemName)
        envelopes <- sequence(subXml.map(I2b2ResultEnvelope.fromXml(breakdownTypes)))
        mappings = envelopes.map(envelope => (envelope.resultType -> envelope))
      } yield Map.empty ++ mappings

      mapAttempt.getOrElse(Map.empty)
    }

    QueryResult(
      resultId = asLong("resultId"),
      instanceId = asLong("instanceId"),
      resultType = extractResultOutputType(xml \ "resultType")(ResultOutputType.fromXml),
      setSize = asLong("setSize"),
      startDate = extractDate("startDate"),
      endDate = extractDate("endDate"),
      description = extract("description"),
      statusType = StatusType.valueOf(asText("status")(xml)).get, //TODO: Avoid fragile .get call
      statusMessage = extract("statusMessage"),
      problem = extractProblemDigest(xml),
      breakdowns = extractBreakdowns("resultEnvelope")
    )
  }

  def fromI2b2(breakdownTypes: Set[ResultOutputType])(xml: NodeSeq): QueryResult = {
    def asLong = extractLong(xml) _

    def asTextOption(path: String*): Option[String] = elemAt(path: _*)(xml).headOption.map(_.text.trim)

    def asXmlGcOption(path: String): Option[XMLGregorianCalendar] = asTextOption(path).filter(!_.isEmpty).flatMap(parseDate)

    val statusText: String = asText("query_status_type", "name")(xml)
    val statusType: StatusType = StatusType.valueOf(statusText).getOrElse{
      throw new IllegalStateException(s"Unknown status name $statusText is not one of ${StatusType.values.mkString(", ")}")
    }
    val statusMessage: Option[String] = asTextOption("query_status_type", "description").map{description =>
      if(description == "ERROR") s"The I2B2 CRC responded with status type $statusType and a message of '$description'"
      else description
    }
    val encodedProblemDigest = extractProblemDigest(xml \ "query_status_type")
    val problemDigest = if (encodedProblemDigest.isDefined) encodedProblemDigest
                        else if (statusType.isError)
                          Some(XmlProblemDigest.create(ErrorStatusFromCrc(statusMessage,xml.toString())))
                        else None

    case class Filling(
                        resultType:Option[ResultOutputType],
                        setSize:Long,
                        startDate:Option[XMLGregorianCalendar],
                        endDate:Option[XMLGregorianCalendar]
                      )

    val filling = if(!statusType.isError) {
      val resultType: Option[ResultOutputType] = extractResultOutputType(xml \ "query_result_type")(ResultOutputType.fromI2b2)
      val setSize = asLong("set_size")
      val startDate = asXmlGcOption("start_date")
      val endDate = asXmlGcOption("end_date")
      Filling(resultType,setSize,startDate,endDate)
    }
    else {
      val resultType = None
      val setSize = 0L
      val startDate = None
      val endDate = None
      Filling(resultType,setSize,startDate,endDate)
    }

    QueryResult(
      resultId = asLong("result_instance_id"),
      instanceId = asLong("query_instance_id"),
      resultType = filling.resultType,
      setSize = filling.setSize,
      startDate = filling.startDate,
      endDate = filling.endDate,
      description = asTextOption("description"),
      statusType = statusType,
      statusMessage = statusMessage,
      problem = problemDigest
    )
  }

  def errorResult(description: Option[String], statusMessage: String, problem:Problem):QueryResult = {
    QueryResult(
      resultId = 0L,
      instanceId = 0L,
      resultType = None,
      setSize = 0L,
      startDate = None,
      endDate = None,
      description = description,
      statusType = StatusType.Error,
      statusMessage = Option(statusMessage),
      problem = Option(problem))
  }

  def errorResult(description: Option[String], statusMessage: String,problem:RawProblem):QueryResult = {

    QueryResult(
      resultId = 0L,
      instanceId = 0L,
      resultType = None,
      setSize = 0L,
      startDate = None,
      endDate = None,
      description = description,
      statusType = StatusType.Error,
      statusMessage = Option(statusMessage),
      problem = Option(problem))
  }

  /**
   * For reconstituting errorResults from a database
   */
  def errorResult(description:Option[String], statusMessage:String, codec:String,stampText:String, summary:String, digestDescription:String,detailsXml:NodeSeq): QueryResult = {
    // This would require parsing the stamp text to change, and without a standard locale that's nigh impossible.
    // If this is replaced with real problems, then this can be addressed then. For now, passing on zero is the best bet.
    val problemDigest = XmlProblemDigest(codec,stampText,summary,digestDescription,detailsXml,0)

    QueryResult(
      resultId = 0L,
      instanceId = 0L,
      resultType = None,
      setSize = 0L,
      startDate = None,
      endDate = None,
      description = description,
      statusType = StatusType.Error,
      statusMessage = Option(statusMessage),
      problem = Option(problemDigest))
  }
}

case class ErrorStatusFromCrc(messageFromCrC:Option[String], xmlResponseFromCrc: String) extends AbstractProblem(ProblemSources.Adapter) {
  override val summary: String = "The I2B2 CRC reported an internal error."
  override val description:String = s"The I2B2 CRC responded with status type ERROR ${messageFromCrC.fold(" but no message")(message => s"and a message of '$message'")}"
  override val detailsText = Some(s"CRC's Response is $xmlResponseFromCrc")
}

