package net.shrine.protocol.version.v1

import io.circe.{Decoder, Encoder}
import net.shrine.problem.{JsonProblemDigest, RawProblem}
import net.shrine.protocol.ResultOutputType
import net.shrine.protocol.version.{JsonText, NodeId, NodeName, QueryId, ResultId, ResultStatus, ResultStatuses, ValueClass}

import scala.util.Try
import scala.collection.immutable.Seq

/**
  * A query result, expected or completed
  *
  * @since 1.26
  * @author dwalend
  */
sealed abstract class Result(
                       id:ResultId,
                       versionInfo: VersionInfo
                       //todo user
                    ) extends Versioned[ResultId] {

  def asJsonText:JsonText = {
    //todo this can maybe be fully generic for all the common currency data, in Versioned
    //todo but the generics didn't work out of the box. Ask on the circe gitter channel SHRINE-2848
    import io.circe.generic.extras.auto.exportEncoder
    import io.circe.syntax.EncoderOps
    import io.circe.generic.extras.Configuration

    implicit val genDevConfig: Configuration = Versioned.genDevConfig

    import io.circe.generic.extras.encoding.UnwrappedEncoder
    implicit def valueClassEncoder[A <: ValueClass[_]](implicit encodeA: UnwrappedEncoder[A]): Encoder[A] = encodeA

    new JsonText(this.asJson.noSpaces)
  }

  def queryId:QueryId
  def adapterNodeId:NodeId
  def adapterNodeName:NodeName
  def status:ResultStatus
  def statusMessage:Option[String]
  def crcQueryInstanceId:Option[Long] //id of this query in the CRC //todo consider adding another type for results that the CRC should know about
}

object Result extends VersionedCompanion {

  val envelopeType: String = classOf[Result].getSimpleName

  def create(query: Query,adapterNode:Node):ResultProgress = ResultProgress(
    queryId = query.id,
    adapterNodeId = adapterNode.id,
    adapterNodeName = adapterNode.name
  )

  //todo this can maybe be fully generic for all the common currency data, maybe it can live in Versioned
  def tryRead(jsonText:JsonText):Try[Result] = Try {
    import io.circe.parser.decode
    import io.circe.generic.extras.auto.exportDecoder
    import io.circe.generic.extras.Configuration

    implicit val genDevConfig: Configuration = Versioned.genDevConfig

    import io.circe.generic.extras.decoding.UnwrappedDecoder
    implicit def valueClassDecoder[A <: ValueClass[_]](implicit decodeA: UnwrappedDecoder[A]): Decoder[A] = decodeA

    decode[Result](jsonText.underlying) match {
      case Right(result) => result
      case Left(x) => throw x //throw errors for now. Eventually do something clever, maybe with circe optics
    }
  }
}

case class ResultProgress(
                           id:ResultId = ResultId.create(),
                           versionInfo: VersionInfo = VersionInfo.create,
                           queryId: QueryId,
                           adapterNodeId:NodeId,
                           adapterNodeName:NodeName, //todo should probably be NodeKey
                           status:ResultStatus = ResultStatuses.IdAssigned,
                           statusMessage:Option[String] = None,
                           crcQueryInstanceId:Option[Long] = None
                         ) extends Result(
  id,
  versionInfo
) {
  def toErrorBeforeCrc(problem: RawProblem): ErrorResult = toError(JsonProblemDigest(problem),ResultStatuses.ErrorInShrine)
  def toCrcError(
                  problem: RawProblem,
                  statusMessage:Option[String] = None
                 ): ErrorResult =
    toError(JsonProblemDigest(problem),ResultStatuses.ErrorFromCrc,statusMessage)

  def toError(
                       problem:JsonProblemDigest,
                       status:ResultStatus,
                       statusMessage:Option[String] = None
                     ) = ErrorResult(
    id = id,
    versionInfo = versionInfo.next,
    queryId = queryId,
    adapterNodeId = adapterNodeId,
    adapterNodeName = adapterNodeName,
    status = status,
    statusMessage = statusMessage,
    problemDigest = problem
  )

  def withStatus(status: ResultStatus,statusMessage:Option[String] = None): ResultProgress = copy(
    versionInfo = versionInfo.next,
    status = status,
    statusMessage = statusMessage
  )

//todo reorder these arguments
  def toCrcResult(
                   count:Int,
                   crcQueryInstanceId:Long,
                   resultType: ResultOutputType,
                   statusMessage:Option[String] = None,
                   breakdowns:Option[Breakdowns] = None
                 ):CrcResult = CrcResult(
    id = id,
    versionInfo = versionInfo.next,
    queryId = queryId,
    adapterNodeName = adapterNodeName,
    adapterNodeId = adapterNodeId,
    crcQueryInstanceId = Some(crcQueryInstanceId),
    count = count,
    resultType = resultType,
    breakdowns = breakdowns
  )
}

case class CrcResult(
                      id:ResultId,
                      versionInfo: VersionInfo,
                      queryId: QueryId,
                      adapterNodeId:NodeId,
                      adapterNodeName:NodeName,
                      status:ResultStatus = ResultStatuses.ResultFromCRC, //Default value for this??
                      statusMessage:Option[String] = None,
                      crcQueryInstanceId:Option[Long],
                      count:Int,
                      resultType: ResultOutputType,
                      breakdowns: Option[Breakdowns] = None
                    ) extends Result(
                                      id,
                                      versionInfo
                                    )

/**
  * Need a little extra structure to support optional breakdown counts in circe
  */
case class Breakdowns(
                      counts:Seq[(String,Seq[(String,Int)])]
                     )

case class ErrorResult(
                        id:ResultId,
                        versionInfo: VersionInfo,
                        queryId: QueryId,
                        adapterNodeId:NodeId,
                        adapterNodeName:NodeName,
                        status:ResultStatus, //Default value for this??
                        statusMessage:Option[String],
                        problemDigest:JsonProblemDigest
                    ) extends Result(
                                      id,
                                      versionInfo
                                    ) {
  override def crcQueryInstanceId: Option[Long] = None
}

