import React from 'react';
import { shallow } from 'enzyme';
import { shallowToJson } from 'enzyme-to-json';

import { History } from 'sandbox/types';
import Paging, { WrappedPaging } from '..';

describe('Paging', () => {
  const getState = () => ({
    history: History(),
  });
  const store = {
    getState,
    dispatch: jest.fn(),
    subscribe: jest.fn(),
  };

  let mockHistory = History({
    rowCount: 150,
  });
  let pagingInstance = new WrappedPaging({
    history: mockHistory,
    dispatch: jest.fn(),
  });

  it('Should render correctly', () => {
    const output = shallow(<Paging store={store} />);
    expect(shallowToJson(output)).toMatchSnapshot();
  });

  it('should have a default limit of 34', () => {
    expect(pagingInstance.props.history.limit).toBe(34);
  });
  it('should have a default rowOffset of 0', () => {
    expect(pagingInstance.props.history.rowOffset).toBe(0);
  });

  it('should have a rowCount of 150', () => {
    expect(pagingInstance.props.history.rowCount).toBe(150);
  });

  it('It should ignore fetchPreviousPage when at index 0', () => {
    pagingInstance.fetchPreviousPage();
    expect(pagingInstance.props.dispatch).not.toHaveBeenCalled();
    expect(pagingInstance.props.history.rowOffset).toBe(0);
  });

  it('fetchNextPage should set the rowOffset to 34', () => {
    pagingInstance.fetchNextPage();
    expect(pagingInstance.props.dispatch).toHaveBeenCalledWith({
      payload: { limit: 34, rowOffset: 34 },
      type: 'FETCH_QUERY_RESULTS',
    });
  });
  it('fetchLastPage should set the rowOffset to 136', () => {
    pagingInstance.fetchLastPage();
    expect(pagingInstance.props.dispatch).toHaveBeenCalledWith({
      payload: { limit: 34, rowOffset: 136 },
      type: 'FETCH_QUERY_RESULTS',
    });
  });

  let mockHistory2 = History({
    rowCount: 150,
    rowOffset: 136,
  });

  let pagingInstance2 = new WrappedPaging({
    history: mockHistory2,
    dispatch: jest.fn(),
  });

  it('Should not allow fetchNextPage to execute the dispatch', () => {
    pagingInstance2.fetchNextPage();
    expect(pagingInstance2.props.dispatch).not.toHaveBeenCalled();
  });

  it('fetchPreviousPage should change the rowOffset from 136 to 102.', () => {
    pagingInstance2.fetchPreviousPage();
    expect(pagingInstance2.props.dispatch).toHaveBeenLastCalledWith({
      payload: { limit: 34, rowOffset: 102 },
      type: 'FETCH_QUERY_RESULTS',
    });
  });
  it('fetchFirstPage should change the rowOffset from 136 to 0.', () => {
    pagingInstance2.fetchFirstPage();
    expect(pagingInstance2.props.dispatch).toHaveBeenLastCalledWith({
      payload: { limit: 34, rowOffset: 0 },
      type: 'FETCH_QUERY_RESULTS',
    });
  });
});
