import { History } from 'sandbox/types';
import { queryHistoryReducer } from '../queryHistoryReducer';
import { List } from 'immutable';

describe('Query History Reducer', () => {
  it('Should yield a default state if none is provided', () => {
    const action = { type: 'UKNOWN_ACTION' };
    const expected = History();
    const actual = queryHistoryReducer(undefined, action);
    expect(actual).toEqual(expected);
  });


  it('Should set isFetching to true when fetch query results is pending', () => {
    const action = { type: 'FETCH_QUERY_RESULTS_PENDING' };
    const expected = History().merge({ isFetching: true });
    const actual = queryHistoryReducer(History(), action);
    expect(actual).toEqual(expected);
  });

  it('Should set the error status to true and provide details when a fetch fails', () => {
    const action = {
      type: 'FETCH_QUERY_RESULTS_REJECTED',
      errorData: {
        status: 'Mock error status',
        statusText: 'Mock status text',
        url: 'Mock url',
      },
    };
    const expected = History().merge({
      error: true,
      url: 'Mock url',
      errorMsg: 'Mock error status Mock status text',
    });
    const actual = queryHistoryReducer(History(), action);
    expect(actual).toEqual(expected);
  });

  it('It should merge query results When query results fetch is completed', () => {
    const action = {
      payload: {
        queryResults: [],
        rowCount: 50,
      },
      type: 'FETCH_QUERY_RESULTS_FULFILLED',
    };
    const expected = History().merge({
      results: List([]),
      rowCount: 50,
    });
    const actual = queryHistoryReducer(History(), action);
    expect(actual).toEqual(expected);
  });

  it('It should have a default limit if none is provided.', () => {
    const action = {
      type: 'FETCH_QUERY_RESULTS',
    };
    const expected = History().limit;
    const { limit: actual } = queryHistoryReducer(History(), action);
    expect(actual).toEqual(expected);
  });

  it('It should update the limit if one is provided.', () => {
    const action = {
      type: 'FETCH_QUERY_RESULTS',
      payload: { limit: 20 },
    };
    const expected = 20;
    const { limit: actual } = queryHistoryReducer(History(), action);
    expect(actual).toEqual(expected);
  });
  

  it('Should always contain the networkId of the currently displayed query', () => {
    const action = {
      type: 'SELECT_QUERY_HISTORY_RESULT',
      payload: { networkId: '12345' },
    };
    const expected = History().merge({ selectedResultId: '12345' });
    const actual = queryHistoryReducer(undefined, action);
    expect(actual).toEqual(expected);
  });

  it('Should set the selectedResultId to null when a query is cleared', () => {
    const action = {
      type: 'CLEAR_QUERY',
    };
    const expected = History().merge({ selectedResultId: null });
    const actual = queryHistoryReducer(undefined, action);
    expect(actual).toEqual(expected);
  });

  it('Should set the selectedResultId to null when a Query Is initially run', () => {
    const action = {
      type: 'RUN_QUERY',
    };
    const expected = History().merge({ selectedResultId: null });
    const actual = queryHistoryReducer(undefined, action);
    expect(actual).toEqual(expected);
  });
});