import { queryPollReducer, initialState } from '../queryPollReducer';

describe('queryPollReducer', () => {
  it('It should yield a default state if none is provided.', () => {
    const action = { type: 'UKNOWN_ACTION' };
    const expected = initialState;
    const actual = queryPollReducer(undefined, action);
    expect(actual).toEqual(expected);
  });

  it('Should update the query name and status message when handling QUERY_STARTED', () => {
    const action = { type: 'QUERY_STARTED', payload: { name: 'mock-name' } };
    const expected = {
      ...initialState,
      name: 'mock-name',
      statusMsg: 'Waiting on response from network',
      isRunning: true,
    };
    const actual = queryPollReducer(initialState, action);
    expect(actual.name).toEqual(expected.name);
    expect(actual.statusMsg).toEqual(expected.statusMsg);
    expect(actual.isRunning).toBe(expected.isRunning);
  });

  it('Should display the correct breakdown message when QUERY_STARTED and breakdowns are selected', () => {
    const action = { type: 'QUERY_STARTED', payload: { name: 'mock-name', data: { mockOption: 'mock-selected-breakdown-option' } } };
    const expected = {
      ...initialState,
      name: 'mock-name',
      statusMsg: 'Waiting on response from network',
      breakdownMsg: 'A Query is currently in progress but a response containing breakdowns has not yet been received. Click the "Query Status" tab for more details.',
      hasBreakdowns: true,
      isRunning: true,
    };
    const actual = queryPollReducer(initialState, action);
    expect(actual.breakdownMsg).toEqual(expected.breakdownMsg);
    expect(actual.hasBreakdowns).toEqual(expected.hasBreakdowns);
  });

  it('Should update the networkId, breakdownMessage and statusMessage when handling QUERY_NETWORK_ID_RECEIVED', () => {
    const action = {
      type: 'QUERY_NETWORK_ID_RECEIVED',
      payload: { networkId: '12345' },
    };
    const expected = {
      networkId: '12345',
      breakdownMsg: 'A Query is currently in progress but a response containing breakdowns was not selected. To view additional query results, rerun the query and select breakdowns prior to running the query.',
      statusMsg: 'Response received from network.  Waiting on results',
    };
    const actual = queryPollReducer({ ...initialState, hasBreakdowns: false }, action);
    expect(actual.networkId).toBe(expected.networkId);
    expect(actual.breakdownMsg).toBe(expected.breakdownMsg);
    expect(actual.statusMsg).toBe(expected.statusMsg);
  });

  it('Should update the networkId, breakdownMessage and statusMessage when handling QUERY_NETWORK_ID_RECEIVED', () => {
    const action = {
      type: 'QUERY_NETWORK_ID_RECEIVED',
      payload: { networkId: '12345' },
    };
    const expected = {
      networkId: '12345',
      breakdownMsg: 'A Query is currently in progress but a response containing breakdowns has not yet been received. Click the "Query Status" tab for more details.',
      statusMsg: 'Response received from network.  Waiting on results',
    };
    const actual = queryPollReducer({ ...initialState, hasBreakdowns: true }, action);
    expect(actual.breakdownMsg).toBe(expected.breakdownMsg);
  });


  it('Should update the Query State when an UPDATE is dispatched', () => {
    const payload = {
      isComplete: true,
      query: {
        queryName: 'mock-query',
        status: 'mock query status',
      },
      results: [{
        adapterNode: 'mock-institution',
        count: 0,
        status: 'MOCK_STATUS',
        statusMessage: 'mock status message',
        internalStatus: 'mock internal status',
        problemDigest: { details: '' },
        breakdowns: [{
          resultType: { i2b2Options: { description: 'mock breakdown description' } },
          results: [{
            dataKey: 'mock data key',
            value: -1,
          }],
        }],
      }],
    };

    const action = {
      type: 'QUERY_UPDATE',
      payload,
    };

    let expected = {
      breakdownMsg: 'The Query has completed but there are no breakdowns to display. To view additional query results, rerun the query and select breakdowns prior to running the query.',
      hasBreakdowns: false,
      breakdowns: {},
      error: false,
      problemDigest: null,
      institutionResults: [{
        count: 0,
        institutionName: 'mock-institution',
        problemDigest: { details: '' },
        status: 'MOCK_STATUS',
        statusMessage: 'mock status message',
        internalStatus: 'mock internal status',
        breakdowns: [{
          resultType: {
            i2b2Options: { description: 'mock breakdown description' },
          },
          results: [{
            dataKey: 'mock data key',
            value: -1,
          }],
        }],
      }],
      isRunning: false,
      name: 'mock-query',
      updated: Number(new Date()),
      status: 'mock query status',
    };
    const actual = queryPollReducer({
      hasBreakdowns: false,
      breakdownMsg: 'The Query has completed but there are no breakdowns to display. To view additional query results, rerun the query and select breakdowns prior to running the query.',
    }, action);
    expected.updated = actual.updated; // Timestamp equality Side effect, allow comparison to pass.
    expect(actual).toEqual(expected);
  });


  it('Should update the breakdown message when an UPDATE is dispatched and the query has breakdowns', () => {
    const payload = {
      isComplete: false,
      query: {
        queryName: 'mock-query',
        status: 'mock query status',
      },
      results: [{
        adapterNode: 'mock-institution',
        count: 0,
        status: 'MOCK_STATUS',
        statusMessage: 'mock status message',
        internalStatus: 'mock internal status',
        problemDigest: { details: '' },
        breakdowns: [{
          resultType: { i2b2Options: { description: 'mock breakdown description' } },
          results: [{
            dataKey: 'mock data key',
            value: -1,
          }],
        }],
      }],
    };

    const action = {
      type: 'QUERY_UPDATE',
      payload,
    };

    let expected = {
      breakdownMsg: 'A Query is currently in progress but a response containing breakdowns has not yet been received. Click the "Query Status" tab for more details.',
      hasBreakdowns: false,
      breakdowns: {},
      error: false,
      institutionResults: [{
        count: 0,
        institutionName: 'mock-institution',
        problemDigest: { details: '' },
        status: 'MOCK_STATUS',
        internalStatus: 'mock internal status',
        statusMessage: 'mock status message',
        breakdowns: [{
          resultType: {
            i2b2Options: { description: 'mock breakdown description' },
          },
          results: [{
            dataKey: 'mock data key',
            value: -1,
          }],
        }],
      }],
      isRunning: true,
      name: 'mock-query',
      updated: Number(new Date()),
      status: 'mock query status',
    };
    const actual = queryPollReducer({
      isComplete: false,
      hasBreakdowns: true,
      breakdownMsg: 'A Query is currently in progress but a response containing breakdowns has not yet been received. Click the "Query Status" tab for more details.',
    }, action);
    expected.updated = actual.updated; // Timestamp equality Side effect, allow comparison to pass.
    expect(actual.breakdownMsg).toEqual(expected.breakdownMsg);
  });
  it('isQueryError should be false when SHRINE does not report a query error', () => {
    const payload = {
      isComplete: true,
      query: {
        queryName: 'mock-query',
        status: 'mock query status',
        isQueryError: false,
      },
      results: [{
        adapterNode: 'mock-institution',
        count: 0,
        status: 'MOCK_STATUS',
        statusMessage: 'mock status message',
        internalStatus: 'mock internal status',
        problemDigest: { details: '' },
        breakdowns: [{
          resultType: { i2b2Options: { description: 'mock breakdown description' } },
          results: [{
            dataKey: 'mock data key',
            value: -1,
          }],
        }],
      }],
    };

    const action = {
      type: 'QUERY_UPDATE',
      payload,
    };

    const actual = queryPollReducer(initialState, action);
    expect(actual.isQueryError).toEqual(false);
  });
  it('isQueryError should be true when SHRINE reports a query error', () => {
    const payload = {
      isComplete: true,
      query: {
        queryName: 'mock-query',
        status: 'mock query status',
        isQueryError: true,
      },
      results: [{
        adapterNode: 'mock-institution',
        count: 0,
        status: 'MOCK_STATUS',
        statusMessage: 'mock status message',
        internalStatus: 'mock internal status',
        problemDigest: { details: '' },
        breakdowns: [{
          resultType: { i2b2Options: { description: 'mock breakdown description' } },
          results: [{
            dataKey: 'mock data key',
            value: -1,
          }],
        }],
      }],
    };

    const action = {
      type: 'QUERY_UPDATE',
      payload,
    };

    const actual = queryPollReducer(initialState, action);
    expect(actual.isQueryError).toEqual(true);
  });

  it('Should create an error state for the view when QUERY_ERROR is dispatched.', () => {
    const action = {
      type: 'QUERY_ERROR',
      errorData: {
        status: 'ERROR',
        statusText: 'Mock error text',
        url: 'mock url',
      },
    };

    let expected = {
      error: true,
      errorMsg: 'ERROR: Mock error text',
      isRunning: false,
      updated: Number(new Date()),
      url: 'mock url',
    };
    const actual = queryPollReducer({}, action);
    expected.updated = actual.updated; // Timestamp equality Side effect, allow comparison to pass.
    expect(actual).toEqual(expected);
  });

  it('Should display the proper breakdown message when isQueryError is set to true', () => {
    const payload = {
      isComplete: true,
      query: {
        queryName: 'mock-query',
        status: 'mock query status',
        isQueryError: true,
      },
      results: [],
    };

    const action = {
      type: 'QUERY_UPDATE',
      payload,
    };

    const result = queryPollReducer(initialState, action);
    const expected = 'The Query is in an error state, so no breakdown results will be received. Click the "Query Status" table for more details.';
    const {breakdownMsg: actual } = result;
    expect(actual).toBe(expected);
  });
});