import {
  QUERY_STARTED,
  QUERY_NETWORK_ID_RECEIVED,
  QUERY_UPDATE,
  CLEAR_QUERY,
  I2B2_NETWORK_ERROR,
} from 'sandbox/actions';
import {
  mapResultsToInstitutionResults,
  mapInstitutionResultsToBreakdowns,
} from 'sandbox/services';
import { defaultState } from 'sandbox/defaultState';
import { QUERY_ERROR } from '../actions';

export const initialState = { ...defaultState.queryData };

const selectBreakdownMessage = hasBreakdowns => (
  hasBreakdowns
    ? 'A Query is currently in progress but a response containing breakdowns has not yet been received. Click the "Query Status" tab for more details.'
    : 'A Query is currently in progress but a response containing breakdowns was not selected. To view additional query results, rerun the query and select breakdowns prior to running the query.'
);

const mapQueryStartedIntoState = (state, { payload }) => {
  const { name } = payload;
  const hasBreakdowns = payload.data && Object.entries(payload.data).length > 0;

  return {
    ...initialState,
    updated: null,
    error: false,
    breakdownMsg: selectBreakdownMessage(hasBreakdowns),
    statusMsg: 'Waiting on response from network',
    isRunning: true,
    hasBreakdowns,
    name,
  };
};

const mapQueryNetworkIdReceivedIntoState = (state, { payload }) => {
  const { name = null, networkId = null } = payload;
  return ({
    ...state,
    error: false,
    isRunning: true,
    updated: null,
    name: name === null ? state.name : name,
    networkId,
    breakdownMsg: selectBreakdownMessage(state.hasBreakdowns),
    statusMsg: 'Response received from network.  Waiting on results',
  });
};

const mapQueryUpdateIntoState = (state, { payload }) => {
  const {
    query,
    results,
    dataVersion,
    isComplete,
  } = payload;
  const institutionResults = mapResultsToInstitutionResults(results);
  const breakdowns = mapInstitutionResultsToBreakdowns(institutionResults);

  let { breakdownMsg } = state;
  if (query.isQueryError || query.isResultsError) {
    breakdownMsg = 'The Query is in an error state, so no breakdown results will be received. Click the "Query Status" table for more details.';
  } else if (isComplete && !state.hasBreakdowns) {
    breakdownMsg = 'The Query has completed but there are no breakdowns to display. To view additional query results, rerun the query and select breakdowns prior to running the query.'
  }

  return {
    ...state,
    problemDigest: query.problemDigest || initialState.problemDigest,
    status: query.status,
    name: query.queryName,
    dataVersion,
    updated: query.changeDate,
    error: false,
    isRunning: isComplete === false,
    institutionResults,
    breakdowns,
    breakdownMsg,
    isQueryError: query.isQueryError,
    isResultsError: query.isResultsError,
  };
};

const mapQueryErrorIntoState = (state, action) => {
  const { status, statusText, url } = action.errorData;
  return {
    ...state,
    updated: null,
    error: true,
    isRunning: false,
    url,
    errorMsg: `${status}: ${statusText}`,
  };
};

export const queryPollReducer = (state = initialState, action) => {
  switch (action.type) {
    case CLEAR_QUERY: { return initialState; }
    case QUERY_STARTED: { return mapQueryStartedIntoState(state, action); }
    case QUERY_NETWORK_ID_RECEIVED: { return mapQueryNetworkIdReceivedIntoState(state, action); }
    case QUERY_UPDATE: { return mapQueryUpdateIntoState(state, action); }
    case I2B2_NETWORK_ERROR:
    case QUERY_ERROR: {
      return mapQueryErrorIntoState(state, action);
    }
    default: { return state; }
  }
};
