import fetch from 'isomorphic-fetch';
import { put, call, apply, takeLatest, take } from 'redux-saga/effects';

import {
  startQuerySaga,
  sseEventListenerSaga,
  fetchSSEAuthTokenSaga,
} from '../queryEventsSaga';


describe('fetchSSEAuthTokenSaga fetch token success', () => {
  const action = {
    type: 'QUERY_NETWORK_ID_RECEIVED',
    payload: {
      networkId: '1234',
    },
  }
  const generator = fetchSSEAuthTokenSaga(action);
  it('Should make a call to queryResultAuth', () => {
    const expected = call(fetch, '/shrine-api/qep/queryResultAuth', {
      headers: {
        Authorization: 'Basic  null',
      },
    });
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
  it('Should parse the text token response', () => {
    const response = {
      ok: true,
      text: () => 'mock token reponse',
    };
    const expected = apply(response, response.text);
    const actual = generator.next(response).value;
    expect(actual).toEqual(expected);
  });
  it('Should dispatch a START_SSE_EVENT_LISTENER action', () => {
    const expected = put({
      type: 'START_SSE_EVENT_LISTENER',
      payload: {
        networkId: action.payload.networkId,
      },
    });
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
  it('Should close the thread', () => {
    const expected = `Polling thread for ${action.payload.networkId} cancelled`;
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
});

describe('fetchSSEAuthTokenSaga fetch token fail', () => {
  const action = {
    type: 'QUERY_NETWORK_ID_RECEIVED',
    payload: {
      networkId: '1234',
    },
  }
  const generator = fetchSSEAuthTokenSaga(action);
  it('Should make a call to queryResultAuth', () => {
    const expected = call(fetch, '/shrine-api/qep/queryResultAuth', {
      headers: {
        Authorization: `Basic  null`,
      },
    });
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
  it('Should dispatch a queryError action on failure', () => {
    const response = {
      ok: false,
      text: () => 'mock token reponse',
    };
    const expected = put({
      type: 'QUERY_ERROR',
      errorData: {
        networkId: '1234',
        status: undefined,
        statusText: undefined,
        url: '/shrine-api/qep/queryResultAuth',
      },
    });
    const actual = generator.next(response).value;
    expect(actual).toEqual(expected);
  });
  it('Should cancel the thread if the fetch is unsuccessful', () => {
    const expected = `Polling thread for ${action.payload.networkId} cancelled`;
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
});

describe('sseEventListenerSaga', () => {
  const action = {
    type: 'START_SSE_EVENT_LISTENER',
    payload: {
      networkId: '1234',
      token: 'mock token response',
    },
  };
  const generator = sseEventListenerSaga(action);
  it('Should call startSSEvents (Start Server Side Events)', () => {
    const startSSEvents = payload => payload;
    const expected = call(startSSEvents, { ...action.payload });
    const actual = generator.next().value;
    expect(actual.toString()).toEqual(expected.toString());
  });
});

describe('startQuerySaga', () => {
  const generator = startQuerySaga();
  it('Should set up a listener for Query NetowrkId and Restart Poll events', () => {
    const expected = takeLatest([
      'QUERY_NETWORK_ID_RECEIVED',
      'RESTART_QUERY_POLL',
    ], fetchSSEAuthTokenSaga);
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
  it('Should set up a listener to Start the Server Side Events', () => {
    const expected = takeLatest('START_SSE_EVENT_LISTENER', sseEventListenerSaga);
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
  it('It should listen for a poll cancellation', () => {
    const expected = take('CANCEL_QUERY_POLL');
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
});