import fetch from 'isomorphic-fetch';
import { put, call, apply, takeLatest, select } from 'redux-saga/effects';

import { fetchQueryResultsPending } from 'sandbox/actions';
import {
  doQueriesFetch,
  queryResultsFetchSaga,
} from '../queryResultsFetchSaga';
import { selectLimit, selectSortBy, selectRowOffset } from '../selectors';

describe('selectLimit', () => {
  it('selectLimit should select the limit property off of a state object', () => {
    const state = { history: { limit: 50 } };
    const expected = 50;
    const actual = selectLimit(state);
    expect(actual).toBe(expected);
  });
});

describe('selectSortBy', () => {
  it('selectLimit should select the limit property off of a state object', () => {
    const state = { history: { sortBy: 'MOCK_SORT_BY' } };
    const expected = 'MOCK_SORT_BY';
    const actual = selectSortBy(state);
    expect(actual).toBe(expected);
  });
});

describe('selectRowOffset', () => {
  it('selectLimit should select the rowOffset property off of a state object', () => {
    const state = { history: { rowOffset: 'MOCK_OFFSET' } };
    const expected = 'MOCK_OFFSET';
    const actual = selectRowOffset(state);
    expect(actual).toBe(expected);
  });
});

describe('Query Results Fetch Saga Successful Fetch', () => {
  const generator = doQueriesFetch();
  it('Should yield a fetch query results pending action', () => {
    const action = { type: 'FETCH_QUERY_RESULTS_PENDING' };
    const expected = put(fetchQueryResultsPending());
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should select the limit from the history state', () => {
    const expected = select(selectLimit);
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should select the sortBy from the history state', () => {
    const expected = select(selectSortBy);
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should select the rowOffset from the history state', () => {
    const expected = select(selectRowOffset);
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should call isomorphic fetch to the specified url', () => {
    const mockUrl =
      '/shrine-api/qep/queryResultPolling?&limit=undefined&sortBy=undefined&skip=undefined';
    const headers = { Authorization: `Basic ${null}` };
    const expected = call(fetch, mockUrl, { headers });
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should process the json from the REST response', () => {
    const response = { json: jest.fn(), ok: true };
    const expected = apply(response, response.json);
    const actual = generator.next(response).value;
    expect(actual).toEqual(expected);
  });

  it('Should dispatch a QUERY_RESULTS_FULFILLED action if fetch was successful.', () => {
    const action = {
      type: 'FETCH_QUERY_RESULTS_FULFILLED',
      payload: {
        allQueries: [],
        rowCount: undefined,
        rowOffset: undefined,
      },
    };
    const expected = put(action);
    const actual = generator.next({ queryResults: [] }).value;
    expect(actual).toEqual(expected);
  });

  it('Should close the fetch thread when completed.', () => {
    const expected = 'Fetch of Query History cancelled';
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
});

describe('Query Results Fetch Saga Unsuccessful Fetch', () => {
  const generator = doQueriesFetch();
  it('Should yield a fetch query results pending action', () => {
    const action = { type: 'FETCH_QUERY_RESULTS_PENDING' };
    const expected = put(fetchQueryResultsPending());
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should select the limit from the history state', () => {
    const expected = select(selectLimit);
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should select the sortBy from the history state', () => {
    const expected = select(selectSortBy);
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should select the rowOffset from the history state', () => {
    const expected = select(selectRowOffset);
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should call isomorphic fetch to the specified url', () => {
    const mockUrl =
      '/shrine-api/qep/queryResultPolling?&limit=undefined&sortBy=undefined&skip=undefined';
    const headers = { Authorization: `Basic ${null}` };
    const expected = call(fetch, mockUrl, { headers });
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });

  it('Should process the json from the REST response', () => {
    const action = {
      type: 'FETCH_QUERY_RESULTS_REJECTED',
      errorData: {
        status: 'mock failed status',
        statusText: 'mock failed status text',
        url: 'mock url',
      },
    };
    const expected = put(action);
    const response = {
      ok: false,
      status: 'mock failed status',
      statusText: 'mock failed status text',
      url: 'mock url',
    };

    const actual = generator.next(response).value;
    expect(actual).toEqual(expected);
  });

  it('Should close the fetch thread when completed.', () => {
    const expected = 'Fetch of Query History cancelled';
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
});

describe('Query Fetch Saga', () => {
  const generator = queryResultsFetchSaga();
  it('should start the query fetch when FETCH_QUERY_RESULTS is dispatched', () => {
    const expected = takeLatest('FETCH_QUERY_RESULTS', doQueriesFetch);
    const actual = generator.next().value;
    expect(actual).toEqual(expected);
  });
});
