import { take, put, takeLatest, apply, call } from 'redux-saga/effects';
import fetch from 'isomorphic-fetch';

import { getStore } from 'sandbox';
import {
  QUERY_NETWORK_ID_RECEIVED,
  RESTART_QUERY_POLL,
  CANCEL_QUERY_POLL,
  queryUpdate,
  queryHistoryUpdate,
  queryError,
} from 'sandbox/actions';
import { getBaseUrl, getAuth } from 'sandbox/services';

let eventSource;
const makeServerEventListeners = () => {
  const store = getStore();
  let eventId;
  const handleQueryResult = (queryResponse) => {
    const queryData = JSON.parse(queryResponse.data);
    const {
      selectedQuery = null,
      allQueries = [],
    } = queryData;
    store.dispatch(queryHistoryUpdate(allQueries));
    if (selectedQuery !== null) {
      store.dispatch(queryUpdate(selectedQuery));

      if (selectedQuery.isComplete) {
        eventSource.removeEventListener(`queryResult-${eventId}`, handleQueryResult);
        eventSource.close();
      }
    }
  };

  const handleAuth = ({ lastEventId }) => {
    if (eventId) eventSource.removeEventListener(`queryResult-${eventId}`, handleQueryResult);
    eventId = lastEventId;
    eventSource.addEventListener(`queryResult-${eventId}`, handleQueryResult);
  };

  const makeHandleError = networkId => (event) => {
    if (event.target.readyState !== EventSource.CLOSED) {
      // store.dispatch({ type: CANCEL_QUERY_POLL });
      // store.dispatch({ type: RESTART_QUERY_POLL, payload: { networkId } });
    }
  };


  return ({ networkId, token }) => {
    const url = `${getBaseUrl()}qep/sse/queryResultSet/${token}?networkId=${networkId}`;

    if (eventSource) {
      eventSource.removeEventListener('auth', handleAuth);
      eventSource.onerror = null;
      eventSource.close();
    }

    eventSource = new EventSource(url);
    eventSource.addEventListener('auth', handleAuth);
    // TODO: figure out why adding any kind of error handler is causing issues.
    // eventSource.onerror(makeHandleError(networkId));
  };
};

const START_SSE_EVENT_LISTENER = 'START_SSE_EVENT_LISTENER';
export function* fetchSSEAuthTokenSaga({ payload }) {
  const { networkId } = payload;
  try {
    const url = `${getBaseUrl()}qep/queryResultAuth`;
    const headers = { Authorization: `Basic  ${getAuth()}` };
    const response = yield call(fetch, url, { headers });
    if (response.ok) {
      const token = yield apply(response, response.text);
      yield put({
        type: START_SSE_EVENT_LISTENER,
        payload: { networkId, token },
      });
    } else {
      const {
        status,
        statusText,
      } = response;
      yield put(queryError({
        status,
        statusText,
        url,
        networkId,
      }));
    }
  } finally {
    const msg = `Polling thread for ${networkId} cancelled`;
    console.log(msg);
    return msg;
  }
}

let serverEventListeners;
export function* sseEventListenerSaga({ payload }) {
  if (!serverEventListeners) serverEventListeners = makeServerEventListeners();
  yield call(serverEventListeners, payload);
}

export function* startQuerySaga() {
  while (true) {
    const fetchSSEAuthTokenThread = yield takeLatest([
      QUERY_NETWORK_ID_RECEIVED,
      RESTART_QUERY_POLL,
    ], fetchSSEAuthTokenSaga);
    const sseEventListenerThread = yield takeLatest(START_SSE_EVENT_LISTENER, sseEventListenerSaga);

    yield take(CANCEL_QUERY_POLL);
    if (fetchSSEAuthTokenThread.isRunning()) {
      yield fetchSSEAuthTokenThread.cancel();
      yield sseEventListenerThread.cancel();
      if (eventSource) eventSource.close();
    }
  }
}
