package net.shrine.hub

import java.util.concurrent.atomic.AtomicReference

import cats.effect.{CancelToken, ContextShift, IO}
import net.shrine.crypto.{BouncyKeyStoreCollection, SignatureNotVerifiedException}
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.hub.data.store.{HubDb, ItemVersionRaceLostException, QueryRow}
import net.shrine.hub.mom.ShrineMomClient
import net.shrine.log.Log
import net.shrine.problem.{AbstractProblem, JsonProblemDigest, ProblemSources}
import net.shrine.protocol.version.v1.{Node, Query, QueryProgress, Researcher, Result, ResultProgress, RunQueryAtHub, RunQueryForResult, Topic, UpdateQueryAtAdapter, UpdateQueryAtQep, UpdateQueryAtQepWithError, UpdateQueryAtQepWithStatus, UpdateQueryReadyForAdapters, UpdateResult}
import net.shrine.protocol.version.{CouldNotVerifySignature, DateStamp, Envelope, JsonText, NodeId, QueryId, QueryStatuses, ResultStatuses}
import net.shrine.source.ConfigSource

import scala.collection.immutable.Iterable
import scala.util.control.NonFatal

object HubDispatcher {

  private val sendToAdaptersAttempts = ConfigSource.config.getInt("shrine.hub.sendToAdaptersAttempts")

  private val atomicCancelToken = new AtomicReference[CancelToken[IO]](null)

  def restart(): IO[Unit] = IO { synchronized {
    stop()
    atomicCancelToken.set(ShrineMomClient.receiveUntilStop(
      HubDb.db.selectTheNetworkIO.map(_.hubQueueName),
      dispatchIO
    ))
  }}

  def stop(): IO[Unit] = IO { synchronized {
    Option(atomicCancelToken.getAndSet(null)).foreach { ct: CancelToken[IO] =>
      Log.debug(s"Stopping HubDispatcher")
      ct.unsafeRunSync()
    }
  }}

  private def dispatchIO(envelope: Envelope,remainingAttempts:Int):IO[Boolean] = {
    envelope.contentsType match {
      case t: String if t == RunQueryAtHub.envelopeType =>
        val command = RunQueryAtHub.tryRead(new JsonText(envelope.contents)).get
        sendQueryToAdaptersIO(command,remainingAttempts)
      case t: String if t == UpdateResult.envelopeType =>
        val result = UpdateResult.tryRead(new JsonText(envelope.contents)).get
        updateResultIO(result).flatMap(_ => IO(true))
      case t: String if t == UpdateQueryAtAdapter.envelopeType =>
        val queryUpdate = UpdateQueryAtAdapter.tryRead(new JsonText(envelope.contents)).get
        updateQueryAtHubAndAdaptersIO(queryUpdate).flatMap(_ => IO(true))
    }
  }

  def updateResultIO(result:Result):IO[Unit] = {

    for{
      _ <- IO(Log.debug(s"Start updating qep about result $result"))
      _ <- HubDb.db.upsertResultIO(result)
      maybeQuery <- HubDb.db.selectQueryIO(result.queryId)
      _ <- maybeQuery.map{ query =>
        ShrineMomClient.sendToNodeIO(
                                      query.id,
                                      envelopeContents = result,
                                      envelopeContentsCompanion = Result,
                                      logString = s"result update from ${result.adapterNodeName} for query ${result.queryId} ${result.status}",
                                      nodeId = query.nodeOfOriginId
                                        )
      }.getOrElse{
        Log.error(s"Query ${result.queryId} not found for $result")
        IO.unit
      }
    } yield ()
  }

  def updateResultIO(updateResult:UpdateResult):IO[Unit] = {
    HubDb.db.updateResultIO(updateResult).
      flatMap { maybeResult =>
        maybeResult.map { result =>
        HubDb.db.selectQueryIO(result.queryId).map{maybeQ:Option[Query] => maybeQ.flatMap((q: Query) => Some((q, result)))}
      }.getOrElse{
          Log.debug(s"Query not found for $updateResult. This is OK for queries QUEUED queries before Shrine 1.26.")
          IO(None)
        }}.
      flatMap { maybeQueryAndResult =>
        maybeQueryAndResult.map{ queryAndResult =>
          val query = queryAndResult._1
          val result = queryAndResult._2
          ShrineMomClient.sendToNodeIO(
            subjectId = result.queryId,
            envelopeContents = result,
            envelopeContentsCompanion = Result,
            logString = s"result update from ${result.adapterNodeName} for query ${result.queryId} ${result.status}",
            nodeId = query.nodeOfOriginId
          )
        }.getOrElse(IO.unit)
      }
  }

  def updateQueryAtHubAndAdaptersIO(updateQuery:UpdateQueryAtAdapter):IO[Unit] = {
    val updateQueryAndFindAdaptersIO: IO[Iterable[Node]] = for{
      _ <- HubDb.db.upsertQueryForUpdateIO(updateQuery)
      adapterNodes <- HubDb.db.selectAllNodesIO.map(_.map(_.get).filter(_.sendQueries))
    } yield adapterNodes
    updateQueryAndFindAdaptersIO.flatMap{ sendQueryUpdateToEveryAdapter(updateQuery,_)}.flatMap(_ => IO.unit)
  }

  val certCollection: BouncyKeyStoreCollection = BouncyKeyStoreCollection.fromConfig

  def sendQueryToAdaptersIO(runQueryAtHub: RunQueryAtHub,remainingAttempts:Int): IO[Boolean] = {

    val queryReceived:QueryProgress = runQueryAtHub.query.withStatus(QueryStatuses.ReceivedAtHub)
    val queryReceivedStateChange = UpdateQueryAtQepWithStatus(queryReceived)

    val queryReadyIO: IO[(Query, Seq[(Node,ResultProgress)])] =
      insertQueryResearcherAndTopic(queryReceived,runQueryAtHub.researcher,runQueryAtHub.topic).     //validate signature, maybe error the query
        flatMap{_ => IO{runQueryAtHub.query.verifySignature(certCollection).toTry.get}}.
        flatMap{ _ =>
        ShrineMomClient.sendToNodeIO(queryReceivedStateChange.queryId,queryReceivedStateChange,UpdateQueryAtQep,s"${queryReceivedStateChange.toString.take(70)} ",queryReceived.nodeOfOriginId)
      }.
      flatMap{ _ => chooseAdapterNodes}.
      flatMap{ adapterNodes => updateToReadyForAdapters(queryReceived,adapterNodes)}

    val sendToAdaptersIO = queryReadyIO.flatMap { queryReadyAndAdapterNodes =>
      val queryReady = queryReadyAndAdapterNodes._1
      val adapterNodesToResult = queryReadyAndAdapterNodes._2

      for {
        _ <- IO(Log.info(s"Will send query ${runQueryAtHub.query.id} to ${adapterNodesToResult.map(_._1.key.underlying).mkString(", ")}"))
        adaptersUsed <- sendToEveryAdapter(queryReady,adapterNodesToResult,runQueryAtHub)
        _ <- IO(Log.info(s"Sent query ${runQueryAtHub.query.id} to ${adaptersUsed.map(_.key.underlying).mkString(", ")}"))
        queryToAdapters <- IO(queryReady.withStatus(QueryStatuses.SentToAdapters))
        _ <- updateQueryState(queryToAdapters)
      } yield true
    }
    //todo worth examining what goes wrong here. If the whole works went off the rails then the query can error. Otherwise it is likely ok to let the results tell the story
    sendToAdaptersIO.handleErrorWith{
      case qaipx:QueryAlreadyInProcessException =>
        Log.debug(s"Query ${qaipx.ivrlx.item.id} is already in progress as a response to a previous message. Further processing bypassed.",qaipx)
        IO(true) //OK to ack. Most likely the Hub is running slow
      case x:SignatureNotVerifiedException =>
        val problem = CouldNotVerifySignature(x,runQueryAtHub.query.id)
        val updateQueryWithError = UpdateQueryAtQepWithError(runQueryAtHub.query.id,JsonProblemDigest(problem),DateStamp.now)
        updateQuery(updateQueryWithError,runQueryAtHub.query.nodeOfOriginId).
          flatMap(_ => IO(true))//OK to ack. No need to look at this message again
      case x:NoAdaptersRunQueriesException =>
        for {
          nodes <- HubDb.db.selectAllNodesIO
          problem = NoAdaptersRunQueriesProblem(x, queryReceived.id, nodes.filter(_.isSuccess).map(_.get).toList)
          updateQueryWithError = UpdateQueryAtQepWithError(runQueryAtHub.query.id,JsonProblemDigest(problem),DateStamp.now)
          _ <- updateQuery(updateQueryWithError,runQueryAtHub.query.nodeOfOriginId)
        } yield true //OK to ack. No need to look at this message again
      case NonFatal(x) => if(remainingAttempts > 0) {
        Log.warn(s"Exception while running query ${runQueryAtHub.query.id}. $remainingAttempts attempts remain", x)
        IO(false)
      } else { //if anything has gone wrong up to this point, the whole query should be an Error. No results are ever coming
        Log.error(s"Exception while running query ${runQueryAtHub.query.id} on last attempt. Will report the query in error", x)
        val problem = HubDispatchOfQueryProblem(x,queryReceived.id)
        val updateQueryWithError = UpdateQueryAtQepWithError(runQueryAtHub.query.id,JsonProblemDigest(problem),DateStamp.now)
        updateQuery(updateQueryWithError,runQueryAtHub.query.nodeOfOriginId).flatMap(_ => IO(true)) //OK to ack. Hub should not see it again anyway
      }
    }
  }

  def chooseAdapterNodes: IO[Seq[Node]] = {
    HubDb.db.selectAllNodesIO.map(_.map(_.get)).
      flatMap{nodes =>
        val adapterNodes = nodes.filter(_.sendQueries)
        if(adapterNodes.isEmpty) IO.raiseError(NoAdaptersRunQueriesException(nodes))
        else IO(adapterNodes.toSeq)
      }
  }

  def sendToEveryAdapter(queryReady:Query,nodesToResultsWithIds: Seq[(Node, ResultProgress)],runQueryAtHub: RunQueryAtHub):IO[List[Node]] = {

    val sendToEachAdapter: List[IO[Node]] = nodesToResultsWithIds.map { adapterNodeAndResultProgress =>

      val adapterNode = adapterNodeAndResultProgress._1
      val resultWithId = adapterNodeAndResultProgress._2
      val runQueryForResult = RunQueryForResult(
        query = queryReady,
        researcher = runQueryAtHub.researcher,
        node = adapterNode,
        topic = runQueryAtHub.topic,
        resultProgress = resultWithId.withStatus(ResultStatuses.SentToAdapter) //todo do this inside RunQueryForResult's companion
      )

      val sendAndUpdateIO = for {
        _ <- ShrineMomClient.sendToNodeIO(
                                            subjectId = runQueryForResult.query.id,
                                            envelopeContents = runQueryForResult,
                                            envelopeContentsCompanion = RunQueryForResult,
                                            logString = s"run query command for ${queryReady.id} to ${adapterNode.momQueueName}",
                                            queueName = adapterNode.momQueueName,
                                            boundedAttempts = Option(sendToAdaptersAttempts)
                                          )
        _ <- IO(Log.debug("Completed sendToNodeIO adapterNode"))
        _ <- updateResultIO(runQueryForResult.resultProgress)
      } yield adapterNode

      sendAndUpdateIO.handleErrorWith{
        case NonFatal(x) =>
          for{
            _ <- updateResultIO(resultWithId.toErrorBeforeCrc(
              HubDispatchOfQueryProblem(x,resultWithId.queryId,Some(adapterNode)))
            )
          } yield adapterNode
      }
    }.toList

    implicit val contextShift: ContextShift[IO] = IO.contextShift(ExecutionContexts.httpClientExecutionContext)

    import cats.implicits._
    sendToEachAdapter.parSequence
  }

  def sendQueryUpdateToEveryAdapter(updateQuery:UpdateQueryAtAdapter, adapterNodes:Iterable[Node]):IO[List[Node]] = {

    import cats.implicits._
    adapterNodes.map { adapterNode =>
    ShrineMomClient.sendToNodeIO(
          subjectId = updateQuery.queryId,
          envelopeContents = updateQuery,
          envelopeContentsCompanion = UpdateQueryAtAdapter,
          logString = s"Sent update query command for ${updateQuery.queryId} to ${adapterNode.momQueueName}",
          queueName = adapterNode.momQueueName
      ).map(_ => adapterNode)
    }.toList.sequence
  }

  def insertQueryResearcherAndTopic[Q <: Query](query:Q,reseacher:Researcher,topic:Topic): IO[Unit] = {
    HubDb.db.upsertQueryResearcherAndTopicIO(query,reseacher,topic).
      flatMap{ qrt: (Option[Query], Option[Researcher], Option[Topic]) =>
        //if no queries are inserted then the query already exists in the database and is already in process
        if(qrt._1.isEmpty) throw QueryAlreadyInProcessException(query.id)

        IO.unit
      }.handleErrorWith{
        //If this query has already been seen at the hub, be done and complete the message
        case ivrlx:ItemVersionRaceLostException[_,_,_]
          if ivrlx.item.isInstanceOf[Query] => //"if" clause to deal with erasure
          throw QueryAlreadyInProcessException(ivrlx.asInstanceOf[ItemVersionRaceLostException[QueryId,Query,QueryRow]])
    }
  }

  def updateToReadyForAdapters(queryReceived: Query,adapterNodes:Seq[Node]): IO[(Query, Seq[(Node, ResultProgress)])] = {
    //todo check previous state
    val queryReady = queryReceived.withStatus(QueryStatuses.ReadyForAdapters)
    val resultProgresses = adapterNodes.map{adapterNode =>
      Result.create(queryReady, adapterNode)
    }
    val update = UpdateQueryReadyForAdapters(queryReady,resultProgresses)

    HubDb.db.upsertQueryReadyForAdaptersIO(update).flatMap { _ =>
      ShrineMomClient.sendToNodeIO(
        update.queryId,
        update,
        UpdateQueryAtQep,
        s"$update",
        queryReady.nodeOfOriginId)
    }.flatMap{_ => IO((queryReady,adapterNodes.zip(update.resultProgresses)))}
  }

  def updateQueryState(query:Query): IO[Unit] = {
    val queryStateChanged = UpdateQueryAtQepWithStatus(query)
    updateQuery(queryStateChanged,query.nodeOfOriginId)
  }

  def updateQuery(updateQuery: UpdateQueryAtQep, nodeId:NodeId): IO[Unit] = {
    HubDb.db.upsertQueryForUpdateIO(updateQuery).
    flatMap{ _ =>
      ShrineMomClient.sendToNodeIO(
        updateQuery.queryId,
        updateQuery,
        UpdateQueryAtQep,
        s"$updateQuery",
        nodeId
      )
    }
  }

  case class NoAdaptersRunQueriesException(nodes:Iterable[Node])
    extends Exception(s"No adapters are configured to run queries of ${nodes.map(_.key).mkString(", ")} ")

  case class QueryAlreadyInProcessException(queryId: QueryId,ivrlx:ItemVersionRaceLostException[QueryId,Query,QueryRow])
    extends Exception(s"Query $queryId is already in the database, already in process",ivrlx)

  object QueryAlreadyInProcessException{
    def apply(queryId: QueryId): QueryAlreadyInProcessException =
      new QueryAlreadyInProcessException(queryId, null)

    def apply(ivrlx: ItemVersionRaceLostException[QueryId, Query, QueryRow]): QueryAlreadyInProcessException =
      new QueryAlreadyInProcessException(ivrlx.item.id, ivrlx)
  }
}

case class HubDispatchOfQueryProblem(x: Throwable, queryId: QueryId, adapter:Option[Node] = None) extends AbstractProblem(ProblemSources.Hub) {

  override val throwable = Some(x)
  override val summary: String = s"Shrine's hub's dispatcher encountered a problem during its last attempt to send a query to ${adapter.fold("the adapters"){node => s"adapter ${node.key}"}}."
  override val description: String =
    s"""HubDispatcher threw an exception while sending query $queryId to the adapters.""".stripMargin
}

case class NoAdaptersRunQueriesProblem(x: Throwable, queryId: QueryId, knownNodes: List[Node])
  extends AbstractProblem(ProblemSources.Hub) {

  override val summary: String = "No adapters are configured to receive queries."
  override val description: String =
    s"""No adapters are configured for Shrine's hub to send query $queryId to.""".stripMargin
  override def detailsText = Some(if (knownNodes.isEmpty) "The hub's collection of adapter nodes is empty."
    else "The following adapter nodes are known but not configured to receive queries:\n" + knownNodes.map(_.key).mkString(", "))
}