package net.shrine.hub

import cats.effect.IO
import fs2.Stream
import net.shrine.hub.data.store.{HubDatabaseNetworkNotFoundException, HubDb}
import net.shrine.log.Log
import net.shrine.messagequeueservice.{MessageQueueService, Queue}
import net.shrine.protocol.version.MomQueueName
import net.shrine.protocol.version.v1.{Network, Node}
import net.shrine.source.ConfigSource

object HubLifecycle {

  def initHubFromConfigIfEmptyIO():IO[Unit] = {
    HubDb.db.selectTheNetworkIO.attempt.flatMap {
      case Left(e: HubDatabaseNetworkNotFoundException) => {
        Log.debug(s"No Network record found. Building one from shrine.conf")

        //get the initial network from shrine.conf's hub section
        val initialConfig = ConfigSource.config.getConfig("shrine.hub.ifNoNetwork")

        //turn the config into network and node entries
        val network: Network = Network.networkFromConfig(initialConfig)
        HubDb.db.upsertNetworkIO(network).flatMap { _ =>
          HubDb.db.selectAllNodesIO.map(_.map(_.get.key)).map(_.toSet)
        }.flatMap { existingNodeKeys =>
          val nodes: Seq[Node] = Node.nodesFromConfig(initialConfig)
          val nodesToUpsert: Seq[Node] = nodes.filterNot(node => existingNodeKeys.contains(node.key))
          val nodeUpserts: Seq[IO[Unit]] = nodesToUpsert.map(node => HubDb.db.upsertNodeIO(node).flatMap(_ => IO.unit))
          nodeUpserts.foldLeft(IO.unit) { (a, b) => a.flatMap(_ => b) }
        }
      }
      case Left(x) => throw x
      case Right(network) => IO(Log.debug(s"${network.networkName} record found. Starting nodes from existing record"))
    }
  }

  def queuesFromDatabase(): IO[Unit] = {

    def createQueues(queues:Set[MomQueueName]): IO[Unit] = {
      queues.toSeq.
        map(_.underlying).
        map(MessageQueueService.service.createQueueIfAbsentIO).
        map(Stream.eval).
        foldLeft[Stream[IO,Queue]](Stream.empty){ (soFar: Stream[IO, Queue],next: Stream[IO,Queue]) =>
          soFar.append(next)
        }.compile.drain
    }

    def deleteQueues(queues:Set[MomQueueName]): IO[Unit] = {
      queues.toSeq.
        map(_.underlying).
        map(MessageQueueService.service.deleteQueueIO).
        map(Stream.eval).
        foldLeft[Stream[IO,Unit]](Stream.empty){ (soFar: Stream[IO, Unit],next: Stream[IO,Unit]) =>
          soFar.append(next)
        }.compile.drain
    }

    //no type annotations in for comprehensions. See https://github.com/oleg-py/better-monadic-for
    for {
      _ <- IO(Log.debug(s"Start queuesFromDatabase"))
      hubQueueName <- HubDb.db.selectTheNetworkIO.attempt.map{
        case Left(e: HubDatabaseNetworkNotFoundException) => None
        case Left(x) => throw x
        case Right(n) => Some(n.hubQueueName)
      }
      _ <- IO(Log.debug(s"hub queue name is $hubQueueName"))
      queueNamesForNodes <- HubDb.db.selectAllNodesIO.map(_.map(_.map(_.momQueueName).get).toList)
      neededQueues <- IO(queueNamesForNodes ++ hubQueueName).map(_.toSet)
      _ <- IO(Log.debug(s"neededQueues is $neededQueues"))
      existingQueues <- MessageQueueService.service.queuesIO.map(_.map(q => new MomQueueName(q.name))).map(_.toSet)
      _ <- IO(Log.debug(s"existingQueues are $existingQueues"))
      queuesToCreate <- IO(neededQueues -- existingQueues)
      queuesToDelete <- IO(existingQueues -- neededQueues)
      _ <- IO(Log.debug(s"Create queues: ${queuesToCreate.mkString(",")}"))
      _ <- createQueues(queuesToCreate)
      _ <- IO(Log.debug(s"Delete queues: ${queuesToDelete.mkString(",")}"))
      _ <- deleteQueues(queuesToDelete)
      _ <- IO(Log.debug(s"Completed queuesFromDatabase"))
    } yield()
  }
}