package net.shrine.hub

import cats.effect.{IO, Timer}
import net.shrine.config.ConfigExtensions
import net.shrine.http4s.catsio.{ExecutionContexts, IOScheduler}
import net.shrine.hub.data.store.HubDb
import net.shrine.problem.{AbstractProblem, ProblemSources}
import net.shrine.protocol.version.v1.{Node, Result, UpdateResultWithError}
import net.shrine.protocol.version.{DateStamp, NodeKey, QueryId, ResultStatuses}
import net.shrine.source.ConfigSource

import scala.concurrent.duration.FiniteDuration

/**
  *
  *
  * @author dwalend
 * @since v0.2.2
 */
object OldResultsPoller {

  val resultsOverdueAfter: FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.hub.result.resultsOverdueAfter")

  val overdueResultsCheckPeriod: FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.hub.result.overdueResultsCheckPeriod")

  implicit val timer: Timer[IO] = IO.timer(ExecutionContexts.crcExecutionContext)

  private val scheduler = IOScheduler.scheduleIndefinitely(
    initialDelay = overdueResultsCheckPeriod,
    interval = overdueResultsCheckPeriod,
    task = () => checkForOverdueResults(),
    name = "OldResultsPoller"
  )

  //call this when a query gets queued
  def restart(): IO[Unit] = scheduler.restart()

  def stop(): IO[Unit] = scheduler.stop()

  def checkForOverdueResults():IO[Unit] = {
    import cats.implicits._
    for {
      nodes <- HubDb.db.selectAllNodesIO
      overdueResults <- HubDb.db.selectOverdueResultsIO(new DateStamp(DateStamp.now.underlying - resultsOverdueAfter.toMillis))
      _ <- overdueResults.map(updateAsOverdue(_,nodes.map(_.get).toSeq)).toList.sequence
    } yield ()
  }

  def updateAsOverdue(result:Result,nodes:Seq[Node]):IO[Unit] = {
    val nodeIdsToKeys = nodes.map(n => n.id -> n.key).toMap
    val problem = QueryAttemptTimeToLiveExceeded(result.queryId,Some(nodeIdsToKeys(result.adapterNodeId)),resultsOverdueAfter)
    val overdueUpdate:UpdateResultWithError = UpdateResultWithError.create(result.queryId,nodeIdsToKeys(result.adapterNodeId),problem,ResultStatuses.ErrorInShrine,Some(problem.summary))
    HubDispatcher.updateResultIO(overdueUpdate)
  }

}


case class QueryAttemptTimeToLiveExceeded(queryId: QueryId,maybeNodeKey:Option[NodeKey],timeToLive:FiniteDuration) extends AbstractProblem(ProblemSources.Hub) {
  override val throwable: Option[Throwable] = None
  override val summary: String = maybeNodeKey.fold(s"Shrine's hub has given up on a query after $timeToLive")(key =>
    s"Shrine's hub has given up on getting a result from $key after $timeToLive")
  override val description: String = maybeNodeKey.fold(s"Shrine's hub has given up on $queryId after $timeToLive")(key =>
    s"Shrine's hub has given up on getting a result from $key for $queryId after $timeToLive")
  override def detailsText: None.type = None
}
