package net.shrine.hub.data.service

import cats.effect.IO
import net.shrine.http4s.auth.BasicOrBearerAuthMiddleware
import net.shrine.hub.data.store.{HubDatabaseNetworkNotFoundException, HubDb, ItemVersionRaceLostException}
import net.shrine.hub.{HubDispatcher, HubLifecycle}
import net.shrine.i2b2.protocol.pm.User
import net.shrine.log.Log
import net.shrine.messagequeueservice.MessageQueueService
import net.shrine.problem.{AbstractProblem, ProblemSources}
import net.shrine.protocol.version.v1.{Network, Node}
import net.shrine.protocol.version.{JsonText, MomQueueName, NodeId, NodeKey, NodeName, UserDomainName}
import org.http4s.dsl.impl.{+&, ->, /, :?, Auth, OptionalQueryParamDecoderMatcher, QueryParamDecoderMatcher}
import org.http4s.dsl.io.{Accepted, Conflict, GET, InternalServerError, NotFound, Ok, PUT, Root, http4sAcceptedSyntax, http4sConflictSyntax, http4sInternalServerErrorSyntax, http4sNotFoundSyntax, http4sOkSyntax}
import org.http4s.headers.`Content-Type`
import org.http4s.{AuthedRoutes, HttpRoutes, Request, Response}

import scala.util.control.NonFatal

object HubDataService extends Auth {

  val jsonContentHeader = `Content-Type`(org.http4s.MediaType.application.json)

  object NodeNameMatcher extends QueryParamDecoderMatcher[String]("nodeName")
  object UserDomainNameMatcher extends QueryParamDecoderMatcher[String]("userDomain")
  object AdminEmailMatcher extends QueryParamDecoderMatcher[String]("adminEmail")

  object OptionalNodeNameMatcher extends OptionalQueryParamDecoderMatcher[String]("nodeName")
  object OptionalUserDomainNameMatcher extends OptionalQueryParamDecoderMatcher[String]("userDomain")
  object OptionalMomQueueNameMatcher extends OptionalQueryParamDecoderMatcher[String]("momQueueName")
  object OptionalSendQueriesMatcher extends OptionalQueryParamDecoderMatcher[Boolean]("sendQueries")

  object OptionalNetworkNameMatcher extends OptionalQueryParamDecoderMatcher[String]("networkName")

  val notAuthedService: HttpRoutes[IO] = HttpRoutes.of[IO] {
    case _@GET -> Root / "ping" => Ok("pong")

    case request@GET -> Root / "nodes" =>
      HubDb.db.selectAllNodesIO.flatMap{tryNodes =>
        Ok(Node.nodesToJson(tryNodes.map(_.get)).underlying,jsonContentHeader)
      }.handleErrorWith {
        x: Throwable => handleGeneralFailure(request, x)
      }

    case request@GET -> Root / "node" / nodeKeyOrId =>
      val maybeNodeIO: IO[Option[Node]] = IO{new NodeId(nodeKeyOrId.toLong)}.flatMap(HubDb.db.selectNodeIO).
        handleErrorWith {
        case _:NumberFormatException =>
          val nodeKey = new NodeKey(nodeKeyOrId)
          HubDb.db.selectNodeByKeyIO(nodeKey)
      }
      maybeNodeIO.flatMap(_.fold(
        NotFound(s"No node with id or key of $nodeKeyOrId found")
      ){ node => Ok(node.asJsonText.underlying,jsonContentHeader)
      }).handleErrorWith {
        x: Throwable => handleGeneralFailure(request, x)
      }

    case request@GET -> Root / "network" =>
      useNetworkOrFail(request, (network: Network) => Ok(network.asJsonText.underlying,jsonContentHeader))
  }

  //noinspection TypeAnnotation because the type is big, scary, and not very interesting
  val authenticatedService = BasicOrBearerAuthMiddleware{
    AuthedRoutes.of[User, IO] {
      case _@GET -> Root / "authPing" as user => Ok(s"pong from $user")

      case request@PUT -> Root / "modifyNode" / nodeKey :?
        OptionalNodeNameMatcher(maybeNodeName) +&
          OptionalUserDomainNameMatcher(maybeUserDomain) +&
          OptionalMomQueueNameMatcher(maybeMomQueueName) +&
          OptionalSendQueriesMatcher(maybeSendQueries) as user =>
        Log.info(s"${user.username}:${user.domain} request to modify $nodeKey")
        HubDb.db.selectNodeByKeyIO(new NodeKey(nodeKey)).flatMap { maybeNode =>
          maybeNode.fold {
            NotFound(s"No node for $nodeKey")
          } { node =>
            val namedNode = maybeNodeName.fold(node) { n => node.copy(name = new NodeName(n)) }
            val domainNode = maybeUserDomain.fold(namedNode) { n => namedNode.copy(userDomainName = new UserDomainName(n)) }
            val queueNode = maybeMomQueueName.fold(domainNode) { n => domainNode.copy(momQueueName = new MomQueueName(n)) }
            val sendNode = maybeSendQueries.fold(queueNode) { b => queueNode.copy(sendQueries = b) }
            val newVersion = sendNode.copy(versionInfo = sendNode.versionInfo.next)

            HubDb.db.upsertNodeIO(newVersion).flatMap(_ => Accepted(newVersion.asJsonText.underlying, jsonContentHeader))
          }
        }.handleErrorWith {
          x: Throwable => handleGeneralFailure(request.req, x)
        }

      case request@PUT -> Root / "createNode" / nodeKey :?
        NodeNameMatcher(nodeName) +&
          UserDomainNameMatcher(userDomain) +&
          AdminEmailMatcher(adminEmail) +&
          OptionalMomQueueNameMatcher(maybeMomQueueName) +&
          OptionalSendQueriesMatcher(maybeSendQueries) as user =>
        Log.info(s"${user.username}:${user.domain} request to create $nodeKey")
        HubDb.db.selectNodeByKeyIO(new NodeKey(nodeKey)).flatMap{maybeNode =>
          maybeNode.fold{
            val node = Node.create(
              name = nodeName,
              key = nodeKey,
              userDomainName = userDomain,
              momQueueName = maybeMomQueueName,
              sendQueries = maybeSendQueries.getOrElse(true),
              adminEmail = adminEmail
            )
            HubDb.db.upsertNodeIO(node).
              flatMap(_ => MessageQueueService.service.createQueueIfAbsentIO(node.momQueueName.underlying)).
              flatMap(_ => Accepted(node.asJsonText.underlying,jsonContentHeader))
          }{
            node => Conflict(s"node ${node.key} already exists")
          }
        }.handleErrorWith {
          x: Throwable => handleGeneralFailure(request.req, x)
        }

      case request@PUT -> Root / "node" as _ =>
        val offeredNode: IO[Node] = request.req.as[String].map(string => Node.tryRead(new JsonText(string)).get)
        offeredNode.flatMap { offeredNode =>
          HubDb.db.upsertNodeIO(offeredNode).
            flatMap(_ => HubLifecycle.queuesFromDatabase()).
            flatMap(_ => IO(offeredNode))
        }.flatMap{ offeredNode =>
          Accepted(offeredNode.asJsonText.underlying,jsonContentHeader)
        }.handleErrorWith {
          case ivrlx: ItemVersionRaceLostException[_,_,_] => Conflict(ivrlx.getMessage)
          case x: Throwable => handleGeneralFailure(request.req, x)
        }

      case request@PUT -> Root / "network" as _ =>
        val offeredNetwork: IO[Network] = request.req.as[String].map(string => Network.tryRead(new JsonText(string)).get)
        offeredNetwork.flatMap { offeredNetwork =>
          HubDb.db.upsertNetworkIO(offeredNetwork).
            flatMap(_ => HubLifecycle.queuesFromDatabase()).
            flatMap(_ => IO(HubDispatcher.restart())).
            flatMap(_ => IO(offeredNetwork))
        }.flatMap{ offeredNetwork =>
          Accepted(offeredNetwork.asJsonText.underlying,jsonContentHeader)
        }.handleErrorWith {
          case ivrlx: ItemVersionRaceLostException[_,_,_] => Conflict(ivrlx.getMessage)
          case x: Throwable => handleGeneralFailure(request.req, x)
        }

      case request@PUT -> Root / "modifyNetwork" :?
        OptionalNetworkNameMatcher(maybeNetworkName) +&
          OptionalMomQueueNameMatcher(maybeMomQueueName) as _ =>
        useNetworkOrFail(request.req, (network: Network) => {
          val namedNetwork = maybeNetworkName.fold(network) { n => network.copy(networkName = n) }
          val queuedNetwork = maybeMomQueueName.fold(namedNetwork) { n => namedNetwork.copy(hubQueueName = new MomQueueName(n)) }
          val newVersion = queuedNetwork.copy(versionInfo = queuedNetwork.versionInfo.next)

          HubDb.db.upsertNetworkIO(newVersion).
            flatMap(_ => Accepted(newVersion.asJsonText.underlying, jsonContentHeader))
        })
    }
  }

  val service: HttpRoutes[IO] = {
    import cats.implicits._ //todo sort out the magical effects of this import. There's at least one implicit type conversion SHRINE-3272
    notAuthedService <+> authenticatedService //IDE does not like <+> , but it compiles just fine
  }

  private def useNetworkOrFail(request: Request[IO], func: Network => IO[Response[IO]]): IO[Response[IO]] = {
    HubDb.db.selectTheNetworkIO.attempt.flatMap {
      case Left(e: HubDatabaseNetworkNotFoundException) => NotFound(s"No network found. Shrine is misconfigured.")
      case Left(x) => handleGeneralFailure(request, x)
      case Right(network) => func(network)
    }
  }

  private def handleGeneralFailure(req: Request[IO], x: Throwable): IO[Response[IO]] = x match {
    case NonFatal(nfx) =>
      val path = req.pathInfo
      HubServerErrorProblem(nfx, path)
      InternalServerError(
        s"""${getClass.getSimpleName} threw an exception while trying to $path.
           |${getClass.getSimpleName} response: ${x.getMessage} Exception: ${nfx.getClass}""".
          stripMargin)
    case _ => throw x
  }
}

case class HubServerErrorProblem(x: Throwable, function: String) extends AbstractProblem(ProblemSources.Hub) {

  override val throwable = Some(x)
  override val summary: String = "SHRINE's hub service encountered a problem."
  override val description: String =
    s"""HubDataService throws an exception while trying to $function,
       |the server's response is: ${x.getMessage} from ${x.getClass}.""".stripMargin
}
