package net.shrine.hub.data.store

import java.sql.SQLException
import java.util.concurrent.TimeoutException

import cats.effect.IO
import com.typesafe.config.Config
import javax.sql.DataSource
import net.shrine.config.ConfigExtensions
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.log.Log
import net.shrine.problem.{AbstractProblem, ProblemSources}
import net.shrine.protocol.version.v1.{Network, Node, Researcher, Result, ResultProgress, Topic, UpdateQueryAtAdapter, UpdateQueryAtQep, UpdateQueryReadyForAdapters, UpdateResult, Versioned, Query => ShrineQuery}
import net.shrine.protocol.version.{DateStamp, Id, ItemVersion, JsonText, NetworkId, NodeId, NodeKey, ProtocolVersion, QueryId, ResearcherId, ResultId, TopicId}
import net.shrine.slick.{CouldNotRunDbIoActionException, TestableDataSourceCreator, TimeoutInDbIoActionException}
import net.shrine.source.ConfigSource
import slick.ast.BaseTypedType
import slick.jdbc.{JdbcProfile, JdbcType}

import scala.collection.immutable.Iterable
import scala.concurrent.duration.{Duration, DurationInt}
import scala.concurrent.{Await, Future, blocking}
import scala.language.postfixOps
import scala.util.{Failure, Try}
import scala.util.control.NonFatal
import scala.reflect.ClassTag

/**
  * Slick DB code for storing common currency data for shrine.
  *
  * @author david
  * @since 1.26
  */
case class HubDb(schemaDef:HubSchema, dataSource: DataSource, timeout:Duration) {
  import schemaDef._
  import schemaDef.jdbcProfile.api._
  import schemaDef.jdbcProfile.backend.DatabaseDef

  implicit val executionContext = ExecutionContexts.databaseExecutionContext

  val database:DatabaseDef = Database.forDataSource(dataSource, None)

  def createTables(): Unit = schemaDef.createTables(database)

  def dropTables(): Unit = schemaDef.dropTables(database)

  private def runTransaction[R](dbio: DBIOAction[R, NoStream, _]): Future[R] = {
    database.run(dbio.transactionally)
  }

  private val databaseContextShift = IO.contextShift(ExecutionContexts.databaseExecutionContext)

  //todo look into withTransactionIsolation
  private def runIO[R](dbio: DBIOAction[R, NoStream, _]): IO[R] = {
    databaseContextShift.evalOn(ExecutionContexts.databaseExecutionContext)(
      IO.fromFuture(IO(blocking(run(dbio))))
    )
  }

  private def runTransactionIO[R](dbio: DBIOAction[R, NoStream, _]): IO[R] = {
    databaseContextShift.evalOn(ExecutionContexts.databaseExecutionContext)(
      IO.fromFuture(IO(blocking(database.run(dbio.transactionally))))
    )
  }

  private def runTransactionBlocking[R](dbio: DBIOAction[R, NoStream, _], timeout: Duration = timeout): R = {
    try {
      Await.result(this.run(dbio.transactionally), timeout)
    } catch {
      case tx:TimeoutException => throw TimeoutInDbIoActionException(dataSource, timeout, tx)
      case NonFatal(x) => throw CouldNotRunDbIoActionException(dataSource, x)
    }
  }

  private def run[R](dbio: DBIOAction[R, NoStream, _]): Future[R] = {
    database.run(dbio)
  }

  private def runBlocking[R](dbio: DBIOAction[R, NoStream, _], timeout: Duration = timeout): R = {
    try {
      Await.result(this.run(dbio), timeout)
    } catch {
      case tx:TimeoutException => throw TimeoutInDbIoActionException(dataSource, timeout, tx)
      case NonFatal(x) => throw CouldNotRunDbIoActionException(dataSource, x)
    }
  }

  /**
    * Only for testing
    */
  private[store] def insertResultRow(resultRow:ResultRow):Int = {
    runBlocking(Results.allRows += resultRow)
  }

  def upsertResultIO(item: Result): IO[Option[Result]] = runTransactionIO(upsertItem(item,Results))

  def upsertQueryIO(item: ShrineQuery): IO[Option[ShrineQuery]] = runTransactionIO(upsertItem(item,Queries))

  //todo get rid of the rest of the blocking methods
  def upsertNodeBlocking(item: Node): Unit = runTransactionBlocking(upsertItem(item,Nodes))

  def upsertNodeIO(item: Node): IO[Option[Node]] = runTransactionIO(upsertItem(item,Nodes))

  def upsertNetworkBlocking(item: Network): Unit = runTransactionBlocking(upsertItem(item,Networks))

  def upsertNetworkIO(item: Network): IO[Option[Network]] = runTransactionIO(upsertItem(item,Networks))

  def upsertResearcherBlocking(item: Researcher): Unit = runTransactionBlocking(upsertItem(item,Researchers))

  def upsertResearcherIO(item: Researcher): IO[Option[Researcher]] = runTransactionIO(upsertItem(item,Researchers))

  def upsertTopicIO(item: Topic): IO[Option[Topic]] = runTransactionIO(upsertItem(item,Topics))

  def upsertQueryResearcherAndTopicIO(query:ShrineQuery, researcher: Researcher, topic: Topic): IO[(Option[ShrineQuery], Option[Researcher], Option[Topic])] = runTransactionIO{
    for{
      q <- upsertItem(query,Queries)
      r <- upsertItem(researcher,Researchers)
      t <- upsertItem(topic,Topics)
    } yield (q,r,t)
  }

  private type UpsertItemType[V] = DBIOAction[Option[V], NoStream, Effect.Read with Effect.Write with Effect.Read with Effect.Write]

  /**
    * @return Some(item) if the new row is inserted
    *         None if not, and any the existing rows with this id and version are the same
    *
    * @throws ItemVersionRaceLostException if the row to insert already exists
    *
    * @throws HubDatabaseAssertException if inserting one row somehow claims to have inserted more than one
    */
  private def upsertItem[
  I <: Id: ClassTag, //Needs the ClassTag to let I survive erasure
  R <: Row[I],
  V <: Versioned[I],
  Items <: ItemTable[I,R]
  ](
     item: V,
     tableCompanion: ItemTableCompanion[I,R,V,Items]
   ): UpsertItemType[V] = {

    val row = tableCompanion.itemToRow(item)
    implicit val IdColumnType: JdbcType[I] with BaseTypedType[I] = MappedColumnType.base[I, Long](
      id => id.underlying,
      long => tableCompanion.longToId(long)
    )

    //increment the table version and isolate this transaction
    incrementTableVersion(tableCompanion.tableName).flatMap { _ =>
      // Find all the rows with the same id and version number
      tableCompanion.allRows.filter(r => r.id === row.id).filter(r => r.itemVersion === row.itemVersion).result.flatMap {
        case Seq() => { // a row with this id and version does not exist yet
          //insert the row only if it is the only row with this id and version
          tableCompanion.allRows += row
        }.map {
          //expect insertCount to be 1
          case ic if ic == 1 => Some(item)
          case insertCount: Int => throw HubDatabaseAssertException(s"""insert into ${tableCompanion.tableName} returned $insertCount, not 1 as expected""")
        }
        case alreadyExists => // at least one row with this id and version does exist already
          // don't insert the row
          // if all the existing rows with this id and version are the same then this isn't a problem
          if (alreadyExists.forall(_ == row)) DBIO.successful(None)
          // However, this is a problem if the new row is different from the others
          // Due to the transaction this should only happen if some other row was written outside this method
          else DBIO.failed(ItemVersionRaceLostException[I, V, R](item, row, alreadyExists))
      }
    }
  }

  private[store] def selectAllResultsHistoryBlocking:Seq[Try[Result]] = {
    runBlocking(Results.allRows.sortBy(x => x.itemVersion.desc).result)
      .map{r => r.toResult}
  }

  /**
    * Only for testing
    */
  private[store] def selectAllResultRowsBlocking:Seq[ResultRow] = {
    runBlocking(Results.allRows.sortBy(x => x.changeDate.desc).result)
  }

  def selectMostRecentResultsForQueryIO(queryId: QueryId):IO[Iterable[Result]] = {
    runIO(Results.allRows.filter(r => r.queryId === queryId).result).map{x => x.groupBy(_.id)
      .map{ idAndResults: (ResultId, Seq[ResultRow]) =>
      idAndResults._2.maxBy(_.changeDate.underlying).toResult
    }.map(_.get)}
  }

  def selectOverdueResultsIO(lastUpdateBefore:DateStamp):IO[Iterable[Result]] = {
    runIO(Results.allRows.result).map{row =>    //all the rows
      row.groupBy(_.id)
      .map{ idAndResults: (ResultId, Seq[ResultRow]) =>
        idAndResults._2.maxBy(_.changeDate.underlying)}  //most recent versions
      .filter{r => r.changeDate.underlying < lastUpdateBefore.underlying} //too old
      .map(_.toResult.get)
        .filterNot(_.status.isFinal)
    }
  }

  def selectMostRecentResultsForQueryBlocking(queryId: QueryId):Seq[Try[Result]] = {
    selectMostRecentResultRowsForQuery(queryId).map{r => r.toResult}
  }

  private def selectMostRecentResultRowsForQuery(queryId: QueryId):Seq[ResultRow] = {
    //do the group by to grab the most recent values - here in code instead of in database
    //in practice this will be fewer than a thousand records, so should be very quick
    //it is possible to do all of this inside the database, but not terribly efficient.
    //See QepQueryDB for an example.
    runBlocking(Results.allRows.filter(r => r.queryId === queryId).result).
      groupBy(_.id).map{ idAndResults: (ResultId, Seq[ResultRow]) =>
        idAndResults._2.maxBy(_.changeDate.underlying)
    }.toSeq
  }

  private[store] def selectAllQueriesBlocking:Seq[Try[ShrineQuery]] = {
    runBlocking(Queries.allRows.result).
      groupBy(_.id).map{ idAndQueries: (QueryId, Seq[QueryRow]) =>
      idAndQueries._2.maxBy(_.itemVersion.underlying)
    }.toSeq.map(_.toQuery)
  }

  def selectLatestQueryByName(queryName: String):Seq[Try[ShrineQuery]] = {
    runBlocking(Queries.allRows.result).
      groupBy(_.id).map{ idAndQueries: (QueryId, Seq[QueryRow]) =>
      idAndQueries._2.maxBy(_.itemVersion.underlying)
    }.filter(_.jsonText.underlying.indexOf("ConnectivityTest_") != -1).take(1).toSeq.map(_.toQuery)
  }

  def selectQueryBlocking(id: QueryId):Try[ShrineQuery] = {
    val noneOrOneQueryRow: Seq[QueryRow] = runBlocking(Queries.allRows.filter(_.id === id).sortBy(_.itemVersion.desc).take(1).result)

    noneOrOneQueryRow match {
      case Seq(queryRow) => queryRow.toQuery
      case x => Failure(HubDatabaseAssertException(s"Expected exactly one query for id $id, selected ${x.size} $x"))
    }
  }

  def selectQueryIO(id: QueryId):IO[Option[ShrineQuery]] = {
    runIO(Queries.allRows.filter(_.id === id).sortBy(_.itemVersion.desc).take(1).result).map{
      case Seq(queryRow) => Some(queryRow.toQuery.get)
      case Seq() => None
      case x => throw HubDatabaseAssertException(s"Expected zero or one query for id $id, selected ${x.size} $x")
    }
  }

  def selectAllNodesBlocking:Seq[Try[Node]] = {
    runBlocking(Nodes.allRows.result).
      groupBy(_.id).map{ idAndNodes: (NodeId, Seq[NodeRow]) =>
      idAndNodes._2.maxBy(_.itemVersion.underlying)
    }.toSeq.map(_.toNode)
  }

  def selectAllNodesIO: IO[Iterable[Try[Node]]] = {
    runIO(Nodes.allRows.result).map{x => x.groupBy(_.id).map{idAndNodes: (NodeId,Seq[NodeRow]) =>
      idAndNodes._2.maxBy(_.itemVersion.underlying).toNode
    }}
  }

  def selectNodeByKeyIO(nodeKey:NodeKey):IO[Option[Node]] = {
    runIO(Nodes.allRows.filter(_.key === nodeKey).sortBy(_.itemVersion.desc).take(1).result).map{
      case Seq() => None
      case Seq(nodeRow) => Some(nodeRow.toNode.get)
      case x => throw HubDatabaseAssertException(s"Expected exactly one node for key $nodeKey, selected ${x.size} $x")
    }
  }

  def selectNodeIO(nodeId:NodeId): IO[Option[Node]] = {
    runIO(Nodes.allRows.filter(_.id === nodeId).sortBy(_.itemVersion.desc).take(1).result).map{
      case Seq() => None
      case Seq(nodeRow) => Some(nodeRow.toNode.get)
      case x => throw HubDatabaseAssertException(s"Expected exactly one node for id $nodeId, selected ${x.size} $x")
    }
  }

  def selectTheNetworkIO:IO[Network] = {
    runIO(Networks.allRows.sortBy(_.itemVersion.desc).take(1).result).map{
      case Seq() => throw HubDatabaseNetworkNotFoundException(s"Expected exactly one network, selected none")
      case Seq(network) => network.toNetwork.get
      case x => throw HubDatabaseAssertException(s"Expected exactly one network, selected ${x.size} $x")
    }
  }

  def selectResearcherByUsername (userName: String):Seq[Try[Researcher]] = {
    runBlocking(Researchers.allRows.result).
      groupBy(_.id).map{ idAndResearchers: (ResearcherId, Seq[ResearcherRow]) =>
      idAndResearchers._2.maxBy(_.itemVersion.underlying)
    }.toSeq.map(_.toResearcher).map( researcher => {
      researcher.filter(_.userName.underlying == userName)
    })
  }

  def selectResearcherByIdBlocking(researcherId: ResearcherId): Try[Researcher] = {
    runBlocking(Researchers.allRows.filter(_.id === researcherId).sortBy(_.itemVersion.desc).take(1).result.head).toResearcher
  }

  def selectResearcherByIdIO(researcherId: ResearcherId): IO[Researcher] = {
    runIO(Researchers.allRows.filter(_.id === researcherId).sortBy(_.itemVersion.desc).take(1).result.head).map(_.toResearcher.get)
  }

  def selectAllResearchersBlocking:Seq[Try[Researcher]] = {
    runBlocking(Researchers.allRows.result).
      groupBy(_.id).map{ idAndResearchers: (ResearcherId, Seq[ResearcherRow]) =>
      idAndResearchers._2.maxBy(_.itemVersion.underlying)
    }.toSeq.map(_.toResearcher)
  }

  def selectTopicByIdBlocking(topicId: TopicId) :Try[Topic] = {
    runBlocking(Topics.allRows.filter(_.id === topicId).sortBy(_.itemVersion.desc).take(1).result.head).toTopic
  }

  def selectTopicByIdIO(topicId: TopicId): IO[Topic] = {
    runIO(Topics.allRows.filter(_.id === topicId).sortBy(_.itemVersion.desc).take(1).result.head).map(_.toTopic.get)
  }

  def selectAllTopics:Seq[Try[Topic]] = {
    runBlocking(Topics.allRows.result).
      groupBy(_.id).map{ idAndTopics: (TopicId, Seq[TopicRow]) =>
      idAndTopics._2.maxBy(_.itemVersion.underlying)
    }.toSeq.map(_.toTopic)
  }

  def updateResultIO(updateResult:UpdateResult): IO[Option[Result]] = runTransactionIO{
    Nodes.allRows.filter(_.key === updateResult.adapterNodeKey).sortBy(_.itemVersion.desc).take(1).result.head.
      flatMap{adapterNode => Results.allRows.filter(_.queryId === updateResult.queryId).filter(_.adapterNodeId === adapterNode.id).sortBy(_.itemVersion.desc).take(1).result.headOption}.
      flatMap{ maybeOldResult =>
        maybeOldResult.map{ oldResult =>
          oldResult.toResult.get match {
            case oldResultProgress:ResultProgress =>
              if(!oldResultProgress.status.before(updateResult.status))
                upsertItem(updateResult.updatedResult(oldResultProgress),Results)
              else {
                Log.warn(s"$updateResult arrived after $oldResultProgress. Ignored.")
                DBIO.successful(None)
              }
            case finalResult:Result if finalResult.status.isFinal =>
              Log.warn(s"$updateResult arrived after $finalResult. Ignored.")
              DBIO.successful(None)
          }
        }.getOrElse{
          Log.warn(s"$updateResult for an unknown result. Ignored.")
          DBIO.successful(None)
        }
      }
  }

  // adding a type annotation - : IO[Option[ShrineQuery]] - fails compiling! todo Check after move to Scala 2.12
  def upsertQueryForUpdateIO(updateQuery:UpdateQueryAtAdapter) = runTransactionIO{

    Queries.allRows.
      filter(_.id === updateQuery.queryId).
      sortBy(_.itemVersion.desc).take(1).result.flatMap {
      case Seq(queryRow: QueryRow) =>
        val updatedQuery = updateQuery.updatedQuery(queryRow.toQuery.get)
        upsertItem(updatedQuery, Queries)
      case Seq() => DBIO.successful(None) //no query was found to update. The hub doesn't know anything about this query
    }
  }

  //todo maybe combine upsertQueryForUpdateIO and upsertQueryForUpdateIO
  def upsertQueryForUpdateIO(updateQuery:UpdateQueryAtQep) = runTransactionIO{

    Queries.allRows.
      filter(_.id === updateQuery.queryId).
      sortBy(_.itemVersion.desc).take(1).result.flatMap {
      case Seq(queryRow: QueryRow) =>
        val updatedQuery = updateQuery.updatedQuery(queryRow.toQuery.get)
        upsertItem(updatedQuery, Queries)
      case Seq() => DBIO.successful(None) //no query was found to update. The hub doesn't know anything about this query
    }
  }

  def upsertQueryReadyForAdaptersIO(updateQuery:UpdateQueryReadyForAdapters) = runTransactionIO{

    Queries.allRows.
      filter(_.id === updateQuery.queryId).
      sortBy(_.itemVersion.desc).take(1).result.flatMap {
      case Seq(queryRow: QueryRow) =>
        val updatedQuery = updateQuery.updatedQuery(queryRow.toQuery.get)
        upsertItem(updatedQuery, Queries)
      case Seq() => DBIO.successful(None) //no query was found to update. The hub doesn't know anything about this query
    }.flatMap{ _ =>
      DBIO.sequence(updateQuery.resultProgresses.map(upsertItem(_,Results)))
    }
  }

  //Reusable query parts
  private def incrementTableVersion(name:String) = {
    for {
      tvs: Seq[TableVersion] <- allTableVersions.filter(_.name === name).result
      updateOut: Int <- if(tvs.isEmpty){
                          allTableVersions += TableVersion(name,0)
                        } else {
                          allTableVersions.insertOrUpdate(tvs.head.next) }
    } yield {
      if(tvs.size > 1) throw IncrementTableVersionException(s"""Expected zero or one TableVersion named "$name", but found ${tvs.size} : ${tvs.mkString(", ")}""")
      if(updateOut != 1) throw IncrementTableVersionException(s"""Expected exactly one row of TableVersions to be updated, not $updateOut""")
      updateOut
    }
  }

  def insertConnectivityTestResultIO(connectivityTestResult: ConnectivityTestResult): IO[Unit] = {
    runIO(allConnectivityTestResults += connectivityTestResult).flatMap(_ => IO.unit)
  }

  def selectRecentConnectivityTestResultIO: IO[ConnectivityTestResult] = {
    runIO(allConnectivityTestResults.sortBy(_.changeDate.desc).take(1).result.head)
  }
}



object HubDb {

  val dataSource:DataSource = TestableDataSourceCreator.dataSource(HubSchema.config)
  //todo can I get rid of this with cats IO?
  val timeout: Duration = HubSchema.config.get("timeout",Duration.apply)

  val db = HubDb(HubSchema.schema,dataSource,timeout)

  val createTablesOnStart: Boolean = HubSchema.config.getBoolean("createTablesOnStart")
  if(createTablesOnStart) db.createTables()

}

/**
  * Separate class to support schema generation without actually connecting to the database.
  *
  * @param jdbcProfile Database profile to use for the schema
  */
case class HubSchema(jdbcProfile: JdbcProfile)  {
  import jdbcProfile.api._

  val allTableVersions = TableQuery[TableVersions]
  val allQueryRows = TableQuery[Queries]

  def ddlForAllTables: jdbcProfile.DDL = {
    allTableVersions.schema ++ 
      Results.allRows.schema ++ 
      Queries.allRows.schema ++
      Nodes.allRows.schema ++ 
      Networks.allRows.schema ++
      Researchers.allRows.schema ++
      Topics.allRows.schema ++
      allConnectivityTestResults.schema
  }

  //to get the schema, use the REPL
  //println(QepQuerySchema.schema.ddlForAllTables.createStatements.mkString(";\n"))

  def createTables(database:Database): Unit = {
    try {
      val future = database.run(ddlForAllTables.create)
      Await.result(future,10 seconds)
    } catch {
      //I'd prefer to check and create schema only if absent. No way to do that with Oracle.
      case x:SQLException => Log.info("Caught exception while creating tables. Recover by assuming the tables already exist.",x)
    }
  }

  def dropTables(database:Database): Unit = {
    val future = database.run(ddlForAllTables.drop)
    //Really wait forever for the cleanup
    Await.result(future,Duration.Inf)
  }

  implicit val ResultIdColumnType: JdbcType[ResultId] with BaseTypedType[ResultId] = MappedColumnType.base[ResultId, Long](
    resultId => resultId.underlying,
    long => new ResultId(long)
  )

  implicit val NodeIdColumnType: JdbcType[NodeId] with BaseTypedType[NodeId] = MappedColumnType.base[NodeId, Long](
    NodeId => NodeId.underlying,
    long => new NodeId(long)
  )

  implicit val NetworkIdColumnType: JdbcType[NetworkId] with BaseTypedType[NetworkId] = MappedColumnType.base[NetworkId, Long](
    NetworkId => NetworkId.underlying,
    long => new NetworkId(long)
  )

  implicit val ResearcherIdColumnType: JdbcType[ResearcherId] with BaseTypedType[ResearcherId] = MappedColumnType.base[ResearcherId, Long](
    ResearcherId => ResearcherId.underlying,
    long => new ResearcherId(long)
  )

  implicit val TopicIdColumnType: JdbcType[TopicId] with BaseTypedType[TopicId] = MappedColumnType.base[TopicId, Long](
    TopicId => TopicId.underlying,
    long => new TopicId(long)
  )

  implicit val QueryIdColumnType: JdbcType[QueryId] with BaseTypedType[QueryId] = MappedColumnType.base[QueryId, Long](
    QueryId => QueryId.underlying,
    long => new QueryId(long)
  )

  implicit val ItemVersionColumnType: JdbcType[ItemVersion] with BaseTypedType[ItemVersion] = MappedColumnType.base[ItemVersion, Int](
    ItemVersion => ItemVersion.underlying,
    int => new ItemVersion(int)
  )

  implicit val ProtocolVersionColumnType: JdbcType[ProtocolVersion] with BaseTypedType[ProtocolVersion] = MappedColumnType.base[ProtocolVersion, Int](
    ProtocolVersion => ProtocolVersion.underlying,
    v => new ProtocolVersion(v)
  )

  implicit val DateStampColumnType: JdbcType[DateStamp] with BaseTypedType[DateStamp] = MappedColumnType.base[DateStamp, Long](
    DateStamp => DateStamp.underlying,
    long => new DateStamp(long)
  )

  implicit val JsonTextColumnType: JdbcType[JsonText] with BaseTypedType[JsonText] = MappedColumnType.base[JsonText, String](
    JsonText => JsonText.underlying,
    text => new JsonText(text)
  )

  implicit val NodeKeyColumnType: JdbcType[NodeKey] with BaseTypedType[NodeKey] = MappedColumnType.base[NodeKey, String](
    NodeKey => NodeKey.underlying,
    text => new NodeKey(text)
  )

  class TableVersions(tag:Tag) extends Table[TableVersion](tag,"TABLE_VERSIONS") {
    def name = column[String]("NAME", O.PrimaryKey)
    def version = column[Int]("VERSION")

    def * = (name,version) <> ((TableVersion.apply _).tupled,TableVersion.unapply)
  }

  abstract class ItemTable[I <: Id,R <:Row[I]](tag:Tag,tableName:String,toId: Long => I) extends Table[R](tag,tableName) {
    def id:Rep[I]
    def protocolVersion = column[ProtocolVersion]("PROTOCOL_VERSION")
    def itemVersion = column[ItemVersion]("ITEM_VERSION")
    def createDate = column[DateStamp]("CREATE_DATE")
    def changeDate = column[DateStamp]("CHANGE_DATE")
    def jsonText = column[JsonText]("JSON")
  }

  /**
    * Gathers up all the specifics for storing this particular kind of common currency data.
    *
    * @tparam I The Id subtype to use
    * @tparam R The Row subtype to use
    * @tparam V The common currency data type
    * @tparam Items The table to store rows in
    */
  trait ItemTableCompanion[
    I <: Id,
    R <: Row[I],
    V <: Versioned[I],
    Items <: ItemTable[I,R]
  ] {
    def tableName:String
    def allRows:TableQuery[Items]
    def itemToRow: V => R
    def longToId: Long => I
  }

  class Results(tag:Tag) extends ItemTable[ResultId,ResultRow](tag,Results.tableName,new ResultId(_)) {
    def id = column[ResultId]("ID")
    def queryId = column[QueryId]("QUERY_ID")
    def adapterNodeId = column[NodeId]("NODE_ID")

    def * = (id,protocolVersion,itemVersion,createDate,changeDate,jsonText,queryId,adapterNodeId) <> ((ResultRow.apply _).tupled,ResultRow.unapply)
  }

  object Results extends ItemTableCompanion[ResultId,ResultRow,Result,Results]{
    override def tableName: String = "RESULTS"
    override def allRows:TableQuery[Results] = TableQuery[Results]
    override def itemToRow: Result => ResultRow = ResultRow.fromResult
    override def longToId: Long => ResultId = new ResultId(_)
  }

  class Queries(tag:Tag) extends ItemTable[QueryId,QueryRow](tag,Queries.tableName,new QueryId(_)) {
    def id = column[QueryId]("ID")
    def researcherId = column[ResearcherId]("RESEARCHER_ID")

    def * = (id,protocolVersion,itemVersion,createDate,changeDate,jsonText,researcherId) <> ((QueryRow.apply _).tupled,QueryRow.unapply)
  }

  object Queries extends ItemTableCompanion[QueryId,QueryRow,ShrineQuery,Queries]{
    override def tableName: String = "QUERIES"
    override def allRows:TableQuery[Queries] = TableQuery[Queries]
    override def itemToRow: ShrineQuery => QueryRow = QueryRow.fromQuery
    override def longToId: Long => QueryId = new QueryId(_)
  }

  class Nodes(tag:Tag) extends ItemTable[NodeId,NodeRow](tag,Nodes.tableName,new NodeId(_)) {
    def id = column[NodeId]("ID")
    def key = column[NodeKey]("KEY")

    def * = (id,protocolVersion,itemVersion,createDate,changeDate,jsonText,key) <> ((NodeRow.apply _).tupled,NodeRow.unapply)
  }

  object Nodes extends ItemTableCompanion[NodeId,NodeRow,Node,Nodes]{
    override def tableName: String = "NODES"
    override def allRows:TableQuery[Nodes] = TableQuery[Nodes]
    override def itemToRow: Node => NodeRow = NodeRow.fromNode
    override def longToId: Long => NodeId = new NodeId(_)
  }

  class Networks(tag:Tag) extends ItemTable[NetworkId,NetworkRow](tag,Networks.tableName,new NetworkId(_)) {
    def id = column[NetworkId]("ID")

    def * = (id,protocolVersion,itemVersion,createDate,changeDate,jsonText) <> ((NetworkRow.apply _).tupled,NetworkRow.unapply)
  }

  object Networks extends ItemTableCompanion[NetworkId,NetworkRow,Network,Networks]{
    override def tableName: String = "NETWORKS"
    override def allRows:TableQuery[Networks] = TableQuery[Networks]
    override def itemToRow: Network => NetworkRow = NetworkRow.fromNetwork
    override def longToId: Long => NetworkId = new NetworkId(_)
  }

  class Researchers(tag:Tag) extends ItemTable[ResearcherId,ResearcherRow](tag,Researchers.tableName,new ResearcherId(_)) {
    def id = column[ResearcherId]("ID")

    def * = (id,protocolVersion,itemVersion,createDate,changeDate,jsonText) <> ((ResearcherRow.apply _).tupled,ResearcherRow.unapply)
  }

  object Researchers extends ItemTableCompanion[ResearcherId,ResearcherRow,Researcher,Researchers]{
    override def tableName: String = "RESEARCHERS"
    override def allRows:TableQuery[Researchers] = TableQuery[Researchers]
    override def itemToRow: Researcher => ResearcherRow = ResearcherRow.fromResearcher
    override def longToId: Long => ResearcherId = new ResearcherId(_)
  }

  class Topics(tag:Tag) extends ItemTable[TopicId,TopicRow](tag,Topics.tableName,new TopicId(_)) {
    def id = column[TopicId]("ID")

    def * = (id,protocolVersion,itemVersion,createDate,changeDate,jsonText) <> ((TopicRow.apply _).tupled,TopicRow.unapply)
  }

  object Topics extends ItemTableCompanion[TopicId,TopicRow,Topic,Topics]{
    override def tableName: String = "TOPICS"
    override def allRows:TableQuery[Topics] = TableQuery[Topics]
    override def itemToRow: Topic => TopicRow = TopicRow.fromTopic
    override def longToId: Long => TopicId = new TopicId(_)
  }

  class ConnectivityTestResults(tag:Tag) extends Table[ConnectivityTestResult](tag,"CONNECTIVITY_TEST_RESULTS") {
    def networkQueryId = column[NetworkQueryId]("NETWORKQUERYID")
    def changeDate = column[Long]("CHANGE_DATE")

    def * = (networkQueryId, changeDate) <> (ConnectivityTestResult.tupled,ConnectivityTestResult.unapply)
  }

  val allConnectivityTestResults = TableQuery[ConnectivityTestResults]
}

object HubSchema {

  val allConfig:Config = ConfigSource.config
  val config:Config = allConfig.getConfig("shrine.queryEntryPoint.audit.database")

  val slickProfile:JdbcProfile = TestableDataSourceCreator.slickDriver(config)

  val schema = HubSchema(slickProfile)
}

case class TableVersion(
                       name:String,
                       count:Int
                       ) {
  def next: TableVersion = this.copy(count = this.count+1)
}

case class IncrementTableVersionException(message:String) extends Exception(message)

/**
  * The common trait for rows of data in the common currency
  */
trait Row[I <: Id] {
  def id:I
  def itemVersion:ItemVersion
}

case class ResultRow (
                      id:ResultId,
                      protocolVersion:ProtocolVersion,
                      itemVersion:ItemVersion,
                      createDate:DateStamp,
                      changeDate:DateStamp,
                      jsonText:JsonText,
                      queryId: QueryId,
                      adapterNodeId: NodeId
                    ) extends Row[ResultId] {

  def toResult: Try[Result] = Result.tryRead(jsonText)
}

object ResultRow {
  def fromResult(result: Result): ResultRow = ResultRow (
      id = result.id,
      protocolVersion = result.versionInfo.protocolVersion,
      itemVersion = result.versionInfo.itemVersion,
      createDate = result.versionInfo.createDate,
      changeDate = result.versionInfo.changeDate,
      jsonText = result.asJsonText,
      queryId = result.queryId,
      adapterNodeId = result.adapterNodeId
  )
}

case class QueryRow (
                       id:QueryId,
                       protocolVersion:ProtocolVersion,
                       itemVersion:ItemVersion,
                       createDate:DateStamp,
                       changeDate:DateStamp,
                       jsonText:JsonText,
                       researcherId: ResearcherId
                     ) extends Row[QueryId] {

  def toQuery: Try[ShrineQuery] = ShrineQuery.tryRead(jsonText)
}

object QueryRow {
  def fromQuery(query: ShrineQuery): QueryRow = QueryRow (
    id = query.id,
    protocolVersion = query.versionInfo.protocolVersion,
    itemVersion = query.versionInfo.itemVersion,
    createDate = query.versionInfo.createDate,
    changeDate = query.versionInfo.changeDate,
    jsonText = query.asJsonText,
    researcherId = query.researcherId
  )
}

case class NodeRow (
                       id:NodeId,
                       protocolVersion:ProtocolVersion,
                       itemVersion:ItemVersion,
                       createDate:DateStamp,
                       changeDate:DateStamp,
                       jsonText:JsonText,
                       key: NodeKey
                     ) extends Row[NodeId] {

  def toNode: Try[Node] = Node.tryRead(jsonText)
}

object NodeRow {
  def fromNode(result: Node): NodeRow = NodeRow (
    id = result.id,
    protocolVersion = result.versionInfo.protocolVersion,
    itemVersion = result.versionInfo.itemVersion,
    createDate = result.versionInfo.createDate,
    changeDate = result.versionInfo.changeDate,
    jsonText = result.asJsonText,
    key = result.key
  )
}

case class NetworkRow (
                     id:NetworkId,
                     protocolVersion:ProtocolVersion,
                     itemVersion:ItemVersion,
                     createDate:DateStamp,
                     changeDate:DateStamp,
                     jsonText:JsonText
                   ) extends Row[NetworkId] {

  def toNetwork: Try[Network] = Network.tryRead(jsonText)
}

object NetworkRow {
  def fromNetwork(result: Network): NetworkRow = NetworkRow (
    id = result.id,
    protocolVersion = result.versionInfo.protocolVersion,
    itemVersion = result.versionInfo.itemVersion,
    createDate = result.versionInfo.createDate,
    changeDate = result.versionInfo.changeDate,
    jsonText = result.asJsonText
  )
}

case class ResearcherRow (
                     id:ResearcherId,
                     protocolVersion:ProtocolVersion,
                     itemVersion:ItemVersion,
                     createDate:DateStamp,
                     changeDate:DateStamp,
                     jsonText:JsonText
                   ) extends Row[ResearcherId] {

  def toResearcher: Try[Researcher] = Researcher.tryRead(jsonText)
}

object ResearcherRow {
  def fromResearcher(result: Researcher): ResearcherRow = ResearcherRow (
    id = result.id,
    protocolVersion = result.versionInfo.protocolVersion,
    itemVersion = result.versionInfo.itemVersion,
    createDate = result.versionInfo.createDate,
    changeDate = result.versionInfo.changeDate,
    jsonText = result.asJsonText
  )
}

case class TopicRow (
                           id:TopicId,
                           protocolVersion:ProtocolVersion,
                           itemVersion:ItemVersion,
                           createDate:DateStamp,
                           changeDate:DateStamp,
                           jsonText:JsonText
                         ) extends Row[TopicId] {

  def toTopic: Try[Topic] = Topic.tryRead(jsonText)
}

object TopicRow {
  def fromTopic(result: Topic): TopicRow = TopicRow (
    id = result.id,
    protocolVersion = result.versionInfo.protocolVersion,
    itemVersion = result.versionInfo.itemVersion,
    createDate = result.versionInfo.createDate,
    changeDate = result.versionInfo.changeDate,
    jsonText = result.asJsonText
  )
}

case class ItemVersionRaceLostException[I <:Id,V <: Versioned[I],R <: Row[I]](item:V,row:R,alreadyExists:Seq[R])
  extends Exception(s"""$row was not inserted for $item. ${alreadyExists.size} rows exist with id of ${item.id}, version ${item.versionInfo.itemVersion} :\n ${alreadyExists.mkString("\n")}""")

case class HubDatabaseAssertException(message:String) extends Exception(message)
case class HubDatabaseNetworkNotFoundException(message:String) extends Exception(message)

case class ConnectivityTestResult(
                                   networkQueryId:NetworkQueryId,
                                   changeDate: Long
                                 )

object ConnectivityTestResult extends ((NetworkQueryId, Long) => ConnectivityTestResult) {

  def apply(networkQueryId: NetworkQueryId): ConnectivityTestResult = {
    new ConnectivityTestResult(
      networkQueryId = networkQueryId,
      changeDate = System.currentTimeMillis()
    )
  }
}