package net.shrine.hub.mom

import java.util.concurrent.TimeUnit

import cats.effect.{CancelToken, ContextShift, IO, Timer}
import com.typesafe.config.Config
import fs2.Stream
import net.shrine.http4s.catsio.{ExecutionContexts, RetryIO}
import net.shrine.hub.data.client.HubClient
import net.shrine.hub.data.store.HubDb
import net.shrine.log.Log
import net.shrine.messagequeueservice.{Message, MessageQueueService, Queue}
import net.shrine.problem.{AbstractProblem, ProblemSources}
import net.shrine.protocol.version.v1.Node
import net.shrine.protocol.version.{Envelope, EnvelopeContents, EnvelopeContentsCompanion, Id, JsonText, MomQueueName, NodeId, ProtocolVersion}
import net.shrine.source.ConfigSource

import scala.concurrent.duration.FiniteDuration
import scala.util.Try
import scala.util.control.NonFatal
import net.shrine.config.ConfigExtensions


/**
  * Common MOM operations in Shrine.
  */
object ShrineMomClient {

  val hubMomConfig: Config = ConfigSource.config.getConfig("shrine.hub.mom")
  val retryDelay: FiniteDuration = hubMomConfig.getFiniteDuration("retryDelay")

  implicit val timer: Timer[IO] = IO.timer(ExecutionContexts.longActionExecutionContext)
  private val retryIfTaskThrows: Throwable => Boolean = {case OkToRetry(_) => true}

  /**
    * Uses the hub's database directly to find the node. Only use from the hub.
    */
  private[hub] def sendToNodeIO(
                    subjectId: Id,
                    envelopeContents: EnvelopeContents,
                    envelopeContentsCompanion: EnvelopeContentsCompanion, //todo automatically grab the companion
                    logString: String,
                    nodeId: NodeId
                  ): IO[Unit] = {
    HubDb.db.selectNodeIO(nodeId).flatMap { node: Option[Node] =>
      sendIO(subjectId, envelopeContents, envelopeContentsCompanion, logString, node.get.momQueueName)
    }
  }

  /**
   * Uses the hub's database directly to find the node. Only use from the hub.
   */
  private[hub] def sendToNodeIO(
                                 subjectId: Id,
                                 envelopeContents: EnvelopeContents,
                                 envelopeContentsCompanion: EnvelopeContentsCompanion, //todo automatically grab the companion
                                 logString: String,
                                 queueName: MomQueueName,
                                 boundedAttempts: Option[Int] = None //None for unbounded
                               ): IO[Unit] = {
      sendIO(
        subjectId,
        envelopeContents,
        envelopeContentsCompanion,
        logString,
        queueName,
        boundedAttempts
      )
  }

  def sendToHubIO(
                   subjectId: Id,
                   envelopeContents: EnvelopeContents,
                   envelopeContentsCompanion: EnvelopeContentsCompanion, //todo automatically grab the companion
                   logString: String,
                   boundedAttempts: Option[Int] = None //None for unbounded
                 ): IO[Unit] = {
    //get the hub's Queue and send the result to the hub
    HubClient.getNetworkIO.flatMap { network =>
      sendIO(subjectId, envelopeContents, envelopeContentsCompanion, logString, network.hubQueueName, boundedAttempts)
    }
  }

  private def sendIO(
              subjectId: Id,
              envelopeContents: EnvelopeContents,
              envelopeContentsCompanion: EnvelopeContentsCompanion, //todo automatically grab the companion
              logString: String,
              queueName: MomQueueName,
              boundedAttempts: Option[Int] = None //None for unbounded
            ): IO[Unit] = {
    val envelope: Envelope = Envelope(envelopeContentsCompanion.envelopeType, subjectId.underlying, envelopeContents.asJsonText.underlying)

    val sIO: IO[Unit] = for {
      _ <- IO(Log.info(s"Will send $logString to $queueName"))
      queue <- MessageQueueService.service.getQueueIO(queueName.underlying)
      _ <- MessageQueueService.service.sendIO(envelope.asJsonText.underlying, queue)
      _ <- IO(Log.info(s"Sent $logString to $queueName"))
    } yield ()

    boundedAttempts.fold{
      RetryIO.keepTrying(task = sIO, delay = retryDelay, retryIfTaskThrows)
    }{attempts =>
      RetryIO.keepTryingBounded(task = sIO, delay = retryDelay, maxRetries = attempts, retryIfTaskThrows)
    }
  }

  def receiveUntilStop(queueNameIO: IO[MomQueueName], dispatch: (Envelope, Int) => IO[Boolean]): CancelToken[IO] = {

    val queueNameStream: Stream[IO, MomQueueName] = Stream.eval(IO.cancelBoundary.flatMap(_ => queueNameIO))

    val queueStream: Stream[IO, Queue] = queueNameStream.flatMap { queueName =>
      Log.info(s"Dispatcher for $queueName started")
      Stream.eval(MessageQueueService.service.getQueueIO(queueName.underlying))
    }
    val dispatchedStream: Stream[IO, Unit] = queueStream.flatMap { queue: Queue => MessageQueueService.service.receiveStream(queue) }.
      flatMap((message: Message) => openEnvelopeAndDispatch(message, dispatch))

    val streamToRun = RetryIO.keepTrying(task = dispatchedStream.compile.drain, delay = retryDelay, retryIfTaskThrows)

    streamToRun.unsafeRunCancelable {
      case Right(_) => Log.debug(s"End of $streamToRun")
      case Left(t) => Log.error("Exception while running dispatch", t)
    }
  }

  private def openEnvelopeAndDispatch(message: Message, dispatch: (Envelope, Int) => IO[Boolean]): Stream[IO, Unit] = {
    Log.debug(s"Message delivery id ${message.deliveryAttemptId} to be dispatched to $dispatch")

    val envelopeTry: Try[Envelope] = Envelope.tryRead(new JsonText(message.contents))

    val envelope = envelopeTry.get
    Log.debug(s"Envelope holds a ${envelope.protocolVersion} ${envelope.contentsType}")

    val dispatchIO: IO[Boolean] = envelope.protocolVersion match {
      case v: ProtocolVersion if v == net.shrine.protocol.version.v1.versionId => dispatch(envelope, message.remainingAttempts)
      case v: ProtocolVersion =>
        // handle the future case just like the current case after warning
        Log.warn(s"Envelope contents version $v is not recognized by this version of shrine. Attempting to read it with ${net.shrine.protocol.version.v1.versionId}")
        dispatch(envelope, message.remainingAttempts)
    }

    val longActionContextShift: ContextShift[IO] = IO.contextShift(ExecutionContexts.longActionExecutionContext)
    val timeLimit = FiniteDuration.apply(message.millisecondsToComplete, TimeUnit.MILLISECONDS)
    val shouldCompleteIO = dispatchIO.timeoutTo(timeLimit, {
      IO(Log.warn(s"openEnvelopeAndDispatch of message ${message.deliveryAttemptId} did not complete in $timeLimit")).flatMap(_ => IO(false))
    })(timer, longActionContextShift)

    val completeIO: IO[Unit] = shouldCompleteIO.flatMap { shouldComplete =>
      if (shouldComplete) message.completeIO().map(_ => Log.debug(s"message ${message.deliveryAttemptId} completed"))
      else IO(Log.debug(s"Message ${message.deliveryAttemptId} not completed"))
    }.handleErrorWith {
      case NonFatal(t) => IO {
        ExceptionDuringDispatch(t, envelope)
        true
      }
    }.flatMap(_ => IO.cancelBoundary)

    Stream.eval(completeIO)
  }
}

case class ExceptionDuringDispatch(x: Throwable, envelope: Envelope) extends AbstractProblem(ProblemSources.Commons) {

  override val throwable = Some(x)
  override val summary: String = s"Exception while dispatching ${envelope.briefString}"
  override val description: String = s"Exception while dispatching ${envelope.briefString}"
}