package net.shrine.hub.data.service

import cats.data.OptionT
import cats.effect.IO
import net.shrine.hub.data.store.HubDb
import net.shrine.protocol.version.{JsonText, NodeKey, NodeName, UserDomainName}
import net.shrine.protocol.version.v1.Node
import net.shrine.hub.data.client.HubServiceRequests
import org.http4s.dsl.io.{Accepted, Conflict, NotFound, Ok}
import org.http4s.headers.Authorization
import org.http4s.{BasicCredentials, Headers, Method, Request, Response, Uri}
import org.junit.{After, Before, Test}
import org.scalatestplus.junit.AssertionsForJUnit

import scala.collection.immutable.Seq
import scala.util.Try

class HubDataServiceTest extends AssertionsForJUnit {

  private val momBaseUri = Uri.unsafeFromString("")

  @Test
  def testPing():Unit = {
    val response: Response[IO] = requestResponse(Request(method = Method.GET, uri = Uri.unsafeFromString("/ping")))
    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("pong")(entityText)
  }

  @Test
  def testNoNodeForId():Unit = {
    val response: Response[IO] = requestResponse(Request(method = Method.GET, uri = Uri.unsafeFromString("/node/1234")))
    assertResult(NotFound)(response.status)
  }

  @Test
  def testNoNodeForKey():Unit = {
    val response: Response[IO] = requestResponse(Request(method = Method.GET, uri = Uri.unsafeFromString("/node/fakeKey")))
    assertResult(NotFound)(response.status)
  }

  @Test
  def testGetNodeForKey():Unit = {
    val expectedNode = Node(new NodeName("Test Node"),new NodeKey("testNode"),new UserDomainName("testnode.domain"), adminEmail= "email@foo")

    HubDb.db.upsertNodeBlocking(expectedNode)

    val response: Response[IO] = requestResponse(HubServiceRequests.getNodeRequest(momBaseUri,expectedNode.key))
    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    val maybeNode: Try[Node] = Node.tryRead(new JsonText(entityText))

    assertResult(maybeNode.get)(expectedNode)
  }

  @Test
  def testGetNodeForId():Unit = {
    val expectedNode = Node(new NodeName("Test Node"),new NodeKey("testNode"),new UserDomainName("testnode.domain"), adminEmail = "email@foo")

    HubDb.db.upsertNodeBlocking(expectedNode)

    val response: Response[IO] = requestResponse(HubServiceRequests.getNodeRequest(momBaseUri,expectedNode.id))
    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    val maybeNode: Try[Node] = Node.tryRead(new JsonText(entityText))

    assertResult(maybeNode.get)(expectedNode)
  }

  @Test
  def testGetNodes():Unit = {
    val expectedNode1 = Node(new NodeName("Test Node 1"),new NodeKey("testNode1"),new UserDomainName("testnode1.domain"), adminEmail = "email@foo")
    val expectedNode2 = Node(new NodeName("Test Node 2"),new NodeKey("testNode2"),new UserDomainName("testnode2.domain"), adminEmail = "email@foo")

    HubDb.db.upsertNodeBlocking(expectedNode2)
    HubDb.db.upsertNodeBlocking(expectedNode1)

    val response: Response[IO] = requestResponse(HubServiceRequests.getNodesRequest(momBaseUri))
    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    val nodes = Node.jsonToNodes(new JsonText(entityText))

    assertResult(Seq(expectedNode1,expectedNode2).toSet)(nodes.toSet)
  }

  @Test
  def testAuthPing():Unit = {
    val response: Response[IO] = requestResponse(Request(
      method = Method.GET,
      uri = Uri.unsafeFromString("/authPing"),
      headers = Headers.of(Authorization(BasicCredentials("ben","kapow")))
    ))
    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assert(entityText.startsWith("pong"))
  }

  @Test
  def testPutNodeForKey():Unit = {
    val expectedNode = Node(new NodeName("Test Node"),new NodeKey("testNode"),new UserDomainName("testnode.domain"), adminEmail = "email@foo")

    val response = requestResponse(HubServiceRequests.putNodeRequest(
      hubServiceUri = momBaseUri,
      username = "ben",
      password = "kapow",
      node = expectedNode)
    )
    assertResult(Accepted)(response.status)
    val entityText = response.as[String].unsafeRunSync
    val maybeNode: Try[Node] = Node.tryRead(new JsonText(entityText))
    assertResult(maybeNode.get)(expectedNode)

    val checkResponse: Response[IO] = requestResponse(HubServiceRequests.getNodeRequest(momBaseUri,expectedNode.key))
    assertResult(Ok)(checkResponse.status)
    val checkEntityText = checkResponse.as[String].unsafeRunSync
    val checkMaybeNode: Try[Node] = Node.tryRead(new JsonText(checkEntityText))
    assertResult(checkMaybeNode.get)(expectedNode)
  }

  @Test
  def testPutNodeForKeyTwiceDuplicate():Unit = {
    val expectedNode = Node(new NodeName("Test Node"),new NodeKey("testNode"),new UserDomainName("testnode.domain"), adminEmail = "email@foo")

    val response = requestResponse(HubServiceRequests.putNodeRequest(momBaseUri,"ben","kapow",expectedNode))
    assertResult(Accepted)(response.status)
    val entityText = response.as[String].unsafeRunSync
    val maybeNode: Try[Node] = Node.tryRead(new JsonText(entityText))
    assertResult(maybeNode.get)(expectedNode)

    val response2 = requestResponse(HubServiceRequests.putNodeRequest(momBaseUri,"ben","kapow",expectedNode))
    assertResult(Accepted)(response2.status)
    val entityText2 = response2.as[String].unsafeRunSync
    val maybeNode2: Try[Node] = Node.tryRead(new JsonText(entityText2))
    assertResult(maybeNode2.get)(expectedNode)

    val checkResponse: Response[IO] = requestResponse(HubServiceRequests.getNodeRequest(momBaseUri,expectedNode.key))
    assertResult(Ok)(checkResponse.status)
    val checkEntityText = checkResponse.as[String].unsafeRunSync
    val checkMaybeNode: Try[Node] = Node.tryRead(new JsonText(checkEntityText))
    assertResult(checkMaybeNode.get)(expectedNode)
  }

  @Test
  def testPutNodeForKeyTwiceBad():Unit = {
    val expectedNode = Node(new NodeName("Test Node"),new NodeKey("testNode"),new UserDomainName("testnode.domain"), adminEmail = "email@foo")

    val response = requestResponse(HubServiceRequests.putNodeRequest(momBaseUri,"ben","kapow",expectedNode))
    assertResult(Accepted)(response.status)
    val entityText = response.as[String].unsafeRunSync
    val maybeNode: Try[Node] = Node.tryRead(new JsonText(entityText))
    assertResult(maybeNode.get)(expectedNode)

    val badExpectedNode = expectedNode.copy(name = new NodeName("No longer test node"))

    val response2 = requestResponse(HubServiceRequests.putNodeRequest(momBaseUri,"ben","kapow",badExpectedNode))
    assertResult(Conflict)(response2.status)

    val checkResponse: Response[IO] = requestResponse(HubServiceRequests.getNodeRequest(momBaseUri,expectedNode.key))
    assertResult(Ok)(checkResponse.status)
    val checkEntityText = checkResponse.as[String].unsafeRunSync
    val checkMaybeNode: Try[Node] = Node.tryRead(new JsonText(checkEntityText))
    assertResult(checkMaybeNode.get)(expectedNode)
  }

  @Test
  def testPutNodeForKeyTwiceChanged():Unit = {
    val expectedNode = Node(new NodeName("Test Node"),new NodeKey("testNode"),new UserDomainName("testnode.domain"), adminEmail = "email@foo")

    val response = requestResponse(HubServiceRequests.putNodeRequest(momBaseUri,"ben","kapow",expectedNode))
    assertResult(Accepted)(response.status)
    val entityText = response.as[String].unsafeRunSync
    val maybeNode: Try[Node] = Node.tryRead(new JsonText(entityText))
    assertResult(maybeNode.get)(expectedNode)

    val goodExpectedNode = expectedNode.copy(
      name = new NodeName("No longer test node"),
      versionInfo = expectedNode.versionInfo.next
    )

    val response2 = requestResponse(HubServiceRequests.putNodeRequest(momBaseUri,"ben","kapow",goodExpectedNode))
    assertResult(Accepted)(response2.status)
    val entityText2 = response2.as[String].unsafeRunSync
    val maybeNode2: Try[Node] = Node.tryRead(new JsonText(entityText2))
    assertResult(maybeNode2.get)(goodExpectedNode)

    val checkResponse: Response[IO] = requestResponse(HubServiceRequests.getNodeRequest(momBaseUri,expectedNode.key))
    assertResult(Ok)(checkResponse.status)
    val checkEntityText = checkResponse.as[String].unsafeRunSync
    val checkMaybeNode: Try[Node] = Node.tryRead(new JsonText(checkEntityText))
    assertResult(checkMaybeNode.get)(goodExpectedNode)
  }

  @Test
  def testCreateNode(): Unit = {

    val response = requestResponse(HubServiceRequests.createNodeRequest(
      hubServiceUri = momBaseUri,
      key = "testNode",
      username = "ben",
      password = "kapow",
      name = "Test's Node.",
      userDomain = "test.node",
      adminEmail = "email@foo")
    )
    assertResult(Accepted)(response.status)

    val responseEntityText = response.as[String].unsafeRunSync
    val maybeNode: Try[Node] = Node.tryRead(new JsonText(responseEntityText))
    assertResult("testNode")(maybeNode.get.key.underlying)
    assertResult(true)(maybeNode.get.sendQueries)

    val checkResponse: Response[IO] = requestResponse(HubServiceRequests.getNodeRequest(momBaseUri,new NodeKey("testNode")))

    assertResult(Ok)(checkResponse.status)
    val checkEntityText = checkResponse.as[String].unsafeRunSync
    val checkMaybeNode: Try[Node] = Node.tryRead(new JsonText(checkEntityText))
    assertResult("testNode")(checkMaybeNode.get.key.underlying)
    assertResult(true)(checkMaybeNode.get.sendQueries)
  }


  @Test
  def testModifyNode(): Unit = {

    val response = requestResponse(HubServiceRequests.createNodeRequest(
      hubServiceUri = momBaseUri,
      key = "testNode",
      username = "ben",
      password = "kapow",
      name = "Test's Node.",
      userDomain = "test.node",
      adminEmail = "email@foo")
    )
    assertResult(Accepted)(response.status)

    val dots = "Lot's of dot's . . ."

    val modifyResponse = requestResponse(HubServiceRequests.modifyNodeRequest(
      hubServiceUri = momBaseUri,
      key = "testNode",
      username = "ben",
      password = "kapow",
      name = Some(dots),
      sendQueries = Some(false)
    ))

    val responseEntityText = modifyResponse.as[String].unsafeRunSync
    val maybeNode: Try[Node] = Node.tryRead(new JsonText(responseEntityText))
    assertResult("testNode")(maybeNode.get.key.underlying)
    assertResult(dots)(maybeNode.get.name.underlying)
    assertResult(false)(maybeNode.get.sendQueries)

    val checkResponse: Response[IO] = requestResponse(HubServiceRequests.getNodeRequest(momBaseUri,new NodeKey("testNode")))

    assertResult(Ok)(checkResponse.status)
    val checkEntityText = checkResponse.as[String].unsafeRunSync
    val checkMaybeNode: Try[Node] = Node.tryRead(new JsonText(checkEntityText))
    assertResult("testNode")(checkMaybeNode.get.key.underlying)
  }

  //todo refactor to a common test package
  private def requestResponse(request: Request[IO]): Response[IO] = {
    val responseOptionIo: OptionT[IO, Response[IO]] = HubDataService.service.run(request)

    val iorio: IO[Response[IO]] = responseOptionIo.map { r: Response[IO] => r }.fold {
      //todo is there some kind of exception or the like available?
      fail(s"No response from service for $request")
    } { r: Response[IO] => r
    }
    iorio.unsafeRunSync()
  }

  @Before
  def beforeEach(): Unit = {
    HubDb.db.createTables()
  }

  @After
  def afterEach(): Unit = {
    HubDb.db.dropTables()
  }

}
