package net.shrine.hub.data.store

import net.shrine.crypto.NewTestKeyStore
import net.shrine.problem.ProblemNotYetEncoded
import net.shrine.protocol.ResultOutputType
import net.shrine.protocol.version.v1.{Breakdowns, CrcResult, Network, Node, Query, Researcher, Result, ResultProgress}
import net.shrine.protocol.version.{DateStamp, ItemVersion, JsonText, MomQueueName, NodeId, NodeKey, NodeName, QueryId, ResearcherId, ResultId, ResultStatuses, TopicId, UserDomainName, UserName}
import org.junit.{After, Before, Test}
import org.scalatestplus.junit.AssertionsForJUnit

import scala.util.Try
import scala.collection.immutable.Seq

/**
  * @author david
  * @since 1.26
  */
class HubDbTest extends AssertionsForJUnit {

  val resultRow = ResultRow(
    id = new ResultId(0),
    protocolVersion = net.shrine.protocol.version.v1.versionId,
    itemVersion = new ItemVersion(1),
    createDate = DateStamp.now,
    changeDate = DateStamp.now,
    jsonText = new JsonText("Test Json Text"),
    queryId = new QueryId(2),
    adapterNodeId = new NodeId(3)
  )

  @Test
  def testInsertResultRow():Unit = {
    HubDb.db.insertResultRow(resultRow)

    assertResult(Seq(resultRow)){HubDb.db.selectAllResultRowsBlocking}
  }

  val resultInProgress = ResultProgress(
    queryId = new QueryId(2),
    adapterNodeId = new NodeId(3),
    adapterNodeName = new NodeName("test node"),
    status = ResultStatuses.ResultFromCRC,
    statusMessage = None
  )

  val crcResult: CrcResult = resultInProgress.toCrcResult(
    count = 2050,
    crcQueryInstanceId = 3L,
    resultType = ResultOutputType.PATIENT_COUNT_XML
  )

  @Test
  def testFirstInsertResult():Unit = {
    HubDb.db.upsertResultIO(crcResult).unsafeRunSync()

    assertResult(Seq(Try(crcResult))){HubDb.db.selectAllResultsHistoryBlocking}
  }

  @Test
  def testFirstInsertResultWithBreakdowns():Unit = {
    val crcResultWithBreakdowns: CrcResult = resultInProgress.toCrcResult(
      count = 2050,
      crcQueryInstanceId = 3L,
      resultType = ResultOutputType.PATIENT_COUNT_XML,
      breakdowns = Some(Breakdowns(Seq("colors" -> Seq(
        "red" -> 300,
        "blue" -> 600,
        "green" -> 30
      )))
    ))

    HubDb.db.upsertResultIO(crcResultWithBreakdowns).unsafeRunSync()

    assertResult(Seq(Try(crcResultWithBreakdowns))){HubDb.db.selectAllResultsHistoryBlocking}
  }

  @Test
  def testResultsRoundTrip():Unit = {
    val error = ResultProgress(
      queryId = new QueryId(2),
      adapterNodeId = new NodeId(3),
      adapterNodeName = new NodeName("test node"),
      status = ResultStatuses.ResultFromCRC,
      statusMessage = None
    ).toCrcError(ProblemNotYetEncoded("Darn"))
    HubDb.db.upsertResultIO(error).unsafeRunSync()

    HubDb.db.upsertResultIO(crcResult).unsafeRunSync()

    assertResult(Seq(Try(error),Try(crcResult))){HubDb.db.selectAllResultsHistoryBlocking}
  }

  @Test
  def testSubsequentInsertResult(): Unit = {
    HubDb.db.upsertResultIO(resultInProgress).unsafeRunSync()
    val nextResult = crcResult
    HubDb.db.upsertResultIO(nextResult).unsafeRunSync()

    assertResult(Seq(Try(nextResult),Try(resultInProgress))){HubDb.db.selectAllResultsHistoryBlocking}
  }

  @Test
  def testDuplicateSubsequentInsertResult(): Unit = {
    HubDb.db.upsertResultIO(resultInProgress).unsafeRunSync()
    val nextResult = crcResult
    HubDb.db.upsertResultIO(nextResult).unsafeRunSync()
    HubDb.db.upsertResultIO(nextResult).unsafeRunSync()

    assertResult(Seq(Try(nextResult),Try(resultInProgress))){HubDb.db.selectAllResultsHistoryBlocking}
  }

  @Test
  def testDuplicateFirstInsertResult(): Unit = {
    HubDb.db.upsertResultIO(resultInProgress).unsafeRunSync()
    HubDb.db.upsertResultIO(resultInProgress).unsafeRunSync()

    assertResult(Seq(Try(resultInProgress))){HubDb.db.selectAllResultsHistoryBlocking}
  }
  @Test
  def testContradictorySubsequentInsertResult(): Unit = {
    HubDb.db.upsertResultIO(resultInProgress).unsafeRunSync()
    val nextResult = crcResult
    val differentNextResult = resultInProgress.copy(statusMessage = Some("Breaking this on purpose"))
    HubDb.db.upsertResultIO(nextResult).unsafeRunSync()

    val caught = intercept[ItemVersionRaceLostException[ResultId,Result,ResultRow]] {
      HubDb.db.upsertResultIO(differentNextResult).unsafeRunSync()
    }

    assertResult(Seq(Try(nextResult),Try(resultInProgress))){HubDb.db.selectAllResultsHistoryBlocking}
  }

  @Test
  def testQueryRoundTrip():Unit = {
    val query = Query.create(
      queryDefinitionXml = "query xml",
      outputTypesXml = "output type xml",
      queryName = "test query",
      nodeOfOriginId = new NodeId(1),
      researcherId = new ResearcherId(2),
      topicId = new TopicId(3),
      projectName = "test project name",
      certCollection = NewTestKeyStore.certCollection
    )

    HubDb.db.upsertQueryIO(query).unsafeRunSync()

    assertResult(Seq(Try(query))){HubDb.db.selectAllQueriesBlocking}
  }

  @Test
  def testNodeRoundTrip():Unit = {
    val node = Node(
      name = new NodeName("testNodeName"),
      key = new NodeKey("testNodeKey"),
      userDomainName = new UserDomainName("testUserDomainName"),
      adminEmail = "email@foo"
    )

    HubDb.db.upsertNodeBlocking(node)

    assertResult(Seq(Try(node))){HubDb.db.selectAllNodesBlocking}
  }

  @Test
  def testSelectNode():Unit = {
    val node = Node(
      name = new NodeName("testNodeName"),
      key = new NodeKey("testNodeKey"),
      userDomainName = new UserDomainName("testUserDomainName"),
      adminEmail = "email@foo"
    )
    HubDb.db.upsertNodeBlocking(node)

    val nextNode = node.copy(versionInfo = node.versionInfo.next,sendQueries = false)
    HubDb.db.upsertNodeBlocking(nextNode)

    assertResult(Seq(Try(nextNode))){HubDb.db.selectAllNodesBlocking}
    assertResult(Some(nextNode)){HubDb.db.selectNodeIO(node.id).unsafeRunSync()}
    assertResult(Some(nextNode)){HubDb.db.selectNodeByKeyIO(node.key).unsafeRunSync()}

    assertResult(None){HubDb.db.selectNodeIO(new NodeId(0)).unsafeRunSync()}
    assertResult(None){HubDb.db.selectNodeByKeyIO(new NodeKey("notthere")).unsafeRunSync()}

  }

  @Test
  def testSelectNodeByKey():Unit = {
    val node = Node(
      name = new NodeName("testNodeName"),
      key = new NodeKey("testNodeKey"),
      userDomainName = new UserDomainName("testUserDomainName"),
        adminEmail = "email@foo"
    )
    HubDb.db.upsertNodeBlocking(node)

    val nextNode = node.copy(versionInfo = node.versionInfo.next,sendQueries = false)
    HubDb.db.upsertNodeBlocking(nextNode)

    assertResult(Seq(Try(nextNode))){HubDb.db.selectAllNodesBlocking}
    assertResult(Some(nextNode)){HubDb.db.selectNodeIO(node.id).unsafeRunSync()}

    assertResult(Some(nextNode)){HubDb.db.selectNodeByKeyIO(node.key).unsafeRunSync()}

    assertResult(None){HubDb.db.selectNodeIO(new NodeId(0)).unsafeRunSync()}
    assertResult(None){HubDb.db.selectNodeByKeyIO(new NodeKey("notthere")).unsafeRunSync()}

  }

  @Test
  def testNetworkRoundTrip():Unit = {
    val network = Network(
      networkName = "Test network",
      hubQueueName = new MomQueueName("testHub"),
      adminEmail = "yourname@example.com"
    )

    HubDb.db.upsertNetworkBlocking(network)

    assertResult(network){HubDb.db.selectTheNetworkIO.unsafeRunSync()}
  }

  @Test
  def testResearcherRoundTrip():Unit = {
    val researcher = Researcher(
      userName = new UserName("testResearcherName"),
      userDomainName = new UserDomainName("testUserDomainName"),
      nodeId = new NodeId(-14)
    )

    HubDb.db.upsertResearcherBlocking(researcher)

    assertResult(Seq(Try(researcher))){HubDb.db.selectAllResearchersBlocking}
  }

  //@Test
  def testInsertConnectivityTestResult():Unit = {
    val timestamp = System.currentTimeMillis()
    val connectivityTestResult = ConnectivityTestResult(1)
    HubDb.db.insertConnectivityTestResultIO(connectivityTestResult).unsafeRunSync()

    val actualConnectivtyTest = HubDb.db.selectRecentConnectivityTestResultIO
    assertResult(connectivityTestResult){actualConnectivtyTest}
  }

  @Before
  def beforeEach(): Unit = {
    HubDb.db.createTables()
  }

  @After
  def afterEach(): Unit = {
    HubDb.db.dropTables()
  }

}