package net.shrine.integration

import java.net.{URL, URLConnection, URLStreamHandler, URLStreamHandlerFactory}
import java.sql.SQLException

import net.shrine.adapter.AbstractQueryRetrievalTestCase.BogusRequest
import net.shrine.adapter._
import net.shrine.adapter.components.QueryNotInDatabase
import net.shrine.adapter.dao.BotDetectedException
import net.shrine.authentication.{NotAuthenticatedException, NotAuthenticatedProblem}
import net.shrine.authorization.{CouldNotInterpretResponseFromPmCell, CouldNotReachPmCell, MissingRequiredRoles}
import net.shrine.http4s.client.legacy.HttpResponse
import net.shrine.problem._
import net.shrine.protocol.QueryResult.StatusType
import net.shrine.protocol._
import net.shrine.protocol.query.QueryDefinition
import net.shrine.qep.querydb.QepDatabaseProblem
import org.junit.runner.RunWith
import org.scalatestplus.junit.JUnitRunner
import org.scalatest.{FlatSpec, Matchers}
import slick.jdbc.H2Profile.api._

import scala.concurrent.duration.FiniteDuration
import scala.xml.{Elem, NodeSeq, SAXParseException}

/**
  * Created by ty on 8/29/16.
  * Tests that we can successfully log every problem in the codebase.
  * Due to the time nature of logging problems, we create the succeed
  * early loop at the bottom to give every problem a chance at being
  * created on time.
  */
@RunWith(classOf[JUnitRunner])
class ProblemCreation extends FlatSpec with Matchers {
  val throwable = new IllegalArgumentException("Boo")
  val credential: Credential = Credential("string", isToken = true)
  val authInfo = AuthenticationInfo("domain", "username", credential)
  val bogus: ShrineRequest = new BogusRequest
  val seconds = new FiniteDuration(10, java.util.concurrent.TimeUnit.SECONDS)
  val queryDefinition = QueryDefinition("string", None)
  val runQueryRequest = new RunQueryRequest("id", seconds, authInfo, 10, None, Set(), queryDefinition)
  val saxxException: SAXParseException = new SAXParseException("hey", null)
  val xmlResponse: String = "<xmlResponse></xmlResponse>"
  val someXml: Elem = <details>Heyo!</details>
  val teapot: HttpResponse = HttpResponse(418, "body")
  val nodeId: XmlNodeName = XmlNodeName("name")
  val queryResult = QueryResult(5l, 5l, None, 5l, None, None, None, StatusType("name", isDone=false), None)
  val readyQueryResponse = ReadQueryResultResponse(5l, queryResult)
  val foo: NonI2b2ableResponse = new Foo()

  "Problems" should "all be successfully created and logged" in {

    URL.setURLStreamHandlerFactory(new BogusUrlFactory)
    val db = Problems.DatabaseConnector
    val queries = Problems.Queries
    db.runBlocking(db.IO.resetTable)
    val problemSize = () => db.runBlocking(queries.size.result)

    problemSize() shouldBe 0


    val problems: Seq[AbstractProblem] = Seq(
      QueryNotFound(10l),
      QueryResultNotAvailable(10l),
      CouldNotRetrieveQueryFromCrc(10l, throwable),
      CrcCouldNotBeInvoked("crcUrl", bogus, CrcInvocationException("url", bogus, throwable)),
      AdapterMappingProblem(AdapterMappingException(runQueryRequest, "message", throwable)),
      AdapterDatabaseProblem(new SQLException("reason", "state", 5)),
      BotDetected(BotDetectedException("domain", "user", 5l, 5l, 5l)),
      CannotParseXmlFromCrc(saxxException, xmlResponse),
      ExceptionWhileLoadingCrcResponse(throwable, xmlResponse),
      ErrorFromCrcBreakdown(ErrorFromCrcException("message")),
      CannotInterpretCrcBreakdownXml(MissingCrCXmlResultException(someXml, throwable)),
      QueryNotInDatabase(I2b2AdminReadQueryDefinitionRequest("project", seconds, authInfo, 5l)),
      // Difficult to test, as I would have to pull it out of the defining code,
      // Change it from an object to a case class, and make sure that I'm not breaking
      // Any breakdown logic by doing so.
      // BreakdownFailure,
      NotAuthenticatedProblem(NotAuthenticatedException("string", "string", "message", throwable)),
      MissingRequiredRoles("pid", Set(), authInfo),
      CouldNotReachPmCell("url", authInfo, throwable),
      CouldNotInterpretResponseFromPmCell("url", authInfo, teapot, throwable),
      ErrorStatusFromCrc(None, "<xml></xml>"),
      QepDatabaseProblem(throwable),
      ProblemNotYetEncoded("summary", Some(throwable)),
      NoI2b2AnalogExists(foo.getClass),
      TestProblem()

    )

    var count = 0
    // give it up to 1 second to finish
    while(problemSize() != problems.length && count < 20) {
      Thread.sleep(50)
      count+=1
    }

    problemSize() shouldBe problems.length
    db.runBlocking(queries.result) should contain theSameElementsAs problems.map(XmlProblemDigest.create(_))
    }

}

class Foo extends ShrineResponse with NonI2b2ableResponse {
  override def toXml: NodeSeq = <details>Yay</details>
}

class BogusUrlFactory extends URLStreamHandlerFactory {
  override def createURLStreamHandler(protocol: String): URLStreamHandler = new BogusUrlHandler
}

class BogusUrlHandler extends URLStreamHandler {
  override def openConnection(u: URL): URLConnection = new BogusUrlConnection(u)
}

class BogusUrlConnection(u: URL) extends URLConnection(u) {
  override def connect(): Unit = {}
}