package net.shrine.messagequeueclient

import cats.effect.IO
import com.typesafe.config.Config
import fs2.Stream
import net.shrine.config.ConfigExtensions
import net.shrine.http4s.client.Http4sHttpClient
import net.shrine.log.Log
import net.shrine.messagequeueservice._
import net.shrine.problem.{AbstractProblem, ProblemSources}
import net.shrine.source.ConfigSource
import org.http4s.{Request, Status, Uri}

import scala.collection.immutable.Seq
import scala.concurrent.duration.Duration
import scala.language.postfixOps
import scala.util.control.NonFatal

/**
  * A simple MessageQueueWebClient that uses MessageQueueWebApi to createQueue,
  * deleteQueue, sendMessage, receiveMessage, getQueues, and sendReceipt
  *
  * @author yifan
  * @since 8/10/17
  */
class MessageQueueWebClient extends MessageQueueService {

  val configPath = "shrine.messagequeue.blockingq"

  val webClientConfig: Config = ConfigSource.config.getConfig("shrine.messagequeue.blockingq")

  val webClientTimeOut: Duration = webClientConfig.get("webClientTimeOut", Duration(_))

  val momBaseUri: Uri = webClientConfig.get("serverUrl", Uri.unsafeFromString)
  
  val timeOutWaitGap: Long = ConfigSource.config.get("shrine.messagequeue.httpClient.timeOutWaitGap", Duration(_)).toMillis

  def filterErrorStatus(uri:Uri, status: Status, bodyString:String): Status = {
    if (status.isSuccess) status
    else {
      throw CouldNotCompleteMomTaskButOKToRetryException(uri, Some(status), Some(bodyString))
      /*
      Survivable status codes discovered during 2.0.0 development, stored for posterity

      //status codes we've observed but which aren't in http4s
      status.code match {
        case 598 => throw CouldNotCompleteMomTaskButOKToRetryException(uri, Some(status), Some(bodyString)) //NetworkReadTimeout
        case 599 => throw CouldNotCompleteMomTaskButOKToRetryException(uri, Some(status), Some(bodyString)) //NetworkConnectTimeout
        case _ => ; //pass through to better-known error codes
      }

      status.responseClass match {
        case x if x == Status.RequestTimeout.responseClass => throw CouldNotCompleteMomTaskButOKToRetryException(uri, Some(status), Some(bodyString))
        case x if x == Status.NotFound.responseClass => throw CouldNotCompleteMomTaskButOKToRetryException(uri, Some(status), Some(bodyString))
        case x if x == Status.InternalServerError.responseClass => throw CouldNotCompleteMomTaskButOKToRetryException(uri, Some(status), Some(bodyString))
        case x if x == Status.ServiceUnavailable.responseClass => throw CouldNotCompleteMomTaskButOKToRetryException(uri, Some(status), Some(bodyString))
        case _ => throw CouldNotCompleteMomTaskDoNotRetryException(uri, Some(status), Some(bodyString))
      }
    */
    }
  }

  def momFetchAndDecodeIO[A](request:Request[IO])(toA: (Status,String) => IO[A]): IO[A] = {
    Http4sHttpClient.webFetchAndDecodeIO(request)(toA).handleErrorWith {
      case NonFatal(x) => throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, None, None, Some(x))
/*
Exceptions discovered during 2.0.0 development, stored for posterity

      case tx:TimeoutException => throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, None, None, Some(tx))
      case cx: ConnectException => throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, None, None, Some(cx))
      case ex: EncoderException => throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, None, None, Some(ex))
      case ccx: ClosedChannelException => throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, None, None, Some(ccx))
      case ux:UnknownHostException if(ux.getMessage.contains("Temporary")) => //Found at UCLA with the message "Temporary failure in name resolution"
                                        throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, None, None, Some(ux))
      case ok: CouldNotCompleteMomTaskButOKToRetryException => throw ok
      case notOk: CouldNotCompleteMomTaskDoNotRetryException => throw notOk
      case jt:java.util.concurrent.TimeoutException => throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, None, None, Some(jt))
      case NonFatal(x) => throw CouldNotCompleteMomTaskDoNotRetryException(request.uri,None,None,Some(x))

 */
    }
  }

  def ping: IO[String] = {
    val request: Request[IO] = ShrineMessageRequests.pingRequest(momBaseUri)
    momFetchAndDecodeIO[String](request){ (status: Status, bodyString:String) =>
      filterErrorStatus(request.uri,status,bodyString)
      if(status.responseClass == Status.Ok.responseClass) {
        IO(bodyString)
      }
      else {
        Log.error(
          s"""Try to $request , HTTPResponse is a success but it does not contain an expected StatusCode
             |Expected StatusCodes: Ok, Actual StatusCode: $status
             |Body: $bodyString""".stripMargin)
        throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, Some(status), Some(bodyString))
      }
    }
  }

  override def createQueueIfAbsentIO(queueName: String): IO[Queue] = {
    val request: Request[IO] = ShrineMessageRequests.createQueueRequest(momBaseUri,queueName)
    momFetchAndDecodeIO[Queue](request){ (status: Status, bodyString:String) =>
      filterErrorStatus(request.uri,status,bodyString)
      if(status.responseClass == Status.Created.responseClass) {
        IO(
          try {
            Queue.fromJson(bodyString)
          } catch {
            case NonFatal(x) =>
              CouldNotInterpretHTTPResponseProblem(x, "create a Queue", queueName, bodyString)
              throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, Some(status), Some(bodyString),Some(x))
          }
        ).map{ queue =>
          Log.info(s"$queue created.")
          queue
        }
      }
      else {
        Log.error(
          s"""Try to create a queue, HTTPResponse is a success but it does not contain an expected StatusCode
             |Expected StatusCodes: Ok, Actual StatusCode: $status
             |Body: $bodyString""".stripMargin)
        throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, Some(status), Some(bodyString))
      }
    }
  }

  override def getQueueIO(queueName: String): IO[Queue] = {
    val request: Request[IO] = ShrineMessageRequests.getQueueRequest(momBaseUri,queueName)
    momFetchAndDecodeIO[Queue](request){ (status: Status, bodyString:String) =>
      filterErrorStatus(request.uri,status,bodyString)
      if(status.responseClass == Status.Ok.responseClass) {
        IO(
          try {
            Queue.fromJson(bodyString)
          } catch {
            case NonFatal(x) =>
              CouldNotInterpretHTTPResponseProblem(x, "get a Queue", queueName, bodyString)
              throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, Some(status), Some(bodyString),Some(x))
          }
        )
      }
      else {
        Log.error(
          s"""Try to get a queue, HTTPResponse is a success but it does not contain an expected StatusCode
             |Expected StatusCodes: Ok, Actual StatusCode: $status
             |Body: $bodyString""".stripMargin)
        throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, Some(status), Some(bodyString))
      }
    }
  }

  private val unit: Unit = ()
  override def deleteQueueIO(queueName: String): IO[Unit] = {
    val request: Request[IO] = ShrineMessageRequests.deleteQueueRequest(momBaseUri,queueName)
    momFetchAndDecodeIO[Unit](request){ (status: Status, bodyString:String) =>
      filterErrorStatus(request.uri,status,bodyString)
      if(status.responseClass == Status.Ok.responseClass) {
        IO(unit)
      }
      else {
        Log.error(
          s"""Try to delete a queue, HTTPResponse is a success but it does not contain an expected StatusCode
             |Expected StatusCodes: Ok, Actual StatusCode: $status
             |Body: $bodyString""".stripMargin)
        throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, Some(status), Some(bodyString))
      }
    }
  }

  override def queuesIO: IO[Seq[Queue]] = {
    val request: Request[IO] = ShrineMessageRequests.getQueuesRequest(momBaseUri)
    momFetchAndDecodeIO[Seq[Queue]](request){ (status: Status, bodyString:String) =>
      filterErrorStatus(request.uri,status,bodyString)
      if(status.responseClass == Status.Ok.responseClass) {
        IO(
          try {
           Queue.seqFromJson(bodyString)
          } catch {
            case NonFatal(x) =>
              CouldNotInterpretHTTPResponseProblem(x, "getQueues", "Any", bodyString)
              throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, Some(status), Some(bodyString), Some(x))
          }
        )
      }
      else {
        Log.error(
          s"""Try to get all queues, HTTPResponse is a success but it does not contain an expected StatusCode
             |Expected StatusCodes: Ok, Actual StatusCode: $status
             |Body: $bodyString""".stripMargin)
        throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, Some(status), Some(bodyString))
      }
    }
  }

  /*
  @deprecated //todo remove for SHRINE-2770, but keep for the retry if we need it

  override def sendFuture(contents: String, to: Queue): Future[Unit] = {
    val waitTimeBeforeResent = webClientConfig.get("waitTimeBeforeResent", Duration(_))

    def sendMessageOnce(): Future[Unit] = {
      Log.debug(s"send to $to '$contents'")
      val sendMessageUrl = s"$momUrl/sendMessage/${to.name}"
      val request: HttpRequest = HttpRequest(
        method = HttpMethods.PUT,
        uri = sendMessageUrl,
        entity = HttpEntity(contents) //todo set contents as XML or json SHRINE-2215
      )
      sprayWebApiFutureWithMOMErrorHandling(request, s"send to ${to.name}").transform({ s: HttpResponse =>
        if (s.status == SprayStatusCodes.Accepted) {
          Log.info(s"Successfully sent Message $contents to Queue $to")
          s
        } else {
          Log.debug(
            s"""Try to sendMessage $contents, HTTPResponse is a success but it does not contain an expected StatusCode
               |Expected StatusCodes: StatusCodes.Accepted, Actual StatusCodes: ${s.status}
               |Response: $s""".stripMargin)
          throw SprayCouldNotCompleteMomTaskDoNotRetryException(s"send a Message to ${to.name}", Some(s.status), Some(s.entity.asString))
        }
      }, { throwable: Throwable =>
        throwable match {
          case SprayCouldNotCompleteMomTaskDoNotRetryException(task, status, content, cause) => {
            if (status.get == SprayStatusCodes.UnprocessableEntity) {
              throw SprayCouldNotCompleteMomTaskButOKToRetryException(s"send a Message to ${to.name}", status, content, cause)
            }
          }
          case _ => //Don't touch
        }
        Log.error(s"Failed to send Message $contents to Queue $to due to Error: $throwable", throwable)
        throwable
      })
    }

    def keepGoing(lastAttempt: Future[Unit]): Try[Boolean] = {
      Try{ Await.result(lastAttempt, waitTimeBeforeResent) }.transform({ s: Unit =>
        Success(false)
      }, { throwable: Throwable =>
        throwable match {
          case okToRetry: SprayCouldNotCompleteMomTaskButOKToRetryException => Success(true)
          case tx:TimeoutException =>
            Log.debug(s"MessageQueueWebClient timed out attempting to send(). Will retry.")
            Success(true)
          case t: Throwable => Failure(t)
        }
      })
    }

    var retrySendResult: Future[Unit] = sendMessageOnce()
    while (keepGoing(retrySendResult).get) {
      Log.debug(s"Last attempt to send $contents to queue $to resulted in keepGoing = $retrySendResult. Sleeping $waitTimeBeforeResent before next attempt ")
      Thread.sleep(waitTimeBeforeResent.toMillis)
      retrySendResult = sendMessageOnce()
    }
    Log.info(s"Finishing send with result $retrySendResult")
    retrySendResult
  }
  */

  override def sendIO(contents: String, to: Queue): IO[Unit] = {

    val request: Request[IO] = ShrineMessageRequests.sendMessageRequest(momBaseUri,to,contents)

    Log.debug(s"send to $to  ${contents.take(70)}")

    momFetchAndDecodeIO[Unit](request){ (status: Status, bodyString:String) =>
      filterErrorStatus(request.uri,status,bodyString)
      if(status.responseClass == Status.Accepted.responseClass) {
        Log.info(s"Successfully sent Message ${contents.take(70)} to Queue $to")
        IO(unit)
      }
      else {
        Log.error(
          s"""Try to sendMessage $contents, HTTPResponse is a success but it does not contain an expected StatusCode
             |Expected StatusCodes: StatusCodes.Accepted, Actual StatusCode: $status
             |Body: $bodyString""".stripMargin)
        throw CouldNotCompleteMomTaskButOKToRetryException(request.uri, Some(status), Some(bodyString))
      }
    }
  }

  override def receiveIO(from: Queue, timeout: Duration): IO[Option[Message]] = {

    val request = ShrineMessageRequests.receiveMessageRequest(momBaseUri,from,timeout)
    momFetchAndDecodeIO[Option[Message]](request){ (status: Status, bodyString:String) =>
      filterErrorStatus(request.uri,status,bodyString)
      IO{messageOptionFromResponse(status,bodyString,request.uri).
        map(msg => MessageQueueClientMessage(msg.deliveryAttemptId, msg.millisecondsToComplete, msg.remainingAttempts, msg.contents))}
    }
  }

  override def receiveStream(from: Queue): Stream[IO, Message] = {
    //todo switch back for SHRINE-3025
    Stream.repeatEval(receiveIO(from,webClientTimeOut)).flatMap(om => Stream.emits(om.toList))

    /*
    val request = ShrineMessageProtocol.receiveMessageStreamRequest(momBaseUri,from)
    Http4sHttpClient.webApiStream(request){ response: Response[IO] =>
      response.body.through(ServerSentEvent.decoder[IO]).
        map{sse =>
          val jsonMessage = MessageAsJson.tryRead(sse.data).get //todo better error handling
          MessageQueueClientMessage(jsonMessage.deliveryAttemptId,jsonMessage.contents)
        }
    }*/
  }

  private def messageOptionFromResponse(status: Status, bodyString:String, uri:Uri): Option[MessageAsJson] = {
    if (status.code == Status.NoContent.code) {
      Log.info(s"No message received from $uri, HTTP Response $status $bodyString")
      None
    } else if (status.code == Status.Ok.code) {
      Log.info(s"Non-empty Message received from $uri")
      try {
        Some(MessageAsJson.tryRead(bodyString).get) //todo better error handling
      } catch {
          case NonFatal(x) =>
            CouldNotInterpretHTTPResponseProblem(x, "create a Message", uri.toString(), bodyString) //todo take a Uri
            throw x
        }
    } else {
      throw CouldNotCompleteMomTaskButOKToRetryException(uri, Some(status), Some(bodyString))
    }
  } //todo this might be able to survive a UnprocessableEntity code for a retry

  case class MessageQueueClientMessage private(
                                                deliveryAttemptId: DeliveryAttemptId,
                                                millisecondsToComplete:Long,
                                                remainingAttempts:Int,
                                                messageContent: String
                                              ) extends Message {

    override def contents: String = messageContent

    override def completeIO(): IO[Unit] = {

      //todo better to have the messageID as a parameter ??
      val request: Request[IO] = ShrineMessageRequests.completeMessageRequest(momBaseUri,deliveryAttemptId)

      momFetchAndDecodeIO[Unit](request) { (status: Status, bodyString:String) =>
        if (status.responseClass == Status.UnprocessableEntity.responseClass) {
          Log.info(s"Try to completeMessage $messageContent, but message does not exist, $contents")
        }
        filterErrorStatus(request.uri, status,bodyString)
          if (status.responseClass == Status.Ok.responseClass) {
            Log.info(s"Message ${this.deliveryAttemptId} completed with $status")
            IO(unit)
          }
          else {
            Log.error(
              s"""Try to completeMessage $deliveryAttemptId, HTTPResponse is a success but it does not contain the expected StatusCode
                 | Expected StatusCodes: Status.OK, Actual StatusCodes: $status

                 | Response: $bodyString""".
                stripMargin)
          throw CouldNotCompleteMomTaskButOKToRetryException(
            request.uri,
            status = Some(status),
            contents = Some(bodyString)
          )
        }
      }
    }
  }
}

//todo rethink how this gets constructed - object vs instance
object MessageQueueWebClient extends MessageQueueService {
  val it = new MessageQueueWebClient

  override def createQueueIfAbsentIO(queueName: String): IO[Queue] = it.createQueueIfAbsentIO(queueName)

  override def getQueueIO(queueName: String): IO[Queue] = it.getQueueIO(queueName)

  override def deleteQueueIO(queueName: String): IO[Unit] = it.deleteQueueIO(queueName)

  override def queuesIO: IO[Seq[Queue]] = it.queuesIO

  override def sendIO(contents: String, to: Queue): IO[Unit] = it.sendIO(contents,to)

  override def receiveIO(from: Queue, timeout: Duration): IO[Option[Message]] = it.receiveIO(from,timeout)

  override def receiveStream(from: Queue): Stream[IO, Message] = it.receiveStream(from)
}

case class CouldNotInterpretHTTPResponseProblem(x: Throwable, task: String, queueName: String, httpResponseString: String) extends AbstractProblem(ProblemSources.Hub) {

  override val throwable = Some(x)
  override val summary: String = s"Unable to $task due to exception"
  override val description: String = s"Unable to $task from queue $queueName due to exception $x, http response: $httpResponseString"
}
