package net.shrine.messagequeuemiddleware

import java.util.concurrent.TimeUnit

import cats.effect.IO
import fs2.Stream
import net.shrine.config.ConfigExtensions
import net.shrine.messagequeueservice.{Message, Queue}
import net.shrine.source.ConfigSource
import org.junit.runner.RunWith
import org.scalatest.concurrent.ScalaFutures
import org.scalatestplus.junit.JUnitRunner
import org.scalatest.time.{Millis, Seconds, Span}
import org.scalatest.{BeforeAndAfterAll, FlatSpec, Matchers}

import scala.collection.immutable.Seq
import scala.concurrent.duration.{Duration, DurationInt}
import scala.language.postfixOps
import scala.util.{Random, Try}

/**
  * Test create, delete queue, send, and receive message, getQueueNames, and acknowledge using MessageQueue service
  */
@RunWith(classOf[JUnitRunner])
class LocalMessageQueueMiddlewareTest extends FlatSpec with BeforeAndAfterAll with ScalaFutures with Matchers {

  val waitForQuickMethod: Duration = Duration("1 minute")
  implicit val defaultPatience: PatienceConfig = PatienceConfig(timeout = Span(5, Seconds), interval = Span(500, Millis))

  "BlockingQueue" should "be able to send and receive just one message" in {

    val configMap: Map[String, String] = Map("shrine.messagequeue.blockingq.messageTimeToLive" -> "7 seconds",
      "shrine.messagequeue.blockingq.messageRedeliveryDelay" -> "2 seconds")

    ConfigSource.atomicConfig.configForBlock(configMap, "LocalMessageQueueMiddlewareTest") {
      val queueName = "receiveOneMessage"

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)

      val queue = LocalMessageQueueMiddleware.createQueueIfAbsentIO(queueName).unsafeRunSync()

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync() == Seq(queue))

      val testContents = "Test message_receiveOneMessage"

      LocalMessageQueueMiddleware.sendIO(testContents, queue).unsafeRunTimed(waitForQuickMethod)

      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ messageOpt =>
        assert(messageOpt.isDefined)
        assert(messageOpt.get.contents == testContents)
        messageOpt.get.completeIO().unsafeRunSync()
      }.unsafeRunSync()

      // receive immediately again, should be no message redelivered yet
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ noMessage =>
        assert(noMessage.isEmpty)
      }.unsafeRunSync()


      LocalMessageQueueMiddleware.deleteQueueIO(queueName).unsafeRunSync()
      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
    }
  }

  "BlockingQueue" should "be able to send and receive a few messages" in {

    val queues: Seq[Queue] = LocalMessageQueueMiddleware.queuesIO.unsafeRunSync()
    queues.foreach({queue: Queue => LocalMessageQueueMiddleware.deleteQueueIO(queue.name).unsafeRunSync()})

    val queueName = "receiveAFewMessages"

    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)

    val queue = LocalMessageQueueMiddleware.createQueueIfAbsentIO(queueName).unsafeRunSync()

    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync() == Seq(queue))

    val testContents1 = "First test message_receiveAFewMessages"
    LocalMessageQueueMiddleware.sendIO(testContents1, queue).unsafeRunTimed(waitForQuickMethod)
    val message: Option[Message] = LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ message =>
      assert(message.isDefined)
      assert(message.get.contents == testContents1)
      message
    }.unsafeRunSync()

    message.get.completeIO().unsafeRunTimed(waitForQuickMethod)

    LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ shouldBeNoMessage =>
      assert(shouldBeNoMessage.isEmpty)
    }.unsafeRunSync()


    val testContents2 = "Second test message_receiveAFewMessages"
    LocalMessageQueueMiddleware.sendIO(testContents2, queue).unsafeRunTimed(waitForQuickMethod)

    val testContents3 = "Third test message_receiveAFewMessages"
    LocalMessageQueueMiddleware.sendIO(testContents3, queue).unsafeRunTimed(waitForQuickMethod)

    val message2: Option[Message] = LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ message2 =>
      assert(message2.isDefined)
      assert(message2.get.contents == testContents2)
      message2
    }.unsafeRunSync()

    message2.get.completeIO().unsafeRunTimed(waitForQuickMethod)

    val message3: Option[Message] = LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ message3 =>
      assert(message3.isDefined)
      assert(message3.get.contents == testContents3)
      message3
    }.unsafeRunSync()

    message3.get.completeIO().unsafeRunTimed(waitForQuickMethod)

    LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ shouldBeNoMessage4 =>
      assert(shouldBeNoMessage4.isEmpty)
    }.unsafeRunSync()

    LocalMessageQueueMiddleware.deleteQueueIO(queueName).unsafeRunSync()
    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
  }

  "BlockingQueue" should "be able to receive a few messages via streaming" in {

    val queues: Seq[Queue] = LocalMessageQueueMiddleware.queuesIO.unsafeRunSync()
    queues.foreach({queue: Queue => LocalMessageQueueMiddleware.deleteQueueIO(queue.name).unsafeRunSync()})

    val queueName = "streamAFewMessages"

    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)

    val queue = LocalMessageQueueMiddleware.createQueueIfAbsentIO(queueName).unsafeRunSync()

    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync() == Seq(queue))

    //put a message in the queue to receive before the stream first starts
    val testContents1 = "First test message_receiveAFewMessages"
    LocalMessageQueueMiddleware.sendIO(testContents1, queue).unsafeRunTimed(waitForQuickMethod)

    val messageStream: Stream[IO, Message] = LocalMessageQueueMiddleware.receiveStream(queue)

    val message: Message = messageStream.take(1).map{ message =>
      assert(message.contents == testContents1)
      message
    }.compile.toList.map(_.head).unsafeRunTimed(waitForQuickMethod).get

    message.completeIO().unsafeRunTimed(waitForQuickMethod)

    //send some messages after the IO Stream has been created
    val testContents2 = "Second test message_receiveAFewMessages"
    LocalMessageQueueMiddleware.sendIO(testContents2, queue).unsafeRunTimed(waitForQuickMethod)

    val testContents3 = "Third test message_receiveAFewMessages"
    LocalMessageQueueMiddleware.sendIO(testContents3, queue).unsafeRunTimed(waitForQuickMethod)

    val messages: List[Message] = messageStream.take(2).compile.toList.unsafeRunTimed(waitForQuickMethod).get
    assert(messages.size == 2)

    val message2: Message = messages.head
    assert(message2.contents == testContents2)
    message2.completeIO().unsafeRunTimed(waitForQuickMethod)

    val message3: Message = messages.tail.head
    assert(message3.contents == testContents3)
    message3.completeIO().unsafeRunTimed(waitForQuickMethod)

    LocalMessageQueueMiddleware.deleteQueueIO(queueName).unsafeRunSync()
    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
  }

  "BlockingQueue" should "be OK if asked to create the same queue twice " in {

    val queues: Seq[Queue] = LocalMessageQueueMiddleware.queuesIO.unsafeRunSync()
    queues.foreach({queue: Queue => LocalMessageQueueMiddleware.deleteQueueIO(queue.name).unsafeRunSync()})

    val queueName = "createSameQueueTwice"
    val queue = LocalMessageQueueMiddleware.createQueueIfAbsentIO(queueName).unsafeRunSync()
    val sameQueue = LocalMessageQueueMiddleware.createQueueIfAbsentIO(queueName).unsafeRunSync()

    assert(queue == sameQueue)
    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync() == Seq(sameQueue))

    val anotherSameQueue = LocalMessageQueueMiddleware.getQueueIO(queueName).unsafeRunSync()
    assert(queue == anotherSameQueue)
    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync() == Seq(sameQueue))

    LocalMessageQueueMiddleware.deleteQueueIO(queueName).unsafeRunSync()
    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
  }

  "BlockingQueue" should "return a failure if deleting a non-existing queue" in {

    val queues: Seq[Queue] = LocalMessageQueueMiddleware.queuesIO.unsafeRunSync()
    queues.foreach({queue: Queue => LocalMessageQueueMiddleware.deleteQueueIO(queue.name).unsafeRunSync()})

    val queueName = "DeletingNonExistingQueue"
    val deleteQueue = Try{LocalMessageQueueMiddleware.deleteQueueIO(queueName).unsafeRunSync()}
    assert(deleteQueue.isFailure)
    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
  }

  "BlockingQueue" should "return a failure if sending message to a non-existing queue" in {

    val queues: Seq[Queue] = LocalMessageQueueMiddleware.queuesIO.unsafeRunSync()
    queues.foreach({queue: Queue => LocalMessageQueueMiddleware.deleteQueueIO(queue.name).unsafeRunSync()})

    val queueName = "SendToNonExistingQueue"
    an [QueueDoesNotExistException] should be thrownBy {
      LocalMessageQueueMiddleware.sendIO("testContent", Queue(queueName)).unsafeRunTimed(waitForQuickMethod)
    }
    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
  }

  "BlockingQueue" should "return a failure if receiving a message to a non-existing queue" in {

    val queues: Seq[Queue] = LocalMessageQueueMiddleware.queuesIO.unsafeRunSync()
    queues.foreach({queue: Queue => LocalMessageQueueMiddleware.deleteQueueIO(queue.name).unsafeRunSync()})

    val queueName = "ReceiveFromNonExistingQueue"
    a [QueueDoesNotExistException] should be thrownBy {
      LocalMessageQueueMiddleware.receiveIO(Queue(queueName), 1 second).unsafeRunSync()
    }

    assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
  }

  "BlockingQueue" should "be able to filter the special characters in queue name" in {
    val queues: Seq[Queue] = LocalMessageQueueMiddleware.queuesIO.unsafeRunSync()
    queues.foreach({queue: Queue => LocalMessageQueueMiddleware.deleteQueueIO(queue.name).unsafeRunSync()})

    val queueName = "test# Qu%eueFilter"

    assert(Queue(queueName).name == "testQueueFilter")
  }

  "BlockingQueue" should "make sure if two InternalMessages are equal, then they have the same hashCode" in {

    val queues: Seq[Queue] = LocalMessageQueueMiddleware.queuesIO.unsafeRunSync()
    queues.foreach({queue: Queue => LocalMessageQueueMiddleware.deleteQueueIO(queue.name).unsafeRunSync()})

    val id: Long = Random.nextLong()
    val createdTime = System.currentTimeMillis()
    val message1: InternalMessage = InternalMessage(id, "message1", createdTime, Queue("to"), 3)
    val message2: InternalMessage = InternalMessage(id, "message1", createdTime, Queue("to"), 1)
    assert(message1 == message2)
    val message3: InternalMessage = InternalMessage(Random.nextLong(), "message3", createdTime, Queue("to"), 3)
    assert(message1 != message3)
  }

  "BlockingQueue" should "be able to send a message again and complete that if it isn't completed by redelivery time" in {

    val configMap: Map[String, String] = Map("shrine.messagequeue.blockingq.messageTimeToLive" -> "7 seconds",
      "shrine.messagequeue.blockingq.messageRedeliveryDelay" -> "2 seconds",
      "shrine.messagequeue.blockingq.messageMaxDeliveryAttempts" -> "4")

    ConfigSource.atomicConfig.configForBlock(configMap, "LocalMessageQueueMiddlewareTest") {
      val queueName = "receiveOneMessage"

      val messageRedeliveryDelay = ConfigSource.config.get("shrine.messagequeue.blockingq.messageRedeliveryDelay", Duration(_)).toMillis

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)

      val queue = LocalMessageQueueMiddleware.createQueueIfAbsentIO(queueName).unsafeRunSync()

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync() == Seq(queue))

      val testContents = "Test message_receiveOneMessage"

      LocalMessageQueueMiddleware.sendIO(testContents, queue).unsafeRunTimed(waitForQuickMethod)

      // first time receive
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ messageOpt =>
        assert(messageOpt.isDefined)
        assert(messageOpt.get.contents == testContents)
        messageOpt
      }.unsafeRunSync()

      // receive immediately again, should be no message redelivered yet
      // 1 second
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ noMessage =>
        assert(noMessage.isEmpty)
      }.unsafeRunSync()

      // receive after the redelivery delay, should have the redelivered message
      TimeUnit.MILLISECONDS.sleep(messageRedeliveryDelay + 1000)
      // > 2 seconds
      val secondAttemptMessage = LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ messageOpt =>
        assert(messageOpt.isDefined)
        assert(messageOpt.get.contents == testContents)
        messageOpt
      }.unsafeRunSync()
      secondAttemptMessage.get.completeIO().unsafeRunTimed(waitForQuickMethod)

      // receive after message is completed, should be no message
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ noMessage =>
        assert(noMessage.isEmpty)
      }.unsafeRunSync()

      LocalMessageQueueMiddleware.deleteQueueIO(queueName).unsafeRunSync()
      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
    }
  }

  "BlockingQueue" should "give up sending a message if too many attempts to send it were made" in {

    val configMap: Map[String, String] = Map("shrine.messagequeue.blockingq.messageTimeToLive" -> "7 seconds",
      "shrine.messagequeue.blockingq.messageRedeliveryDelay" -> "2 seconds",
      "shrine.messagequeue.blockingq.messageMaxDeliveryAttempts" -> "2")

    ConfigSource.atomicConfig.configForBlock(configMap, "LocalMessageQueueMiddlewareTest") {
      val queueName = "receiveOneMessage"

      //1 attempt
      val messageRedeliveryDelay = ConfigSource.config.get("shrine.messagequeue.blockingq.messageRedeliveryDelay", Duration(_)).toMillis

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)

      val queue = LocalMessageQueueMiddleware.createQueueIfAbsentIO(queueName).unsafeRunSync()

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync() == Seq(queue))

      val testContents = "Test message_receiveOneMessage"

      LocalMessageQueueMiddleware.sendIO(testContents, queue).unsafeRunTimed(waitForQuickMethod)

      // first time receive
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ messageOpt =>
        assert(messageOpt.isDefined)
        assert(messageOpt.get.contents == testContents)
        messageOpt
      }.unsafeRunSync()

      // receive immediately again, should be no message redelivered yet
      // 1 second
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ noMessage =>
        assert(noMessage.isEmpty)
      }.unsafeRunSync()

      // receive after the redelivery delay, should have the redelivered message
      TimeUnit.MILLISECONDS.sleep(messageRedeliveryDelay + 1000)
      // > 2 seconds
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ messageOpt =>
        assert(messageOpt.isDefined)
        assert(messageOpt.get.contents == testContents)
        messageOpt
      }.unsafeRunSync()

      // receive immediately again, should be no message redelivered yet
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ noMessage =>
        assert(noMessage.isEmpty)
      }.unsafeRunSync()

      // > 3 seconds

      // receive after the redelivery delay again, reached maxRedelivery attempt, should be no message
      TimeUnit.MILLISECONDS.sleep(messageRedeliveryDelay + 1000)
      // > 4 seconds
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ noMessage =>
        assert(noMessage.isEmpty)
      }.unsafeRunSync()

      LocalMessageQueueMiddleware.deleteQueueIO(queueName).unsafeRunSync()
      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
    }
  }

  "BlockingQueue" should "give up sending a message if it it lives past its expiration time" in {

    val configMap: Map[String, String] = Map("shrine.messagequeue.blockingq.messageTimeToLive" -> "2 seconds",
      "shrine.messagequeue.blockingq.messageRedeliveryDelay" -> "1 second")

    ConfigSource.atomicConfig.configForBlock(configMap, "LocalMessageQueueMiddlewareTest") {
      val queueName = "receiveOneMessage"

      //2 seconds
      val messageTimeToLive = ConfigSource.config.get("shrine.messagequeue.blockingq.messageTimeToLive", Duration(_)).toMillis

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)

      val queue = LocalMessageQueueMiddleware.createQueueIfAbsentIO(queueName).unsafeRunSync()

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync() == Seq(queue))

      val testContents = "Test message_receiveOneMessage"

      LocalMessageQueueMiddleware.sendIO(testContents, queue).unsafeRunTimed(waitForQuickMethod)

      // receive after the messageTimeToLive
      TimeUnit.MILLISECONDS.sleep(messageTimeToLive + 1000)
      // > 3 seconds
      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ noMessage =>
        assert(noMessage.isEmpty)
      }.unsafeRunSync()


      LocalMessageQueueMiddleware.deleteQueueIO(queueName).unsafeRunSync()
      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
    }
  }

  "BlockingQueue" should "give up sending a message if it it lives past its expiration time even if another message is pending" in {

    val configMap: Map[String, String] = Map("shrine.messagequeue.blockingq.messageTimeToLive" -> "2 seconds",
      "shrine.messagequeue.blockingq.messageRedeliveryDelay" -> "1 second")

    ConfigSource.atomicConfig.configForBlock(configMap, "LocalMessageQueueMiddlewareTest") {
      val queueName = "receiveOneAfterMissingAnotherMessage"

      //2 seconds
      val messageTimeToLive = ConfigSource.config.get("shrine.messagequeue.blockingq.messageTimeToLive", Duration(_)).toMillis

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)

      val queue = LocalMessageQueueMiddleware.createQueueIfAbsentIO(queueName).unsafeRunSync()

      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync() == Seq(queue))

      val testContents = "Test message_receiveOneMessage that should time out"

      LocalMessageQueueMiddleware.sendIO(testContents, queue).unsafeRunTimed(waitForQuickMethod)
      TimeUnit.MILLISECONDS.sleep(messageTimeToLive/2 )

      val testOtherContents = "Test message_receiveOneMessage that should arrive"
      LocalMessageQueueMiddleware.sendIO(testOtherContents, queue).unsafeRunTimed(waitForQuickMethod)
      TimeUnit.MILLISECONDS.sleep(messageTimeToLive/2 + 100)

      //receive only the second message sent
      LocalMessageQueueMiddleware.receiveIO(queue,1 second).map{ message: Option[Message] =>
        assert(message.isDefined)
        assert(message.exists(_.contents == testOtherContents))
      }.unsafeRunSync()

      LocalMessageQueueMiddleware.receiveIO(queue, 1 second).map{ noMessage =>
        assert(noMessage.isEmpty)
      }.unsafeRunSync()

      LocalMessageQueueMiddleware.deleteQueueIO(queueName).unsafeRunSync()
      assert(LocalMessageQueueMiddleware.queuesIO.unsafeRunSync().isEmpty)
    }
  }
}