package net.shrine.messagequeuemiddleware

import cats.data.OptionT
import cats.effect.IO
import fs2.{Pure, Stream}
import net.shrine.messagequeuemiddleware.LocalMessageQueueMiddleware.LocalMessage
import net.shrine.messagequeueservice.{MessageAsJson, Queue, ShrineMessageRequests}
import net.shrine.source.ConfigSource
import org.http4s.EntityDecoder.text
import org.http4s.dsl.io.{Accepted, Created, NoContent, NotFound, Ok}
import org.http4s.{Method, Request, Response, ServerSentEvent, Uri}
import org.junit.runner.RunWith
import org.scalatest.FlatSpec
import org.scalatestplus.junit.JUnitRunner

import scala.collection.immutable.Seq
import scala.concurrent.duration.Duration

/**
  * Test basic functions of MessageQueueWebApi
  * Created by yifan on 7/27/17.
  */


@RunWith(classOf[JUnitRunner])
class MessageQueueWebApiTest extends FlatSpec { //todo maybe switch to just Spec or RefSpec

  private val expectedQueue: Queue = Queue("testQueueInWebApi")
  private val messageContent = "test Content"

  private val momBaseUri = Uri.unsafeFromString("")

  //todo refactor to a common test package
  private def requestResponse(request: Request[IO]): Response[IO] = {
    val responseOptionIo: OptionT[IO, Response[IO]] = MessageQueueWebApi().service.run(request)

    val iorio: IO[Response[IO]] = responseOptionIo.map { r: Response[IO] => r }.fold {
      //todo is there some kind of exception or the like available?
      fail("No response from service")
    } { r: Response[IO] => r
    }
    iorio.unsafeRunSync()
  }

  "MessageQueueWebApi" should "reply to any request with NotFound if it is not enabled" in {
    val response: Response[IO] = requestResponse(Request(method = Method.GET, uri = Uri.unsafeFromString("/ping")))//iorio.unsafeRunSync()
    assertResult(NotFound)(response.status)
  }

  val configMap: Map[String, String] = Map(
    "shrine.messagequeue.blockingqWebApi.enabled" -> "true",
    "shrine.messagequeue.blockingq.messageTimeToLive" -> "4 days",
    "shrine.messagequeue.blockingq.messageRedeliveryDelay" -> "3 seconds",
    "shrine.messagequeue.blockingq.messageMaxDeliveryAttempts" -> "2"
  )

  "MessageQueueWebApi" should "reply to ping with pong" in {
    ConfigSource.atomicConfig.configForBlock(configMap, "MessageQueueWebApiTest") {
      val response: Response[IO] = requestResponse(Request(method = Method.GET, uri = Uri.unsafeFromString("/ping")))
      assertResult(Ok)(response.status)
      val entityText = response.as[String].unsafeRunSync
      assertResult("pong")(entityText)
    }
  }

  "MessageQueueWebApi" should "reply to parallel with the time waited in ms" in {
    ConfigSource.atomicConfig.configForBlock(configMap, "MessageQueueWebApiTest") {
      val response: Response[IO] = requestResponse(Request(method = Method.GET, uri = Uri.unsafeFromString("/parallel?timeOutSeconds=1")))
      assertResult(Ok)(response.status)
      val entityText = response.as[String].unsafeRunSync
      assert(entityText.startsWith("1"),s"$entityText should start with 1, but doesn't")
    }
  }

  "MessageQueueWebApi" should "reply to a request to an unknown url path with status NotFound" in {
    ConfigSource.atomicConfig.configForBlock(configMap, "MessageQueueWebApiTest") {
      val response: Response[IO] = requestResponse(Request(method = Method.GET, uri = Uri.unsafeFromString("/unknownPath")))
      assertResult(NotFound)(response.status)
    }
  }

  "MessageQueueWebApi" should "create, retrieve, and delete queues in" in {
    ConfigSource.atomicConfig.configForBlock(configMap, "MessageQueueWebApiTest") {

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.getQueuesRequest(momBaseUri))
        assertResult(Ok)(response.status)
        val bodyString = response.as[String].unsafeRunSync
        val queues = Queue.seqFromJson(bodyString)

        assertResult(Seq.empty)(queues)
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.createQueueRequest(momBaseUri,"testQueueInWebApi"))
        assertResult(Created)(response.status)
        val bodyString = response.as[String].unsafeRunSync
        val queue: Queue = Queue.fromJson(bodyString)

        assertResult(expectedQueue)(queue)
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.getQueuesRequest(momBaseUri))
        assertResult(Ok)(response.status)
        val bodyString = response.as[String].unsafeRunSync
        val queues = Queue.seqFromJson(bodyString)

        assertResult(Seq(expectedQueue))(queues)
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.getQueueRequest(momBaseUri,"testQueueInWebApi"))
        assertResult(Ok)(response.status)
        val bodyString = response.as[String].unsafeRunSync
        val queue: Queue = Queue.fromJson(bodyString)

        assertResult(expectedQueue)(queue)
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.deleteQueueRequest(momBaseUri,"testQueueInWebApi"))
        assertResult(Ok)(response.status)
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.getQueuesRequest(momBaseUri))
        assertResult(Ok)(response.status)
        val bodyString = response.as[String].unsafeRunSync
        val queues = Queue.seqFromJson(bodyString)

        assertResult(Seq.empty)(queues)
      }
    }
  }

  "MessageQueueWebApi" should "send, receive, and ack messages in" in {
    ConfigSource.atomicConfig.configForBlock(configMap, "MessageQueueWebApiTest") {

      val twoSeconds = Duration.create(2,"seconds")
      val queue = {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.createQueueRequest(momBaseUri,"testQueueInWebApi"))
        assertResult(Created)(response.status)
        val bodyString = response.as[String].unsafeRunSync
        val queue: Queue = Queue.fromJson(bodyString)

        assertResult(expectedQueue)(queue)
        queue
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.receiveMessageRequest(momBaseUri,queue,twoSeconds))
        assertResult(NoContent)(response.status)
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.sendMessageRequest(momBaseUri,queue,messageContent))
        assertResult(Accepted)(response.status)
      }

      val message = {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.receiveMessageRequest(momBaseUri,queue,twoSeconds))
        assertResult(Ok)(response.status)

        val entityText = response.as[String].unsafeRunSync
        val message: LocalMessage = MessageAsJson.tryRead(entityText).map(LocalMessage(_)).get
        assertResult(message.contents)(messageContent)

        message
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.completeMessageRequest(momBaseUri,message.deliveryAttemptId))
        assertResult(Ok)(response.status)
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.receiveMessageRequest(momBaseUri,queue,twoSeconds))
        assertResult(NoContent)(response.status)
      }

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.deleteQueueRequest(momBaseUri,queue.name))
        assertResult(Ok)(response.status)
      }
    }
  }

  "MessageQueueWebApi" should "send, receive via stream, and ack messages in" in {
    ConfigSource.atomicConfig.configForBlock(configMap, "MessageQueueWebApiTest") {

      val twoSeconds = Duration.create(2,"seconds")
      val queue = {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.createQueueRequest(momBaseUri,"testQueueInWebApi"))
        assertResult(Created)(response.status)
        val bodyString = response.as[String].unsafeRunSync
        val queue: Queue = Queue.fromJson(bodyString)

        assertResult(expectedQueue)(queue)
        queue
      }

      val stream: Stream[IO, LocalMessage] = {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.receiveMessageStreamRequest(momBaseUri,queue))
        assertResult(Ok)(response.status)

        response.body.
          through(ServerSentEvent.decoder).
          map{sse =>
            MessageAsJson.tryRead(sse.data).map(LocalMessage(_)).get
          }
      }

      val count = 10
      (1 to count).foreach{ i =>
        val response: Response[IO] = requestResponse(ShrineMessageRequests.sendMessageRequest(momBaseUri,queue,messageContent+i))
        assertResult(Accepted)(response.status)
      }

      stream.map{message =>
        assert(message.contents.startsWith(messageContent))

        val response: Response[IO] = requestResponse(ShrineMessageRequests.completeMessageRequest(momBaseUri,message.deliveryAttemptId))
        assertResult(Ok)(response.status)
      }.take(count).compile.drain.unsafeRunTimed(twoSeconds)

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.deleteQueueRequest(momBaseUri,queue.name))
        assertResult(Ok)(response.status)
      }
    }
  }

  "MessageQueueWebApi" should "send, receive via stream, and ack messages of very long strings" in {
    ConfigSource.atomicConfig.configForBlock(configMap, "MessageQueueWebApiTest") {

      val twoSeconds = Duration.create(2,"seconds")
      val queue = {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.createQueueRequest(momBaseUri,"testQueueInWebApi"))
        assertResult(Created)(response.status)
        val entityText = response.as[String].unsafeRunSync
        val queue: Queue = Queue.fromJson(entityText)

        assertResult(expectedQueue)(queue)
        queue
      }

      val stream: Stream[IO, LocalMessage] = {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.receiveMessageStreamRequest(momBaseUri,queue))
        assertResult(Ok)(response.status)

        response.body.
          through(ServerSentEvent.decoder).
          map{sse =>
            MessageAsJson.tryRead(sse.data).map(LocalMessage(_)).get
          }
      }

      val iters = 14
      val expectedMessages = (1 to iters).map{ i =>
        val number = scala.math.pow(2,i).toInt
        s"$i $number ${"*".*(number)}"
      }

      expectedMessages.foreach{ expected =>
        val response: Response[IO] = requestResponse(ShrineMessageRequests.sendMessageRequest(momBaseUri,queue,expected))
        assertResult(Accepted)(response.status)
      }

      val expectedMessageStream: Stream[Pure, String] = Stream.emits(expectedMessages)

      val zippedStream = expectedMessageStream.zip(stream)

      zippedStream.map{expectedAndMessage: (String, LocalMessage) =>
        val expected: String = expectedAndMessage._1
        val message = expectedAndMessage._2

        assertResult(expected.length)(message.contents.length)
        assertResult(expected)(message.contents)

        val response: Response[IO] = requestResponse(ShrineMessageRequests.completeMessageRequest(momBaseUri,message.deliveryAttemptId))
        assertResult(Ok)(response.status)
      }.take(iters).compile.drain.unsafeRunTimed(twoSeconds)

      {
        val response: Response[IO] = requestResponse(ShrineMessageRequests.deleteQueueRequest(momBaseUri,queue.name))
        assertResult(Ok)(response.status)
      }
    }
  }

}