package net.shrine.messagequeueservice

import cats.effect.IO
import fs2.Stream
import net.shrine.source.ConfigSource
import org.http4s.{Status, Uri}

import scala.collection.immutable.Seq
import scala.concurrent.duration.Duration

/**
  * This API mostly imitates AWS SQS' API via an embedded HornetQ. See http://docs.aws.amazon.com/sdk-for-java/v1/developer-guide/examples-sqs.html
  *
  * @author david
  * @since 7/18/17
  */
// in 1.23 all but the server side will use the client RemoteHornetQ implementation (which will call to the server at the hub)
//todo in 1.24, create an AwsSqs implementation of the trait

//todo rename with IO postfixes
trait MessageQueueService {
  /*
SQS returns CreateQueueResult, which contains queueUrl: String
 */
  def createQueueIfAbsentIO(queueName:String): IO[Queue]

  /*
SQS has getQueue, which contains queueUrl: String
 */
  def getQueueIO(queueName:String):IO[Queue]

  /*
  SQS takes in DeleteMessageRequest, which contains a queueUrl: String and a ReceiptHandle: String
  returns a DeleteMessageResult, toString for debugging
   */
  def deleteQueueIO(queueName:String): IO[Unit]

  /*
Returns the names of the queues created on this server.
 */
  def queuesIO: IO[Seq[Queue]]

  /*
SQS sendMessage(String queueUrl, String messageBody) => SendMessageResult
   */
  def sendIO(contents:String, to:Queue): IO[Unit]

  /*
SQS ReceiveMessageResult receiveMessage(String queueUrl)
   */
  def receiveIO(from:Queue, timeout:Duration): IO[Option[Message]]

  /*
SQS I don't think has a streaming equivalent in their Java API so use something like
    Stream.repeatEval(receive(from,webClientTimeOut)).flatMap(om => Stream.emits(om.toList))
   */
  def receiveStream(from:Queue):Stream[IO,Message]

  // SQS has DeleteMessageResult deleteMessage(String queueUrl, String receiptHandle) instead of our Message.complete

}

object MessageQueueService {

  lazy val service:MessageQueueService = {
    import scala.reflect.runtime.universe.runtimeMirror

    val momClassName = ConfigSource.config.getString("shrine.messagequeue.implementation")
    val classLoaderMirror = runtimeMirror(getClass.getClassLoader)
    val module = classLoaderMirror.staticModule(momClassName)

    classLoaderMirror.reflectModule(module).instance.asInstanceOf[MessageQueueService]
  }
}

case class Queue(var name:String) {
  // filter all (Unicode) characters that are not letters
  // filter neither letters nor (decimal) digits, replaceAll("[^\\p{L}]+", "")
  name = name.filterNot((c:Char) => c.isWhitespace).
    replaceAll("[^\\p{L}\\p{Nd}]+", "").
    filterNot(_ == '.') //And your little . , too
  if (name.length == 0) {
    throw new IllegalArgumentException("ERROR: A valid Queue name must contain at least one letter!")
  }

  def asJsonString:String = {
    import io.circe.generic.extras.Configuration
    import io.circe.generic.extras.auto.exportEncoder
    import io.circe.syntax.EncoderOps

    implicit val genDevConfig: Configuration = Queue.genDevConfig

    this.asJson.noSpaces
  }

}

object Queue {

  import io.circe.generic.extras.Configuration
  val genDevConfig: Configuration =
    Configuration.default.withDiscriminator("encodedClass")

  def fromJson(jsonString:String):Queue = {
    import io.circe.parser.decode
    import io.circe.generic.extras.auto.exportDecoder

    implicit val genDevConfig: Configuration = Queue.genDevConfig

    decode[Queue](jsonString) match {
      case Right(envelope) => envelope
      case Left(x) => throw x
    }
  }

  def seqFromJson(jsonString:String):Seq[Queue] = {
    import io.circe.parser.decode
    import io.circe.generic.extras.auto.exportDecoder

    implicit val genDevConfig: Configuration = Queue.genDevConfig

    decode[Seq[Queue]](jsonString) match {
      case Right(envelope) => envelope
      case Left(x) => throw x
    }
  }

}

case class CouldNotCompleteMomTaskButOKToRetryException(uri:Uri,
                                                         status:Option[Status] = None,
                                                         contents:Option[String] = None,
                                                         cause:Option[Throwable] = None
                                                        ) extends Exception(s"Could not $uri due to status code $status with message '$contents'",cause.orNull)

case class CouldNotCompleteMomTaskDoNotRetryException(uri:Uri,
                                                       status:Option[Status] = None,
                                                       contents:Option[String] = None,
                                                       cause:Option[Throwable] = None
                                                      ) extends Exception(s"Could not $uri due to status code $status with message '$contents' ",cause.orNull)
