package net.shrine.messagequeueservice

import cats.effect.IO
import org.http4s.{Method, Request, Uri}

import scala.concurrent.duration.Duration
import scala.util.{Random, Try}

/**
  * Container for common parts for moving messages around in Shrine's custom message service
  */
object ShrineMessageRequests {

  def pingRequest(momBaseUri: Uri): Request[IO] = Request(
    method = Method.GET,
    uri = momBaseUri.withPath(s"${momBaseUri.path}/ping")
  )

  def createQueueRequest(momBaseUri: Uri,queueName:String): Request[IO] = Request(
    method = Method.PUT,
    uri = momBaseUri.withPath(s"${momBaseUri.path}/createQueue/$queueName")
  )

  def getQueueRequest(momBaseUri: Uri,queueName:String): Request[IO] = Request(
    method = Method.GET,
    uri = momBaseUri.withPath(s"${momBaseUri.path}/getQueue/$queueName")
  )

  def deleteQueueRequest(momBaseUri: Uri,queueName:String): Request[IO] = Request(
    method = Method.PUT,
    uri = momBaseUri.withPath(s"${momBaseUri.path}/deleteQueue/$queueName")
  )

  def getQueuesRequest(momBaseUri: Uri): Request[IO] = Request(
    method = Method.GET,
    uri = momBaseUri.withPath(s"${momBaseUri.path}/getQueues")
  )

  def sendMessageRequest(momBaseUri: Uri,to:Queue,contents:String): Request[IO] = Request(
    method = Method.PUT,
    uri = momBaseUri.withPath(s"${momBaseUri.path}/sendMessage/${to.name}")
  ).withEntity(contents)

  def receiveMessageRequest(momBaseUri: Uri,from:Queue,timeout:Duration): Request[IO] = Request[IO](
    method = Method.GET,
    uri = momBaseUri.
      withPath(s"${momBaseUri.path}/receiveMessage/${from.name}").
      withQueryParam("timeOutSeconds",timeout.toSeconds)
  )

  def receiveMessageStreamRequest(momBaseUri: Uri,from:Queue): Request[IO] = Request[IO](
    method = Method.GET,
    uri = momBaseUri.
      withPath(s"${momBaseUri.path}/receiveStream/${from.name}")
  )

  def completeMessageRequest(momBaseUri: Uri, deliveryAttemptId:DeliveryAttemptId): Request[IO] = Request[IO](
    method = Method.PUT,
    uri = momBaseUri.withPath(s"${momBaseUri.path}/acknowledge/${deliveryAttemptId.underlying}")
  )

}

class DeliveryAttemptId(val underlying:Long) extends AnyVal {
  override def toString: String = s"DeliveryAttemptId(${underlying})"
}

object DeliveryAttemptId {
  def create():DeliveryAttemptId = new DeliveryAttemptId(Random.nextLong())
}

/**
  * For converting between Scala and json
  */
case class MessageAsJson(
                          deliveryAttemptId: DeliveryAttemptId,
                          millisecondsToComplete:Long,
                          remainingAttempts:Int,
                          contents: String
                        ) {
  def asJsonText:String = {
    import io.circe.generic.extras.Configuration
    import io.circe.generic.extras.auto.exportEncoder
    import io.circe.syntax.EncoderOps

    implicit val genDevConfig: Configuration = MessageAsJson.genDevConfig

    this.asJson.noSpaces
  }
}

object MessageAsJson {
  import io.circe.generic.extras.Configuration
  val genDevConfig: Configuration =
    Configuration.default.withDiscriminator("encodedClass")

  def tryRead(jsonText: String):Try[MessageAsJson] = Try {
    import io.circe.parser.decode
    import io.circe.generic.extras.auto.exportDecoder
    import io.circe.generic.extras.Configuration

    implicit val genDevConfig: Configuration = MessageAsJson.genDevConfig

    decode[MessageAsJson](jsonText) match {
      case Right(envelope) => envelope
      case Left(x) => throw x //throw errors to pick up in the Try
    }
  }
}
