package net.shrine.networkhealth

import java.util.concurrent.Executors

import cats.effect.{IO, Timer}
import net.shrine.config.ConfigExtensions
import net.shrine.crypto.BouncyKeyStoreCollection
import net.shrine.http4s.catsio.IOScheduler
import net.shrine.hub.HubDispatcher
import net.shrine.hub.data.store.{ConnectivityTestResult, HubDb}
import net.shrine.log.Log
import net.shrine.networkhealth.template.NetworkHealthEmailer
import net.shrine.protocol.version.{QueryId, ResearcherId}
import net.shrine.protocol.version.v1.{ConnectivityTest, CrcResult, Node, Query, Researcher, Result, RunQueryAtHub, Topic}
import net.shrine.source.ConfigSource

import scala.collection.immutable.Iterable
import scala.concurrent.ExecutionContext
import scala.concurrent.duration.{Duration, FiniteDuration}
import scala.util.control.NonFatal


object NetworkHealthScheduler {

  lazy implicit val timer : Timer[IO]  = IO.timer(ExecutionContext.fromExecutor(Executors.newScheduledThreadPool(2)))

  // If no scheduler is supposed to be configured, this will be None
  private val scheduler: Option[IOScheduler] = scheduleConnectivityTest

  def restart(): IO[Unit] = scheduler.fold(IO.unit){_.restart()}

  def stop(): IO[Unit] = scheduler.fold(IO.unit){_.stop()}

  private def scheduleConnectivityTest: Option[IOScheduler] = {
    val connectivityTestDataOption = ConnectivityTestData.fromConfig
    val certCollection: BouncyKeyStoreCollection = BouncyKeyStoreCollection.fromConfig

    connectivityTestDataOption.map{ connectivityTestData =>
      Log.debug(s"Starting the connectivity test scheduler")
      IOScheduler.scheduleIndefinitely(
        connectivityTestData.delay,
        connectivityTestData.interval,
        () => connectivityTest(connectivityTestData, certCollection),
        name = "NetworkHealthScheduler"
      )
    }
  }

  private def runQuery(queryId: QueryId,
                       originalQuery: Query,
                       researcher: Researcher,
                       topic: Topic,
                       certCollection: BouncyKeyStoreCollection): IO[Query] = {
    val query = Query.create(
      queryDefinitionXml = originalQuery.queryDefinitionXml,
      outputTypesXml = originalQuery.outputTypesXml,
      queryName = s"ConnectivityTest_${System.currentTimeMillis()}",
      nodeOfOriginId = originalQuery.nodeOfOriginId,
      researcherId = researcher.id,
      topicId = originalQuery.topicId,
      projectName = originalQuery.projectName,
      certCollection = certCollection
    )

    for {
      _ <- IO(Log.debug(s"About to run the connectivity test with new query id ${query.id}"))
      _ <- HubDispatcher.sendQueryToAdaptersIO(RunQueryAtHub(query, researcher, topic), 0)
      _ <- IO(Log.info("Finished starting the connectivity test"))
    } yield query
  }

  private def updateConnectivityTestResults(query: Query): IO[Unit] = for {
    _ <- IO(Log.debug("Updating the database with the latest connectivity test results"))
    connectivityTest = ConnectivityTestResult(networkQueryId = query.id.underlying)
    // Add the connectivity test data to the table now that the time limit has been reached
    _ <- HubDb.db.insertConnectivityTestResultIO(connectivityTest)
  } yield ()

  private def sendEmailsIO(query: Query, queryResults: Iterable[Result]): IO[Unit] = {
    import cats.implicits._

    queryResults.map {
      case x: CrcResult => IO(Log.debug(s"Successfully ran connectivity test on node ${x.adapterNodeName}"))
      case result => for {
          network <- HubDb.db.selectTheNetworkIO.map(Some(_)).handleErrorWith(_ => IO(None))
          maybeNode <- HubDb.db.selectNodeIO(result.adapterNodeId)
        } yield {
          maybeNode.foreach { node: Node =>
            if (node.adminEmail.nonEmpty) {
              Log.debug(s"Sending email for connectivity test for node ${node.name} to ${node.adminEmail}")
              val toList: List[String] = List(node.adminEmail)
              val ccList: List[String] = network.fold(List[String]())(n => List(n.adminEmail))

              NetworkHealthEmailer(ConfigSource.config).sendConnectivityIssueEmail(query, result, toList, ccList = ccList)
            }
            else Log.error(s"Could not find email for connectivity test for ${node.name}")
          }
        }
    }.toList.sequence_
  }

  private def checkAndReportErrorsIO(query : Query): IO[Unit] = for {
      queryResultsIO <- HubDb.db.selectMostRecentResultsForQueryIO(query.id)
      _ <- IO(Log.debug(s"Checking for and reporting any errors with running the connectivity test query with id ${query.id}"))
      _ <- sendEmailsIO(query, queryResultsIO)
    } yield ()

  private def testAfterTimeLimit(query: Query): IO[Unit] = for {
      _ <- IO(Log.debug(s"Connectivity test time limit reached... performing update"))
      _ <- updateConnectivityTestResults(query)
      _ <- checkAndReportErrorsIO(query)
    } yield ()

  private def connectivityTest(testData: ConnectivityTestData,certCollection: BouncyKeyStoreCollection): IO[Unit] = for {
      _ <- IO(Log.info("Starting the connectivity test"))
      query <- HubDb.db.selectQueryIO(testData.queryId).map(_.getOrElse(throw ConnectivityTestException("Error retrieving query from the database")))
      researcher <- HubDb.db.selectResearcherByIdIO(testData.researcherId)
      topic <- HubDb.db.selectTopicByIdIO(query.topicId)
      queryRan <- runQuery(testData.queryId, query, researcher , topic, certCollection)
      _ <- IOScheduler
        .scheduleOnce(testData.timeLimit, () => testAfterTimeLimit(queryRan), name = "NetworkHealthScheduler.testAfterTimeLimit")
        .restart()
    } yield ()
}

case class ConnectivityTestData(queryId: QueryId, researcherId: ResearcherId, interval: FiniteDuration, timeLimit: FiniteDuration, delay: FiniteDuration)

object ConnectivityTestData {
  def fromConfig: Option[ConnectivityTestData] = {
    import io.circe.config.syntax.CirceConfigOps
    import io.circe.generic.auto.exportDecoder

    val connectivityTestConfig = ConfigSource.config.getConfigOrEmpty("shrine.hub.networkHealth.connectivityTest")

    if(connectivityTestConfig.hasPath("queryId") && connectivityTestConfig.getString("queryId").nonEmpty) {
      connectivityTestConfig.as[ConnectivityTest] match {
        case Right(connectivityTest) => fromConnectivityTest(connectivityTest)
        case Left(e) => Log.error(s"An error occurred while loading the network health connectivity test config: $e"); None
      }
    }
    else {
      Log.debug("No connectivity test configured")
      None
    }
  }

  //todo use config values or something.
  def fromConnectivityTest(connectivityTest: ConnectivityTest): Option[ConnectivityTestData] = {

    def parseDuration(durationString:String):FiniteDuration = {
      Some(Duration(durationString)).collect({ case d: FiniteDuration => d }).getOrElse(throw new NumberFormatException(s"$durationString is not a FiniteDuration"))
    }

    val interval = parseDuration(connectivityTest.interval)
    val timeLimit = parseDuration(connectivityTest.timeLimit)
    val delay = parseDuration(connectivityTest.delay)
    val queryId = connectivityTest.queryId.toLong
    HubDb.db.selectQueryIO(new QueryId(queryId)).map {
      case Some(q) =>
        Log.debug(s"Found query with id $queryId and about to schedule the network health connectivity test")
        Some(ConnectivityTestData(new QueryId(queryId), q.researcherId, interval, timeLimit, delay))
      case None => Log.error(s"Unable to find the query id specified for the connectivity test"); None
    }.handleErrorWith {
      case NonFatal(nfx) => Log.error("An error occurred while loading the networkHealth config in shrine.conf",nfx); IO(None)
    }.unsafeRunSync()
  }
}
