package net.shrine.networkhealth

import java.io.{ByteArrayInputStream, ByteArrayOutputStream, IOException, InputStream}
import java.util.jar.JarFile
import java.net.URL
import java.nio.file.attribute.FileTime
import java.util.zip.{ZipEntry, ZipOutputStream}

import cats.effect.{ContextShift, IO}
import fs2.io.readInputStream
import net.shrine.config.ConfigExtensions
import net.shrine.protocol.version.v1.{CrcResult, Result}
import net.shrine.source.ConfigSource
import org.http4s.{Header, HttpRoutes, Request, Response, StaticFile}
import org.http4s.dsl.impl.{->, /, /:, Auth}
import org.http4s.dsl.io.{GET, NotFound, Ok, Root}
import org.http4s.dsl.io.{http4sNotFoundSyntax, http4sOkSyntax}
import io.circe.generic.auto.exportEncoder
import io.circe.syntax.EncoderOps
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.hub.data.store.HubDb
import net.shrine.log.Log
import net.shrine.protocol.version.QueryId
import org.http4s.server.Router

import scala.collection.mutable.ListBuffer
import scala.util.control.NonFatal


class NetworkHealthService extends Auth {

  implicit val concurrentEffect: ContextShift[IO] = IO.contextShift(ExecutionContexts.longActionExecutionContext)

  def staticResource(file: String, request: Request[IO]): IO[Response[IO]] =
    StaticFile.fromResource("/" + file, ExecutionContexts.longActionExecutionContext, Some(request)).getOrElseF(NotFound(s"File not found $file"))


  def router: HttpRoutes[IO] = {
    Router[IO](
      "" -> service
    )
  }

  val service: HttpRoutes[IO] = HttpRoutes.of[IO] {

    case _@GET -> Root / "ping" => Ok("pong")

    case _@GET -> Root / "webpage" / "data" / path if List(".json").exists(path.endsWith) => {

      path match {
        case "network_status.json" => Ok(getNetworkHealthData)
        case _ => NotFound(path)
      }
    }

    case req@GET -> "webpage" /: path if List(".css", ".json", ".js", ".map", ".html", ".jpg").exists(path.toList.mkString.endsWith) => {
      val newPath: String = path.toString.drop(1)

      staticResource("webpage/" + newPath, req)
    }

    case _@GET -> Root / "webpage" / "download" => {

      val contentTypeHeader: Header = Header("Content-Type", "application/zip")
      val inputStreamIO: IO[InputStream] = downloadWebPage
      Ok(readInputStream(inputStreamIO, chunkSize = 2048, ExecutionContexts.longActionExecutionContext, closeAfterUse = true), contentTypeHeader)
    }
  }

  def downloadWebPage: IO[InputStream] = {
    var path: URL = getClass.getClassLoader.getResource("/webpage")
    val jarFile = path.getPath.drop(5).dropRight(10)
    val byteArrayIO: IO[Array[Byte]] = createWebPageZip(jarFile)

    byteArrayIO.map(byteArray => new ByteArrayInputStream(byteArray))
  }

  def createWebPageZip(jarFile: String): IO[Array[Byte]] = {

    val (zipStream, byteArrayInputStream) = ZipUtil.createZipStream(jarFile)

    addJsonToZip(zipStream).map( _ => byteArrayInputStream.toByteArray)

  }

  def addJsonToZip(zipStream : ZipOutputStream): IO[Unit] = {
    //add the dynamically generated json data as a file
    val entry = new ZipEntry("webpage/data/network_status.json")
    entry.setCreationTime(FileTime.fromMillis(System.currentTimeMillis()))
    entry.setComment("Created by SHRINE")
    zipStream.putNextEntry(entry)

    val networkJsonIO: IO[String] = getNetworkHealthData

    IO(zipStream).bracket(zip => {
      networkJsonIO.map(networkJson => {
        zip.write(networkJson.getBytes)
      })
    }) {zip => IO(zip.close())}
  }

  def getNetworkHealthData: IO[String] = {

    for{
      connectivityTestResult <- HubDb.db.selectRecentConnectivityTestResultIO
      queryResults <- HubDb.db.selectMostRecentResultsForQueryIO(new QueryId(connectivityTestResult.networkQueryId))
    } yield  {

      val nodeStatusList: ListBuffer[NodeStatus] = new ListBuffer[NodeStatus]()

      queryResults.foreach { queryResult: Result => {
          val statusColor = if(queryResult.isInstanceOf[CrcResult]) "green" else "red"
          nodeStatusList += NodeStatus(queryResult.adapterNodeName.underlying, statusColor)
        }
      }

      val allNodesSorted = nodeStatusList.sortBy(_.name)

      val config = ConfigSource.config
      val networkHealthConfig = config.getConfig("shrine.hub.networkHealth.webpage")
      val networkName: String = networkHealthConfig.getString("networkName")
      val note: Option[String] = networkHealthConfig.getOption("webPageNote", _.getString)
      val alertMsg: Option[String] = networkHealthConfig.getOption("alertMessage", _.getString)

      val lastUpdated: Long = connectivityTestResult.changeDate

      val networkHealthData: NetworkHealthData = NetworkHealthData(allNodesSorted.toList, networkName, lastUpdated, alertMsg, note)

      networkHealthData.asJson.toString()
    }
  }

  //more static data for testing purposes
  def getTestErrorStackTrace: String = {
     """com.sun.jersey.api.client.ClientHandlerExceptionjava.net.ConnectException: Connection refused (Connection refused)
             |com.sun.jersey.client.urlconnection.URLConnectionClientHandler.handle(URLConnectionClientHandler.java:155)
             |com.sun.jersey.api.client.Client.handle(Client.java:652)
             |com.sun.jersey.api.client.WebResource.handle(WebResource.java:682)
             |com.sun.jersey.api.client.WebResource.access$200(WebResource.java:74)
             |com.sun.jersey.api.client.WebResource$Builder.post(WebResource.java:560)
             |net.shrine.http4s.client.legacy.JerseyHttpClient.post(JerseyHttpClient.scala:56)
             |net.shrine.http4s.client.legacy.Poster.post(Poster.scala:12)
             |net.shrine.adapter.client.RemoteAdapterClient$$anonfun$query$2$$anonfun$apply$1.apply(RemoteAdapterClient.scala:53)
             |net.shrine.adapter.client.RemoteAdapterClient$$anonfun$query$2$$anonfun$apply$1.apply(RemoteAdapterClient.scala:51)
             |scala.concurrent.impl.ExecutionContextImpl$DefaultThreadFactory$$anon$2$$anon$4.block(ExecutionContextImpl.scala:48)
             |scala.concurrent.forkjoin.ForkJoinPool.managedBlock(ForkJoinPool.java:3640)
             |scala.concurrent.impl.ExecutionContextImpl$DefaultThreadFactory$$anon$2.blockOn(ExecutionContextImpl.scala:45)
             |scala.concurrent.package$.blocking(package.scala:123)
             |net.shrine.adapter.client.RemoteAdapterClient$$anonfun$query$2.apply(RemoteAdapterClient.scala:51)
             |net.shrine.adapter.client.RemoteAdapterClient$$anonfun$query$2.apply(RemoteAdapterClient.scala:51)
             |scala.concurrent.impl.Future$PromiseCompletingRunnable.liftedTree1$1(Future.scala:24)
             |scala.concurrent.impl.Future$PromiseCompletingRunnable.run(Future.scala:24)
             |scala.concurrent.impl.ExecutionContextImpl$AdaptedForkJoinTask.exec(ExecutionContextImpl.scala:121)
             |scala.concurrent.forkjoin.ForkJoinTask.doExec(ForkJoinTask.java:260)
             |scala.concurrent.forkjoin.ForkJoinPool$WorkQueue.runTask(ForkJoinPool.java:1339)
             |scala.concurrent.forkjoin.ForkJoinPool.runWorker(ForkJoinPool.java:1979)
             |scala.concurrent.forkjoin.ForkJoinWorkerThread.run(ForkJoinWorkerThread.java:107)java.net.ConnectExceptionConnection refused (Connection refused)
             |java.net.PlainSocketImpl.socketConnect(Native Method)
             |java.net.AbstractPlainSocketImpl.doConnect(AbstractPlainSocketImpl.java:350)
             |java.net.AbstractPlainSocketImpl.connectToAddress(AbstractPlainSocketImpl.java:206)
             |java.net.AbstractPlainSocketImpl.connect(AbstractPlainSocketImpl.java:188)
             |java.net.SocksSocketImpl.connect(SocksSocketImpl.java:392)
             |java.net.Socket.connect(Socket.java:589)
             |sun.security.ssl.SSLSocketImpl.connect(SSLSocketImpl.java:668)
             |sun.net.NetworkClient.doConnect(NetworkClient.java:175)
             |sun.net.www.http.HttpClient.openServer(HttpClient.java:432)
             |sun.net.www.http.HttpClient.openServer(HttpClient.java:527)
             |sun.net.www.protocol.https.HttpsClient.<init>(HttpsClient.java:264)
             |sun.net.www.protocol.https.HttpsClient.New(HttpsClient.java:367)
             |sun.net.www.protocol.https.AbstractDelegateHttpsURLConnection.getNewHttpClient(AbstractDelegateHttpsURLConnection.java:191)
             |sun.net.www.protocol.http.HttpURLConnection.plainConnect0(HttpURLConnection.java:1138)
             |sun.net.www.protocol.http.HttpURLConnection.plainConnect(HttpURLConnection.java:1032)
             |sun.net.www.protocol.https.AbstractDelegateHttpsURLConnection.connect(AbstractDelegateHttpsURLConnection.java:177)
             |sun.net.www.protocol.http.HttpURLConnection.getOutputStream0(HttpURLConnection.java:1316)
             |sun.net.www.protocol.http.HttpURLConnection.getOutputStream(HttpURLConnection.java:1291)
             |sun.net.www.protocol.https.HttpsURLConnectionImpl.getOutputStream(HttpsURLConnectionImpl.java:250)
             |com.sun.jersey.client.urlconnection.URLConnectionClientHandler$1$1.getOutputStream(URLConnectionClientHandler.java:238)
             |com.sun.jersey.api.client.CommittingOutputStream.commitStream(CommittingOutputStream.java:117)
             |com.sun.jersey.api.client.CommittingOutputStream.write(CommittingOutputStream.java:89)
             |sun.nio.cs.StreamEncoder.writeBytes(StreamEncoder.java:221)
             |sun.nio.cs.StreamEncoder.implFlushBuffer(StreamEncoder.java:291)
             |sun.nio.cs.StreamEncoder.implFlush(StreamEncoder.java:295)
             |sun.nio.cs.StreamEncoder.flush(StreamEncoder.java:141)
             |java.io.OutputStreamWriter.flush(OutputStreamWriter.java:229)
             |java.io.BufferedWriter.flush(BufferedWriter.java:254)
             |com.sun.jersey.core.util.ReaderWriter.writeToAsString(ReaderWriter.java:191)
             |com.sun.jersey.core.provider.AbstractMessageReaderWriterProvider.writeToAsString(AbstractMessageReaderWriterProvider.java:128)
             |com.sun.jersey.core.impl.provider.entity.StringProvider.writeTo(StringProvider.java:88)
             |com.sun.jersey.core.impl.provider.entity.StringProvider.writeTo(StringProvider.java:58)
             |com.sun.jersey.api.client.RequestWriter.writeRequestEntity(RequestWriter.java:300)
             |com.sun.jersey.client.urlconnection.URLConnectionClientHandler._invoke(URLConnectionClientHandler.java:217)
             |com.sun.jersey.client.urlconnection.URLConnectionClientHandler.handle(URLConnectionClientHandler.java:153)
             |com.sun.jersey.api.client.Client.handle(Client.java:652)
             |com.sun.jersey.api.client.WebResource.handle(WebResource.java:682)
             |com.sun.jersey.api.client.WebResource.access$200(WebResource.java:74)
             |com.sun.jersey.api.client.WebResource$Builder.post(WebResource.java:560)
             |net.shrine.http4s.client.legacy.JerseyHttpClient.post(JerseyHttpClient.scala:56)
             |net.shrine.http4s.client.legacy.Poster.post(Poster.scala:12)
             |net.shrine.adapter.client.RemoteAdapterClient$$anonfun$query$2$$anonfun$apply$1.apply(RemoteAdapterClient.scala:53)
             |net.shrine.adapter.client.RemoteAdapterClient$$anonfun$query$2$$anonfun$apply$1.apply(RemoteAdapterClient.scala:51)
             |scala.concurrent.impl.ExecutionContextImpl$DefaultThreadFactory$$anon$2$$anon$4.block(ExecutionContextImpl.scala:48)
             |scala.concurrent.forkjoin.ForkJoinPool.managedBlock(ForkJoinPool.java:3640)
             |scala.concurrent.impl.ExecutionContextImpl$DefaultThreadFactory$$anon$2.blockOn(ExecutionContextImpl.scala:45)
             |scala.concurrent.package$.blocking(package.scala:123)
             |net.shrine.adapter.client.RemoteAdapterClient$$anonfun$query$2.apply(RemoteAdapterClient.scala:51)
             |net.shrine.adapter.client.RemoteAdapterClient$$anonfun$query$2.apply(RemoteAdapterClient.scala:51)
             |scala.concurrent.impl.Future$PromiseCompletingRunnable.liftedTree1$1(Future.scala:24)
             |scala.concurrent.impl.Future$PromiseCompletingRunnable.run(Future.scala:24)
             |scala.concurrent.impl.ExecutionContextImpl$AdaptedForkJoinTask.exec(ExecutionContextImpl.scala:121)
             |scala.concurrent.forkjoin.ForkJoinTask.doExec(ForkJoinTask.java:260)
             |scala.concurrent.forkjoin.ForkJoinPool$WorkQueue.runTask(ForkJoinPool.java:1339)
             |scala.concurrent.forkjoin.ForkJoinPool.runWorker(ForkJoinPool.java:1979)
             |scala.concurrent.forkjoin.ForkJoinWorkerThread.run(ForkJoinWorkerThread.java:107)""".stripMargin
  }
}

object ZipUtil {

  def createZipStream(jar: String): (ZipOutputStream, ByteArrayOutputStream) = {
    val byteArrayOutputStream: ByteArrayOutputStream = new ByteArrayOutputStream(2048)

    val zipStream = new ZipOutputStream(byteArrayOutputStream)

    try {
      val jf = new JarFile(jar)

      val iter = jf.entries
      while ( iter.hasMoreElements) {
        val entry = iter.nextElement
        if(entry.getName.contains("webpage"))
        {
          addToZipFile(entry.getName, entry.getLastModifiedTime.toMillis , jf.getInputStream(entry), zipStream)
        }
      }
    } catch {
      case NonFatal(e) => Log.error("Error while zipping.", e)
        if (zipStream != null) zipStream.close()
    }

    (zipStream, byteArrayOutputStream)
  }

  private def addToZipFile(inputFileName: String, lastModified: Long, inputStream: InputStream, zipStream: ZipOutputStream): Unit = {
    try {
      // create a new ZipEntry, which is basically another file
      // within the archive. We omit the path from the filename
      val entry = new ZipEntry(inputFileName)
      entry.setCreationTime(FileTime.fromMillis(lastModified))
      entry.setComment("Created by SHRINE")
      zipStream.putNextEntry(entry)
      Log.info("Generated new entry for: " + inputFileName)

      // Now we copy the existing file into the zip archive. To do
      // this we write into the zip stream, the call to putNextEntry
      // above prepared the stream, we now write the bytes for this
      // entry.
      val readBuffer = new Array[Byte](2048)
      var amountRead = 0
      var written = 0
      amountRead = inputStream.read(readBuffer)
      while(amountRead > 0){
        zipStream.write(readBuffer, 0, amountRead)
        written += amountRead
        amountRead = inputStream.read(readBuffer)
      }
      Log.info("Stored " + written + " bytes to " + inputFileName)
    } catch {
      case e: IOException =>
        throw new Exception("Unable to process " + inputFileName, e)
    } finally if (inputStream != null) inputStream.close()
  }
}

case class NodeStatus(name: String, statusColor: String)

case class NetworkHealthData(nodes: List[NodeStatus], networkName: String, timestamp: Long, alertMessage: Option[String], note: Option[String])

case class ConnectivityTestException(message: String) extends Exception(message)