package net.shrine.networkhealth.template

import java.text.SimpleDateFormat
import java.util.{Date, SimpleTimeZone}

import cats.effect.IO
import com.typesafe.config.Config
import net.shrine.config.ConfigExtensions
import courier.{Content, Envelope, Mailer, Text}
import javax.mail.internet.InternetAddress
import net.shrine.email.ConfiguredMailer
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.source.ConfigSource
import net.shrine.log.Log
import net.shrine.problem.{AbstractProblem, ProblemSources}
import net.shrine.networkhealth.template.{txt => twirlTextTemplates}
import net.shrine.protocol.query.QueryDefinition
import net.shrine.protocol.version.v1.{ErrorResult, Query, Result}

import scala.util.{Failure, Success, Try}
import scala.util.control.NonFatal


case class NetworkHealthEmailer(emailData: EmailData,
                                templateData: NetworkHealthTemplateData){

  def sendConnectivityIssueEmail(query: Query, result: Result, toList: List[String], ccList: List[String] = List()): Unit = {

    Log.info(s"Preparing to send connectivity issue email")

    val errorDataList: List[ErrorData]= createErrorDataList(query, result)

    val now = System.currentTimeMillis()
    val utcDate: String = getUTCFormattedDate(now)

    var nodeName: String = result.adapterNodeName.underlying

    val subject: String = s"Action Required: $nodeName Connectivity Test Issue - $utcDate"
    val to: List[InternetAddress] = toList.map(new InternetAddress(_))

    val cc: List[InternetAddress] = ccList.map(new InternetAddress(_))

    //Render the Twirl template - the template file name should be formatted as [name of template].scala.[format] where format can be html, txt, js, or xml
    // The template is rendered by calling the function [format].[name of template].(...)
    // The twirl template will need to be compiled first in order for the IDE to resolve the code below
    var renderedTemplate: String = twirlTextTemplates.connectivityissue(utcDate, nodeName, templateData, errorDataList).toString().stripPrefix("\n")

    //build up the email body
    val emailBody: Content = Text(renderedTemplate)

    val envelope:Envelope = Envelope.from(emailData.from).to(to: _* ).cc(cc: _*).subject(subject).content(emailBody)

    Log.debug(s"About to send $envelope .")

    Try ({
      //send the email
      IO.fromFuture[Unit](IO{emailData.mailer(envelope)(ExecutionContexts.longActionExecutionContext)}).unsafeRunSync()

      Log.info(s"Sent and logged $envelope .")
    }).recover({
      case NonFatal(x) =>
        Log.error(s"Error sending connectivity test email: $x")
    })
  }

  def createErrorDataList(query: Query, result: Result): List[ErrorData] = {

    val queryTerm: String = QueryDefinition.fromI2b2(query.queryDefinitionXml) match {
      case Success(qt) => qt.expr.map(_.toXml.lastOption).fold("")(_.get.text)
      case Failure(e) => ""
    }

    val errorData: ErrorData = result match {
      case er: ErrorResult =>
        val timestamp: String = er.problemDigest.stampText

        ErrorData(queryTerm, query.id.underlying.toString, query.queryName, timestamp,
          er.problemDigest.summary, er.problemDigest.description, er.problemDigest.codec, er.problemDigest.detailsXml.toString())
      case pr =>
        val timestamp: String = pr.versionInfo.changeDate.underlying.toString

        ErrorData(queryTerm, query.id.underlying.toString, query.queryName, timestamp)
    }

    val errorDataList: List[ErrorData] = List(errorData)
    errorDataList
  }

  def getUTCFormattedDate(timestamp: Long) : String = {
    val date = new Date(timestamp)
    val format = new SimpleDateFormat("MMM dd, yyyy HH:mm:ss z")
    format.setTimeZone(new SimpleTimeZone(SimpleTimeZone.UTC_TIME, "UTC"))
    format.format(date)
  }
}

object NetworkHealthEmailer {

  def apply(config:Config): NetworkHealthEmailer = {
    val config = ConfigSource.config
    val emailConfig = config.getConfig("shrine.hub.networkHealth.email")

    val ed: EmailData = EmailData(
      emailConfig.get("from", new InternetAddress(_)),
      ConfiguredMailer.createMailerFromConfig(config.getConfig("shrine.email")))

    val td = NetworkHealthTemplateData(emailConfig.getString("networkName"),
                          emailConfig.getString("networkSignature"),
                          emailConfig.getString("grantDescription"),
                          emailConfig.getString("helpContactName"),
                          emailConfig.getString("helpContactEmail"))
    NetworkHealthEmailer(emailData =  ed, templateData = td)
  }
}

case class NetworkHealthTemplateData(networkName: String, networkSignature: String, grantDescription: String, helpContactName: String, helpContactEmail: String)

case class EmailData(from:InternetAddress,
                     mailer:Mailer)

case class ErrorData(queryTerm: String, networkId: String, name: String, timestamp: String, summary: String = "", description: String= "",
                     codec: String = "", details: String= "")


case class CouldNotSendNetworkHealthEmail(envelope:Envelope, ex:Throwable) extends AbstractProblem(ProblemSources.Dsa) {
  override def summary: String = "The Network Health Service was not able to send email."
  override def description: String = s"The network health service was not able to ${envelope.to} due to ${throwable.get}"
  override def throwable = Some(ex)
  override def detailsText = Some(s"Could not send $envelope")
}