package net.shrine.networkhealth

import java.util.Date

import cats.data.OptionT
import cats.effect.IO
import net.shrine.hub.data.store.{ConnectivityTestResult, HubDb}
import net.shrine.config.ConfigExtensions
import org.http4s.EntityDecoder.text
import org.http4s.dsl.io.Ok
import org.http4s.{EntityDecoder, Method, Request, Response, Uri}
import org.junit.runner.RunWith
import org.scalatest.FlatSpec
import org.scalatestplus.junit.JUnitRunner
import io.circe.generic.auto.exportEncoder
import io.circe.syntax.EncoderOps
import net.shrine.crypto.NewTestKeyStore
import net.shrine.log.Log
import net.shrine.problem.RawProblem
import net.shrine.protocol.ResultOutputType.I2b2Options
import net.shrine.protocol.{ErrorStatusFromCrc, ResultOutputType}
import net.shrine.protocol.version.{MomQueueName, NodeId, NodeKey, NodeName, QueryId, ResultStatuses, TopicId, UserDomainName, UserName}
import net.shrine.protocol.version.v1.{ConnectivityTest, CrcResult, Network, NetworkHealthCheck, Node, Query, Researcher, ResultProgress}
import net.shrine.source.ConfigSource

import scala.util.{Failure, Success, Try}

/**
  * Test web api of Network Health
  */
@RunWith(classOf[JUnitRunner])
class NetworkHealthServiceTest extends FlatSpec { //todo maybe switch to just Spec or RefSpec

  private def requestResponse(request: Request[IO]): Response[IO] = {
    val responseOptionIo: OptionT[IO, Response[IO]] = new NetworkHealthService().router.run(request)

    val iorio: IO[Response[IO]] = responseOptionIo.map { r: Response[IO] => r }.fold {
      //todo is there some kind of exception or the like available?
      fail("No response from service")
    } { r: Response[IO] => r
    }
    iorio.unsafeRunSync()
  }

  "NetworkHealthService" should "reply to ping with pong" in {
    val response: Response[IO] = requestResponse(Request(method = Method.GET, uri = Uri.unsafeFromString("/ping")))
    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("pong")(entityText)
  }

  "NetworkHealthService" should "return network health json" in {
    import io.circe.generic.auto.exportDecoder
    import org.http4s.circe.jsonOf

    HubDb.db.createTables()
    initData()

    val nodeStatusList: List[NodeStatus] = List(
      NodeStatus("test node", "green")
    )

    val networkHealthConfig = ConfigSource.config.getConfig("shrine.hub.networkHealth.webpage")
    val networkName = networkHealthConfig.getString("networkName")

    val note: Option[String] = networkHealthConfig.getOption("webPageNote", _.getString)
    val alertMsg: Option[String] = networkHealthConfig.getOption("alertMessage", _.getString)

    val currentDate = new Date()

    val timestamp = currentDate.toInstant.toEpochMilli
    val expectedData: NetworkHealthData = NetworkHealthData(nodeStatusList, networkName, timestamp, alertMsg, note)

    val response: Response[IO] = requestResponse(Request(method = Method.GET, uri = Uri.unsafeFromString(s"/webpage/data/network_status.json")))
    assertResult(Ok)(response.status)

    implicit val networkDecoder: EntityDecoder[IO, NetworkHealthData] = jsonOf[IO, NetworkHealthData]
    val entityData: NetworkHealthData = response.as[NetworkHealthData].unsafeRunSync
    val actualData = entityData.copy(timestamp = expectedData.timestamp)
    HubDb.db.dropTables()

    assertResult(expectedData.asJson.toString())(actualData.asJson.toString)
  }

  def initData() {
    import io.circe.config.syntax.CirceConfigOps
    import io.circe.generic.auto.exportDecoder

    HubDb.db.upsertResultIO(crcResult).unsafeRunSync()

    val queryNetworkId = "2"

    val connectivityTestResult : ConnectivityTestResult = ConnectivityTestResult(crcResult.queryId.underlying)
    HubDb.db.insertConnectivityTestResultIO(connectivityTestResult).unsafeRunSync()
  }

  def runQuery(): Unit = {

    var nodeOfOrigin : Node  = HubDb.db.selectAllNodesIO.unsafeRunSync().head.get
    val allResearchers: Seq[Try[Researcher]] = HubDb.db.selectAllResearchersBlocking
    allResearchers.headOption.get match {
      case Success(researcher) => {
        val query = Query.create(
          queryDefinitionXml = "query xml",
          outputTypesXml = "output type xml",
          queryName="test query",
          nodeOfOriginId = nodeOfOrigin.id,
          researcherId = researcher.id,
          topicId = new TopicId(1),
          projectName = "SHRINE",
          certCollection = NewTestKeyStore.certCollection
        )
        HubDb.db.upsertQueryIO(query).unsafeRunSync()

        val allNodes: Seq[Try[Node]] = HubDb.db.selectAllNodesBlocking //Not sorted !!
        allNodes.foreach {
          case Success(node) => {
            val resultInProgress = ResultProgress(
              queryId = query.id,
              adapterNodeId = node.id,
              adapterNodeName = node.name,
              statusMessage = None
            )
            val isError = node.name.underlying.startsWith("U")
            if(isError) {
              val problem : RawProblem = ErrorStatusFromCrc(Some("This is a test"), "")
              val crcErrorResult = resultInProgress.toCrcError(problem)
              HubDb.db.upsertResultIO(crcErrorResult).unsafeRunSync()
            }else{

              var resultType =  ResultOutputType("myresult", isBreakdown=false, I2b2Options("i2b2 description"), None)
              var crcResult1 = resultInProgress.toCrcResult(10, 10L, resultType)
              HubDb.db.upsertResultIO(crcResult1).unsafeRunSync()
            }
          }
          case Failure(n) => Log.warn("Not able to retrieve node information to record results")
        }
      }
      case Failure(n) => Log.warn("Not able to retrieve researcher information to record results")
    }
  }

  val resultInProgress = ResultProgress(
    queryId = new QueryId(2),
    adapterNodeId = new NodeId(3),
    adapterNodeName = new NodeName("test node"),
    status = ResultStatuses.ResultFromCRC,
    statusMessage = None
  )

  val crcResult: CrcResult = resultInProgress.toCrcResult(
    count = 2050,
    crcQueryInstanceId = 3L,
    resultType = ResultOutputType.PATIENT_COUNT_XML
  )

}