package net.shrine.qep

import cats.effect.{Fiber, IO}
import com.typesafe.config.Config
import net.shrine.api.steward.db.StewardDatabase
import net.shrine.audit.NetworkQueryId
import net.shrine.authentication.AuthenticationResult.{Authenticated, NotAuthenticated}
import net.shrine.authentication.{Authenticator, NotAuthenticatedException}
import net.shrine.authorization.AuthorizationResult.{Authorized, NotAuthorized}
import net.shrine.authorization.QueryAuthorizationService
import net.shrine.crypto.BouncyKeyStoreCollection
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.hub.data.client.HubClient
import net.shrine.hub.mom.ShrineMomClient
import net.shrine.i2b2.protocol.pm.User
import net.shrine.log.{Log, Loggable}
import net.shrine.problem.{AbstractProblem, ProblemNotYetEncoded, ProblemSources}
import net.shrine.protocol.{AggregatedRunQueryResponse, BaseShrineRequest, FlagQueryRequest, FlagQueryResponse, I2b2RequestHandler, ReadApprovedQueryTopicsRequest, ReadQueryDefinitionRequest, ReadQueryDefinitionResponse, ReadResultOutputTypesRequest, ReadResultOutputTypesResponse, RenameQueryRequest, RenameQueryResponse, ResultOutputType, RunQueryRequest, ShrineResponse, UnFlagQueryRequest, UnFlagQueryResponse, XmlNodeName}
import net.shrine.protocol.query.QueryDefinition
import net.shrine.protocol.version.{QueryId, QueryStatuses}
import net.shrine.protocol.version.v1.{Query, QueryProgress, Researcher, RunQueryAtHub, Topic, UpdateQueryAtAdapter, UpdateQueryAtAdapterWithFlagging, UpdateQueryAtAdapterWithName, UpdateQueryAtQepWithError, UpdateQueryAtQepWithStatus}
import net.shrine.qep.audit.QepAuditDb
import net.shrine.qep.querydb.{QepQuery, QepQueryDb}
import net.shrine.util.XmlDateHelper

import scala.util.{Failure, Try}
import scala.util.control.NonFatal
import scala.xml.NodeSeq
/**
 * @author clint
 * @since Feb 19, 2014
 */
final case class I2b2QepService(
                                 commonName:String,
                                 authenticator: Authenticator,
                                 authorizationService: QueryAuthorizationService,
                                 breakdownTypes: Set[ResultOutputType],
                                 collectQepAudit:Boolean,
                                 xmlNodeName: XmlNodeName,
                                 certCollection: BouncyKeyStoreCollection,
                                 sendToHubAttempts:Int
                               ) extends I2b2RequestHandler with Loggable {

  def readResultOutputTypes(request: ReadResultOutputTypesRequest): ShrineResponse = {
    info(s"doReadResultOutputTypes($request)")

    authenticateAndThen(request) { _ =>
      val resultOutputTypes = ResultOutputType.nonErrorTypes ++ breakdownTypes

      ReadResultOutputTypesResponse(resultOutputTypes)
    }
  }

  override def runQuery(runQueryRequest: RunQueryRequest, shouldBroadcast: Boolean): ShrineResponse = {
    authenticateAndThen(runQueryRequest, Some(updateUser)) { authResult =>

      // inject modified, authorized runQueryRequest
      //inject QEP NodeId
      authorizeAndThen(runQueryRequest.copy(nodeId = Some(xmlNodeName))) { authorizedRequest =>
        debug(s"doRunQuery authorizedRequest is $authorizedRequest")

        val runQueryIO = for {
          localNode <- HubClient.getLocalNodeIO
          researcher <- IO(Researcher.createRepeatableForV26(
            userName = authorizedRequest.authn.username,
            userDomainName = authorizedRequest.authn.domain,
            nodeId = localNode.id
          ))
          topic <- IO(Topic.createRepeatableForV26(
            researcherId = researcher.id,
            name = runQueryRequest.topicId.get,
            description = runQueryRequest.topicName.getOrElse("") //todo topic descriptions are blank here ?? SHRINE-3632
          ))
          queryWithId <- IO(Query.create(
            id = runQueryRequest.networkQueryId,
            queryDefinitionXml = runQueryRequest.queryDefinition.toI2b2String,
            outputTypesXml = runQueryRequest.i2b2OutputList.toString(),
            queryName = runQueryRequest.queryDefinition.name,
            nodeOfOriginId = localNode.id,
            researcherId = researcher.id,
            topicId = topic.id,
            projectName = runQueryRequest.projectId,
            certCollection
          ))
          _ <- QepQueryDb.db.insertQueryIO(queryWithId, researcher)
          _ <- IO(if (collectQepAudit) QepAuditDb.db.insertQepQuery(authorizedRequest, commonName)) //todo context shift to database SHRINE-3632
          _ <- asynchSendToHub(queryWithId, researcher, topic)
        } yield ()

        runQueryIO.unsafeRunSync()

        AggregatedRunQueryResponse(
          queryId = authorizedRequest.networkQueryId,
          createDate = XmlDateHelper.now,
          userId = authResult.username,
          groupId = authResult.domain,
          requestXml = authorizedRequest.queryDefinition,
          queryInstanceId = authorizedRequest.networkQueryId,
          results = Seq.empty,
          statusTypeName = "RECEIVED_BY_QEP" //todo figure out the right statuses for 1.23. See SHRINE-2148 SHRINE-3632
        )
      }
    }
  }

  override def readQueryDefinition(request: ReadQueryDefinitionRequest, shouldBroadcast: Boolean): ReadQueryDefinitionResponse = {
    authenticateAndThen(request) { _ =>
      info(s"doReadQueryDefinition($request,$shouldBroadcast)")
      import net.shrine.util.StringEnrichments._

      val tryBaseResp: Try[ReadQueryDefinitionResponse] = for {
        queryFromDb: QepQuery <- Try {
          QepQueryDb.db.selectQueryById(request.queryId).getOrElse(throw NoQueryForIdException(request.queryId))
        }
        nodeSeq: NodeSeq <- queryFromDb.queryXml.tryToXml
        queryDefinition: QueryDefinition <- queryDefinitionFromStored(nodeSeq)
      } yield {
        ReadQueryDefinitionResponse(
          masterId = queryFromDb.networkId,
          name = queryFromDb.queryName,
          userId = queryFromDb.userName,
          createDate = XmlDateHelper.toXmlGregorianCalendar(queryFromDb.dateCreated),
          queryDefinition = queryDefinition.toI2b2.toString()
        )
      }
      //todo nope - fix for SHRINE-3632
      tryBaseResp.transform({ rqdr =>
        Try {
          rqdr
        }
      }, { x =>
        x match {
          case NonFatal(t) => CouldNotBuildResponseFromQepCache(t, request.queryId)
          case _ => //don't touch
        }
        Failure(x)
      }).get
    }
  }

  private def queryDefinitionFromStored(nodeSeq: NodeSeq): Try[QueryDefinition] = {
    RunQueryRequest.fromXml(breakdownTypes)(nodeSeq).map(_.queryDefinition).recoverWith { case _ => QueryDefinition.fromI2b2(nodeSeq) }
  }

  private def updateUser(user: User): Unit = {
    StewardDatabase.db.upsertUser(user)
  }

  private def sendUpdateQuery(updateQuery:UpdateQueryAtAdapter):Unit = {

    ShrineMomClient.sendToHubIO(updateQuery.queryId,updateQuery,UpdateQueryAtAdapter,s"${updateQuery.getClass.getSimpleName} for ${updateQuery.queryId}").
      unsafeRunAsync{ cb: Either[Throwable, Unit] =>
        cb.fold(
          {t => Log.error("Exception while running AbstractQepService.sendUpdateQuery", t)},
          _ => Log.debug(s"sent $updateQuery"))
      }
  }

  private def asynchSendToHub(queryWithId:QueryProgress,researcher: Researcher,topic: Topic): IO[Fiber[IO, Unit]] = {
    val recordAndSend = for {
      queryToHub <- IO(queryWithId.withStatus(QueryStatuses.SentToHub))
      _ <- ShrineMomClient.sendToHubIO(
        subjectId = queryToHub.id,
        envelopeContents = RunQueryAtHub(queryToHub,researcher,topic),
        envelopeContentsCompanion = RunQueryAtHub,
        logString = s"Sent query command ${queryWithId.id}",
        Option(sendToHubAttempts)
      )
      _ <- QepQueryDb.db.updateQepQueryIO(UpdateQueryAtQepWithStatus(queryToHub))
    } yield ()

    recordAndSend.handleErrorWith{ x =>
      QepQueryDb.db.updateQepQueryIO(
        UpdateQueryAtQepWithError(queryWithId,ProblemNotYetEncoded("The QEP encountered an unforeseen problem while sending a run query request to the hub",x))
      )
    }.start(IO.contextShift(ExecutionContexts.longActionExecutionContext))
  }

  private def authorizeAndThen[T](request: RunQueryRequest)(body: RunQueryRequest => T): T = {
    debug(s"authorizeAndThen($request) with $authorizationService")

    val authorizedRequest = authorizationService.authorizeRunQueryRequest(request) match {
      case na: NotAuthorized => throw na.toException
      case authorized: Authorized => request.copy(topicIdAndName = authorized.topicIdAndName)
    }

    body(authorizedRequest)
  }

  private def authenticateAndThen[T](request: BaseShrineRequest, onSuccess: Option[User => Unit] = None)(f: Authenticated => T): T = {
    val authResult = authenticator.authenticate(request.authn, onSuccess)

    authResult match {
      case a: Authenticated => f(a)
      case na:NotAuthenticated => throw NotAuthenticatedException(na)
    }
  }

  override def renameQuery(request: RenameQueryRequest, shouldBroadcast: Boolean):RenameQueryResponse = {
    authenticateAndThen(request) { _ =>
      info(s"doRenameQuery($request,$shouldBroadcast)")
      sendUpdateQuery(UpdateQueryAtAdapterWithName(new QueryId(request.networkQueryId),request.queryName))
      QepQueryDb.db.renamePreviousQuery(request)
      RenameQueryResponse(request.networkQueryId, request.queryName)
    }
  }

  override def readApprovedQueryTopics(request: ReadApprovedQueryTopicsRequest, shouldBroadcast: Boolean):ShrineResponse =
    authenticateAndThen(request, Some(updateUser)) { _ =>
    info(s"doReadApprovedQueryTopics($request,$shouldBroadcast)")
    //TODO: XXX: HACK: Would like to remove the cast
    authorizationService.readApprovedTopics(request) match {
      case Left(errorResponse) => errorResponse
      case Right(validResponse) => validResponse
    }
  }

  override def flagQuery(request: FlagQueryRequest, shouldBroadcast: Boolean = true): FlagQueryResponse.type = {
    authenticateAndThen(request) { _ =>
      sendUpdateQuery(UpdateQueryAtAdapterWithFlagging(new QueryId(request.networkQueryId), flagged = true, request.message))
      QepQueryDb.db.insertQepQueryFlag(request)
      FlagQueryResponse
    }
  }

  override def unFlagQuery(request: UnFlagQueryRequest, shouldBroadcast: Boolean = true): UnFlagQueryResponse.type = {
    authenticateAndThen(request) { _ =>
      sendUpdateQuery(UpdateQueryAtAdapterWithFlagging(new QueryId(request.networkQueryId), flagged = false))
      QepQueryDb.db.insertQepQueryFlag(request)
      UnFlagQueryResponse
    }
  }

}

object I2b2QepService {

  def apply(
             qepConfig: Config,
            commonName: String,
            authenticator: Authenticator,
            authorizationService: QueryAuthorizationService,
            breakdownTypes: Set[ResultOutputType],
            nodeId: XmlNodeName,
            certCollection: BouncyKeyStoreCollection,
           ): I2b2QepService = {
    I2b2QepService(
      commonName,
      authenticator,
      authorizationService,
      breakdownTypes,
      qepConfig.getBoolean("audit.collectQepAudit"),
      nodeId ,
      certCollection,
      qepConfig.getInt("sendToHubAttempts")
    )
  }
}

case class NoQueryForIdException(networkQueryId: NetworkQueryId) extends Exception(s"$networkQueryId not found in QEP Cache.")

case class CouldNotBuildResponseFromQepCache(t: Throwable,networkQueryId: NetworkQueryId) extends AbstractProblem(ProblemSources.Qep) {
  override val throwable = Some(t)
  override def summary: String = s"${t.getClass.getSimpleName} while building response from the QEP cache."
  override def description: String = "The QEP generated an exception while building a response locally."
  override def detailsText = Some(s"networkQueryId is $networkQueryId")
}