package net.shrine.qep

import com.typesafe.config.Config
import net.shrine.authentication.{AuthenticationType, Authenticator, PmAuthenticator}
import net.shrine.authorization.{AllowsAllAuthorizationService, AuthorizationType, QueryAuthorizationService, StewardQueryAuthorizationService}
import net.shrine.config.ConfigExtensions
import net.shrine.crypto.BouncyKeyStoreCollection
import net.shrine.http4s.client.legacy.Poster
import net.shrine.log.Loggable
import net.shrine.protocol.{ResultOutputType, XmlNodeName}

/**
  * @author david 
  * @since 1.22
  */
//todo maybe this can go with SHRINE-3633
case class QueryEntryPointComponents(
                                     i2b2Service: I2b2QepService
                                    )

object QueryEntryPointComponents extends Loggable {
  def apply(
             qepConfig:Config,
             certCollection: BouncyKeyStoreCollection,
             breakdownTypes: Set[ResultOutputType],
             pmPoster: Poster, //todo could really have its own
             nodeId:XmlNodeName
           ):QueryEntryPointComponents = {

    val commonName: String = certCollection.myEntry.commonName.getOrElse {
      val hostname = java.net.InetAddress.getLocalHost.getHostName
      warn(s"No common name available from ${certCollection.myEntry}. Using $hostname instead.")
      hostname
    }

    val authenticator: Authenticator = AuthStrategy.determineAuthenticator(qepConfig, pmPoster)
    val authorizationService: QueryAuthorizationService = AuthStrategy.determineQueryAuthorizationService(qepConfig)

    debug(s"authorizationService set to $authorizationService")

    QueryEntryPointComponents(
      I2b2QepService(
        qepConfig,
        commonName,
        authenticator,
        authorizationService,
        breakdownTypes,
        nodeId,
        certCollection
      )
    )
  }
}

/**
  * @author clint
  * @since Jul 1, 2014
  */

object AuthStrategy {

  import AuthenticationType._
  import AuthorizationType._

  def determineAuthenticator(qepConfig:Config, pmPoster: Poster): Authenticator = {

    //todo put these default values in reference.conf if you decide to use one
    val defaultAuthenticationType: AuthenticationType = AuthenticationType.Pm
    val authType = qepConfig.getOption("authenticationType",_.getString).flatMap(AuthenticationType.valueOf).getOrElse(defaultAuthenticationType)

    authType match {
      case NoAuthentication => AllowsAllAuthenticator
      case Pm => PmAuthenticator(pmPoster)
      case _ => throw new IllegalArgumentException(s"Unknown authentication type '$authType'")
    }
  }

  def determineQueryAuthorizationService(qepConfig:Config): QueryAuthorizationService = {

    val authorizationType = AuthorizationType.valueOf(qepConfig.getString("authorizationType")).get

    authorizationType match {
      case ShrineSteward => StewardQueryAuthorizationService(qepConfig.getConfig("shrineSteward"))
      case NoAuthorization => AllowsAllAuthorizationService
      case _ => throw new IllegalArgumentException(s"Unknown authorization type '$authorizationType'")
    }
  }
}