package net.shrine.qep.querydb

import java.sql.SQLException
import java.util.concurrent.TimeoutException

import cats.effect.IO
import com.typesafe.config.Config
import javax.sql.DataSource
import net.shrine.audit.{Checksum, NetworkQueryId, QueryName, Time, UserName}
import net.shrine.config.ConfigExtensions
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.i2b2.protocol.pm.User
import net.shrine.log.Log
import net.shrine.problem.{AbstractProblem, ProblemSources, XmlProblemDigest}
import net.shrine.protocol.query.QueryDefinition
import net.shrine.protocol.version.v1.{Breakdowns, CrcResult, ErrorResult, QueryError, Researcher, Result, ResultProgress, UpdateQueryAtQep, UpdateQueryAtQepWithError, UpdateQueryReadyForAdapters, Query => ProtocolQuery}
import net.shrine.protocol.version.{QueryStatuses, ResultStatuses}
import net.shrine.protocol.{DefaultBreakdownResultOutputTypes, FlagQueryRequest, I2b2ResultEnvelope, QueryMaster, QueryResult, ReadPreviousQueriesRequest, ReadPreviousQueriesResponse, RenameQueryRequest, ResultOutputType, ResultOutputTypes, UnFlagQueryRequest}
import net.shrine.slick.{CouldNotRunDbIoActionException, TestableDataSourceCreator, TimeoutInDbIoActionException}
import net.shrine.source.ConfigSource
import net.shrine.util.{Sort, XmlDateHelper}
import slick.ast.BaseTypedType
import slick.jdbc.{JdbcProfile, JdbcType}

import scala.collection.immutable.{Iterable, ListMap}
import scala.concurrent.duration.{Duration, DurationInt}
import scala.concurrent.{Await, ExecutionContextExecutorService, Future, blocking}
import scala.language.postfixOps
import scala.util.control.NonFatal
import scala.xml.XML

/**
  * DB code for the QEP's query instances and query results.
  *
  * @author david
  * @since 1/19/16
  */
case class QepQueryDb(schemaDef:QepQuerySchema,dataSource: DataSource,timeout:Duration) {
  import schemaDef._
  import jdbcProfile.api._
  import schemaDef.jdbcProfile.backend.DatabaseDef

  implicit val executionContext: ExecutionContextExecutorService = ExecutionContexts.databaseExecutionContext

  val database:DatabaseDef = Database.forDataSource(dataSource, None)

  def createTables(): Unit = schemaDef.createTables(database)

  def dropTables(): Unit = schemaDef.dropTables(database)

  private def run[R](dbio: DBIOAction[R, NoStream, _]): Future[R] = {
    database.run(dbio)
  }

  private def runBlocking[R](dbio: DBIOAction[R, NoStream, _], timeout: Duration = timeout): R = {
    try {
      Await.result(this.run(dbio), timeout)
    } catch {
      case tx:TimeoutException => throw TimeoutInDbIoActionException(dataSource, timeout, tx)
      case NonFatal(x) => throw CouldNotRunDbIoActionException(dataSource, x)
    }
  }

  private val databaseContextShift = IO.contextShift(ExecutionContexts.databaseExecutionContext)

  private def runIO[R](dbio: DBIOAction[R, NoStream, _]): IO[R] = {
    databaseContextShift.evalOn(ExecutionContexts.databaseExecutionContext)(
      IO.fromFuture(IO(blocking(run(dbio))))
    )
  }

  private def runTransaction[R](dbio: DBIOAction[R, NoStream, _]): Future[R] = {
    database.run(dbio.transactionally)
  }

  private def runTransactionIO[R](dbio: DBIOAction[R, NoStream, _]): IO[R] = {
    databaseContextShift.evalOn(ExecutionContexts.databaseExecutionContext)(
      IO.fromFuture(IO(blocking(database.run(dbio.transactionally))))
    )
  }

  private def runTransactionBlocking[R](dbio: DBIOAction[R, NoStream, _], timeout: Duration = timeout): R = {
    try {
      Await.result(this.run(dbio.transactionally), timeout)
    } catch {
      case tx:TimeoutException => throw TimeoutInDbIoActionException(dataSource, timeout, tx)
      case NonFatal(x) => throw CouldNotRunDbIoActionException(dataSource, x)
    }
  }

  //used by AbstractQepService's doRunQuery, which is what the "i2b2" h4s service wraps around to run queries
  def insertQueryIO(query:ProtocolQuery, researcher:Researcher):IO[Int] = {
    Log.debug(s"insertQepQuery $query")

    for {
      rowsChanged <- runIO(allQepQueryQuery += QepQuery(query,researcher))
      _ <- QepQueryDbChangeNotifier.triggerChangesFor(query.id.underlying)
    } yield rowsChanged
  }


  //used by Receiver
  def updateQepQueryIO(sqsaq: UpdateQueryAtQep): IO[Unit] = {
    Log.debug(s"updateQepQuery $sqsaq")

    val problems = sqsaq match {
      case errorUpdate: UpdateQueryAtQepWithError => Seq(QueryProblemDigestRow(errorUpdate))
      case _ => Seq.empty
    }

    val resultRows: Seq[QueryResultRow] = sqsaq match {
      case hasResultProgress: UpdateQueryReadyForAdapters => hasResultProgress.resultProgresses.map(QueryResultRow(_))
      case _ => Seq.empty
    }

    runTransactionIO {

      for{
        maybeOldQepQuery <- mostRecentVisibleQepQueries.filter(_.networkId === sqsaq.queryId.underlying).take(1).result
        _ <- allQepQueryQuery += {
          maybeOldQepQuery match {
            case Seq(oldQepQuery) => oldQepQuery.updateWith(sqsaq)
            case Seq() => throw NoQueryForQueryIdException(sqsaq.queryId.underlying)
            case _ => throw new IllegalStateException(s"${maybeOldQepQuery.size} queries found for ${sqsaq.queryId.underlying}")
            }
          }
        _ <- allQueryProblemDigestRows ++= problems
        _ <- allQueryResultRows ++= resultRows
      } yield ()
    }.flatMap(_ => QepQueryDbChangeNotifier.triggerChangesFor(sqsaq.queryId.underlying)).handleErrorWith{
      case nqfix:NoQueryForQueryIdException => IO(s"${Log.info(nqfix.getMessage)} . Assuming this is fine for a network health query.")
    }
  }

  //todo only used in test code
  def insertQepQuery(qepQuery: QepQuery):Unit = {
    runBlocking(allQepQueryQuery += qepQuery)
    QepQueryDbChangeNotifier.triggerChangesFor(qepQuery.networkId).unsafeRunSync()
  }

  val allQepQueryQuery = TableQuery[QepQueries]

  def selectAllQepQueries:Seq[QepQuery] = {
    runBlocking(mostRecentVisibleQepQueries.result)
  }

  def selectResultsRowIO(networkQueryId: NetworkQueryId, user:User): IO[QueryAndResults] = {

    runTransactionIO(
      for {
        query: QepQuery <- mostRecentVisibleQepQueries.filter(_.networkId === networkQueryId).take(1).result.head
        maybeQueryProblem: Option[QueryProblemDigestRow] <- mostRecentQueryProblemDigestRows.filter(_.networkId === networkQueryId).take(1).result.headOption
        queryResults: Seq[QueryResultRow] <- mostRecentQueryResultRows.filter(_.networkQueryId === networkQueryId).result
        breakdowns: Seq[QepQueryBreakdownResultsRow] <- mostRecentBreakdownResultsRows.filter(_.networkQueryId === networkQueryId).result
        problems: Seq[QepProblemDigestRow] <- mostRecentProblemDigestRows.filter(_.networkQueryId === networkQueryId).result
        maybeQueryFlag: Option[QepQueryFlag] <- mostRecentQueryFlags.filter(_.networkId === networkQueryId).take(1).result.headOption
      } yield {

        if (user.sameUserAs(query.userName, query.userDomain)) {
          val breakdownTypeToResults: Map[ResultOutputType, Seq[QepQueryBreakdownResultsRow]] = breakdowns.groupBy(_.resultType)

          val sortQepQueryBreakdownResultsRowAlphaNumerically =
            Sort.orderingFor[QepQueryBreakdownResultsRow, String](_.dataKey, Sort.compareAlphaNumerically)

          val sortResultOutputTypeTupleAlphaNumerically =
            Sort.orderingFor[(ResultOutputType, _), String](_._1.name, Sort.compareAlphaNumerically)

          val sortedBreakdowns: Map[ResultOutputType, Seq[QepQueryBreakdownResultsRow]] = ListMap(
            breakdownTypeToResults
              .toList
              .map(tup => (tup._1, tup._2.sorted(sortQepQueryBreakdownResultsRowAlphaNumerically)))  // sort each value
              .sorted(sortResultOutputTypeTupleAlphaNumerically) : _*                                // sort map by keys
          )

          def seqOfOneProblemRowToProblemDigest(problemSeq:Seq[QepProblemDigestRow]):XmlProblemDigest = {
            if(problemSeq.size == 1) problemSeq.head.toProblemDigest
            else throw new IllegalStateException(s"problemSeq size was not 1. $problemSeq")
          }

          val adapterNodesToProblemDigests: Map[String, XmlProblemDigest] = problems.groupBy(_.adapterNode).map(nodeToProblem => nodeToProblem._1 -> seqOfOneProblemRowToProblemDigest(nodeToProblem._2) )

          val fullResult: Seq[FullQueryResult] = queryResults.map(r => FullQueryResult(
            r,
            sortedBreakdowns,
            adapterNodesToProblemDigests.get(r.adapterNode)
          ))

          val flag: Option[QueryFlag] = maybeQueryFlag.map(QueryFlag(_))

          val queryAndProblems = FullQuery(maybeQueryProblem, query, flag)

          val queryAndResults = QueryAndResults(queryAndProblems, fullResult)
          queryAndResults
        }
        else {
          throw RequestingUsernameDoesNotMatchQueryException(query.networkId, query.userName, user)
        }
      }
    )
  }

  def selectFullQuery(networkQueryId: NetworkQueryId): IO[Option[FullQuery]] = {

    runTransactionIO(
      for {
        query: Seq[QepQuery] <- mostRecentVisibleQepQueries.filter(_.networkId === networkQueryId).result
        problems: Seq[QueryProblemDigestRow] <- mostRecentQueryProblemDigestRows.filter(_.networkId === networkQueryId).result
      } yield {
        query.map(FullQuery(problems.lastOption, _)).lastOption
      }
    )
  }

  def selectPreviousQueries(request: ReadPreviousQueriesRequest):ReadPreviousQueriesResponse = {
    val previousQueries: Seq[QepQuery] = selectPreviousQueriesByUserAndDomain(
      request.authn.username,
      request.authn.domain,
      None,
      Some(request.fetchSize))
    val flags:Map[NetworkQueryId,QepQueryFlag] = selectMostRecentQepQueryFlagsFor(previousQueries.map(_.networkId).to[Set])
    val queriesAndFlags = previousQueries.map(x => (x,flags.get(x.networkId)))

    ReadPreviousQueriesResponse(queriesAndFlags.map(x => x._1.toQueryMaster(x._2)))
  }

  def countPreviousQueriesByUserAndDomain(userName: UserName, domain: String):Int = {
    val q = mostRecentVisibleQepQueries.filter(r => r.userName === userName && r.userDomain === domain)

    runBlocking(q.size.result)
  }

  def selectQueryById(networkQueryId: NetworkQueryId): Option[QepQuery] =
    runBlocking(mostRecentVisibleQepQueries.filter(_.networkId === networkQueryId).result).lastOption

  def selectPreviousQueriesByUserAndDomain(userName: UserName, domain: String, skip:Option[Int] = None, limit:Option[Int] = None,
                                         sortBy: Option[String] = None):Seq[QepQuery] = {

    Log.debug(s"start selectPreviousQueriesByUserAndDomain $userName $domain")

    val q = mostRecentVisibleQepQueries.filter(r => r.userName === userName && r.userDomain === domain)

    val qWithSort =
      sortBy.getOrElse("dateCreated.desc") match {
      case "dateCreated.asc" => q.sortBy(_.dateCreated.asc)
      case "dateCreated.desc" => q.sortBy(_.dateCreated.desc)
      case "queryName.asc" => q.sortBy(_.queryName.asc)
      case "queryName.desc" => q.sortBy(_.queryName.desc)
      case s => throw new Exception(s"Unknown sortBy option: $s")
    }

    val qWithSkip = skip.fold(qWithSort)(qWithSort.drop)
    val qWithLimit = limit.fold(qWithSkip)(qWithSkip.take)

    val result = runBlocking(qWithLimit.result)

    Log.debug(s"finished selectPreviousQueriesByUserAndDomain with $result")

    result
  }

  def renamePreviousQuery(request:RenameQueryRequest):Unit = {

    val networkQueryId = request.networkQueryId
    runTransactionBlocking(
      for {
        queryResults <- mostRecentVisibleQepQueries.filter(_.networkId === networkQueryId).result
        _ <- allQepQueryQuery ++= queryResults.map(_.copy(queryName = request.queryName,changeDate = System.currentTimeMillis()))
      } yield ()
    )
    QepQueryDbChangeNotifier.triggerChangesFor(networkQueryId).unsafeRunSync()
  }

  //todo should not ever be called. only used in AbstractQepService. Dead code
  private[qep] def insertQepQueryFlag(flagQueryRequest: FlagQueryRequest):Unit = {
    insertQepQueryFlag(QepQueryFlag(flagQueryRequest))
  }

  //todo should not ever be called. only used in AbstractQepService. Dead code
  private[qep] def insertQepQueryFlag(unflagQueryRequest: UnFlagQueryRequest):Unit = {
    insertQepQueryFlag(QepQueryFlag(unflagQueryRequest))
  }

  private[querydb] def insertQepQueryFlag(qepQueryFlag: QepQueryFlag):Unit = {
    runBlocking(allQepQueryFlags += qepQueryFlag)
    QepQueryDbChangeNotifier.triggerChangesFor(qepQueryFlag.networkQueryId).unsafeRunSync()
  }

  def selectMostRecentQepQueryFlagsFor(networkIds:Set[NetworkQueryId]):Map[NetworkQueryId,QepQueryFlag] = {
    val flags:Seq[QepQueryFlag] = runBlocking(mostRecentQueryFlags.filter(_.networkId inSet networkIds).result)

    flags.map(x => x.networkQueryId -> x).toMap
  }

  def selectMostRecentQepQueryFlagFor(networkQueryId: NetworkQueryId): Option[QepQueryFlag] =
    runBlocking(mostRecentQueryFlags.filter(_.networkId === networkQueryId).result).lastOption

  //only used in tests
  def insertQepResultRow(qepQueryRow:QueryResultRow): Int = {
    val rowsChanged = runBlocking(allQueryResultRows += qepQueryRow)
    QepQueryDbChangeNotifier.triggerChangesFor(qepQueryRow.networkQueryId).unsafeRunSync()
    rowsChanged
  }

  def insertQueryResultIO(result:Result): IO[Unit] = {

    val queryResultRow = QueryResultRow(result)
    val problems: Seq[QepProblemDigestRow] = result match {
      case errorResult: ErrorResult => Seq(QepProblemDigestRow(errorResult))
      case _ => Seq.empty
    }
    val breakdowns: Seq[QepQueryBreakdownResultsRow] = result match {
      case crcResult: CrcResult => QepQueryBreakdownResultsRow.breakdownRowsFor(crcResult)
      case _ => Seq.empty
    }
    runTransactionIO(
      for {
        _ <- allQueryResultRows += queryResultRow
        _ <- allProblemDigestRows ++= problems
        _ <- allBreakdownResultsRows ++= breakdowns
      } yield ()
    ).flatMap{_ => QepQueryDbChangeNotifier.triggerChangesFor(result.queryId.underlying)}
  }

  //Only used in dead code and tests
  private[qep] def insertQueryResult(networkQueryId:NetworkQueryId,result:QueryResult): Unit = {

    val adapterNode = result.description.getOrElse(throw new IllegalStateException("description is empty, does not have an adapter node"))

    val queryResultRow = QueryResultRow(networkQueryId,result)
    val breakdowns: Iterable[QepQueryBreakdownResultsRow] = result.breakdowns.flatMap(QepQueryBreakdownResultsRow.breakdownRowsFor(networkQueryId,adapterNode,result.resultId,_))
    val problem: Seq[QepProblemDigestRow] = result.problem.map(p => QepProblemDigestRow(networkQueryId,adapterNode,p.codec,p.stampText,p.summary,p.description,p.detailsXml.toString,System.currentTimeMillis())).to[Seq]

    runTransactionBlocking(
      for {
      _ <- allQueryResultRows += queryResultRow
      _ <- allBreakdownResultsRows ++= breakdowns
      _ <- allProblemDigestRows ++= problem
      } yield QepQueryDbChangeNotifier.triggerChangesFor(networkQueryId).unsafeRunSync()
    )
  }

  private[querydb] def selectMostRecentQepResultRowsFor(networkId:NetworkQueryId): Seq[QueryResultRow] = {
    runBlocking(mostRecentQueryResultRows.filter(_.networkQueryId === networkId).result)
  }

  def selectMostRecentFullQueryResultsFor(networkId:NetworkQueryId): Seq[FullQueryResult] = {

    val (queryResults, breakdowns,problems):(Seq[QueryResultRow],Seq[QepQueryBreakdownResultsRow],Seq[QepProblemDigestRow]) = runTransactionBlocking(
      for {
        queryResults: Seq[QueryResultRow] <- mostRecentQueryResultRows.filter(_.networkQueryId === networkId).result
        breakdowns: Seq[QepQueryBreakdownResultsRow] <- mostRecentBreakdownResultsRows.filter(_.networkQueryId === networkId).result
        problems: Seq[QepProblemDigestRow] <- mostRecentProblemDigestRows.filter(_.networkQueryId === networkId).result
      } yield (queryResults, breakdowns, problems)
    )

    val breakdownTypeToResults: Map[ResultOutputType, Seq[QepQueryBreakdownResultsRow]] = breakdowns.groupBy(_.resultType)

    def seqOfOneProblemRowToProblemDigest(problemSeq:Seq[QepProblemDigestRow]):XmlProblemDigest = {
      if(problemSeq.size == 1) problemSeq.head.toProblemDigest
      else throw new IllegalStateException(s"problemSeq size was not 1. $problemSeq")
    }

    val adapterNodesToProblemDigests: Map[String, XmlProblemDigest] = problems.groupBy(_.adapterNode).map(nodeToProblem => nodeToProblem._1 -> seqOfOneProblemRowToProblemDigest(nodeToProblem._2) )

    queryResults.map(r => FullQueryResult(
      r,
      breakdownTypeToResults,
      adapterNodesToProblemDigests.get(r.adapterNode)
    ))
  }

  def selectMostRecentQepResultsFor(networkId:NetworkQueryId): Seq[QueryResult] = {
    val fullQueryResults = selectMostRecentFullQueryResultsFor(networkId)

    fullQueryResults.map(_.toQueryResult)
  }

  //only used in tests
  private[querydb] def insertQueryBreakdown(breakdownResultsRow:QepQueryBreakdownResultsRow): Int = {
    runBlocking(allBreakdownResultsRows += breakdownResultsRow)
  }

  def selectAllBreakdownResultsRows: Seq[QepQueryBreakdownResultsRow] = {
    runBlocking(allBreakdownResultsRows.result)
  }

  def selectDistinctAdaptersWithResults:Seq[String] = {
   runBlocking(allQueryResultRows.map(_.adapterNode).distinct.result).sorted
  }

  def selectPreviousQueriesWithResultStatusIO(userName: UserName, domain: String, skip:Option[Int] = None, limit:Option[Int] = None,
                                              sortBy: Option[String] = None, selectedQueryId: Option[NetworkQueryId] = None,
                                              previousQueryIds: Option[List[NetworkQueryId]] = None): IO[(Int, Seq[QueryWithResultStatus])] = {

    val q = mostRecentVisibleQepQueries.filter(r => r.userName === userName && r.userDomain === domain)

    val qWithSort = sortBy.getOrElse("dateCreated.desc") match {
      case "dateCreated.asc" => q.sortBy(_.dateCreated.asc)
      case "dateCreated.desc" => q.sortBy(_.dateCreated.desc)
      case "queryName.asc" => q.sortBy(_.queryName.asc)
      case "queryName.desc" => q.sortBy(_.queryName.desc)
      case s => throw new Exception(s"Unknown sortBy option: $s")
    }

    val qWithSkip = skip.fold(qWithSort)(qWithSort.drop)
    val qWithLimit = limit.fold(qWithSkip)(qWithSkip.take)

    val qWithIdFilter = previousQueryIds.fold(qWithLimit)(qIdList => qWithLimit.filter(_.networkId inSet qIdList))

    def createQueryStateObservedChecksumSet(queries: Seq[QepQuery], allQueryResults: Seq[QueryResultRow]): Seq[Checksum] = {

      val queryStatesObserved: Seq[Long] = queries.map(query => {
        val queryResults = allQueryResults.filter(_.networkQueryId == query.networkId)
        QueryStateObserved(query, queryResults).checksum
      })

      queryStatesObserved
    }

    /** Specialized fast version that takes advantage of the relatively small fixed number of queries */
    def mostRecentRelevantQueryResultRows(queries: Seq[QepQuery]): DBIOAction[Seq[QueryResultRow], NoStream, Effect.Read] = {
      allQueryResultRows.filter(_.networkQueryId inSet queries.map(q => q.networkId)).result.map{rows: Seq[QueryResultRow] =>
        rows.groupBy(_.resultId).map(grouped => grouped._2.maxBy(_.changeDate)).toSeq
      }
    }

    runTransactionIO(
      for {
          queries: Seq[QepQuery] <- qWithIdFilter.result
          results: Seq[QueryResultRow] <- mostRecentRelevantQueryResultRows(queries)
          queryStatesObserved: Seq[QueryStateObserved] <- allQueryStatesObserved.filter(_.checksum inSet createQueryStateObservedChecksumSet(queries, results)).result
          queryFlags: Seq[QepQueryFlag] <- mostRecentQueryFlags.filter(_.networkId inSet queries.map(q => q.networkId)).result
          queryCount: Int <- allQepQueryQuery.filter(_.userName === userName ).filter(_.userDomain === domain).filter(_.deleted === false).map(_.networkId).distinct.size.result
      }
      yield {
        val queryFlagMap: Map[NetworkQueryId, QueryFlag] = queryFlags.map(q =>  q.networkQueryId ->  QueryFlag(q)).toMap
        val queryWithResult = queries.flatMap(query => {
          val queryResults: Seq[QueryResultRow] = results.filter(_.networkQueryId == query.networkId)
          val checksum: Checksum = QueryStateObserved(query, queryResults).checksum
          val observed: Boolean = selectedQueryId.fold(queryStatesObserved.exists(_.checksum == checksum)){queryId =>
            if(query.networkId == queryId)
              true
            else {
              queryStatesObserved.exists(_.checksum == checksum)
            }
          }

          Seq(QueryWithResultStatus(query, queryResults, queryFlagMap.get(query.networkId), observed))
        })
        (queryCount, queryWithResult)
      }
    )
  }

  def insertQueryStateObservedIO(queryStateObserved: QueryStateObserved): IO[Unit] = {
    runIO(allQueryStatesObserved += queryStateObserved).flatMap(_ => IO.unit)
  }

  def selectQueryStateObservedByChecksum(queryStateObserved: QueryStateObserved):IO[Option[QueryStateObserved]] = {
    runIO(allQueryStatesObserved.filter(_.checksum === queryStateObserved.checksum).take(1).result).map{
      case Seq(queryRow) => Some(queryRow)
      case Seq() => None
      case x => throw QepAuditDatabaseAssertException(s"Expected zero or one query for id ${queryStateObserved.networkQueryId}, selected ${x.size} $x")
    }
  }
}

case class QepAuditDatabaseAssertException(message:String) extends Exception(message)


object QepQueryDb {

  val dataSource:DataSource = TestableDataSourceCreator.dataSource(QepQuerySchema.config)
  val timeout: Duration = QepQuerySchema.config.get("timeout",Duration.apply)

  val db = QepQueryDb(QepQuerySchema.schema,dataSource,timeout)

  val createTablesOnStart: Boolean = QepQuerySchema.config.getBoolean("createTablesOnStart")
  if(createTablesOnStart) QepQueryDb.db.createTables()

}

/**
  * Separate class to support schema generation without actually connecting to the database.
  *
  * @param jdbcProfile Database profile to use for the schema
  */
case class QepQuerySchema(jdbcProfile: JdbcProfile) {
  import jdbcProfile.api._

  def ddlForAllTables: jdbcProfile.DDL = {
    allQepQueryQuery.schema ++ allQepQueryFlags.schema ++ allQueryResultRows.schema ++ allBreakdownResultsRows.schema ++ allProblemDigestRows.schema ++ allQueryStatesObserved.schema ++ allQueryProblemDigestRows.schema
  }

  //to get the schema, use the REPL
  //println(QepQuerySchema.schema.ddlForAllTables.createStatements.mkString(";\n"))

  def createTables(database:Database): Unit = {
    try {
      val future = database.run(ddlForAllTables.create)
      Await.result(future,10 seconds)
    } catch {
      //I'd prefer to check and create schema only if absent. No way to do that with Oracle.
      case x:SQLException => Log.info("Caught exception while creating tables. Recover by assuming the tables already exist.",x)
    }
  }

  def dropTables(database:Database): Unit = {
    val future = database.run(ddlForAllTables.drop)
    //Really wait forever for the cleanup
    Await.result(future,Duration.Inf)
  }

  class QepQueries(tag:Tag) extends Table[QepQuery](tag,"previousQueries") {
    def networkId = column[NetworkQueryId]("networkId")
    def userName = column[UserName]("userName")
    def userDomain = column[String]("domain")
    def queryName = column[QueryName]("queryName")
    def expression = column[Option[String]]("expression")
    def dateCreated = column[Time]("dateCreated")
    def deleted = column[Boolean]("deleted")
    def queryXml = column[String]("queryXml")
    def changeDate = column[Long]("changeDate")
    def status = column[String]("status")

    def * = (networkId,userName,userDomain,queryName,expression,dateCreated,deleted,queryXml,changeDate,status) <> (QepQuery.tupled,QepQuery.unapply)

  }

  val allQepQueryQuery = TableQuery[QepQueries]
  val mostRecentQepQueryQuery: Query[QepQueries, QepQuery, Seq] = for(
    queries <- allQepQueryQuery if !allQepQueryQuery.filter(_.networkId === queries.networkId).filter(_.changeDate > queries.changeDate).exists
  ) yield queries
  val mostRecentVisibleQepQueries = mostRecentQepQueryQuery.filter(_.deleted === false)

  class QepQueryFlags(tag:Tag) extends Table[QepQueryFlag](tag,"queryFlags") {
    def networkId = column[NetworkQueryId]("networkId")
    def flagged = column[Boolean]("flagged")
    def flagMessage = column[String]("flagMessage")
    def changeDate = column[Long]("changeDate")

    def * = (networkId,flagged,flagMessage,changeDate) <> (QepQueryFlag.tupled,QepQueryFlag.unapply)
  }

  val allQepQueryFlags = TableQuery[QepQueryFlags]
  val mostRecentQueryFlags: Query[QepQueryFlags, QepQueryFlag, Seq] = for(
    queryFlags <- allQepQueryFlags if !allQepQueryFlags.filter(_.networkId === queryFlags.networkId).filter(_.changeDate > queryFlags.changeDate).exists
  ) yield queryFlags

  implicit val qepQueryResultTypesColumnType: JdbcType[ResultOutputType] with BaseTypedType[ResultOutputType] = MappedColumnType.base[ResultOutputType,String] ({
    resultType: ResultOutputType => QepQueryBreakdownResultsRow.queryResultTypesToString(resultType)
  },{
    string: String => QepQueryBreakdownResultsRow.stringsToQueryResultTypes(string)
  })

  class QepQueryResults(tag:Tag) extends Table[QueryResultRow](tag,"queryResults") {
    def resultId = column[Long]("resultId")
    def networkQueryId = column[NetworkQueryId]("networkQueryId")
    def instanceId = column[Long]("instanceId")
    def adapterNode = column[String]("adapterNode")
    def resultType = column[Option[ResultOutputType]]("resultType")
    def size = column[Long]("size")
    def startDate = column[Option[Long]]("startDate")
    def endDate = column[Option[Long]]("endDate")
    def status = column[String]("status")
    def statusMessage = column[Option[String]]("statusMessage")
    def changeDate = column[Long]("changeDate")

    def * = (resultId,networkQueryId,instanceId,adapterNode,resultType,size,startDate,endDate,status,statusMessage,changeDate) <> (QueryResultRow.tupled,QueryResultRow.unapply)
  }

  val allQueryResultRows = TableQuery[QepQueryResults]

  //Most recent query result rows for each queryId from each adapter
  val mostRecentQueryResultRows: Query[QepQueryResults, QueryResultRow, Seq] = for(
    queryResultRows <- allQueryResultRows if !allQueryResultRows.filter(_.networkQueryId === queryResultRows.networkQueryId).filter(_.adapterNode === queryResultRows.adapterNode).filter(_.changeDate > queryResultRows.changeDate).exists
  ) yield queryResultRows

  class QepQueryBreakdownResults(tag:Tag) extends Table[QepQueryBreakdownResultsRow](tag,"queryBreakdownResults") {
    def networkQueryId = column[NetworkQueryId]("networkQueryId")
    def adapterNode = column[String]("adapterNode")
    def resultId = column[Long]("resultId")
    def resultType = column[ResultOutputType]("resultType")
    def dataKey = column[String]("dataKey")
    def value = column[Long]("value")
    def changeDate = column[Long]("changeDate")

    def * = (networkQueryId,adapterNode,resultId,resultType,dataKey,value,changeDate) <> (QepQueryBreakdownResultsRow.tupled,QepQueryBreakdownResultsRow.unapply)
  }

  val allBreakdownResultsRows = TableQuery[QepQueryBreakdownResults]
  //Most recent query result rows for each queryId from each adapter
  val mostRecentBreakdownResultsRows: Query[QepQueryBreakdownResults, QepQueryBreakdownResultsRow, Seq] = for(
    breakdownResultsRows <- allBreakdownResultsRows if !allBreakdownResultsRows.
      filter(_.networkQueryId === breakdownResultsRows.networkQueryId).
      filter(_.adapterNode === breakdownResultsRows.adapterNode).
      filter(_.resultId === breakdownResultsRows.resultId).
      filter(_.resultType === breakdownResultsRows.resultType).
      filter(_.dataKey === breakdownResultsRows.dataKey).
      filter(_.changeDate > breakdownResultsRows.changeDate).exists
  ) yield breakdownResultsRows

  /*
case class ProblemDigest(codec: String, stampText: String, summary: String, description: String, detailsXml: NodeSeq) extends XmlMarshaller {
    */
  class QepResultProblemDigests(tag:Tag) extends Table [QepProblemDigestRow](tag,"queryResultProblemDigests") {
    def networkQueryId = column[NetworkQueryId]("networkQueryId")
    def adapterNode = column[String]("adapterNode")
    def codec = column[String]("codec")
    def stamp = column[String]("stamp")
    def summary = column[String]("summary")
    def description = column[String]("description")
    def details = column[String]("details")
    def changeDate = column[Long]("changeDate")

    def * = (networkQueryId,adapterNode,codec,stamp,summary,description,details,changeDate) <> (QepProblemDigestRow.tupled,QepProblemDigestRow.unapply)
  }

  val allProblemDigestRows = TableQuery[QepResultProblemDigests]
  val mostRecentProblemDigestRows: Query[QepResultProblemDigests, QepProblemDigestRow, Seq] = for(
    problemDigests <- allProblemDigestRows if !allProblemDigestRows.filter(_.networkQueryId === problemDigests.networkQueryId).filter(_.adapterNode === problemDigests.adapterNode).filter(_.changeDate > problemDigests.changeDate).exists
  ) yield problemDigests


  class QueryStatesObserved(tag:Tag) extends Table[QueryStateObserved](tag,"RESULTS_OBSERVED") {
    def networkId = column[NetworkQueryId]("NETWORKQUERYID")
    def checksum = column[Checksum]("CHECKSUM")
    def observedTime = column[Long]("OBSERVEDTIME")

    def * = (networkId,checksum,observedTime) <> (QueryStateObserved.tupled,QueryStateObserved.unapply)
  }

  val allQueryStatesObserved = TableQuery[QueryStatesObserved]

  class QueryProblemDigests(tag:Tag) extends Table[QueryProblemDigestRow](tag,"QUERY_PROBLEM_DIGESTS") {
    def networkId = column[NetworkQueryId]("NETWORKQUERYID")
    def codec = column[String]("CODEC")
    def stamp = column[String]("STAMP")
    def summary = column[String]("SUMMARY")
    def description = column[String]("DESCRIPTION")
    def details = column[String]("DETAILS")
    def changeDate = column[Long]("CHANGEDATE")

    def * = (networkId,codec,stamp,summary,description,details,changeDate) <> (QueryProblemDigestRow.tupled,QueryProblemDigestRow.unapply)
  }

  val allQueryProblemDigestRows = TableQuery[QueryProblemDigests]
  val mostRecentQueryProblemDigestRows: Query[QueryProblemDigests, QueryProblemDigestRow, Seq] = for(
    problemDigests <- allQueryProblemDigestRows if !allQueryProblemDigestRows.filter(_.networkId === problemDigests.networkId).filter(_.changeDate > problemDigests.changeDate).exists
  ) yield problemDigests

}

object QepQuerySchema {

  val allConfig:Config = ConfigSource.config
  val config:Config = allConfig.getConfig("shrine.queryEntryPoint.audit.database")

  val slickProfile:JdbcProfile = TestableDataSourceCreator.slickDriver(config)

  import net.shrine.config.ConfigExtensions
  val moreBreakdowns: Set[ResultOutputType] = config.getOptionConfigured("breakdownResultOutputTypes",ResultOutputTypes.fromConfig).getOrElse(Set.empty)

  val schema = QepQuerySchema(slickProfile)
}



case class QepQuery(
                     networkId:NetworkQueryId,
                     userName: UserName,
                     userDomain: String,
                     queryName: QueryName,
                     expression: Option[String],
                     dateCreated: Time,
                     deleted: Boolean,
                     queryXml: String,
                     changeDate: Time,
                     status: String = QueryStatuses.BeforeV26.statusName
                   ){

  def toQueryMaster(qepQueryFlag:Option[QepQueryFlag]):QueryMaster = {

    QueryMaster(
      queryMasterId = networkId.toString,
      networkQueryId = networkId,
      name = queryName,
      userId = userName,
      groupId = userDomain,
      createDate = XmlDateHelper.toXmlGregorianCalendar(dateCreated),
      flagged = qepQueryFlag.map(_.flagged),
      flagMessage = qepQueryFlag.map(_.flagMessage)
    )
  }

  def updateWith(update:UpdateQueryAtQep):QepQuery = {
    this.copy(status = update.queryStatus.statusName,changeDate = update.changeDate.underlying)
  }
}

object QepQuery extends ((NetworkQueryId,UserName,String,QueryName,Option[String],Time,Boolean,String,Time,String) => QepQuery) {

  def apply(query: ProtocolQuery,researcher:Researcher):QepQuery = {
    if(query.researcherId != researcher.id) throw new IllegalArgumentException(s"researcher id ${researcher.id} does not equal query researcher id ${query.researcherId} for $researcher and $query")

    new QepQuery(
      networkId = query.id.underlying,
      userName = researcher.userName.underlying,
      userDomain = researcher.userDomainName.underlying,
      queryName = query.queryName,
      expression = QueryDefinition.fromI2b2(query.queryDefinitionXml).get.expr.map(_.toString),
      dateCreated = query.versionInfo.createDate.underlying,
      deleted = false,
      queryXml = query.queryDefinitionXml,
      changeDate = query.versionInfo.changeDate.underlying,
      status = query.status.statusName
    )
  }
}

case class QepQueryFlag(
                       networkQueryId: NetworkQueryId,
                       flagged:Boolean,
                       flagMessage:String,
                       changeDate:Long
                       )

object QepQueryFlag extends ((NetworkQueryId,Boolean,String,Long) => QepQueryFlag) {
  def apply(flagQueryRequest: FlagQueryRequest):QepQueryFlag = {
    QepQueryFlag(
      networkQueryId = flagQueryRequest.networkQueryId,
      flagged = true,
      flagMessage = flagQueryRequest.message.getOrElse(""),
      changeDate = System.currentTimeMillis()
    )
  }

  def apply(unflagQueryRequest: UnFlagQueryRequest):QepQueryFlag = {
    QepQueryFlag(
      networkQueryId = unflagQueryRequest.networkQueryId,
      flagged = false,
      flagMessage = "",
      changeDate = System.currentTimeMillis()
    )
  }

}

//todo replace with a class per state
case class FullQueryResult(
                            resultId:Long,
                            networkQueryId:NetworkQueryId,
                            instanceId:Long,
                            adapterNode:String,
                            resultType:Option[ResultOutputType],
                            count:Long,
                            startDate:Option[Long],
                            endDate:Option[Long],
                            status:String,
                            statusMessage:Option[String],
                            changeDate:Long,
                            breakdownTypeToResults:Map[ResultOutputType,Seq[QepQueryBreakdownResultsRow]],
                            problemDigest:Option[XmlProblemDigest]
                          ) {

  def toQueryResult: QueryResult = {
    def resultEnvelopesFrom(breakdownTypeToResults:Map[ResultOutputType,Seq[QepQueryBreakdownResultsRow]]): Map[ResultOutputType, I2b2ResultEnvelope] = {
      def resultEnvelopeFrom(resultType:ResultOutputType,breakdowns:Seq[QepQueryBreakdownResultsRow]):I2b2ResultEnvelope = {
        val data = breakdowns.map(b => b.dataKey -> b.value).toMap
        I2b2ResultEnvelope(resultType,data)
      }
      breakdownTypeToResults.map(r => r._1 -> resultEnvelopeFrom(r._1,r._2))
    }

    QueryResult(
      resultId = resultId,
      instanceId = instanceId,
      resultType = resultType,
      setSize = count,
      startDate = startDate.map(XmlDateHelper.toXmlGregorianCalendar),
      endDate = endDate.map(XmlDateHelper.toXmlGregorianCalendar),
      description = Some(adapterNode),
      statusType = QueryResult.StatusType.valueOf(status).getOrElse(throw new IllegalStateException(s"Don't know a StatusType for $status")),
      statusMessage = statusMessage,
      breakdowns = resultEnvelopesFrom(breakdownTypeToResults),
      problem = problemDigest
    )
  }
}

object FullQueryResult {
  def apply(row:QueryResultRow,
            breakdownTypeToResults:Map[ResultOutputType,Seq[QepQueryBreakdownResultsRow]],
            problemDigest:Option[XmlProblemDigest]):FullQueryResult = {
    FullQueryResult(resultId = row.resultId,
      networkQueryId = row.networkQueryId,
      instanceId = row.instanceId,
      adapterNode = row.adapterNode,
      resultType = row.resultType,
      count = row.size,
      startDate = row.startDate,
      endDate = row.endDate,
      status = row.status,
      statusMessage = row.statusMessage,
      changeDate = row.changeDate,
      breakdownTypeToResults = breakdownTypeToResults,
      problemDigest = problemDigest
    )
  }
}

case class QueryResultRow(
                           resultId:Long,
                           networkQueryId:NetworkQueryId,
                           instanceId:Long,
                           adapterNode:String,
                           resultType:Option[ResultOutputType],
                           size:Long,
                           startDate:Option[Long],
                           endDate:Option[Long],
                           status:String,
                           statusMessage:Option[String],
                           changeDate:Long
                         ) {
}

object QueryResultRow extends ((Long,NetworkQueryId,Long,String,Option[ResultOutputType],Long,Option[Long],Option[Long],String,Option[String],Long) => QueryResultRow)
{
  //todo more dead code to remove
  def apply(networkQueryId:NetworkQueryId,result:QueryResult):QueryResultRow = {
    new QueryResultRow(
      resultId = result.resultId,
      networkQueryId = networkQueryId,
      instanceId = result.instanceId,
      adapterNode = result.description.getOrElse(s"$result has None in its description field, instead of the name of an adapter node."),
      resultType = result.resultType,
      size = result.setSize,
      startDate = result.startDate.map(_.toGregorianCalendar.getTimeInMillis),
      endDate = result.endDate.map(_.toGregorianCalendar.getTimeInMillis),
      status = result.statusType.name,
      statusMessage = result.statusMessage,
      changeDate = System.currentTimeMillis()
    )
  }

  def apply(result:Result):QueryResultRow = {
    result match {
      case resultProgress:ResultProgress => QueryResultRow(
        resultId = resultProgress.id.underlying,
        networkQueryId = resultProgress.queryId.underlying,
        instanceId = resultProgress.crcQueryInstanceId.getOrElse(-1L), //todo change instanceId to an optional field
        adapterNode = resultProgress.adapterNodeName.underlying,
        resultType = None,
        size = -1L, //todo change size to an optional field
        startDate = Some(resultProgress.versionInfo.createDate.underlying),
        endDate = None,
        status = resultProgress.status.statusName,
        statusMessage = resultProgress.statusMessage,
        changeDate = resultProgress.versionInfo.changeDate.underlying
      )
      case errorResult:ErrorResult => QueryResultRow(
        resultId = errorResult.id.underlying,
        networkQueryId = errorResult.queryId.underlying,
        instanceId = errorResult.crcQueryInstanceId.getOrElse(-1L), //todo change instanceId to an optional field
        adapterNode = errorResult.adapterNodeName.underlying,
        resultType = None,
        size = -1L, //todo change size to an optional field
        startDate = Some(errorResult.versionInfo.createDate.underlying),
        endDate = Some(errorResult.versionInfo.createDate.underlying),
        status = errorResult.status.statusName,
        statusMessage = errorResult.statusMessage,
        changeDate = errorResult.versionInfo.changeDate.underlying
      )
      case crcResult:CrcResult => QueryResultRow(
        resultId = crcResult.id.underlying,
        networkQueryId = crcResult.queryId.underlying,
        instanceId = crcResult.crcQueryInstanceId.getOrElse(-1L), //todo change instanceId to an optional field
        adapterNode = crcResult.adapterNodeName.underlying,
        resultType = Some(crcResult.resultType),
        size = crcResult.count, //todo change size to an optional field
        startDate = Some(crcResult.versionInfo.createDate.underlying),
        endDate = Some(crcResult.versionInfo.createDate.underlying),
        status = crcResult.status.statusName,
        statusMessage = crcResult.statusMessage,
        changeDate = crcResult.versionInfo.changeDate.underlying
      )
    }
  }


  def apply(networkQueryId:NetworkQueryId,result:FullQueryResult):QueryResultRow = {
    new QueryResultRow(
      resultId = result.resultId,
      networkQueryId = networkQueryId,
      instanceId = result.instanceId,
      adapterNode = result.adapterNode,
      resultType = result.resultType,
      size = result.count,
      startDate = result.startDate,
      endDate = result.endDate,
      status = result.status,
      statusMessage = result.statusMessage,
      changeDate = result.changeDate
    )
  }
}

case class QepQueryBreakdownResultsRow(
                                        networkQueryId: NetworkQueryId,
                                        adapterNode:String,
                                        resultId:Long,
                                        resultType: ResultOutputType,
                                        dataKey:String,
                                        value:Long,
                                        changeDate:Long
                                      )

object QepQueryBreakdownResultsRow extends ((NetworkQueryId,String,Long,ResultOutputType,String,Long,Long) => QepQueryBreakdownResultsRow){

  //todo more dead code
  def breakdownRowsFor(networkQueryId:NetworkQueryId,
                       adapterNode:String,
                       resultId:Long,
                       breakdown:(ResultOutputType,I2b2ResultEnvelope)): Iterable[QepQueryBreakdownResultsRow] = {
    breakdown._2.data.map(b => QepQueryBreakdownResultsRow(networkQueryId,adapterNode,resultId,breakdown._1,b._1,b._2,System.currentTimeMillis()))
  }

  def breakdownRowsFor(crcResult: CrcResult): Seq[QepQueryBreakdownResultsRow] = {
    crcResult.breakdowns.map{ bd: Breakdowns =>
      bd.counts.flatMap { axis =>
        axis._2.map { nameCount =>
          QepQueryBreakdownResultsRow(
            networkQueryId = crcResult.queryId.underlying,
            adapterNode = crcResult.adapterNodeName.underlying,
            resultId = crcResult.crcQueryInstanceId.get,
            resultType = stringsToQueryResultTypes(axis._1),
            dataKey = nameCount._1,
            value = nameCount._2,
            changeDate = crcResult.versionInfo.changeDate.underlying
          )
        }
      }
    }.getOrElse(Seq.empty)
  }

  val qepQueryResultTypes: Set[ResultOutputType] = DefaultBreakdownResultOutputTypes.toSet ++ ResultOutputType.values ++ QepQuerySchema.moreBreakdowns
  val stringsToQueryResultTypes: Map[String, ResultOutputType] = qepQueryResultTypes.map(x => (x.name,x)).toMap
  val queryResultTypesToString: Map[ResultOutputType, String] = stringsToQueryResultTypes.map(_.swap)

}

case class QepProblemDigestRow(
                                networkQueryId: NetworkQueryId,
                                adapterNode: String,
                                codec: String,
                                stampText: String,
                                summary: String,
                                description: String,
                                details: String,
                                changeDate:Long
                              ){
  def toProblemDigest: XmlProblemDigest = {
    XmlProblemDigest(
      codec,
      stampText,
      summary,
      description,
      if(!details.isEmpty) XML.loadString(details)
      else <details/>,
      //TODO: FIGURE OUT HOW TO GET AN ACTUAL EPOCH INTO HERE
      0
    )
  }
}

object QepProblemDigestRow extends ((NetworkQueryId,String,String,String,String,String,String,Long) => QepProblemDigestRow) {
  def apply(errorResult:ErrorResult):QepProblemDigestRow = new QepProblemDigestRow(
    networkQueryId = errorResult.queryId.underlying,
    adapterNode = errorResult.adapterNodeName.underlying,
    codec = errorResult.problemDigest.codec,
    stampText = errorResult.problemDigest.stampText,
    summary = errorResult.problemDigest.summary,
    description = errorResult.problemDigest.description,
    details = errorResult.problemDigest.detailsXml.toString(),
    changeDate = errorResult.versionInfo.changeDate.underlying
  )
}

case class QepDatabaseProblem(x:Exception) extends AbstractProblem(ProblemSources.Qep){
  override val summary = "A problem encountered while using a database."

  override val throwable = Some(x)

  override val description: QueryName = x.getMessage
}

case class QueryWithResultStatus(
                                  networkId:NetworkQueryId,
                                  userName: UserName,
                                  userDomain: String,
                                  queryName: QueryName,
                                  expression: Option[String],
                                  dateCreated: Time,
                                  deleted: Boolean,
                                  queryXml: String,
                                  changeDate: Time,
                                  status: String = QueryStatuses.BeforeV26.statusName,
                                  queryFlag: Option[QueryFlag],
                                  isComplete: Boolean,
                                  isQueryError: Boolean,
                                  isResultsError: Boolean,
                                  observed: Boolean
                                )

object QueryWithResultStatus {
  def apply(query: QepQuery,
            results: Seq[QueryResultRow],
            queryFlag: Option[QueryFlag],
            observed: Boolean
            ): QueryWithResultStatus = {

    new QueryWithResultStatus(
      networkId = query.networkId,
      userName = query.userName,
      userDomain = query.userDomain,
      queryName = query.queryName,
      expression = query.expression,
      dateCreated = query.dateCreated,
      deleted = query.deleted,
      queryXml = query.queryXml,
      changeDate = query.changeDate,
      status = query.status,
      queryFlag = queryFlag,
      isComplete = isComplete(results),
      isQueryError = isQueryError(query, results),
      isResultsError = isResultsError(query, results),
      observed = observed
    )
  }

  def isComplete(results: Seq[QueryResultRow]): Boolean = {
    if (results.isEmpty) false
    else results.forall{ p =>
      ResultStatuses.namesToStatuses.get(p.status).map(_.isFinal).getOrElse {
      QueryResult.StatusType.valueOf(p.status).get.isDone}
    }
  }

  def isQueryError(query:  QepQuery, results: Seq[QueryResultRow]): Boolean = {
    val isQueryError: Boolean = QueryStatuses.namesToStatuses.get(query.status).exists(_.isError)

    isQueryError
  }

  def isResultsError(query:  QepQuery, results: Seq[QueryResultRow]): Boolean = {

    if(results.isEmpty) false
    else results.forall { p =>
      ResultStatuses.namesToStatuses.get(p.status).map(_.isError).getOrElse {
        QueryResult.StatusType.valueOf(p.status).get.isError
      }
    }
  }
}

case class QueryStateObserved(
                               networkQueryId:NetworkQueryId,
                               checksum: Checksum,
                               observedTime: Time
                             )

object QueryStateObserved extends ((NetworkQueryId,Checksum,Time) => QueryStateObserved) {

  private def create(networkQueryId: NetworkQueryId, checksum: Checksum): QueryStateObserved = {
    new QueryStateObserved(
      networkQueryId = networkQueryId,
      checksum = checksum,
      observedTime = System.currentTimeMillis()
    )
  }

  def apply(queryAndResults: QueryAndResults): QueryStateObserved = {

    val query: QepQuery = queryAndResults.fullQuery.query
    val queryResults: Seq[ObservedResult] = queryAndResults.fullResults.map(QueryResultRow(query.networkId, _)).sortBy(_.adapterNode).map(ObservedResult(_))

    val queriesAndResult = QueriesAndResult(ObservedQuery(query), queryResults)
    val checksum = queriesAndResult.hashCode()
    val networkQueryId: NetworkQueryId = query.networkId

    create(networkQueryId = networkQueryId, checksum = checksum)
  }

  def apply(query: QepQuery, queryResults: Seq[QueryResultRow]): QueryStateObserved = {

    val queriesAndResult = QueriesAndResult(ObservedQuery(query), queryResults.sortBy(_.adapterNode).map(ObservedResult(_)))
    val checksum = queriesAndResult.hashCode()
    val networkQueryId: NetworkQueryId = query.networkId

    create(networkQueryId = networkQueryId, checksum = checksum)
  }

  case class ObservedQuery(networkId:NetworkQueryId,
                           status: String = QueryStatuses.BeforeV26.statusName
                          )
  object ObservedQuery {
    def apply(query: QepQuery): ObservedQuery = {
      ObservedQuery (
        networkId = query.networkId,
        status = query.status
      )
    }
  }

  case class ObservedResult(resultId:Long,
                             networkQueryId:NetworkQueryId,
                             adapterNode:String,
                             status:String
                          )
  object ObservedResult {
    def apply(result: QueryResultRow): ObservedResult = {
      ObservedResult(
        resultId = result.resultId,
        networkQueryId = result.networkQueryId,
        adapterNode = result.adapterNode,
        status = result.status
      )
    }
  }

  case class QueriesAndResult(query: ObservedQuery, results: Seq[ObservedResult])
}


case class QueryProblemDigestRow(
                                networkQueryId: NetworkQueryId,
                                codec: String,
                                stampText: String,
                                summary: String,
                                description: String,
                                details: String,
                                changeDate:Long
                              ){
  def toProblemDigest: XmlProblemDigest = {
    XmlProblemDigest(
      codec,
      stampText,
      summary,
      description,
      if(!details.isEmpty) XML.loadString(details)
      else <details/>,
      changeDate
    )
  }
}

object QueryProblemDigestRow extends ((NetworkQueryId,String,String,String,String,String,Long) => QueryProblemDigestRow) {
  def apply(queryError:QueryError):QueryProblemDigestRow = new QueryProblemDigestRow(
    networkQueryId = queryError.id.underlying,
    codec = queryError.problemDigest.codec,
    stampText = queryError.problemDigest.stampText,
    summary = queryError.problemDigest.summary,
    description = queryError.problemDigest.description,
    details = queryError.problemDigest.detailsXml.toString(),
    changeDate = queryError.problemDigest.epoch
  )

  def apply(update:UpdateQueryAtQepWithError):QueryProblemDigestRow = new QueryProblemDigestRow(
    networkQueryId = update.queryId.underlying,
    codec = update.problemDigest.codec,
    stampText = update.problemDigest.stampText,
    summary = update.problemDigest.summary,
    description = update.problemDigest.description,
    details = update.problemDigest.detailsXml.toString(),
    changeDate = update.problemDigest.epoch
  )
}

case class FullQuery(query: QepQuery,
                     flag:Option[QueryFlag],
                     problemDigest: Option[XmlProblemDigest]
                     )

object FullQuery{
  def apply(problemDigest: Option[QueryProblemDigestRow]= None, query: QepQuery, flag:Option[QueryFlag]= None): FullQuery = {
    val xmlProblemDigest: Option[XmlProblemDigest] = problemDigest.map(_.toProblemDigest)
    new FullQuery(query, flag, xmlProblemDigest)
  }
}

case class QueryAndResults(
                            fullQuery:FullQuery,
                            fullResults: Seq[FullQueryResult]
                     )

case class QueryFlag(
                      flagged:Boolean,
                      flagMessage:String,
                      changeDate:Long
                    )

object QueryFlag{
  def apply(qepQueryFlag: QepQueryFlag): QueryFlag = QueryFlag(qepQueryFlag.flagged, qepQueryFlag.flagMessage, qepQueryFlag.changeDate)
}



case class BreakdownResultsForType(resultType:ResultOutputType,results:Seq[BreakdownResult])

object BreakdownResultsForType {
  def apply(adapterName: String, breakdownType: ResultOutputType, breakdowns: Seq[QepQueryBreakdownResultsRow]): BreakdownResultsForType = {
    val breakdownResults = breakdowns.filter(_.adapterNode == adapterName).map(row => BreakdownResult(row.dataKey,row.value,row.changeDate))

    BreakdownResultsForType(breakdownType,breakdownResults)
  }
}

case class BreakdownResult(dataKey:String,value:Long,changeDate:Long)

case class RequestingUsernameDoesNotMatchQueryException(networkId: Long, requestingUser: String, actualUser: User)
  extends Exception(s"Query $networkId belongs to user ${actualUser.username} but requested by user $requestingUser")

case class NoQueryForQueryIdException(networkId: Long)
  extends Exception(s"No query found for $networkId")
