package net.shrine.qep.querydb

import net.shrine.crypto.NewTestKeyStore
import net.shrine.i2b2.protocol.pm.User
import net.shrine.problem.{AbstractProblem, JsonProblemDigest, ProblemSources, TestProblem, XmlProblemDigest}
import net.shrine.protocol.QueryResult.StatusType
import net.shrine.protocol.version.{NodeKey, NodeName, QueryStatuses, UserDomainName}
import net.shrine.protocol.version.v1.{Node, Query, Researcher, Topic, UpdateQueryAtQepWithError}
import net.shrine.protocol.{Credential, DefaultBreakdownResultOutputTypes, I2b2ResultEnvelope, QueryResult, ResultOutputType}
import net.shrine.util.{ShouldMatchersForJUnit, XmlDateHelper}
import org.junit.{After, Before, Test}

/**
  * @author david 
  * @since 1/20/16
  */
class QepQueryDbTest extends ShouldMatchersForJUnit {

  val qepQuery = QepQuery(
    networkId = 1L,
    userName = "ben",
    userDomain = "testDomain",
    queryName = "testQuery1",
    expression = Some("testExpression"),
    dateCreated = System.currentTimeMillis(),
    deleted = false,
    queryXml = "testXML",
    changeDate = System.currentTimeMillis(),
    status = QueryStatuses.ReadyForAdapters.statusName
  )

  val secondQepQuery = QepQuery(
    networkId = 2L,
    userName = "dave",
    userDomain = "testDomain",
    queryName = "testQuery2",
    expression = Some("testExpression"),
    deleted = false,
    dateCreated = System.currentTimeMillis(),
    queryXml = "testXML",
    changeDate = System.currentTimeMillis() - 1000,
    status = QueryStatuses.ReadyForAdapters.statusName
  )

  val thirdQepQuery = QepQuery(
    networkId = 3L,
    userName = "ben",
    userDomain = "testDomain",
    queryName = "testQuery3",
    expression = Some("testExpression"),
    dateCreated = System.currentTimeMillis(),
    deleted = false,
    queryXml = "testXML",
    changeDate = System.currentTimeMillis() - 2000,
    status = QueryStatuses.ReadyForAdapters.statusName
  )


  val flag = QepQueryFlag(
    networkQueryId = 1L,
    flagged = true,
    flagMessage = "This query is flagged",
    changeDate = System.currentTimeMillis()
  )

  val testNode = Node(
    name = new NodeName("testNodeName"),
    key = new NodeKey("testNodeKey"),
    userDomainName = new UserDomainName("testUserDomainName"),
    adminEmail = "email@foo"
  )

  val researcherUserName: String = "ben"
  val researcherFullName: String  = researcherUserName

  val researcherUser: User = User(
    fullName = researcherUserName,
    username = researcherFullName,
    domain = "domain",
    credential = new Credential("researcher's password",false),
    params = Map(),
    rolesByProject = Map()
  )

  @Test
  def testInsertQepQuery() {

    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepQuery(secondQepQuery)

    val results = QepQueryDb.db.selectAllQepQueries
    results should equal(Seq(qepQuery,secondQepQuery))
  }

  @Test
  def testSelectQueryById() {

    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepQuery(secondQepQuery)

    val result1: Option[QepQuery] = QepQueryDb.db.selectQueryById(qepQuery.networkId)
    result1 should equal(Some(qepQuery))

    val result2: Option[QepQuery] = QepQueryDb.db.selectQueryById(secondQepQuery.networkId)
    result2 should equal(Some(secondQepQuery))

    val result3: Option[QepQuery] = QepQueryDb.db.selectQueryById(-1L)
    result3 should equal(None)
  }

  @Test
  def testSelectQepQueriesForUser() {

    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepQuery(secondQepQuery)
    QepQueryDb.db.insertQepQuery(thirdQepQuery)

    val results = QepQueryDb.db.selectPreviousQueriesByUserAndDomain("ben","testDomain", sortBy= Some("queryName.asc"))
    results should equal(Seq(qepQuery,thirdQepQuery))
  }

  @Test
  def testSelectQepQueriesForUserWithLimit() {

    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepQuery(secondQepQuery)
    QepQueryDb.db.insertQepQuery(thirdQepQuery)

    val results = QepQueryDb.db.selectPreviousQueriesByUserAndDomain("ben","testDomain",None,Some(100))
    results should equal(Seq(thirdQepQuery, qepQuery))
  }

  @Test
  def testSelectQepQueriesForUserWithSkipAndLimit() {

    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepQuery(secondQepQuery)
    QepQueryDb.db.insertQepQuery(thirdQepQuery)

    val results = QepQueryDb.db.selectPreviousQueriesByUserAndDomain("ben","testDomain",Some(1),Some(100))
    results should equal(Seq(qepQuery))
  }

  @Test
  def testSelectQueryFlag() {

    val results1 = QepQueryDb.db.selectMostRecentQepQueryFlagFor(1L)
    results1 should equal(None)

    QepQueryDb.db.insertQepQueryFlag(flag)

    val results2 = QepQueryDb.db.selectMostRecentQepQueryFlagFor(1L)
    results2 should equal(Some(flag))
  }

  @Test
  def testSelectQueryFlags() {

    val results1 = QepQueryDb.db.selectMostRecentQepQueryFlagsFor(Set(1L,2L))
    results1 should equal(Map.empty)

    QepQueryDb.db.insertQepQueryFlag(flag)

    val results2 = QepQueryDb.db.selectMostRecentQepQueryFlagsFor(Set(1L,2L))
    results2 should equal(Map(1L -> flag))
  }

  val qepResultRowFromExampleCom = QueryResultRow(
    resultId = 10L,
    networkQueryId = 1L,
    instanceId = 100L,
    adapterNode = "example.com",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 30L,
    startDate = Some(System.currentTimeMillis() - 60),
    endDate = Some(System.currentTimeMillis() - 30),
    status = QueryResult.StatusType.Finished.name,
    statusMessage = None,
    changeDate = System.currentTimeMillis()
  )

  @Test
  def testInsertQueryResultRow() {

    QepQueryDb.db.insertQepResultRow(qepResultRowFromExampleCom)

    val results = QepQueryDb.db.selectMostRecentQepResultRowsFor(1L)
    results should equal(Seq(qepResultRowFromExampleCom))
  }

  val queryResult = QueryResult(
    resultId = 20L,
    instanceId = 200L,
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    setSize = 2000L,
    startDate = Some(XmlDateHelper.now),
    endDate = Some(XmlDateHelper.now),
    description = Some("example.com"),
    statusType = StatusType.Finished,
    statusMessage = None
  )

  @Test
  def testInsertQueryResult(): Unit = {
    QepQueryDb.db.insertQueryResult(2L,queryResult)

    val results = QepQueryDb.db.selectMostRecentQepResultsFor(2L)

    results should equal(Seq(queryResult))
  }

  val qepResultRowFromExampleComInThePast = QueryResultRow(
    resultId = 8L,
    networkQueryId = 1L,
    instanceId = 100L,
    adapterNode = "example.com",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 0L,
    startDate = qepResultRowFromExampleCom.startDate,
    endDate = None,
    status = QueryResult.StatusType.Processing.name,
    statusMessage = None,
    changeDate = qepResultRowFromExampleCom.changeDate - 40
  )

  val qepResultRowFromGeneralHospital = QueryResultRow(
    resultId = 100L,
    networkQueryId = 1L,
    instanceId = 100L,
    adapterNode = "generalhospital.org",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 100L,
    startDate = Some(System.currentTimeMillis() - 60),
    endDate = Some(System.currentTimeMillis() - 30),
    status = QueryResult.StatusType.Finished.name,
    statusMessage = None,
    changeDate = System.currentTimeMillis()
  )

  @Test
  def testGetMostRecentResultRows() {

    QepQueryDb.db.insertQepResultRow(qepResultRowFromExampleComInThePast)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromGeneralHospital)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromExampleCom)

    val results = QepQueryDb.db.selectMostRecentQepResultRowsFor(1L)
    results.to[Set] should equal(Set(qepResultRowFromExampleCom,qepResultRowFromGeneralHospital))
  }

  def maleRow(breakdownChangeDate:Long) = QepQueryBreakdownResultsRow(
    networkQueryId = 1L,
    adapterNode = "example.com",
    resultId = 100L,
    resultType = DefaultBreakdownResultOutputTypes.PATIENT_GENDER_COUNT_XML,
    dataKey = "male",
    value = 388,
    changeDate = breakdownChangeDate
  )

  def femaleRow(breakdownChangeDate:Long) = QepQueryBreakdownResultsRow(
    networkQueryId = 1L,
    adapterNode = "example.com",
    resultId = 100L,
    resultType = DefaultBreakdownResultOutputTypes.PATIENT_GENDER_COUNT_XML,
    dataKey = "female",
    value = 390,
    changeDate = breakdownChangeDate
  )

  def unknownRow(breakdownChangeDate:Long) = QepQueryBreakdownResultsRow(
    networkQueryId = 1L,
    adapterNode = "example.com",
    resultId = 100L,
    resultType = DefaultBreakdownResultOutputTypes.PATIENT_GENDER_COUNT_XML,
    dataKey = "unknown",
    value = 4,
    changeDate = breakdownChangeDate
  )

  @Test
  def testInsertBreakdownRows(): Unit = {
    val breakdownChangeDate = System.currentTimeMillis()

    val expectedMaleRow = maleRow(breakdownChangeDate)
    val expectedFemaleRow = femaleRow(breakdownChangeDate)
    val expectedUnknownRow = unknownRow(breakdownChangeDate)

    QepQueryDb.db.insertQueryBreakdown(maleRow(breakdownChangeDate))
    QepQueryDb.db.insertQueryBreakdown(femaleRow(breakdownChangeDate))
    QepQueryDb.db.insertQueryBreakdown(unknownRow(breakdownChangeDate))

    val results = QepQueryDb.db.selectAllBreakdownResultsRows
    results.to[Set] should equal(Set(expectedMaleRow,expectedFemaleRow,expectedUnknownRow))
  }

  val breakdowns: Map[ResultOutputType, I2b2ResultEnvelope] = Map(
    DefaultBreakdownResultOutputTypes.PATIENT_GENDER_COUNT_XML -> I2b2ResultEnvelope(DefaultBreakdownResultOutputTypes.PATIENT_GENDER_COUNT_XML,Map(
      "male" -> 388,"female" -> 390,"unknown" -> 4)))

  @Test
  def testInsertQueryResultWithBreakdowns(): Unit = {
    val queryResultWithBreakdowns = queryResult.copy(breakdowns = breakdowns)

    QepQueryDb.db.insertQueryResult(2L,queryResultWithBreakdowns)

    val results = QepQueryDb.db.selectMostRecentQepResultsFor(2L)

    //Had a sporadic failure 8-9-2016 on David's lap top. Did not fail on rerun.
    results should equal(Seq(queryResultWithBreakdowns))
  }

  @Test
  def testInsertQueryResultWithBreakdownsWithDifferentTimestamps(): Unit = {
    val queryResultWithBreakdowns = queryResult.copy(breakdowns = breakdowns)

    QepQueryDb.db.insertQueryResult(1L,queryResult)

    val breakdownChangeDate = System.currentTimeMillis()

    QepQueryDb.db.insertQueryBreakdown(maleRow(breakdownChangeDate))
    QepQueryDb.db.insertQueryBreakdown(femaleRow(breakdownChangeDate))
    QepQueryDb.db.insertQueryBreakdown(unknownRow(breakdownChangeDate+1))

    val results = QepQueryDb.db.selectMostRecentQepResultsFor(1L)

    //Had a sporadic failure 8-9-2016 on David's lap top. Did not fail on rerun.
    results should equal(Seq(queryResultWithBreakdowns))
  }

  @Test
  def testInsertQueryResultWithProblem(): Unit = {
    val queryResultWithProblem = queryResult.copy(statusType = StatusType.Error,problem = Some(XmlProblemDigest.create(TestProblem())))

    QepQueryDb.db.insertQueryResult(2L,queryResultWithProblem)

    val results = QepQueryDb.db.selectMostRecentQepResultsFor(2L)

    results should equal(Seq(queryResultWithProblem))
  }

  @Test
  def testInsertQepQueryWithError() {
    val researcher = Researcher.createRepeatableForV26(
      userName = "ben",
      userDomainName = "testQuery",
      nodeId = testNode.id
    )

    val topic = Topic.createRepeatableForV26(
      researcherId = researcher.id,
      name = "myTopic",
      description = "topic description"
    )

    val networkId = 3L
    val queryWithId: Query = Query.create(
      id = networkId,
      queryDefinitionXml = "<query/>",
      outputTypesXml = "<outputypes/>",
      queryName = "my query name",
      nodeOfOriginId = testNode.id,
      researcherId = researcher.id,
      topicId = topic.id,
      projectName = "my test project",
      NewTestKeyStore.certCollection
    )

    QepQueryDb.db.insertQueryIO(queryWithId,researcher).unsafeRunSync()

    case class ErrorRunningQuery(t: Throwable) extends AbstractProblem(ProblemSources.Qep){
      override def summary: String = "Error running query"
      override def description: String = "An error occurred when running query"
    }

    val problem = ErrorRunningQuery(new Exception("This is a query with an error"))

    val queryWithError = queryWithId.toError(JsonProblemDigest(problem))
    val storeQueryStatus = UpdateQueryAtQepWithError(queryWithId.id,JsonProblemDigest(problem),queryWithError.versionInfo.changeDate)

    QepQueryDb.db.updateQepQueryIO(storeQueryStatus).unsafeRunSync()

    val problemDigestRow = QueryProblemDigestRow(queryWithError)
    val expectedResult = FullQuery(Some(problemDigestRow),QepQuery(queryWithError,researcher))
    val results: Option[FullQuery] = QepQueryDb.db.selectFullQuery(networkId).unsafeRunSync()
    results.get should equal(expectedResult)
  }

  @Before
  def beforeEach(): Unit = {
    QepQueryDb.db.createTables()
  }

  @After
  def afterEach(): Unit = {
    QepQueryDb.db.dropTables()
  }
}