package net.shrine.api.dashboard

import cats.effect.IO
import com.typesafe.config.Config
import io.circe.generic.auto.exportEncoder
import io.circe.syntax.EncoderOps
import net.shrine.adapter.mappings.AdapterMappings
import net.shrine.authentication.PmAuthenticator
import net.shrine.authorization.steward.{Role, UserName, researcherRole, stewardRole}
import net.shrine.circe.CirceXml
import net.shrine.config.ConfigExtensions
import net.shrine.crypto.{BouncyKeyStoreCollection, DownStreamCertCollection, HubCertCollection, KeyStoreDescriptor, KeyStoreDescriptorParser, KeyStoreEntry, UtilHasher}
import net.shrine.http4s.auth.BasicOrBearerAuthMiddleware
import net.shrine.http4s.client.legacy.{EndpointConfig, OntClient, Poster, PosterOntClient}
import net.shrine.i2b2.protocol.pm.User
import net.shrine.log.{Log, Loggable}
import net.shrine.problem.XmlProblemDigest
import net.shrine.protocol.query.Term
import net.shrine.protocol.HiveCredentials
import net.shrine.source.ConfigSource
import net.shrine.util.{PeerToPeerModel, SingleHubModel, Versions}
import org.http4s.dsl.impl.{+&, ->, /, :?, Auth, OptionalQueryParamDecoderMatcher}
import org.http4s.dsl.io.{GET, NotFound, Ok, Root, http4sNotFoundSyntax, http4sOkSyntax}
import org.http4s.server.Router
import org.http4s.{AuthedRoutes, HttpRoutes, Response, Status}

import scala.collection.immutable.{Map, Set}
import scala.util.control.NonFatal

case class DashboardService() extends Auth {

  def router: HttpRoutes[IO] = {
    Router[IO](
      "" -> rootRoutes,
      "/user" -> privateUserRoutes,
      "/admin" -> adminRoutes
    )
  }

  private val privateUserRoutes = AuthenticationFailureMiddleWare(BasicOrBearerAuthMiddleware(knownUserRoutes))

  def knownUserRoutes: AuthedRoutes[User, IO] = AuthedRoutes.of[User, IO] {

    case GET -> Root / "whoami" as user => Ok(OutboundUser.createFromUser(user).asJson.toString())

    case x => NotFound(s"The dashboard service does not respond to $x")
  }

  private def adminRoutes: HttpRoutes[IO] = {
    Router[IO](
      "/status" -> privateStatusRoutes
    )
  }

  private val privateStatusRoutes = BasicOrBearerAuthMiddleware(authAdminRoutes)

  private val config: Config = ConfigSource.config

  def authAdminRoutes: AuthedRoutes[User, IO] = AuthedRoutes.of[User, IO] {

    // _ s are all "user" which is only used to authenticate
    case GET -> Root / "summary" as _ => Ok(Summary(config).asJson.toString())

    case GET -> Root / "queryAdapterTest" as _ => Ok(QueryAdapterTestResult().asJson.toString())

    case GET -> Root / "i2b2" as _ => Ok(I2b2().asJson.toString())

    case GET -> Root / "optionalParts" as _ => Ok(OptionalParts().asJson.toString())

    case GET -> Root / "adapter" as _ => Ok(Adapter(config).asJson.toString())

    case GET -> Root / "qep" as _ => Ok(Qep(keyStoreDescriptor).asJson.toString())

    case GET -> Root / "hub" as _ => Ok(Hub().asJson.toString())

    case GET -> Root / "keystore" as _ => Ok(KeyStoreReport(config.getConfig("shrine.keystore"),keyStoreDescriptor).asJson.toString())

    case GET -> Root / "config" as _  => Ok(ParsedConfig(config).asJson.toString())

      //classpath is a synonym for config
    case GET -> Root / "classpath" as _  => Ok(ParsedConfig(config).asJson.toString())

    case GET -> Root / "problems" :? OptionalOffsetPreModParam(offset) +& OptionalnPreMaxParam(nPreMax) +& OptionalEpochParam(epoch) as _ => queryProblems(offset, nPreMax, epoch)

    case x => NotFound(s"The dashboard service does not respond to $x")
  }

  val dashboardInfo: String  =
    """
      |The SHRINE dashboard service.
      |
      |This API gives admins general information about the SHRINE network.
      |
    """.stripMargin

  object OptionalOffsetPreModParam extends OptionalQueryParamDecoderMatcher[Int]("offset")
  object OptionalnPreMaxParam extends OptionalQueryParamDecoderMatcher[Int]("n")
  object OptionalEpochParam extends OptionalQueryParamDecoderMatcher[Long]("epoch")

  private def rootRoutes: HttpRoutes[IO] = HttpRoutes.of[IO] {
    case GET -> Root => Ok(dashboardInfo)

    case _ @ GET -> Root / "ping" => Ok("pong")

    case _ @ GET -> Root / "version" => getVersion

    case x => NotFound(s"The dashboard service does not respond to $x")
  }

  def getVersion: IO[Response[IO]] = {
    val response: AppVersion = AppVersion()

    Ok(response.asJson.toString())
  }

  //noinspection ConvertExpressionToSAM
  def queryProblems(offsetPreMod: Option[Int] = Some(0), nPreMax: Option[Int] = Some(20), epoch: Option[Long]): IO[Response[IO]] = {
    import io.circe.parser.parse
    import io.circe.syntax.EncoderOps
    import io.circe.{Encoder, Json}
    import org.http4s.circe.jsonEncoder

    implicit val encodeXmlProblemDigest: Encoder[Seq[XmlProblemDigest]] = new Encoder[Seq[XmlProblemDigest]] {
      final def apply(seq: Seq[XmlProblemDigest]): Json = {

        val res = for(xml:XmlProblemDigest <- seq) yield {
          val json  = CirceXml.toJson(xml.detailsXml)
          val jsonString: String = json.toString()
          val detailsJson = parse(jsonString).getOrElse(Json.Null)

          Json.obj(
            ("codec", Json.fromString(xml.codec)),
            ("stampText", Json.fromString(xml.stampText)),
            ("summary", Json.fromString(xml.summary)),
            ("description", Json.fromString(xml.description)),
            ("detailsXml", detailsJson)
          )
        }
        Json.fromValues(res)
      }
    }

    implicit val ProblemResponse: Encoder[DashboardStatusHelper.ProblemResponse] = new Encoder[DashboardStatusHelper.ProblemResponse] {
      final def apply(problemResponse: DashboardStatusHelper.ProblemResponse): Json = {

        Json.obj(
        ("size", Json.fromInt(problemResponse.size)),
        ("offset", Json.fromInt(problemResponse.offset)),
        ("n", Json.fromInt(problemResponse.n)),
        ("problems", problemResponse.problems.asJson)
        )
      }
    }

    val problemResponse: DashboardStatusHelper.ProblemResponse = DashboardStatusHelper.getProblems(offsetPreMod, nPreMax, epoch)
    val json: Json = problemResponse.asJson
    IO(Response(status = Status.Ok).withEntity(json))
  }

  lazy val keyStoreDescriptor:KeyStoreDescriptor = {
    val config = ConfigSource.config
    KeyStoreDescriptorParser.apply(
      config.getConfig("shrine.keystore"),
      config.getConfig("shrine.hub"),
      config.getConfig("shrine.queryEntryPoint"))
  }
}

// Replaces StewardQueryAuthorizationService so that we never transmit a password
//todo remove or rework in SHRINE-3055
case class Steward(stewardBaseUrl: String = "Not needed", qepUsername: String = "Not used in 1.26", password: String = "REDACTED")

case class Qep(
                maxQueryWaitTimeMillis: Long,
                create: Boolean,
                attachSigningCert: Boolean,
                authorizationType: String,
                authenticationType: String,
                steward: Option[Steward],
                broadcasterUrl: Option[String],
                trustModel: String,
                trustModelIsHub: Boolean
              )

object Qep {
  val key = "shrine.queryEntryPoint."

  def apply(keyStoreDescriptor:KeyStoreDescriptor): Qep = {
    val qepConfig = ConfigSource.config.getConfig("shrine.queryEntryPoint")
    val createQep = qepConfig.getBoolean("create")

    if(createQep) {
      new Qep(
        maxQueryWaitTimeMillis = 0L,
        create = createQep,
        //todo: delete attatchSingingCert
        attachSigningCert = qepConfig.getBoolean("attachSigningCert"),
        authorizationType = qepConfig.getString("authorizationType"),
        authenticationType = PmAuthenticator.getClass.getSimpleName,
        steward = if(qepConfig.getString("authorizationType") == "shrine-steward") Some(Steward())
                  else None,
        broadcasterUrl = None,
        trustModel = keyStoreDescriptor.trustModel.description,
        trustModelIsHub = keyStoreDescriptor.trustModel match {
          case _: SingleHubModel => true
          case PeerToPeerModel    => false
        }
      )
    }
    else {
      new Qep(
        maxQueryWaitTimeMillis = 0L,
        create = false,
        //todo: delete attatchSingingCert
        attachSigningCert = false,
        authorizationType = "",
        authenticationType = "",
        steward = None,
        broadcasterUrl = None,
        trustModel = keyStoreDescriptor.trustModel.description,
        trustModelIsHub = true
      )
    }
  }
}

case class AbbreviatedKeyStoreEntry(alias: String, cn: String, md5: String)

case class KeyStoreReport(
                           fileName: String,
                           password: String = "REDACTED",
                           privateKeyAlias: Option[String],
                           owner: Option[String],
                           issuer: Option[String],
                           expires: Long,
                           md5Signature: String,
                           sha256Signature: String,
                           caTrustedAlias: Option[String],
                           caTrustedSignature: Option[String],
                           remoteSiteStatuses: Seq[String] = Seq.empty,
                           isHub: Boolean,
                           abbreviatedEntries: Seq[AbbreviatedKeyStoreEntry]
                         )

object KeyStoreReport {

  def apply(keystoreConfig: Config,keyStoreDescriptor:KeyStoreDescriptor): KeyStoreReport = {
    val certCollection: BouncyKeyStoreCollection = BouncyKeyStoreCollection.fromFileRecoverWithClassPath(keyStoreDescriptor)
    val maybeCaEntry: Option[KeyStoreEntry] = certCollection match {
      case DownStreamCertCollection(_, caEntry) => Some(caEntry)
      case HubCertCollection(_, caEntry)        => Some(caEntry)
    }

    def sortFormat(input: String): Option[String] = {
      if (input.isEmpty) None
      else {
        def isLong(str: String) = str.split('=').headOption.getOrElse(str).length > 2
        // Just an ugly sort for formatting purposes. I want The long key last, and otherwise just
        // Sort them lexicographically.
        Some(input.split(", ").sortBy(a => (isLong(a), a)).mkString(", "))
      }
    }

    new KeyStoreReport(
      fileName = keyStoreDescriptor.file,
      privateKeyAlias = keyStoreDescriptor.privateKeyAlias,
      owner = sortFormat(certCollection.myEntry.cert.getSubjectDN.getName),
      issuer = sortFormat(certCollection.myEntry.cert.getIssuerDN.getName),
      expires = certCollection.myEntry.cert.getNotAfter.getTime,
      md5Signature = UtilHasher.encodeCert(certCollection.myEntry.cert, "MD5"),
      sha256Signature = UtilHasher.encodeCert(certCollection.myEntry.cert, "SHA-256"),
      caTrustedAlias = maybeCaEntry.map(_.aliases.first),
      caTrustedSignature = maybeCaEntry.map(entry => UtilHasher.encodeCert(entry.cert, "MD5")),
      isHub = keyStoreDescriptor.trustModel == SingleHubModel(true),
      abbreviatedEntries = certCollection.allEntries.map(entry => AbbreviatedKeyStoreEntry(
        entry.aliases.first,
        entry.commonName.getOrElse("No common name"),
        UtilHasher.encodeCert(entry.cert, "MD5"))).toList
    )
  }
}

case class ParsedConfig(configMap:Map[String, String])

object ParsedConfig {
  def isPassword(key: String): Boolean = {
    if (key.toLowerCase.contains("password")) true
    else false
  }

  def apply(config:Config):ParsedConfig = {
    import scala.collection.JavaConverters._

    val entries: Set[(String, String)] = config.entrySet.asScala.to[Set].map(x => (x.getKey, x.getValue.render())).filterNot(x => isPassword(x._1))
    val sortedMap: Map[String, String] = entries.toMap
    ParsedConfig(sortedMap)
  }
}

case class Adapter(crcEndpointUrl: String,
                   setSizeObfuscation: Boolean,
                   adapterMappingsFilename: Option[String],
                   adapterMappingsDate: Option[Long]
                  )

object Adapter {
  def apply(config: Config): Adapter = {
    val crcEndpointUrl = config.getString("shrine.adapter.crcEndpoint.url")
    val setSizeObfuscation = config.getBoolean("shrine.adapter.setSizeObfuscation")

    val maybeAdapterMappingsFile: Option[String] = Option(config.getString("shrine.adapter.adapterMappingsFileName"))

    Adapter(crcEndpointUrl, setSizeObfuscation, maybeAdapterMappingsFile, mappingFileInfo.map(_._2))
  }

  def mappingFileInfo: Option[(String, Long, String)] = {
    val maybeAdapterMappingsFile: Option[String] = Option(ConfigSource.config.getString("shrine.adapter.adapterMappingsFileName"))

    maybeAdapterMappingsFile.map{ adapterMappingsFile =>

      val adapterMappings: AdapterMappings = AdapterMappings(adapterMappingsFile)
      (adapterMappings.source,adapterMappings.lastModified,adapterMappings.version)
    }
  }
}

case class OutboundUser(userName:UserName,fullName:String,roles:Set[Role] = Set(researcherRole))

object OutboundUser extends Loggable {

  def createResearcher(userName:UserName,fullName:String): OutboundUser = OutboundUser(userName,fullName)
  def createSteward(userName:UserName,fullName:String): OutboundUser = OutboundUser(userName,fullName,Set(researcherRole,stewardRole))
  def createFromUser(user:User): OutboundUser = if( user.params.toList.contains((stewardRole,"true"))) createSteward(user.username,user.fullName)
  else createResearcher(user.username,user.fullName)
  /** If the user is unknown but has a user id, best effort is to guess this user is a researcher with userName as their full name*/
  def createUnknownUser(userName:UserName): OutboundUser = {
    info(s"Creating an OutboundUser for unknown userName $userName")
    createResearcher(userName,userName)
  }

}

case class Summary(
                    shrineVersion: String,
                    shrineBuildDate: String,
                    ontologyVersion: String,
                    ontologyVersionTerm: String,
                    ontologyTerm: String,
                    adapterMappingsFileName: Option[String],
                    adapterMappingsDate: Option[Long]
                  )

object Summary {

  def apply(shrineConfig:Config): Summary = {
    val termVal = shrineConfig.getString("shrine.networkStatusQuery")
    val term = Term(termVal, termVal)

    val adapterMappingInfo = Adapter.mappingFileInfo

    val ontologyVersion = try {

      val certCollection: BouncyKeyStoreCollection = BouncyKeyStoreCollection.fromConfig

      val ontEndpoint: EndpointConfig = shrineConfig.getConfigured("shrine.ontEndpoint", EndpointConfig(_))
      val ontPoster: Poster = Poster(certCollection,ontEndpoint)

      import scala.concurrent.duration.DurationInt

      //TODO: XXX: Un-hard-code max wait time param
      val ontClient: OntClient = new PosterOntClient(shrineConfig.getConfigured("shrine.hiveCredentials", HiveCredentials(_, HiveCredentials.ONT)), 1.minute, ontPoster)

      new OntClientOntologyMetadata(ontClient).ontologyVersion
    }
    catch {
      case NonFatal(x) =>
        Log.info("Problem while getting ontology version", x)
        s"Unavailable due to: ${x.getMessage}"
    }

    Summary(
      shrineVersion = Versions.version,
      shrineBuildDate = Versions.buildDate,
      //todo in scala 2.12, do better
      ontologyVersion = ontologyVersion,
      ontologyVersionTerm = OntClientOntologyMetadata.versionContainerTerm,
      ontologyTerm = term.value,
      adapterMappingsFileName = adapterMappingInfo.map(_._1),
      adapterMappingsDate = adapterMappingInfo.map(_._2)
    )
  }
}

case class QueryAdapterTestResult (isHub: Boolean,
                                   queryResult:Option[String],// was Option[SingleNodeResult], which made circe stumble, but it is already hacked to always be None
                                   adapterOk: Boolean,
                                   keystoreOk: Boolean,
                                   hubOk: Boolean,
                                   qepOk: Boolean)

object QueryAdapterTestResult {

  def apply(): QueryAdapterTestResult = new QueryAdapterTestResult(false, None, true, true, true, true)

}

case class I2b2(pmUrl: String,
                crcUrl: Option[String],
                ontUrl: String,
                i2b2Domain: String,
                username: String,
                crcProject: String,
                ontProject: String)

object I2b2 {
  val shrineConfig: Config = ConfigSource.config.getConfig("shrine")//.getConfigured("pmEndpoint", EndpointConfig(_))

  val ontHiveCredentials: HiveCredentials = shrineConfig.getConfigured("hiveCredentials", HiveCredentials(_, HiveCredentials.ONT))
  val crcHiveCredentials: HiveCredentials = shrineConfig.getConfigured("hiveCredentials", HiveCredentials(_, HiveCredentials.CRC))
  def apply(): I2b2 = new I2b2(
    pmUrl = shrineConfig.getConfigured("pmEndpoint", EndpointConfig(_)).url.toString,
    crcUrl = Option(shrineConfig.getString("adapter.crcEndpoint.url")),
    ontUrl = shrineConfig.getConfigured("ontEndpoint", EndpointConfig(_)).url.toString,
    i2b2Domain = ontHiveCredentials.domain,
    username = ontHiveCredentials.username,
    crcProject = crcHiveCredentials.projectId,
    ontProject = ontHiveCredentials.projectId
  )
}

case class DownstreamNode(name: String, url: String)

case class OptionalParts(isHub: Boolean,
                         stewardEnabled: Boolean,
                         downstreamNodes: Seq[DownstreamNode])

object OptionalParts {
  val shrineConfig: Config = ConfigSource.config.getConfig("shrine")//.getConfigured("pmEndpoint", EndpointConfig(_))

  def apply(): OptionalParts = {
    OptionalParts(
      shrineConfig.getBoolean("hub.create"),
      shrineConfig.getString("queryEntryPoint.authorizationType") == "shrine-steward",
      Seq.empty
    )
  }
}

case class Hub(
                create: Boolean,
                downstreamNodes: Seq[DownstreamNode])

object Hub {
  def apply(): Hub = {
    val create = ConfigSource.config.getBoolean("shrine.hub.create")
    val downstreamNodes = Seq.empty
    Hub(create, downstreamNodes)
  }
}

object AuthenticationFailureMiddleWare {

  //noinspection TypeAnnotation
  def apply(service: HttpRoutes[IO]) = {
    //convert 401 unauthorized to 200
    //This is to work around issue with dashboard app displaying a browser dialog
    // prompting again for username and password
    service.map {
      case Status.Unauthorized(resp) => resp.withStatus(Status.Ok).withEntity("AuthenticationFailed")
      case resp => resp
    }
  }
}

case class AppVersion(
                       currentVersion:String,
                       buildDate:String
                     )

object AppVersion {
  def apply(): AppVersion = AppVersion(Versions.version, Versions.buildDate)
}

final case class OntClientOntologyMetadata(client: OntClient) {
  def ontologyVersion: String = {
    import OntClientOntologyMetadata._
    val versionTermOption = for {
      versionTerm <- client.childrenOf(versionContainerTerm).headOption
      lastPathPart <- dropTrailingSlash(versionTerm).split(backslashRegex).lastOption
    } yield lastPathPart

    versionTermOption.getOrElse("UNKNOWN")
  }
}

object OntClientOntologyMetadata {
  private val backslash = """\"""
  private val backslashRegex = """\\"""

  def dropTrailingSlash(s: String): String = if (s.endsWith(backslash)) s.dropRight(1) else s

  val versionContainerTerm = """\\SHRINE\SHRINE\ONTOLOGYVERSION"""
}