package net.shrine.api.dashboard

import cats.data.OptionT
import cats.effect.{ConcurrentEffect, ContextShift, IO}
import cats.implicits.{catsSyntaxEither => _}
import com.typesafe.config.Config
import net.shrine.http4s.auth.BasicAuthentication
import net.shrine.source.ConfigSource
import org.http4s.dsl.io.{Ok, Unauthorized}
import org.http4s.headers.{Accept, Authorization}
import org.http4s.{BasicCredentials, Headers, MediaType, Method, Request, Response, Status, Uri}
import org.junit.runner.RunWith
import org.scalatest.{FlatSpec, Ignore}
import org.scalatestplus.junit.JUnitRunner

import scala.collection.JavaConverters.asScalaSetConverter
import scala.collection.immutable.{Map, Set}
import scala.concurrent.ExecutionContext.Implicits.global
/**
  * Test web api of DashboardService
  */
@RunWith(classOf[JUnitRunner])
class DashboardServiceTest extends FlatSpec {

  def extractResponse(request: Request[IO], dashboardService: DashboardService = DashboardService()): Response[IO] = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val responseOptionIo: OptionT[IO, Response[IO]] = dashboardService.router.run(request)

    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  "DashboardService" should " reply to ping with pong" in {
    val pingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString("/ping"))

    val response: Response[IO] = extractResponse(pingRequest)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("pong")(entityText)
  }

  "DashboardService" should " return OK to a whoami request for a valid user" in {

    val username = "admin"
    val password = "pwd"
    val configRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/user/whoami"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val dashboardService = DashboardService()

    val response: Response[IO] = extractResponse(configRequest, dashboardService)

    assertResult(Ok)(response.status)
    response.as[String].unsafeRunSync
  }

  "DashboardService" should  "return a 200 and AuthenticationFailure for a user/whoami request with bad credentials" in {

    val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/user/whoami"),
      headers = Headers.of(Authorization(BasicCredentials("admin", "wrongPassword"))))
    val response: Response[IO] = extractResponse(request)

    val responseText = response.as[String].unsafeRunSync()

    assertResult("""AuthenticationFailed""")(responseText)
    assertResult(Ok)(response.status)
  }

  /*
//Needs a real ontology service
  "DashboardService" should " return summary status" in {

    val username = "admin"
    val password = "pwd"

    val request: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/admin/status/summary"),
      headers = Headers.of(
        Accept(MediaType.application.json),
        Authorization(BasicCredentials(username, password)))
    )

    val dashboardService = new DashboardService()

    val response: Response[IO] = extractResponse(request, dashboardService)
    val entityText = response.as[String].unsafeRunSync

    assertResult(Ok)(response.status)
  }
  */

  "DashboardService" should " not return config if user is not authorized" in {
    val configRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString("/admin/status/config"))

    val response: Response[IO] = extractResponse(configRequest)

    assertResult(Unauthorized)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult(BasicAuthentication.unauthorizedMsg)(entityText)
  }

  "DashboardService" should " return config if user is authorized" in {
    import cats.syntax.either.catsSyntaxEither
    import io.circe.generic.auto.exportEncoder
    import io.circe.syntax.EncoderOps
    import io.circe.Json
    import io.circe.parser.parse
    import org.http4s.circe.jsonEncoder

    val username = "admin"
    val password = "pwd"
    val configRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/admin/status/config"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val dashboardService = DashboardService()

    val expectedResponse: String = """{
                                   |      "keyValues" : {
                                     |    "keystore.file" : "shrine.keystore",
                                     |    "queryEntryPoint.trustModelIsHub" : "true",
                                     |    "keystore.privateKeyAlias" : "shrine-test",
                                     |    "hub.downstreamNodes.CHB" : "http://chb:8080/rest",
                                     |    "authenticate.usersource.admin.username" : "admin",
                                     |    "problem.database.createTablesOnStart" : "false",
                                     |    "problem.problemHandler" : "net.shrine.problem.LogAndDatabaseProblemHandler$",
                                     |    "problem.timeout" : "30",
                                     |    "queryEntryPoint.broadcasterServiceEndpoint.url" : "https://localhost:8080/shrine/test",
                                     |    "keystore.caCertAliases" : "[\n    # shrine.conf @ file:/Users/marc-danie/IdeaProjects/shrine/services/dashboard/target/test-classes/shrine.conf: 56\n    shrine-test-ca\n]",
                                     |    "problem.database.dataSourceFrom" : "JNDI",
                                     |    "authenticate.usersource.qep.username" : "qep",
                                     |    "hub.downstreamNodes.AGR" : "http://agr:8080/rest",
                                     |    "problem.database.jndiDataSourceName" : "java:comp/env/jdbc/problemDB",
                                     |    "authenticate.usersource.type" : "ConfigUserSource",
                                     |    "keystore.keyStoreType" : "JKS",
                                     |    "authenticate.usersource.researcher.username" : "ben",
                                     |    "dashboard.gruntWatch" : "false",
                                     |    "authenticate.usersource.steward.username" : "dave",
                                     |    "problem.database.slickProfileClassName" : "slick.jdbc.MySQLProfile$",
                                     |    "hub.downstreamNodes.PHS" : "http://example:8080/rest"
                                     |  }
                                     |}""".stripMargin

    val expectedJson: Json = parse(expectedResponse).getOrElse(Json.Null)

    val response: Response[IO] = extractResponse(configRequest, dashboardService)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    val resultJson: Json = parse(expectedResponse).getOrElse(Json.Null)

    assertResult(expectedJson)(resultJson)
  }

  "DashboardService" should " return qep if user is authorized" in {

    val username = "admin"
    val password = "pwd"
    val configRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/admin/status/qep"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val dashboardService = DashboardService()

    val response: Response[IO] = extractResponse(configRequest, dashboardService)

    assertResult(Ok)(response.status)
    response.as[String].unsafeRunSync
  }

  "DashboardService" should " return keystore if user is authorized" in {

    val username = "admin"
    val password = "pwd"
    val configRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/admin/status/keystore"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val dashboardService = DashboardService()

    val response: Response[IO] = extractResponse(configRequest, dashboardService)

    assertResult(Ok)(response.status)
    response.as[String].unsafeRunSync
  }

  "DashboardService" should " return adapter if user is authorized" in {

    val username = "admin"
    val password = "pwd"
    val configRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/admin/status/adapter"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val dashboardService = DashboardService()

    val response: Response[IO] = extractResponse(configRequest, dashboardService)

    assertResult(Ok)(response.status)
    response.as[String].unsafeRunSync
  }

  "DashboardService" should " return queryAdapterTest results if a user is authorized" in {

    val username = "admin"
    val password = "pwd"
    val configRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/admin/status/queryAdapterTest"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val dashboardService = DashboardService()

    val response: Response[IO] = extractResponse(configRequest, dashboardService)

    assertResult(Ok)(response.status)
    response.as[String].unsafeRunSync
  }

  "DashboardService" should " return i2b2 results if a user is authorized" in {

    val username = "admin"
    val password = "pwd"
    val configRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/admin/status/i2b2"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val dashboardService = DashboardService()

    val response: Response[IO] = extractResponse(configRequest, dashboardService)

    assertResult(Ok)(response.status)
    response.as[String].unsafeRunSync
  }

  "DashboardService" should " return optionalParts if a user is authorized" in {

    val username = "admin"
    val password = "pwd"
    val configRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/admin/status/optionalParts"),
      headers = Headers.of(Authorization(BasicCredentials(username, password))))

    val dashboardService = DashboardService()

    val response: Response[IO] = extractResponse(configRequest, dashboardService)

    assertResult(Ok)(response.status)
    response.as[String].unsafeRunSync
  }

}

