package net.shrine.api.i2b2

import java.sql.SQLException

import cats.effect.IO
import fs2.Stream
import net.shrine.api.util.ShrineConfiguration
import net.shrine.authentication.NotAuthenticatedException
import net.shrine.authorization.AuthorizationException
import net.shrine.config.ConfigExtensions
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.http4s.client.Http4sHttpClient
import net.shrine.http4s.client.legacy.EndpointConfig
import net.shrine.log.Log
import net.shrine.problem.{AbstractProblem, ProblemNotYetEncoded, ProblemSources}
import net.shrine.protocol.{ErrorResponse, HandleableI2b2Request, I2b2RequestHandler, ResultOutputType, ShrineRequest, ShrineResponse}
import net.shrine.qep.querydb.QepDatabaseProblem
import net.shrine.serialization.I2b2Marshaller
import net.shrine.slick.DbIoActionException
import net.shrine.source.ConfigSource
import net.shrine.util.XmlUtil
import org.http4s.dsl.impl.{->, /}
import org.http4s.dsl.io.{BadRequest, GET, InternalServerError, NotFound, Ok, POST, Root, http4sBadRequestSyntax, http4sInternalServerErrorSyntax, http4sNotFoundSyntax, http4sOkSyntax}
import org.http4s.headers.`Content-Type`
import org.http4s.{HttpRoutes, Request, Response, Uri}
import org.xml.sax.SAXParseException

import scala.util.Try
import scala.util.control.NonFatal
import scala.xml.{NodeSeq, XML}

case class I2B2Service() {

  val i2b2ServiceInfo: String  =
    """
      |The SHRINE I2B2 service.
      |
      |This API makes calls to I2B2.
      |
    """.stripMargin

  val service: HttpRoutes[IO] = HttpRoutes.of[IO] {
    case GET -> Root / "ping" => Ok("pong")

    case req @ POST -> Root / "ontologyService"  => ontologyService(req)
    case req @ POST -> Root / "pmService"  => pmService(req)
    case req @ POST -> Root / "request"  => i2b2Request(req)
    case x => NotFound(s"The I2B2 service does not respond to $x")
  }

  def i2b2Request(request: Request[IO]): IO[Response[IO]] ={

    val breakdownTypes: Set[ResultOutputType] =  ShrineConfiguration.breakdownTypes


    def prettyPrint(xml: NodeSeq): String = XmlUtil.prettyPrint(xml.head).trim

    request.as[String].flatMap[Response[IO]] { bodyText: String =>

      def errorResponse(e: Throwable): ErrorResponse = e match {
        case nax:NotAuthenticatedException => ErrorResponse(nax.problem)
        case ax:AuthorizationException => ErrorResponse(ax.problem)
        case dbioax:DbIoActionException => ErrorResponse(QepDatabaseProblem(dbioax))
        case sqlx:SQLException => ErrorResponse(QepDatabaseProblem(sqlx))
        case saxx:SAXParseException => ErrorResponse(QepCouldNotInterpretRequest(bodyText,saxx))
        case _ => ErrorResponse(ProblemNotYetEncoded("The QEP encountered an unforeseen problem while processing an i2b2 request",e))
      }

      def handleRequest(shrineRequest: ShrineRequest with HandleableI2b2Request): Try[IO[Response[IO]]] = Try {
        Log.info(s"Running request from user: ${shrineRequest.authn.username} of type ${shrineRequest.requestType.toString}")

        val shouldBroadcast = true
        val i2b2RequestHandler: I2b2RequestHandler = ShrineConfiguration.queryEntryPointComponents.map(x => x.i2b2Service).get

        //todo this context shift should be deeper -where we know we're talking to either the PM cell or ONT cell SHRINE-3632
        val httpClientContextShift = IO.contextShift(ExecutionContexts.httpClientExecutionContext)
        val shrineResponseIO: IO[ShrineResponse] = httpClientContextShift.evalOn(ExecutionContexts.httpClientExecutionContext)(
          IO(shrineRequest.handleI2b2(i2b2RequestHandler, shouldBroadcast)).handleErrorWith(t => {
            IO{
              Log.error("Error processing request: ", t)
              errorResponse(t)
            }
          })
        )

        val responseStringIO: IO[String] = shrineResponseIO.map { shrineResponse =>
          if(shrineResponse.toXmlString.contains("ERROR")) {
            val xmlShrineResponse = prettyPrint(shrineResponse.toXml)
            Log.error(s"The I2B2 request $request returned a response with an error: $xmlShrineResponse")
          }
          //TODO: Revisit this.  For now, we bail if we get something that isn't i2b2able
          val responseString: String = shrineResponse match {
            case i2b2able: I2b2Marshaller => prettyPrint(i2b2able.toI2b2)
            case _ => throw new InvalidI2B2Response(s"Unexpected error: $shrineResponse has no i2b2 representation")
          }
          responseString
        }
        responseStringIO.flatMap(Ok(_))
      }.recover {
        case NonFatal(e) =>
          Log.error("Error processing request: ", e)
          Ok(prettyPrint(errorResponse(e).toI2b2))
      }

      def handleParseError(e: Throwable): Try[IO[Response[IO]]] = Try {
        Log.debug(s"Failed to unmarshal i2b2 request $bodyText")

        Log.error("Couldn't understand request: ", e)

        //NB: The legacy webclient can't deal with non-200 status codes.
        //It also can't deal with ErrorResponses in several cases, but we have nothing better to return for now.
        //TODO: Return a 400 status here, once we're using the new web client
        Ok(prettyPrint(errorResponse(e).toI2b2))
      }

      val builder: IO[Response[IO]] = HandleableI2b2Request.fromI2b2String(breakdownTypes)(bodyText).transform(handleRequest, handleParseError).get
      builder
    }
  }

  def ontologyService(request: Request[IO]): IO[Response[IO]] = {
    val ontEndpointConfig = ConfigSource.config.getConfigured("shrine.ontEndpoint", EndpointConfig(_))
    val baseOntologyUri = Uri.unsafeFromString(ontEndpointConfig.url.toString)

    routeXMLUsingBaseUri(request, baseOntologyUri)
  }

  def pmService(request: Request[IO]): IO[Response[IO]] = {
    val pmEndpointConfig = ConfigSource.config.getConfigured("shrine.pmEndpoint", EndpointConfig(_))
    val basePMUri = Uri.unsafeFromString(pmEndpointConfig.url.toString)

    routeXMLUsingBaseUri(request, basePMUri)
  }

  def routeXMLUsingBaseUri(request: Request[IO], baseUri: Uri): IO[Response[IO]] = {

    request.as[String].flatMap[Response[IO]]{ bodyText: String =>

      val bodyXml = XML.loadString(bodyText)

      val redirectUrl: String = (bodyXml \\ "redirect_url").headOption.map(_.text.trim).getOrElse {
        Log.error("Error parsing redirect_url tag")

        throw new RedirectUrlException("Error parsing redirect_url tag")
      }

      if (redirectUrl == null || redirectUrl.isEmpty) {
        Log.error("Detected missing redirect_url tag")

        throw new RedirectUrlException("Detected missing redirect_url tag")
      }

      val redirectPath: Uri.Path = Uri.unsafeFromString(redirectUrl).path
      val basePath:Uri.Path = baseUri.path
      val newPath = redirectPath.replace(basePath, "")

      val ontologyRequestUrl = Uri.unsafeFromString(baseUri.toString().concat(newPath))

      val xmlContentHeader = `Content-Type`(org.http4s.MediaType.text.xml)
      val i2b2Request: Request[IO] = request.withUri(ontologyRequestUrl).withEntity(bodyText).withContentType(xmlContentHeader)

      Http4sHttpClient.webFetchResponse(i2b2Request).flatMap{ response: Response[IO] => {
          response.status.code  match {
            case code if code < 300 => IO(response)
            case x =>
              val message = response.bodyAsText.flatMap[IO, String]{text: String => {
                val errorMessage = s"""${getClass.getSimpleName} got an unexpected response while callling the I2B2 service at ${i2b2Request.uri}
                                 |Response Code : $x
                                 |Response: $text""".stripMargin
                  Log.error(errorMessage)
                  Stream(errorMessage)
                }
              }
              IO(Response[IO](status = response.status).withEntity(message))
          }
        }
      }
    }.handleErrorWith{
      case r: RedirectUrlException => BadRequest(r.getMessage)
      case x: Throwable => handleGeneralFailure(request, x)
    }
  }

  private def handleGeneralFailure(req: Request[IO], x: Throwable): IO[Response[IO]] = x match {
    case NonFatal(nfx) =>
      val path = req.pathInfo
      val errorMessage =  s"""${getClass.getSimpleName} threw an exception while trying to $path.
                      |${getClass.getSimpleName} response: ${x.getMessage}. Exception: ${nfx.getClass}""".
        stripMargin
      Log.error(errorMessage,nfx)
      InternalServerError(errorMessage)
    case _ => throw x
  }
}

final class RedirectUrlException(message: String) extends Exception(message)

final class InvalidI2B2Response(message: String) extends Exception(message)

case class QepCouldNotInterpretRequest(i2b2Request:String,x:Exception) extends AbstractProblem(ProblemSources.Qep){
  override val summary = "The QEP could not interpret a request."
  override val throwable = Some(x)
  override val description: String = x.getMessage
  override val detailsText = Some(s"Request contents:'$i2b2Request'")
}