package net.shrine.api.i2b2

  import cats.data.OptionT
  import cats.effect.{ConcurrentEffect, ContextShift, IO}
  import cats.implicits.{catsSyntaxEither => _}
  import net.shrine.api.steward.{InboundShrineQuery, InboundTopicRequest, OutboundTopic, OutboundUser, TopicState}
  import net.shrine.api.steward.db.StewardDatabase
  import net.shrine.api.steward.{QueryContents, stewardRole}
  import net.shrine.i2b2.protocol.pm.User
  import net.shrine.protocol.Credential
  import org.http4s.dsl.io.Ok
  import org.http4s.{Method, Request, Response, Uri}
  import org.junit.runner.RunWith
  import org.scalatest.FlatSpec
  import org.scalatestplus.junit.JUnitRunner

  import scala.concurrent.ExecutionContext.Implicits.global

@RunWith(classOf[JUnitRunner])
class I2B2ServiceTest extends FlatSpec  {

  def extractResponse(request: Request[IO]): Response[IO] = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val responseOptionIo: OptionT[IO, Response[IO]] = I2B2Service().service.run(request)

    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  "I2B2Service" should "reply to ping with pong" in {
    val pingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString("/ping"))

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("pong")(entityText)
  }

  "I2B2Service" should "return breakdown result output types" in {
    val resultOutputTypeRequest: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString("/request"))
      .withEntity("""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                    |<ns6:request xmlns:ns4="http://www.i2b2.org/xsd/cell/crc/psm/1.1/"
                    |  xmlns:ns7="http://www.i2b2.org/xsd/cell/crc/psm/querydefinition/1.1/"
                    |  xmlns:ns3="http://www.i2b2.org/xsd/cell/crc/pdo/1.1/"
                    |  xmlns:ns5="http://www.i2b2.org/xsd/hive/plugin/"
                    |  xmlns:ns2="http://www.i2b2.org/xsd/hive/pdo/1.1/"
                    |  xmlns:ns6="http://www.i2b2.org/xsd/hive/msg/1.1/">
                    |	<message_header>
                    |		<proxy>
                    |            <redirect_url>https://localhost</redirect_url>
                    |        </proxy>
                    |		<sending_application>
                    |			<application_name>i2b2_QueryTool</application_name>
                    |			<application_version>1.6</application_version>
                    |		</sending_application>
                    |		<sending_facility>
                    |			<facility_name>PHS</facility_name>
                    |		</sending_facility>
                    |		<receiving_application>
                    |			<application_name>i2b2_DataRepositoryCell</application_name>
                    |			<application_version>1.6</application_version>
                    |		</receiving_application>
                    |		<receiving_facility>
                    |			<facility_name>PHS</facility_name>
                    |		</receiving_facility>
                    |		<message_type>
                    |			<message_code>Q04</message_code>
                    |			<event_type>EQQ</event_type>
                    |		</message_type>
                    |		<security>
                    |			<domain>i2b2demo</domain>
                    |			<username>myuser</username>
                    |			<password is_token="true" token_ms_timeout="1800000">SessionKey:XXXXXXX</password>
                    |		</security>
                    |		<message_control_id>
                    |			<message_num>N092v60Hr0I0V9g72BP58</message_num>
                    |			<instance_num>0</instance_num>
                    |		</message_control_id>
                    |		<processing_id>
                    |			<processing_id>P</processing_id>
                    |			<processing_mode>I</processing_mode>
                    |		</processing_id>
                    |		<accept_acknowledgement_type>messageId</accept_acknowledgement_type>
                    |		<project_id>SHRINE</project_id>
                    |	</message_header>
                    |	<request_header>
                    |		<result_waittime_ms>180000</result_waittime_ms>
                    |	</request_header>
                    |	<message_body>
                    |		<ns4:psmheader>
                    |            <user login="shrine">shrine</user>
                    |            <patient_set_limit>0</patient_set_limit>
                    |            <estimated_time>0</estimated_time>
                    |            <request_type>CRC_QRY_getResultType</request_type>
                    |		</ns4:psmheader>
                    |	</message_body>
                    |</ns6:request>""".stripMargin)

    val expectedResult = """<ns4:response xmlns:ns11="http://www.i2b2.org/xsd/hive/msg/result/1.1/" xmlns:ns10="http://www.i2b2.org/xsd/cell/ont/1.1/" xmlns:ns9="http://www.i2b2.org/xsd/cell/crc/psm/analysisdefinition/1.1/" xmlns:ns8="http://www.i2b2.org/xsd/cell/crc/psm/querydefinition/1.1/" xmlns:ns7="http://sheriff.shrine.net/" xmlns:ns6="http://www.i2b2.org/xsd/cell/pm/1.1/" xmlns:ns5="http://www.i2b2.org/xsd/cell/crc/psm/1.1/" xmlns:ns4="http://www.i2b2.org/xsd/hive/msg/1.1/" xmlns:ns3="http://www.i2b2.org/xsd/cell/crc/pdo/1.1/" xmlns:ns2="http://www.i2b2.org/xsd/hive/pdo/1.1/">
                           |  <message_header>
                           |    <i2b2_version_compatible>1.1</i2b2_version_compatible>
                           |    <hl7_version_compatible>2.4</hl7_version_compatible>
                           |    <sending_application>
                           |      <application_name>SHRINE</application_name>
                           |      <application_version>1.3-compatible</application_version>
                           |    </sending_application>
                           |    <sending_facility>
                           |      <facility_name>SHRINE</facility_name>
                           |    </sending_facility>
                           |    <datetime_of_message>2011-04-08T16:21:12.251-04:00</datetime_of_message>
                           |    <security/>
                           |    <project_id xsi:nil="true" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"/>
                           |  </message_header>
                           |  <response_header>
                           |    <result_status>
                           |      <status type="DONE">DONE</status>
                           |    </result_status>
                           |  </response_header>
                           |  <message_body>
                           |    <ns4:response xsi:type="ns4:result_type_responseType" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                           |      <status>
                           |        <condition type="DONE">DONE</condition>
                           |      </status>
                           |      <query_result_type>
                           |        <result_type_id>1</result_type_id>
                           |        <name>PATIENT_COUNT_XML</name>
                           |        <display_type>CATNUM</display_type>
                           |        <visual_attribute_type>LA</visual_attribute_type>
                           |        <description>Number of patients</description>
                           |      </query_result_type>
                           |      <query_result_type>
                           |        <result_type_id>2</result_type_id>
                           |        <name>PATIENT_HEIGHT_COUNT_XML</name>
                           |        <display_type>CATNUM</display_type>
                           |        <visual_attribute_type>LA</visual_attribute_type>
                           |        <description>Age patient breakdown</description>
                           |      </query_result_type>
                           |      <query_result_type>
                           |        <result_type_id>3</result_type_id>
                           |        <name>PATIENT_GENDER_COUNT_XML</name>
                           |        <display_type>CATNUM</display_type>
                           |        <visual_attribute_type>LA</visual_attribute_type>
                           |        <description>Gender patient breakdown</description>
                           |      </query_result_type>
                           |    </ns4:response>
                           |  </message_body>
                           |</ns4:response>""".stripMargin
    val response: Response[IO]= extractResponse(resultOutputTypeRequest)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult(expectedResult)(entityText)
  }

  "I2B2Service" should "return list of approved topics for a user" in {
    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createFiveTopics()
    createSixQueries()

    val resultOutputTypeRequest: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString("/request"))
      .withEntity("""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                    |
                    |<ns2:request xmlns:ns2="http://www.i2b2.org/xsd/hive/msg/1.1/" xmlns:ns7="http://sheriff.shrine.net/">
                    |
                    |<message_header>
                    |
                    |	<proxy>
                    |            <redirect_url>/shrine/rest/i2b2/request</redirect_url>
                    |        </proxy>
                    |
                    |
                    |	<i2b2_version_compatible>1.1</i2b2_version_compatible>
                    |
                    |	<hl7_version_compatible>2.4</hl7_version_compatible>
                    |
                    |	<sending_application>
                    |
                    |		<application_name>SHRINE Sheriff</application_name>
                    |
                    |		<application_version>1.1</application_version>
                    |
                    |	</sending_application>
                    |
                    |	<sending_facility>
                    |
                    |		<facility_name>i2b2 Hive</facility_name>
                    |
                    |	</sending_facility>
                    |
                    |	<receiving_application>
                    |
                    |		<application_name>SHRINE Sheriff Cell</application_name>
                    |
                    |		<application_version>1.4</application_version>
                    |
                    |	</receiving_application>
                    |
                    |	<receiving_facility>
                    |
                    |		<facility_name>SHRINE</facility_name>
                    |
                    |	</receiving_facility>
                    |
                    |	<datetime_of_message>2007-04-09T15:19:18.906-04:00</datetime_of_message>
                    |
                    |	<security>
                    |
                    |		<domain>i2b2demo</domain>
                    |
                    |		<username>ben</username>
                    |
                    |		<password is_token="true" token_ms_timeout="1800000">SessionKey:XXXXXXX</password>
                    |
                    |	</security>
                    |
                    |	<message_control_id>
                    |
                    |		<message_num>Ov9rXnjTjY6Oq11mFBOVt</message_num>
                    |
                    |		<instance_num>0</instance_num>
                    |
                    |	</message_control_id>
                    |
                    |	<processing_id>
                    |
                    |		<processing_id>P</processing_id>
                    |
                    |		<processing_mode>I</processing_mode>
                    |
                    |	</processing_id>
                    |
                    |	<accept_acknowledgement_type>AL</accept_acknowledgement_type>
                    |
                    |	<application_acknowledgement_type>AL</application_acknowledgement_type>
                    |
                    |	<country_code>US</country_code>
                    |
                    |	<project_id>SHRINE</project_id>
                    |
                    |</message_header>
                    |
                    |<request_header>
                    |
                    |	<result_waittime_ms>180000</result_waittime_ms>
                    |
                    |</request_header>
                    |
                    |<message_body>
                    |
                    |<ns7:sheriff_header xsi:type="ns7:sheriffHeaderType" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"/>
                    |
                    | <ns7:sheriff_request xsi:type="ns7:sheriffRequestType" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"/>
                    |
                    |</message_body>
                    |
                    |</ns2:request>""".stripMargin)

    val expectedResult = """<ns4:response xmlns:ns11="http://www.i2b2.org/xsd/hive/msg/result/1.1/" xmlns:ns10="http://www.i2b2.org/xsd/cell/ont/1.1/" xmlns:ns9="http://www.i2b2.org/xsd/cell/crc/psm/analysisdefinition/1.1/" xmlns:ns8="http://www.i2b2.org/xsd/cell/crc/psm/querydefinition/1.1/" xmlns:ns7="http://sheriff.shrine.net/" xmlns:ns6="http://www.i2b2.org/xsd/cell/pm/1.1/" xmlns:ns5="http://www.i2b2.org/xsd/cell/crc/psm/1.1/" xmlns:ns4="http://www.i2b2.org/xsd/hive/msg/1.1/" xmlns:ns3="http://www.i2b2.org/xsd/cell/crc/pdo/1.1/" xmlns:ns2="http://www.i2b2.org/xsd/hive/pdo/1.1/">
                           |  <message_header>
                           |    <i2b2_version_compatible>1.1</i2b2_version_compatible>
                           |    <hl7_version_compatible>2.4</hl7_version_compatible>
                           |    <sending_application>
                           |      <application_name>SHRINE</application_name>
                           |      <application_version>1.3-compatible</application_version>
                           |    </sending_application>
                           |    <sending_facility>
                           |      <facility_name>SHRINE</facility_name>
                           |    </sending_facility>
                           |    <datetime_of_message>2011-04-08T16:21:12.251-04:00</datetime_of_message>
                           |    <security/>
                           |    <project_id xsi:nil="true" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"/>
                           |  </message_header>
                           |  <response_header>
                           |    <result_status>
                           |      <status type="DONE">DONE</status>
                           |    </result_status>
                           |  </response_header>
                           |  <message_body>
                           |    <ns7:sheriff_response xsi:type="ns7:sheriffResponseType" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                           |      <sheriffEntry>
                           |        <approval>Approved</approval>
                           |        <queryName>UncontroversialKidneys</queryName>
                           |        <queryTopicID>1</queryTopicID>
                           |      </sheriffEntry>
                           |    </ns7:sheriff_response>
                           |  </message_body>
                           |</ns4:response>""".stripMargin
      val response: Response[IO]= extractResponse(resultOutputTypeRequest)

      assertResult(Ok)(response.status)
      val entityText = response.as[String].unsafeRunSync
      assertResult(expectedResult)(entityText)
  }


  val researcherUserName = "ben"
  val researcherFullName = researcherUserName

  val stewardUserName = "dave"
  val stewardFullName = stewardUserName

  val researcherUser = User(
    fullName = researcherFullName,
    username = researcherUserName,
    domain = "domain",
    credential = new Credential("ben's password",false),
    params = Map(),
    rolesByProject = Map()
  )

  val stewardUser = User(
    fullName = stewardFullName,
    username = stewardUserName,
    domain = "domain",
    credential = new Credential("dave's password",false),
    params = Map(stewardRole -> "true"),
    rolesByProject = Map()
  )

  val researcherOutboundUser = OutboundUser.createFromUser(researcherUser)
  val uncontroversialTopic = OutboundTopic(1,"UncontroversialKidneys","Study kidneys without controversy",researcherOutboundUser,0L,TopicState.pending.name,researcherOutboundUser,0L)
  private val queryContent: QueryContents = "<queryDefinition><name>18-34 years old@18:31:51</name><expr><term>\\\\SHRINE\\SHRINE\\Demographics\\Age\\18-34 years old\\</term></expr></queryDefinition>"

  def createFiveTopics(): Unit = {
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("slightlyControversial","who cares?")).unsafeRunSync()
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("controversial","controversial")).unsafeRunSync()
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("moderatelyControversial","more controversial than that")).unsafeRunSync()
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("veryControversial","Just say no")).unsafeRunSync()
  }

  def createSixQueries() = {
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()
    StewardDatabase.db.changeTopicStateIO(1,TopicState.approved,stewardUserName).unsafeRunSync()

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("Forbidden topic","No way is the data steward going for this")).unsafeRunSync()
    StewardDatabase.db.changeTopicStateIO(2,TopicState.rejected,stewardUserName).unsafeRunSync()

    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(0,"A test query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(2),InboundShrineQuery(1,"B forbidden query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(2," C test query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(2),InboundShrineQuery(3,"4 forbidden query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(4,"7 test query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(2),InboundShrineQuery(5,"% forbidden query",queryContent))
  }
}
