package net.shrine.api.qep

import java.util.concurrent.TimeUnit

import cats.data.{Kleisli, OptionT}
import cats.effect.IO
import javax.xml.datatype.XMLGregorianCalendar
import net.shrine.audit.{NetworkQueryId, QueryName, Time}
import net.shrine.authorization.AuthorizationResult.{Authorized, NotAuthorized}
import net.shrine.authorization.QueryAuthorizationService
import net.shrine.http4s.auth.{BasicAuthMiddleware, BasicOrBearerAuthMiddleware, UserSessionToken}
import net.shrine.i2b2.protocol.pm.User
import net.shrine.log.Log
import net.shrine.problem.JsonProblemDigest
import net.shrine.protocol.query.{And, Constrained, Expression, Panel, PanelTiming, QueryDefinition, QueryTiming, SimpleExpression, Term, ValueConstraint}
import net.shrine.protocol.version.{QueryStatuses, ResultStatuses}
import net.shrine.protocol.{BroadcastMessage, QueryResult, ResultOutputType}
import net.shrine.qep.AuthStrategy
import net.shrine.qep.querydb.{BreakdownResultsForType, FullQuery, FullQueryResult, QepQueryDb, QepQueryDbChangeNotifier, QueryAndResults, QueryFlag, QueryStateObserved, QueryWithResultStatus}
import net.shrine.source.ConfigSource
import net.shrine.util.{XmlDateHelper, XmlUtil}
import org.http4s.dsl.impl.{+&, ->, /, :?, Auth, OptionalQueryParamDecoderMatcher}
import org.http4s.dsl.io.{BadRequest, Forbidden, GET, NotFound, Ok, POST, RangeNotSatisfiable, Root, http4sBadRequestSyntax, http4sForbiddenSyntax, http4sNotFoundSyntax, http4sOkSyntax, http4sRangeNotSatisfiableSyntax}
import org.http4s.headers.`Content-Type`
import org.http4s.server.{AuthMiddleware, Router}
import org.http4s.{AuthedRequest, AuthedRoutes, EntityDecoder, HttpRoutes, Response, Status}

import scala.concurrent.duration.FiniteDuration
import scala.util.Try
import scala.xml.NodeSeq

case class QepService() extends Auth {

  val qepServiceInfo: String  =
    """
      |The SHRINE qep service.
      |
      |This API retrieves data from the SHRINE QEP.
      |
    """.stripMargin


  def router: HttpRoutes[IO] = {
    Router[IO](
      "/ping" -> pingService,
      "/login" -> JsonContentService(loginService),
      "" -> JsonContentService(privateQepService)
    )
  }

  object OptionalNetworkIdParam extends OptionalQueryParamDecoderMatcher[String]("networkId")
  object OptionalLimitParam extends OptionalQueryParamDecoderMatcher[Int]("limit")
  object OptionalSkipParam extends OptionalQueryParamDecoderMatcher[Int]("skip")
  object OptionalSortByParam extends OptionalQueryParamDecoderMatcher[String]("sortBy")
  object OptionalTimeoutSecondsParam extends OptionalQueryParamDecoderMatcher[Long]("timeoutSeconds")

  val pingService: HttpRoutes[IO] = HttpRoutes.of[IO] {
    case GET -> Root => Ok("pong")

    case x => NotFound(s"The QEP service does not respond to $x")
  }

  /**
    * Login service: Authenticates user using Basic auth and returns an object containing the user name and session Id.
     */

  import io.circe.syntax._
  import io.circe.generic.auto._

  val authQepLoginRoute: AuthedRoutes[User, IO] = AuthedRoutes.of[User, IO] {
    case GET -> Root as user =>
      val resultJson = UserSessionToken(
        username = user.username,
        sessionId = user.credential.value
      ).asJson.toString()
      Ok (resultJson)
  }

  val loginService:  HttpRoutes[IO] = BasicAuthMiddleware(authQepLoginRoute)

  /**
    * Main QEP service -- i.e. querying
    */

  val authQepRoutes: AuthedRoutes[User, IO] = AuthedRoutes.of[User, IO] {
    case GET -> Root / "authPing" as _ => Ok("pong")

   case _ @ GET -> Root / "queryResult"
      :? OptionalTimeoutSecondsParam(timeoutSec) //todo timeoutSec is never used. Is that OK?
      +& OptionalNetworkIdParam(networkId)
      +& OptionalLimitParam(limit)
      +& OptionalSkipParam(skip)
      +& OptionalSortByParam(sortBy) as user => retrieveQueryResults(user, networkId, limit, skip, sortBy)

    //SHRINE-3219 - replace queryResult with queryResultPolling and also rename
    case _ @ GET -> Root / "queryResultPolling"
      :? OptionalTimeoutSecondsParam(timeoutSec)
      +&OptionalNetworkIdParam(networkId)
      +& OptionalLimitParam(limit)
      +& OptionalSkipParam(skip)
      +& OptionalSortByParam(sortBy) as user => pollForQueryResults(user, networkId, timeoutSec, limit, skip, sortBy)

   case req @ POST -> Root / "queryStatus" as user => retrieveQueryStatus(req)

   case request@POST -> Root / "runQuery" as user =>
     import io.circe.generic.auto.exportDecoder
     import org.http4s.circe.jsonOf
     implicit val basicQueryDecoder: EntityDecoder[IO, BasicQuery] = jsonOf[IO, BasicQuery]

     val basicQueryIO = request.req.as[BasicQuery]

     basicQueryIO.flatMap(authorizeAndRunQuery(_, user)(QueryRunner.runQuery) ).handleErrorWith
     {
       case x: InvalidValueConstraintException =>
         Log.error(s"Error running query: ${x.getMessage}")
         BadRequest(s"Error running query: ${x.getMessage}")
     }

   case x => NotFound(s"The QEP service does not respond to $x")
  }

  val privateQepService = BasicOrBearerAuthMiddleware(authQepRoutes)


  def getPreviousQueryDataIO(user: User,
                             limitOption : Option[Int],
                             skipOption : Option[Int],
                             sortByOption: Option[String],
                             selectedQueryId: Option[NetworkQueryId] = None,
                             queryIdListOption: Option[List[NetworkQueryId]] = None): IO[QueriesAndCount] ={
    val queriesAndCountIO: IO[(Int, Seq[QueryWithResultStatus])] = QepQueryDb.db.selectPreviousQueriesWithResultStatusIO(
      userName = user.username,
      domain = user.domain,
      skip = skipOption,
      limit = limitOption,
      sortBy = sortByOption,
      selectedQueryId = selectedQueryId,
      queryIdListOption
    )

    queriesAndCountIO.map(queriesAndCount => {
      val queryCells: Seq[QueryCell] = queriesAndCount._2.map(q =>  {
        QueryCell(q, q.queryFlag)
      })

      QueriesAndCount(queriesAndCount._1, queryCells)
    })
  }

  def getQueryResultAsStringIO(user: User,
                     limitOption : Option[Int],
                     skipOption : Option[Int],
                     sortByOption: Option[String],
                     queryAndResults: QueryAndResults): IO[String] = {
    import io.circe.generic.auto.exportEncoder
    import io.circe.syntax.EncoderOps

    for {
      queryStatedObserved <- IO(QueryStateObserved(queryAndResults))
      existingQueryStateObserved <- QepQueryDb.db.selectQueryStateObservedByChecksum(queryStatedObserved)
      _ <- existingQueryStateObserved.fold(QepQueryDb.db.insertQueryStateObservedIO(queryStatedObserved))(_ => IO.unit)
      queriesAndCount  <- getPreviousQueryDataIO(user, limitOption, skipOption, sortByOption, Some(queryAndResults.fullQuery.query.networkId))
    }
    yield{
      val queryResults: ResultsRow = ResultsRow(queryAndResults)
      val queriesAndResult = PreviousQueriesAndResult(rowCount = queriesAndCount.queryRowCount, rowOffset = skipOption.getOrElse(0), allQueries = queriesAndCount.queryCells, selectedQuery = Some(queryResults))

      val resultJson = queriesAndResult.asJson.toString().filter(_ != '\n')

      var finalResult: String = resultJson
      if(queriesAndResult.selectedQuery.get.isComplete){
        finalResult = finalResult.concat("\n")
      }

      finalResult
    }
  }

  def getQueryResultIO(user: User,
                       limitOption : Option[Int],
                       skipOption : Option[Int],
                       sortByOption: Option[String],
                       queryAndResults: QueryAndResults): IO[(PreviousQueriesAndResult, Boolean)] = {
    for {
      queryStatedObserved <- IO(QueryStateObserved(queryAndResults))
      existingQueryStateObserved <- QepQueryDb.db.selectQueryStateObservedByChecksum(queryStatedObserved)
      isSelectedQueryStateObserved <- existingQueryStateObserved.fold{QepQueryDb.db.insertQueryStateObservedIO(queryStatedObserved).flatMap(_ => IO(false))}(_ => IO(true))
      queriesAndCount  <- getPreviousQueryDataIO(user, limitOption, skipOption, sortByOption, Some(queryAndResults.fullQuery.query.networkId))
    } yield{
        val queryResults = ResultsRow(queryAndResults)
        val queriesAndResult = PreviousQueriesAndResult(queriesAndCount.queryRowCount, skipOption.getOrElse(0), queriesAndCount.queryCells, Some(queryResults))

        (queriesAndResult, isSelectedQueryStateObserved)
    }
  }

  def pollForQueryResults(
                          user: User,
                          networkId: Option[String] = None,
                          timeoutSeconds: Option[Long] = None,
                          limitOption : Option[Int] = None,
                          skipOption : Option[Int] = None,
                          sortByOption: Option[String] = None,
                          queryIdListOption: Option[List[NetworkQueryId]] = None): IO[Response[IO]] = {
    import io.circe.generic.auto.exportEncoder
    import io.circe.syntax.EncoderOps

    val maximumQueryLimit = 200
    val defaultQueryLimit = 50
    val validatedQueryLimit = limitOption.getOrElse(defaultQueryLimit)

    if(validatedQueryLimit > maximumQueryLimit) RangeNotSatisfiable(s"Shrine cannot supply $validatedQueryLimit Queries. Request $maximumQueryLimit or fewer using the limit parameter.")
    else {
      val networkQueryId: Option[NetworkQueryId] = networkId.map(_.toLong)

      networkQueryId match {
        case Some(nId) => {
          val timeoutMilliseconds: Long = timeoutSeconds.getOrElse(15L) * 1000
          val requestStartTime = System.currentTimeMillis()
          val deadline = requestStartTime + timeoutMilliseconds

          val queryAndResultsIO: IO[QueryAndResults] = QepQueryDb.db.selectResultsRowIO(nId, user)
          val previousQueriesAndResultWithObservedIO: IO[(PreviousQueriesAndResult, Boolean)] = queryAndResultsIO.flatMap { queryAndResults => {
            val queryResultIO = getQueryResultIO(user, limitOption, skipOption, sortByOption, queryAndResults)
            queryResultIO
          }
          }

          previousQueriesAndResultWithObservedIO.flatMap(previousQueriesAndResult => {

            if (shouldRespondNow(deadline, previousQueriesAndResult._1, previousQueriesAndResult._2)) {
              Log.debug(s"Responding now for $nId")
              val queryResultString: String = previousQueriesAndResult._1.toJsonText
              Log.debug(s"Will respond to request for $networkQueryId immediately with $queryResultString")

              Ok(queryResultString)
            }
            else {
              val timeout = FiniteDuration(deadline - System.currentTimeMillis(), TimeUnit.MILLISECONDS)
              QepQueryDbChangeNotifier.setTriggerForChange(nId, timeout, replyWithQueryResults(
                nId,
                user,
                limitOption,
                skipOption,
                sortByOption
              ))
            }
          })
        }
        case None => {
          val queriesAndCountIO: IO[QueriesAndCount] = getPreviousQueryDataIO(user, limitOption, skipOption, sortByOption, queryIdListOption = queryIdListOption)

          queriesAndCountIO.flatMap(queriesAndCount => {
            val queriesNoResult = PreviousQueriesAndResult(queriesAndCount.queryRowCount, skipOption.getOrElse(0), allQueries = queriesAndCount.queryCells)
            Ok(queriesNoResult.asJson.toString())
          })
        }
      }
    }
  }


  private def replyWithQueryResults(
                                    networkQueryId: NetworkQueryId,
                                    user: User,
                                    limitOption : Option[Int],
                                    skipOption : Option[Int],
                                    sortByOption: Option[String]
                                   ): IO[Response[IO]] = {

    //todo just one trip to the database SHRINE-3389
    QepQueryDb.db.selectResultsRowIO(networkQueryId, user).
      flatMap(getQueryResultAsStringIO(user, limitOption, skipOption, sortByOption, _)).
      flatMap(Ok(_)) //todo context shift for reply ?
  }

  /**
    * @param deadline time when a response must go
    * @param previousQueriesAndResult either the PreviousQueriesAndResult or something is not right
    * @return true to respond now, false to dither
    */
  def shouldRespondNow(deadline: Long,
                       previousQueriesAndResult: PreviousQueriesAndResult,
                       selectedResultObserved: Boolean
                      ):Boolean = {

    val previousQueriesUpdated = previousQueriesAndResult.allQueries.exists(q => !q.observed && !q.isQueryComplete)

    val currentTime = System.currentTimeMillis()

    val shouldRespond =  (currentTime >= deadline) || (previousQueriesUpdated || !selectedResultObserved)

    Log.debug(s"Should respond for query ${previousQueriesAndResult.selectedQuery.get.query.networkId} is $shouldRespond")

    shouldRespond
  }

  def retrieveQueryResults(user: User, networkId: Option[String], limitOption : Option[Int], skipOption : Option[Int],
                           sortByOption: Option[String], queryIdListOption: Option[List[NetworkQueryId]]= None): IO[Response[IO]] = {
    val networkQueryId: Option[NetworkQueryId] = networkId.map(_.toLong)

    val result = getStatusIO(networkQueryId, user, limitOption, skipOption, sortByOption, queryIdListOption)
    Ok(result)
  }

  def getStatusIO(networkQueryId: Option[NetworkQueryId], user:User, limitOption : Option[Int], skipOption : Option[Int], sortByOption: Option[String],
                  queryIdList: Option[List[NetworkQueryId]]): IO[String] = {
    import io.circe.generic.auto.exportEncoder
    import io.circe.syntax.EncoderOps

    networkQueryId match {

      case Some(nid) =>
        for {
          troubleOrResultsRow <- QepQueryDb.db.selectResultsRowIO(networkQueryId.get, user)
          queryStatedObserved <- IO(QueryStateObserved(troubleOrResultsRow))
          existingQueryStateObserved <- QepQueryDb.db.selectQueryStateObservedByChecksum(queryStatedObserved)
          _ <- existingQueryStateObserved.fold(QepQueryDb.db.insertQueryStateObservedIO(queryStatedObserved))(_ => IO.unit)
          queriesAndCount <- getPreviousQueryDataIO(user, limitOption, skipOption, sortByOption, networkQueryId, queryIdList)
        }
        yield {
          val queryResults = ResultsRow(troubleOrResultsRow)
          val queriesAndResult = PreviousQueriesAndResult(queriesAndCount.queryRowCount, skipOption.getOrElse(0), queriesAndCount.queryCells, Some(queryResults))
          val resultJson = queriesAndResult.asJson.toString().filter(_ != '\n')
          resultJson
        }
      case None =>
        val queriesAndCountIO: IO[QueriesAndCount]  = getPreviousQueryDataIO(user, limitOption, skipOption, sortByOption, queryIdListOption = queryIdList)
        queriesAndCountIO.flatMap(queriesAndCount => {
          val queriesAndResult = PreviousQueriesAndResult(queriesAndCount.queryRowCount, skipOption.getOrElse(0), allQueries = queriesAndCount.queryCells)
          IO(queriesAndResult.asJson.toString())
        })
    }
  }

  def retrieveQueryStatus(authedRequest: AuthedRequest[IO, User]): IO[Response[IO]] = {
    import io.circe.generic.auto.exportDecoder
    import org.http4s.circe.jsonOf
    val user = authedRequest.authInfo

    implicit val queryIdListDecoder: EntityDecoder[IO, QueryIdList] = jsonOf[IO, QueryIdList]
    val queryIdListIO: IO[QueryIdList] = authedRequest.req.as[QueryIdList]


    queryIdListIO.flatMap(qIdlist => {
      pollForQueryResults(user, queryIdListOption = Some(qIdlist.queryIds))
    })
  }

  private def constructPanel(conceptGroup: ConceptGroup, panelNumber: Int, panelTiming: Option[String]): Panel = {

    val isExcluded: Boolean = conceptGroup.isExcluded
    val minOccurrences: Int = conceptGroup.options.occurrences
    val start: Option[XMLGregorianCalendar] = conceptGroup.options.startDate.map(XmlDateHelper.toXmlGregorianCalendar)
    val end: Option[XMLGregorianCalendar] = conceptGroup.options.endDate.map(XmlDateHelper.toXmlGregorianCalendar)
    val timing: PanelTiming = panelTiming.fold(PanelTiming.Any)(p => PanelTiming.valueOfOrElse(p)(PanelTiming.Any))
    val concepts: List[Concept] = conceptGroup.concepts

    val termsToConstraints: Seq[(SimpleExpression, Option[Constrained])] = for {
      conceptItem <- concepts
      term: SimpleExpression = Term(conceptItem.path, conceptItem.displayName)
      termOption = Option(term).collect { case t: Term => t }
      modifierOption = None
      valueConstraintOption: Option[ConceptValueConstraint] = conceptItem.valueConstraint
      constrainedOption = {
        if (modifierOption.isEmpty && valueConstraintOption.isEmpty) {
          None
        }
        else {
          termOption.map(t => Constrained(t, modifierOption, valueConstraintOption.map(_.toValueConstraint)))
        }
      }
    } yield {
      term -> constrainedOption
    }

    val termsWithNoConstraints: Seq[SimpleExpression] = termsToConstraints.collect { case (expr, None) => expr }
    val termsWithSomeConstraints: Seq[Constrained] = termsToConstraints.collect { case (_, Some(constrained)) => constrained }

    Panel(panelNumber, isExcluded, minOccurrences, start, end, timing, termsWithNoConstraints, termsWithSomeConstraints)
  }

  private def constructQueryDefinition(basicQuery: BasicQuery): QueryDefinition = {

    val name = basicQuery.name

    val timing = basicQuery.queryTiming.map(q => QueryTiming.valueOf(q))

    val panelsAttempt = basicQuery.conceptGroups.zipWithIndex.map(conceptGroupWithindex => {
      constructPanel(conceptGroupWithindex._1, panelNumber = conceptGroupWithindex._2, panelTiming = None)
    })

    val exprsAttempt = Try(panelsAttempt.map(_.toExpression))

    val consolidatedExprAttempt: Try[Option[Expression]] = exprsAttempt.map {
      case es if es.isEmpty => None
      case es if es.size == 1 => Some(es.head)
      case es => Some(And(es: _*))
    }

    consolidatedExprAttempt.map(consolidated => QueryDefinition(name, consolidated.map(_.normalize))).get
  }

  def authorizeAndRunQuery(basicQuery: BasicQuery, user: User)(body: RunQueryData => IO[RunQueryResponse]): IO[Response[IO]] = {

    import io.circe.generic.auto.exportEncoder
    import io.circe.syntax.EncoderOps

    val topicId = Some(basicQuery.topic.id.toString)
    val topicName = Some(basicQuery.topic.name)

    val queryDefinition = constructQueryDefinition(basicQuery)

    val networkQueryId = BroadcastMessage.Ids.next

    val runQueryData = RunQueryData(user,
      networkQueryId,
      basicQuery.projectId,
      queryDefinition,
      basicQuery.resultOutputs,
      topicId,
      topicName)

    val qepConfig = ConfigSource.config.getConfig("shrine.queryEntryPoint")

    val authorizationService: QueryAuthorizationService = AuthStrategy.determineQueryAuthorizationService(qepConfig)

    Log.debug(s"auditAuthorizeAndThen($user, $networkQueryId, $queryDefinition, $topicId) with $authorizationService")

    authorizationService.authorizeRunQuery(user, networkQueryId, queryDefinition, topicId, topicName) match {
      case na: NotAuthorized => Forbidden(na.reason)
      case _: Authorized => body(runQueryData).flatMap(response => Ok(response.asJson.toString()))
    }
  }

  case class QueryIdList(queryIds: List[NetworkQueryId])
}

case class ResultsTable(
                         rowCount:Int,
                         rowOffset:Int,
                         queries:Seq[QueryCell]
                       )

case class PreviousQueriesAndResult(rowCount:Int,
                                    rowOffset:Int,
                                    selectedQuery: Option[ResultsRow],
                                    isAllQueriesComplete: Boolean,
                                    allQueries:Seq[QueryCell]
                            )
{
  def toJsonText: String = {
    import io.circe.generic.auto.exportEncoder
    import io.circe.syntax.EncoderOps

    val resultJson = this.asJson.toString().filter(_ != '\n')

    var finalResult: String = resultJson
    if(this.selectedQuery.get.isComplete){
      finalResult = finalResult.concat("\n")
    }

    finalResult
  }
}
object PreviousQueriesAndResult {

  def apply(
            rowCount:Int,
            rowOffset:Int,
            allQueries:Seq[QueryCell],
            selectedQuery: Option[ResultsRow] = None
            ): PreviousQueriesAndResult = {

    val allQueriesComplete: Boolean  = allQueries.forall(_.isQueryComplete)
    PreviousQueriesAndResult(
      rowCount,
      rowOffset,
      selectedQuery,
      allQueriesComplete,
      allQueries
    )
  }
}

object JsonContentService {

  def apply(service: HttpRoutes[IO]) = {
    val jsonContentHeader = `Content-Type`(org.http4s.MediaType.application.json)
    service.map {
      case Status.Successful(resp) => resp.putHeaders(jsonContentHeader)
      case resp => resp
    }
  }
}

case class QueryCell(networkId:String,
                     queryName: QueryName,
                     dateCreated: Time,
                     changeDate: Time,
                     flag:Option[QueryFlag],
                     status: String,
                     internalStatus: String,
                     isQueryError: Boolean,
                     isResultsError: Boolean,
                     isQueryComplete: Boolean,
                     problemDigest: Option[JsonProblemDigest] = None,
                     observed: Boolean
                    )

object QueryCell {

  def apply(queryWithResultStatus: QueryWithResultStatus, flag: Option[QueryFlag]): QueryCell = QueryCell(
    networkId = queryWithResultStatus.networkId.toString,
    queryName = queryWithResultStatus.queryName,
    dateCreated = queryWithResultStatus.dateCreated,
    changeDate = queryWithResultStatus.changeDate,
    flag = flag,
    status = queryStatusToUiString(queryWithResultStatus.status.toString, queryWithResultStatus.isComplete),
    internalStatus = queryWithResultStatus.status.toString,
    observed =  queryWithResultStatus.observed,
    isQueryError = queryWithResultStatus.isQueryError,
    isResultsError = queryWithResultStatus.isResultsError,
    isQueryComplete = queryWithResultStatus.isComplete || queryWithResultStatus.isQueryError

  )

  def apply(fullQuery: FullQuery, isComplete: Boolean, isQueryError: Boolean, isResultsError: Boolean, observed: Boolean ): QueryCell = {

    QueryCell(
      networkId = fullQuery.query.networkId.toString,
      queryName = fullQuery.query.queryName,
      dateCreated = fullQuery.query.dateCreated,
      changeDate = fullQuery.query.changeDate,
      flag = fullQuery.flag,
      status = queryStatusToUiString(fullQuery.query.status.toString, isComplete),
      internalStatus = fullQuery.query.status.toString,
      problemDigest = fullQuery.problemDigest.map(JsonProblemDigest(_)),
      observed = observed,
      isQueryError = isQueryError,
      isResultsError = isResultsError,
      isQueryComplete = isComplete || isQueryError
    )
  }

  def queryStatusToUiString(internalStatus: String, isComplete: Boolean): String = {

    val maybeStatus = QueryStatuses.namesToStatuses.get(internalStatus)
    maybeStatus.map{q =>
      if(q == QueryStatuses.SentToAdapters && isComplete) "Completed"
      else q.uiString
    }.getOrElse("Unknown")
  }
}
case class ResultsRow(
                       query:QueryCell,
                       results: Seq[QResult],
                       isComplete: Boolean,
                       dataVersion:Time
                     )
object ResultsRow {
  def apply(queryResults: QueryAndResults): ResultsRow = {
    val networkQueryId = queryResults.fullQuery.query.networkId
    val results: Seq[QResult] = queryResults.fullResults.map(QResult(networkQueryId, _))

    val isComplete = if (results.isEmpty) false
    else results.forall(_.isComplete)

    val isQueryError = QueryStatuses.namesToStatuses.get(queryResults.fullQuery.query.status).exists(_.isError)
    val isResultsError = if(results.isEmpty) false else results.forall(_.isError)

    val queryCell: QueryCell = QueryCell(queryResults.fullQuery, isComplete, isQueryError, isResultsError, observed = true)

    val dataVersion = (Seq(queryCell.changeDate) ++ results.map(_.changeDate)).max //the latest change date

    ResultsRow(queryCell, results, isComplete, dataVersion)
  }
}

case class QResult(
                    resultId:Long,
                    networkQueryId:NetworkQueryId,
                    instanceId:Long,
                    adapterNode:String,
                    resultType:Option[ResultOutputType],
                    count:Long,
                    status:String,
                    internalStatus: String,
                    statusMessage:Option[String],
                    changeDate:Long,
                    breakdowns: Seq[BreakdownResultsForType],
                    problemDigest:Option[JsonProblemDigest]
                  ) {

  def isComplete: Boolean = {
    ResultStatuses.namesToStatuses.get(internalStatus).map(_.isFinal).getOrElse {
      QueryResult.StatusType.valueOf(internalStatus).get.isDone
    }
  }

  def isError: Boolean = {
    ResultStatuses.namesToStatuses.get(internalStatus).map(_.isError).getOrElse {
      QueryResult.StatusType.valueOf(internalStatus).get.isError
    }
  }
}

object QResult {
  def apply(networkQueryId:NetworkQueryId,fullQueryResult: FullQueryResult): QResult = {
    //Try to catch and log SHRINE-2239 in the wild
    if(networkQueryId != fullQueryResult.networkQueryId) {
      val isx = new IllegalStateException(s"networkQueryId $networkQueryId != fullQueryResult.networkQueryId ${fullQueryResult.networkQueryId}")
      isx.fillInStackTrace()
      Log.debug("SHRINE-2239 trapped ",isx)
    }

    new QResult(
      resultId = fullQueryResult.resultId,
      networkQueryId = networkQueryId,
      instanceId  = fullQueryResult.instanceId,
      adapterNode = fullQueryResult.adapterNode,
      resultType = fullQueryResult.resultType,
      count = fullQueryResult.count,
      status = resultStatusToUiString(fullQueryResult.status),
      internalStatus = fullQueryResult.status.toString,
      statusMessage = fullQueryResult.statusMessage,
      changeDate = fullQueryResult.changeDate,
      breakdowns = fullQueryResult.breakdownTypeToResults.map(tToR => BreakdownResultsForType(fullQueryResult.adapterNode,tToR._1,tToR._2)).to[Seq],
      problemDigest = fullQueryResult.problemDigest.map(JsonProblemDigest(_))
    )
  }

  def resultStatusToUiString(internalStatus: String): String = {
    val legacyStatusMap: Map[String, String] = Map(
      "FINISHED" -> "Completed",
      //Anything from 1.25.4 or earlier is likely never coming back
      "ERROR" -> "Site Error",
      "PROCESSING" -> "Site Error",
      "RUNNING" -> "Site Error",
      "QUEUED" -> "Site Error",
      "INCOMPLETE" -> "Site Error",
      "HELD" -> "Site Error",
      "SMALL_QUEUE" -> "Site Error",
      "TIMEDOUT" -> "Site Error",
      "MEDIUM_QUEUE" -> "Site Error",
      "LARGE_QUEUE" -> "Site Error",
      "NO_MORE_QUEUE" -> "Site Error",
      "MEDIUM_QUEUE_RUNNING" -> "Site Error",
      "LARGE_QUEUE_RUNNING" -> "Site Error",
      "HUB_WILL_SUBMIT" -> "Site Error"
    ).withDefaultValue("Unknown")

    ResultStatuses.namesToStatuses.get(internalStatus).fold{legacyStatusMap(internalStatus)}{status => status.statusName}
  }
}

case class QueriesAndCount(queryRowCount: Int, queryCells: Seq[QueryCell])

case class QueryIdList(queryIds: List[NetworkQueryId])


case class BasicQuery(name:String, projectId: String, topic: TopicIdAndName, resultOutputs: List[ResultOutput], conceptGroups: List[ConceptGroup], queryTiming: Option[String])

case class ConceptGroup(concepts:  List[Concept],
                        panelTiming: Option[String],
                        isExcluded: Boolean,
                        options: ConceptGroupOptions)

case class Concept(displayName: String,
                   path: String, valueConstraint: Option[ConceptValueConstraint])


case class ConceptGroupOptions(startDate: Option[Long],
                               endDate: Option[Long],
                               occurrences: Int)

case class TopicIdAndName(id: Long, name: String)

case class ResultOutput(name: String)

object ResultOutput {

  def i2b2OutputList(resultOutput: List[ResultOutput]): NodeSeq = XmlUtil.stripWhitespace {
    <result_output_list>
      {for {
      (outputType, i) <- resultOutput.zipWithIndex
      priorityIndex = i + 1
    } yield {
        <result_output priority_index={priorityIndex.toString} name={outputType.name.toString.toLowerCase}/>
    }}
    </result_output_list>
  }
}

case class RunQueryData(user: User,
                        networkQueryId: Long,
                        projectId: String,
                        queryDefinition: QueryDefinition,
                        resultOutputs: List[ResultOutput],
                        topicId: Option[String],
                        topicName: Option[String])
case class RunQueryResponse(networkQueryId: NetworkQueryId, status: String)

case class ConceptValueConstraint(valueType: String, operator: String, value: String, unit: Option[String]) {

  validateLabValues(valueType, unit, operator, value)

  def toValueConstraint: ValueConstraint = {
      ValueConstraint(valueType, unit, operator, value)
  }

  def toDoubleOption(s: String):Option[Double] = {
    try {
      Some(s.toDouble)
    } catch {
      case e: NumberFormatException => None
    }
  }

  def validateLabValues(valueType: String,  unit: Option[String], operator: String, value: String): Unit =  {

    if(!(valueType == "FLAG" || valueType == "NUMBER")) {
      throw InvalidValueConstraintException(s"Unrecognized value type $valueType. Must be either FLAG or NUMBER.")
    }

    if(valueType == "NUMBER" && toDoubleOption(value).isEmpty) {
      throw InvalidValueConstraintException(s"Value type is NUMBER but value is not a valid number: $value.")
    }

    if(!(operator == "EQ" || operator == "LT" || operator == "GT" || operator == "GE"
      || operator == "LE" || operator == "BETWEEN")) {
      throw InvalidValueConstraintException(s"Unrecognized operator $operator. Must be either EQ, LT, GT, GE, LE, or BETWEEN.")
    }
  }
}

case class InvalidValueConstraintException(message: String) extends IllegalArgumentException(message)