package net.shrine.api.qep

import cats.effect.IO
import net.shrine.crypto.BouncyKeyStoreCollection
import net.shrine.hub.data.client.HubClient
import net.shrine.hub.mom.ShrineMomClient
import net.shrine.log.Log
import net.shrine.protocol.BroadcastMessage
import net.shrine.protocol.version.QueryStatuses
import net.shrine.protocol.version.v1.{Query, Researcher, RunQueryAtHub, Topic, UpdateQueryAtQepWithStatus}
import net.shrine.qep.audit.{QepAuditDb, QepQueryAuditData}
import net.shrine.qep.querydb.QepQueryDb
import net.shrine.source.ConfigSource

object QueryRunner {

  def runQuery(runQueryData: RunQueryData): IO[RunQueryResponse] = {

    val certCollection: BouncyKeyStoreCollection = BouncyKeyStoreCollection.fromConfig
    val commonName: String = certCollection.myEntry.commonName.getOrElse {
      val hostname = java.net.InetAddress.getLocalHost.getHostName
      Log.warn(s"No common name available from ${certCollection.myEntry}. Using $hostname instead.")
      hostname
    }

    val collectQepAudit = ConfigSource.config.getBoolean("shrine.queryEntryPoint.audit.collectQepAudit")

    val i2b2OutputListStr = ResultOutput.i2b2OutputList(runQueryData.resultOutputs).toString()
    val networkQueryId = BroadcastMessage.Ids.next
    val qepQueryAuditData = QepQueryAuditData(
      commonName,
      runQueryData.user.username,
      networkQueryId,
      runQueryData.queryDefinition.name,
      runQueryData.topicId,
      runQueryData.topicName
    )
    val runQueryIO = for {
      localNode <- HubClient.getLocalNodeIO
      researcher <- IO(Researcher.createRepeatableForV26(
        userName = runQueryData.user.username,
        userDomainName = runQueryData.user.domain,
        nodeId = localNode.id
      ))
      topic <- IO(Topic.createRepeatableForV26(
        researcherId = researcher.id,
        name = runQueryData.topicId.get,
        description = runQueryData.topicName.getOrElse("")
      ))
      queryWithId <- IO(Query.create(
        id = networkQueryId,
        queryDefinitionXml = runQueryData.queryDefinition.toI2b2String,
        outputTypesXml = i2b2OutputListStr,
        queryName = runQueryData.queryDefinition.name,
        nodeOfOriginId = localNode.id,
        researcherId = researcher.id,
        topicId = topic.id,
        projectName = runQueryData.projectId,
        certCollection
      ))
      _ <- QepQueryDb.db.insertQueryIO(queryWithId, researcher)
      _ <- if (collectQepAudit) QepAuditDb.db.insertQepQueryIO(qepQueryAuditData) else IO.unit
      queryToHub <- IO(queryWithId.withStatus(QueryStatuses.SentToHub))
      _ <- ShrineMomClient.sendToHubIO(
        subjectId = queryToHub.id,
        envelopeContents = RunQueryAtHub(queryToHub, researcher, topic),
        envelopeContentsCompanion = RunQueryAtHub,
        logString = s"Sent query command ${queryWithId.id}"
      )
      _ <- QepQueryDb.db.updateQepQueryIO(UpdateQueryAtQepWithStatus(queryToHub))
    } yield ()

    runQueryIO.flatMap { _ =>
      IO(RunQueryResponse(networkQueryId, "RECEIVED_BY_QEP"))
    }
  }


}
