package net.shrine.api.qep

import java.util.concurrent.TimeUnit

import cats.data.OptionT
import cats.effect.{ConcurrentEffect, ContextShift, IO, Timer}
import cats.implicits.{catsSyntaxEither => _}
import io.circe.Decoder.Result
import io.circe.{DecodingFailure, Json}
import io.circe.generic.auto.exportEncoder
import io.circe.syntax.EncoderOps
import net.shrine.audit.NetworkQueryId
import net.shrine.crypto.NewTestKeyStore
import net.shrine.http4s.auth.BasicAuthentication
import net.shrine.http4s.auth.UserSessionToken
import net.shrine.http4s.catsio.ExecutionContexts
import net.shrine.i2b2.protocol.pm.User
import net.shrine.problem.{AbstractProblem, JsonProblemDigest, ProblemSources}
import net.shrine.protocol.version.v1.{Node, Query, Researcher, Topic, UpdateQueryAtQepWithError}
import net.shrine.protocol.version.{DateStamp, NodeKey, NodeName, QueryStatuses, ResultStatuses, UserDomainName}
import net.shrine.protocol.{Credential, QueryResult, ResultOutputType}
import net.shrine.qep.querydb.{QepQuery, QepQueryDb, QueryAndResults, QueryResultRow, QueryWithResultStatus}
import net.shrine.util.ShouldMatchersForJUnit
import org.http4s.circe.jsonOf
import org.http4s.dsl.io.{BadRequest, Ok, RangeNotSatisfiable, Unauthorized}
import org.http4s.headers.Authorization
import org.http4s.{BasicCredentials, EntityDecoder, Headers, Method, Request, Response, Uri}
import org.junit.runner.RunWith
import org.scalatestplus.junit.JUnitRunner
import org.scalatest.{BeforeAndAfter, FlatSpec}

import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.duration.FiniteDuration


@RunWith(classOf[JUnitRunner])
class QepServiceTest extends FlatSpec with BeforeAndAfter with ShouldMatchersForJUnit {

  implicit val jsonDecoder: EntityDecoder[IO, Json] = jsonOf[IO, Json]

  def extractResponse(request: Request[IO]): Response[IO] = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val responseOptionIo: OptionT[IO, Response[IO]] = QepService().router.run(request)

    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  "QepService" should " reply to ping with pong" in {
    val pingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/ping"))

    val response: Response[IO] = extractResponse(pingRequest)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("pong")(entityText)
  }

  "A user with valid credentials" should "be able to get a session Id from the login service" in {
    val username = "admin"
    val password = "pwd"
    val loginRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/login"),
      headers = Headers.of(Authorization(BasicCredentials(username, password)))
    )
    val response: Response[IO]= extractResponse(loginRequest)
    val result: Json = response.as[Json].unsafeRunSync
    val expectedResult = UserSessionToken(username, "pwd").asJson
    assertResult(Ok)(response.status)
    assertResult(expectedResult)(result)
  }

  "A user with an invalid password" should "not be able to get a session id from the login service" in {
    val username = "admin"
    val password = "bad pwd"
    val loginRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString("/login"),
      headers = Headers.of(Authorization(BasicCredentials(username, password)))
    )
    val response: Response[IO]= extractResponse(loginRequest)
    assertResult(Unauthorized)(response.status)
  }

  "QepService" should "return an OK and previous queries  for a queryResult request" in {

    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromBch)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromMgh)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromDfci)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromPartners)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
    uri = Uri.unsafeFromString(s"/queryResult"),
    headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(queryResultTableRequest)

    val result: Json = response.as[Json].unsafeRunSync
    val queryRowCount: Int = QepQueryDb.db.countPreviousQueriesByUserAndDomain(
      userName = researcherUser.username,
      domain = researcherUser.domain
    )

    val queriesAndResult: PreviousQueriesAndResult = getQueryDataIO(researcherUser).flatMap(queriesAndResult => {

      IO(PreviousQueriesAndResult(queryRowCount, 0, allQueries = queriesAndResult.queryCells))
    }).unsafeRunSync()
    val expectedJson: Json = queriesAndResult.asJson
    assertResult(Ok)(response.status)
    assertResult(expectedJson)(result)
  }

  "QepService" should "not return table of data for a queryResultsTable request if user is not authorized" in {
    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResult"))

    val response: Response[IO] = extractResponse(queryResultTableRequest)
    val entityText = response.as[String].unsafeRunSync

    assertResult(Unauthorized)(response.status)
    assertResult(BasicAuthentication.unauthorizedMsg)(entityText)
  }

  "QepService" should "return an OK and a table of data for a queryResults request with different skip and limit values" in {
    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromBch)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromDfci)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromPartners)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResult?skip=2&limit=2"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(queryResultTableRequest)

    val result: Json = response.as[Json].unsafeRunSync

    val skip = 2
    val queriesAndResult: PreviousQueriesAndResult = getQueryDataIO(researcherUser, skipOption = Some(skip), limitOption = Some(2)).flatMap(queriesAndResult => {

      IO(PreviousQueriesAndResult(queriesAndResult.queryRowCount, skip, allQueries = queriesAndResult.queryCells))
    }).unsafeRunSync()

    val expectedJson: Json = queriesAndResult.asJson
    assertResult(Ok)(response.status)
    assertResult(expectedJson)(result)
  }

  "QepService" should "return an OK and a table of pevious queries request with different sort by" in {
    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepQuery(qep2Query)
    QepQueryDb.db.insertQepQuery(qep3Query)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromMgh)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromDfci)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromPartners)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromBch)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromHarvard)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResult?limit=2&sortBy=queryName.asc"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(queryResultTableRequest)

    val result: Json = response.as[Json].unsafeRunSync

    val queriesAndResult: PreviousQueriesAndResult = getQueryDataIO(researcherUser, limitOption = Some(2), sortByOption = Some("queryName.asc")).flatMap(queriesAndResult => {

      IO(PreviousQueriesAndResult(queriesAndResult.queryRowCount, 0, allQueries = queriesAndResult.queryCells))
    }).unsafeRunSync()

    val expectedJson: Json = queriesAndResult.asJson
    assertResult(Ok)(response.status)
    assertResult(expectedJson)(result)
  }

  "QepService" should "return an OK and specified query result and list of queries for a queryResult request" in {
    val networkId = qep2Query.networkId
    QepQueryDb.db.insertQepQuery(qep2Query)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromHarvard)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromNU)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResult?networkId=$networkId"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    //extractResponse(queryResultTableRequest)

    val response: Response[IO] = extractResponse(queryResultTableRequest)

    assertResult(Ok)(response.status)

    val result: Json = response.as[Json].unsafeRunSync

    val expectedJson = getQueryResultIO(researcherUser,networkId).unsafeRunSync()

    assertResult(expectedJson)(result)
  }

  "QepService" should "return an OK and specified query result and list of queries for a queryResultPolling request" in {
    val networkId = qepQuery.networkId
    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromBch)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromMgh)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromDfci)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromPartners)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling?networkId=$networkId&timeoutSeconds=1"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(queryResultTableRequest)

    assertResult(Ok)(response.status)

    val result: Json = response.as[Json].unsafeRunSync

    val expectedJson: String = """{
                                 |  "rowCount" : 1,
                                 |  "rowOffset" : 0,
                                 |  "selectedQuery" : {
                                 |    "query" : {
                                 |      "networkId" : "1",
                                 |      "queryName" : "TestQuery",
                                 |      "dateCreated" : ,
                                 |      "changeDate" : ,
                                 |      "flag" : null,
                                 |      "status" : "Completed",
                                 |      "internalStatus" : "Sent to Adapters",
                                 |      "isQueryError" : false,
                                 |      "isResultsError" : false,
                                 |      "isQueryComplete" : true,
                                 |      "problemDigest" : null,
                                 |      "observed" : true
                                 |    },
                                 |    "results" : [
                                 |      {
                                 |        "resultId" : 10,
                                 |        "networkQueryId" : 1,
                                 |        "instanceId" : 100,
                                 |        "adapterNode" : "BCH",
                                 |        "resultType" : {
                                 |          "name" : "PATIENT_COUNT_XML",
                                 |          "isBreakdown" : false,
                                 |          "i2b2Options" : {
                                 |            "description" : "Number of patients",
                                 |            "displayType" : "CATNUM"
                                 |          },
                                 |          "id" : 4
                                 |        },
                                 |        "count" : 3000,
                                 |        "status" : "Completed",
                                 |        "internalStatus" : "FINISHED",
                                 |        "statusMessage" : null,
                                 |        "changeDate" : ,
                                 |        "breakdowns" : [
                                 |        ],
                                 |        "problemDigest" : null
                                 |      },
                                 |      {
                                 |        "resultId" : 10,
                                 |        "networkQueryId" : 1,
                                 |        "instanceId" : 100,
                                 |        "adapterNode" : "MGH",
                                 |        "resultType" : {
                                 |          "name" : "PATIENT_COUNT_XML",
                                 |          "isBreakdown" : false,
                                 |          "i2b2Options" : {
                                 |            "description" : "Number of patients",
                                 |            "displayType" : "CATNUM"
                                 |          },
                                 |          "id" : 4
                                 |        },
                                 |        "count" : 30,
                                 |        "status" : "Completed",
                                 |        "internalStatus" : "FINISHED",
                                 |        "statusMessage" : null,
                                 |        "changeDate" : ,
                                 |        "breakdowns" : [
                                 |        ],
                                 |        "problemDigest" : null
                                 |      },
                                 |      {
                                 |        "resultId" : 10,
                                 |        "networkQueryId" : 1,
                                 |        "instanceId" : 100,
                                 |        "adapterNode" : "DFCI",
                                 |        "resultType" : {
                                 |          "name" : "PATIENT_COUNT_XML",
                                 |          "isBreakdown" : false,
                                 |          "i2b2Options" : {
                                 |            "description" : "Number of patients",
                                 |            "displayType" : "CATNUM"
                                 |          },
                                 |          "id" : 4
                                 |        },
                                 |        "count" : 30000,
                                 |        "status" : "Completed",
                                 |        "internalStatus" : "FINISHED",
                                 |        "statusMessage" : null,
                                 |        "changeDate" : ,
                                 |        "breakdowns" : [
                                 |        ],
                                 |        "problemDigest" : null
                                 |      },
                                 |      {
                                 |        "resultId" : 10,
                                 |        "networkQueryId" : 1,
                                 |        "instanceId" : 100,
                                 |        "adapterNode" : "Partners",
                                 |        "resultType" : {
                                 |          "name" : "PATIENT_COUNT_XML",
                                 |          "isBreakdown" : false,
                                 |          "i2b2Options" : {
                                 |            "description" : "Number of patients",
                                 |            "displayType" : "CATNUM"
                                 |          },
                                 |          "id" : 4
                                 |        },
                                 |        "count" : 300,
                                 |        "status" : "Completed",
                                 |        "internalStatus" : "FINISHED",
                                 |        "statusMessage" : null,
                                 |        "changeDate" : ,
                                 |        "breakdowns" : [
                                 |        ],
                                 |        "problemDigest" : null
                                 |      }
                                 |    ],
                                 |    "isComplete" : true,
                                 |    "dataVersion" :
                                 |  },
                                 |  "isAllQueriesComplete" : true,
                                 |  "allQueries" : [
                                 |    {
                                 |      "networkId" : "1",
                                 |      "queryName" : "TestQuery",
                                 |      "dateCreated" : ,
                                 |      "changeDate" : ,
                                 |      "flag" : null,
                                 |      "status" : "Completed",
                                 |      "internalStatus" : "Sent to Adapters",
                                 |      "isQueryError" : false,
                                 |      "isResultsError" : false,
                                 |      "isQueryComplete" : true,
                                 |      "problemDigest" : null,
                                 |      "observed" : true
                                 |    }
                                 |  ]
                                 |}""".stripMargin

    val resultJson: String = result.toString()
    val dateCreatedPattern = "\"dateCreated\" : [0-9]+".r
    val changedDatePattern = "\"changeDate\" : [0-9]+".r
    val dataVersionPattern = "\"dataVersion\" : [0-9]+".r
    val resultJsonNoDateCreated = dateCreatedPattern.replaceAllIn(resultJson, "\"dateCreated\" : ")
    val resultJsonNoChangeDate = changedDatePattern.replaceAllIn(resultJsonNoDateCreated, "\"changeDate\" : ")
    val resultJsonNoDataVersion = dataVersionPattern.replaceAllIn(resultJsonNoChangeDate, "\"dataVersion\" :")
    assertResult(expectedJson)(resultJsonNoDataVersion)
  }

  "QepService" should "return a RangeNotSatisfiable if the requester specifies a very high limit" in {
    val networkId = qepQuery.networkId
    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromBch)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromMgh)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromDfci)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromPartners)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling?networkId=$networkId&timeoutSeconds=1&limit=300"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(queryResultTableRequest)

    assertResult(RangeNotSatisfiable)(response.status)
  }

    "QepService" should "return an OK and specified query with isResultsError if all results have error" in {
    val networkId = qepQueryWithResultErrors.networkId
    QepQueryDb.db.insertQepQuery(qepQueryWithResultErrors)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromNUError)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromHarvardError)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling?networkId=$networkId&timeoutSeconds=1"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(queryResultTableRequest)

    assertResult(Ok)(response.status)

    val result: Json = response.as[Json].unsafeRunSync

    val expectedJson: String = """{
                                 |  "rowCount" : 1,
                                 |  "rowOffset" : 0,
                                 |  "selectedQuery" : {
                                 |    "query" : {
                                 |      "networkId" : "50",
                                 |      "queryName" : "TestQuery",
                                 |      "dateCreated" : ,
                                 |      "changeDate" : ,
                                 |      "flag" : null,
                                 |      "status" : "Completed",
                                 |      "internalStatus" : "Sent to Adapters",
                                 |      "isQueryError" : false,
                                 |      "isResultsError" : true,
                                 |      "isQueryComplete" : true,
                                 |      "problemDigest" : null,
                                 |      "observed" : true
                                 |    },
                                 |    "results" : [
                                 |      {
                                 |        "resultId" : 501,
                                 |        "networkQueryId" : 50,
                                 |        "instanceId" : 120,
                                 |        "adapterNode" : "NU",
                                 |        "resultType" : {
                                 |          "name" : "PATIENT_COUNT_XML",
                                 |          "isBreakdown" : false,
                                 |          "i2b2Options" : {
                                 |            "description" : "Number of patients",
                                 |            "displayType" : "CATNUM"
                                 |          },
                                 |          "id" : 4
                                 |        },
                                 |        "count" : 80000,
                                 |        "status" : "Site Error",
                                 |        "internalStatus" : "ERROR",
                                 |        "statusMessage" : null,
                                 |        "changeDate" : ,
                                 |        "breakdowns" : [
                                 |        ],
                                 |        "problemDigest" : null
                                 |      },
                                 |      {
                                 |        "resultId" : 502,
                                 |        "networkQueryId" : 50,
                                 |        "instanceId" : 100,
                                 |        "adapterNode" : "HARVARD",
                                 |        "resultType" : {
                                 |          "name" : "PATIENT_COUNT_XML",
                                 |          "isBreakdown" : false,
                                 |          "i2b2Options" : {
                                 |            "description" : "Number of patients",
                                 |            "displayType" : "CATNUM"
                                 |          },
                                 |          "id" : 4
                                 |        },
                                 |        "count" : 30000,
                                 |        "status" : "Error In Shrine",
                                 |        "internalStatus" : "Error In Shrine",
                                 |        "statusMessage" : null,
                                 |        "changeDate" : ,
                                 |        "breakdowns" : [
                                 |        ],
                                 |        "problemDigest" : null
                                 |      }
                                 |    ],
                                 |    "isComplete" : true,
                                 |    "dataVersion" :
                                 |  },
                                 |  "isAllQueriesComplete" : true,
                                 |  "allQueries" : [
                                 |    {
                                 |      "networkId" : "50",
                                 |      "queryName" : "TestQuery",
                                 |      "dateCreated" : ,
                                 |      "changeDate" : ,
                                 |      "flag" : null,
                                 |      "status" : "Completed",
                                 |      "internalStatus" : "Sent to Adapters",
                                 |      "isQueryError" : false,
                                 |      "isResultsError" : true,
                                 |      "isQueryComplete" : true,
                                 |      "problemDigest" : null,
                                 |      "observed" : true
                                 |    }
                                 |  ]
                                 |}""".stripMargin

    val resultJson: String = result.toString()
    val dateCreatedPattern = "\"dateCreated\" : [0-9]+".r
    val changedDatePattern = "\"changeDate\" : [0-9]+".r
    val dataVersionPattern = "\"dataVersion\" : [0-9]+".r
    val resultJsonNoDateCreated = dateCreatedPattern.replaceAllIn(resultJson, "\"dateCreated\" : ")
    val resultJsonNoChangeDate = changedDatePattern.replaceAllIn(resultJsonNoDateCreated, "\"changeDate\" : ")
    val resultJsonNoDataVersion = dataVersionPattern.replaceAllIn(resultJsonNoChangeDate, "\"dataVersion\" :")
    assertResult(expectedJson)(resultJsonNoDataVersion)
  }


  "QepService" should "return an OK and specified query with status isError if there is a Query error" in {
    val networkId = qepQueryWithErrors.networkId
    QepQueryDb.db.insertQepQuery(qepQueryWithErrors)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling?networkId=$networkId&timeoutSeconds=1"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(queryResultTableRequest)

    assertResult(Ok)(response.status)

    val result: Json = response.as[Json].unsafeRunSync

    val expectedJson: String = """{
                                 |  "rowCount" : 1,
                                 |  "rowOffset" : 0,
                                 |  "selectedQuery" : {
                                 |    "query" : {
                                 |      "networkId" : "50",
                                 |      "queryName" : "TestQuery",
                                 |      "dateCreated" : ,
                                 |      "changeDate" : ,
                                 |      "flag" : null,
                                 |      "status" : "Network Error",
                                 |      "internalStatus" : "HubError",
                                 |      "isQueryError" : true,
                                 |      "isResultsError" : false,
                                 |      "isQueryComplete" : true,
                                 |      "problemDigest" : null,
                                 |      "observed" : true
                                 |    },
                                 |    "results" : [
                                 |    ],
                                 |    "isComplete" : false,
                                 |    "dataVersion" :
                                 |  },
                                 |  "isAllQueriesComplete" : true,
                                 |  "allQueries" : [
                                 |    {
                                 |      "networkId" : "50",
                                 |      "queryName" : "TestQuery",
                                 |      "dateCreated" : ,
                                 |      "changeDate" : ,
                                 |      "flag" : null,
                                 |      "status" : "Network Error",
                                 |      "internalStatus" : "HubError",
                                 |      "isQueryError" : true,
                                 |      "isResultsError" : false,
                                 |      "isQueryComplete" : true,
                                 |      "problemDigest" : null,
                                 |      "observed" : true
                                 |    }
                                 |  ]
                                 |}""".stripMargin

    val resultJson: String = result.toString()
    val dateCreatedPattern = "\"dateCreated\" : [0-9]+".r
    val changedDatePattern = "\"changeDate\" : [0-9]+".r
    val dataVersionPattern = "\"dataVersion\" : [0-9]+".r
    val resultJsonNoDateCreated = dateCreatedPattern.replaceAllIn(resultJson, "\"dateCreated\" : ")
    val resultJsonNoChangeDate = changedDatePattern.replaceAllIn(resultJsonNoDateCreated, "\"changeDate\" : ")
    val resultJsonNoDataVersion = dataVersionPattern.replaceAllIn(resultJsonNoChangeDate, "\"dataVersion\" :")
    assertResult(expectedJson)(resultJsonNoDataVersion)
  }

  "QepService" should "immediately return an OK for a queryResultPolling result without a version" in {
    val networkId = qepQuery.networkId
    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromBch)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromMgh)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromDfci)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromPartners)

    val queryResultTableRequest: Request[IO] = Request(
      method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling?networkId=$networkId"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow")))
    )

    val responseOptionIo: OptionT[IO, Response[IO]] = QepService().router.run(queryResultTableRequest)

    val responseIO = responseOptionIo.map { r: Response[IO] => r }.fold {
        fail("No response from service")
      } { r: Response[IO] => r
    }

    val response = responseIO.unsafeRunSync()

    assertResult(Ok)(response.status)

    val resultJson: Json = response.as[Json].unsafeRunSync

    import io.circe.generic.auto.exportDecoder
    val queriesAndResults = resultJson.as[PreviousQueriesAndResult]
    queriesAndResults.fold({df:DecodingFailure => fail(df)},{ _ => })
  }

  "QepService" should "return an OK and the same queryResultPolling result if nothing changes" in {
    val networkId = qepQuery.networkId
    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromBch)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromMgh)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromDfci)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromPartners)

    def queryResultTableRequest(lastVersion:Long = -1L): Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling?networkId=$networkId&timeoutSeconds=1"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val responseOptionIo: OptionT[IO, Response[IO]] = QepService().router.run(queryResultTableRequest())
    val responseIO = responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }
    val response = responseIO.unsafeRunSync()
    assertResult(Ok)(response.status)
    val resultJson: Json = response.as[Json].unsafeRunSync

    import io.circe.generic.auto.exportDecoder
    val queriesAndResults: Result[PreviousQueriesAndResult] = resultJson.as[PreviousQueriesAndResult]

    val latestChange = queriesAndResults.fold({df:DecodingFailure => fail(df)},{ previousQueriesAndResults:PreviousQueriesAndResult =>
      //todo the structure still doesn't expose a version with a checksum, so figure out the latest timestamp and use that
      val maxQueryChange = previousQueriesAndResults.allQueries.maxBy(_.changeDate).changeDate
      val maxResultChange = previousQueriesAndResults.selectedQuery.fold(0L){resultsRow =>
        resultsRow.results.maxBy(_.changeDate).changeDate
      }
      Math.max(maxQueryChange,maxResultChange)
    })

    val responseOptionIo2: OptionT[IO, Response[IO]] = QepService().router.run(queryResultTableRequest(latestChange))
    val responseIO2 = responseOptionIo2.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }

    val query2StartTime = System.currentTimeMillis()

    val response2 = responseIO2.unsafeRunSync()
    assertResult(Ok)(response2.status)
    val resultJson2: Json = response2.as[Json].unsafeRunSync

    import io.circe.generic.auto.exportDecoder
    val queriesAndResults2: Result[PreviousQueriesAndResult] = resultJson2.as[PreviousQueriesAndResult]

    assertResult(queriesAndResults)(queriesAndResults2)
    assert(System.currentTimeMillis() - query2StartTime >= 1000L,"Time between request and reply was less than one second.")
  }

  "QepService" should "return an OK and an updated queryResultPolling result if a result changes" in {
    val networkId = qepQuery.networkId
    QepQueryDb.db.insertQepQuery(qepQuery)
    QepQueryDb.db.insertQepResultRow(incompleteResultRowFromBch)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromMgh)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromDfci)
    QepQueryDb.db.insertQepResultRow(qepResultRowFromPartners)

    def queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling?networkId=$networkId"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val responseOptionIo: OptionT[IO, Response[IO]] = QepService().router.run(queryResultTableRequest)
    val responseIO = responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }
    val response = responseIO.unsafeRunSync()
    assertResult(Ok)(response.status)
    val resultJson: Json = response.as[Json].unsafeRunSync

    import io.circe.generic.auto.exportDecoder
    val queriesAndResults: Result[PreviousQueriesAndResult] = resultJson.as[PreviousQueriesAndResult]

    val latestChange = queriesAndResults.fold({df:DecodingFailure => fail(df)},{ previousQueriesAndResults:PreviousQueriesAndResult =>
      //todo the structure still doesn't expose a version with a checksum, so figure out the latest timestamp and use that
      val maxQueryChange = previousQueriesAndResults.allQueries.maxBy(_.changeDate).changeDate
      val maxResultChange = previousQueriesAndResults.selectedQuery.fold(0L){resultsRow =>
        resultsRow.results.maxBy(_.changeDate).changeDate
      }
      Math.max(maxQueryChange,maxResultChange)
    })

    //trigger a change after a two-second delay. On my laptop the service is taking about half a second to just make the request
    val timer: Timer[IO] = IO.timer(ExecutionContexts.longActionExecutionContext)
    IO.sleep(FiniteDuration(2L,TimeUnit.SECONDS))(timer).flatMap{_ =>
      IO(QepQueryDb.db.insertQepResultRow(qepResultRowFromBch))
    }.unsafeRunAsyncAndForget()

    val responseOptionIo2: OptionT[IO, Response[IO]] = QepService().router.run(queryResultTableRequest)
    val responseIO2 = responseOptionIo2.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }

    val query2StartTime = System.currentTimeMillis()

    val response2 = responseIO2.unsafeRunSync()
    assertResult(Ok)(response2.status)
    val resultJson2: Json = response2.as[Json].unsafeRunSync

    import io.circe.generic.auto.exportDecoder
    val queriesAndResults2: Result[PreviousQueriesAndResult] = resultJson2.as[PreviousQueriesAndResult]

    assert(System.currentTimeMillis() - query2StartTime < 15000L,"Time between request and reply was greater than 15 seconds.")
    assert(System.currentTimeMillis() - query2StartTime > 1000L,"Time between request and reply was less than 1 second.")
    assert(queriesAndResults != queriesAndResults2)
  }

 /* Uncomment and fix once the observed flag is added to the QueryCell
  "QepService" should "return an OK and whether query result was previously viewed" in {
    val networkId = qep2Query.networkId
    QepQueryDb.db.insertQepQuery(qep2Query)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling?networkId=$networkId&timeoutSeconds=1"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(queryResultTableRequest)
    val result: Json = response.as[Json].unsafeRunSync

    val observedInitial = false
    val observedResult = result.findAllByKey("selectedQuery").head.findAllByKey("observed").head.asBoolean.get
    assertResult(Ok)(response.status)
    assertResult(observedInitial)(observedResult)

    val secondResponse: Response[IO] = extractResponse(queryResultTableRequest)
    val observed = true
    val secondResult: Json = secondResponse.as[Json].unsafeRunSync
    val secondObservedResult = secondResult.findAllByKey("selectedQuery").head.findAllByKey("observed").head.asBoolean.get
    assertResult(Ok)(secondResponse.status)
    assertResult(observed)(secondObservedResult)
  }*/

  "QepService" should "return an OK and list of previous queries for queryResultPolling" in {
    QepQueryDb.db.insertQepQuery(qep2Query)

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(queryResultTableRequest)
    val result: Json = response.as[Json].unsafeRunSync

    assertResult(Ok)(response.status)

    val queriesAndResult: PreviousQueriesAndResult = getQueryDataIO(researcherUser).flatMap(queriesAndResult => {

      IO(PreviousQueriesAndResult(queriesAndResult.queryRowCount, 0, allQueries = queriesAndResult.queryCells))
    }).unsafeRunSync()

    val expectedJsonStr: Json = queriesAndResult.asJson
    assertResult(expectedJsonStr)(result)
  }

  "QepService" should "return problem digest for query with an error" in {

    val testNode = Node(
      name = new NodeName("testNodeName"),
      key = new NodeKey("testNodeKey"),
      userDomainName = new UserDomainName("testUserDomainName"),
      adminEmail = "email@foo"
    )

    val networkId = 4L

    val researcher = Researcher.createRepeatableForV26(
      userName = researcherUser.username,
      userDomainName = researcherUser.domain,
      nodeId = testNode.id
    )

    val topic = Topic.createRepeatableForV26(
      researcherId = researcher.id,
      name = "myTopic",
      description = "topic description"
    )

    val queryWithId = Query.create(
      id = networkId,
      queryDefinitionXml = "<query/>",
      outputTypesXml = "<outputypes/>",
      queryName = "my query name",
      nodeOfOriginId = testNode.id,
      researcherId = researcher.id,
      topicId = topic.id,
      projectName = "my test project",
      NewTestKeyStore.certCollection
    )

    case class ErrorRunningQuery(t: Throwable) extends AbstractProblem(ProblemSources.Qep){
      override def summary: String = "Error running query"

      override def description: String = "An error occurred when running query"
    }

    val problem = ErrorRunningQuery(new Exception("This is a query with an error"))

    QepQueryDb.db.insertQueryIO(queryWithId,researcher).unsafeRunSync()

    val storeQueryStatus = UpdateQueryAtQepWithError(queryWithId.id,JsonProblemDigest(problem),DateStamp.now)

    QepQueryDb.db.updateQepQueryIO(storeQueryStatus).unsafeRunSync()

    val queryResultTableRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/queryResultPolling?networkId=$networkId&timeoutSeconds=1"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    extractResponse(queryResultTableRequest)
    val response: Response[IO] = extractResponse(queryResultTableRequest)
    val result: Json = response.as[Json].unsafeRunSync


    val expectedJson: Json = getQueryResultIO(researcherUser,networkId).unsafeRunSync()

    assertResult(expectedJson)(result)
    assertResult(expectedJson.findAllByKey("selectedQuery").head.findAllByKey("problemDigest").head)(result.findAllByKey("selectedQuery").head.findAllByKey("problemDigest").head)
  }

  "QepService" should " run a valid query" in {
    val body = """{
                 | "name": "Query_Rest",
                 | "projectId": "SHRINE",
                 | "topic": {
                 |  "id": 1,
                 |  "name": "Test Topic"
                 | },
                 | "queryTiming": "ANY",
                 | "resultOutputs": [
                 |	{
                 |		"name": "patient_count_xml"
                 |	}
                 | ],
                 | "conceptGroups": [
                 |   {
                 |     "concepts": [
                 |       {
                 |         "displayName": "Female",
                 |         "path": "\\\\SHRINE\\SHIRNE\\Demographics\\Sex\\Female\\",
                 |         "valueConstraint": {
                 |        	"valueType": "NUMBER",
                 |        	"unit": "mmol/L",
                 |        	"operator" : "LT",
                 |        	"value": "10"
                 |         }
                 |       }
                 |     ],
                 |     "isExcluded": false,
                 |     "options": {
                 |       "occurrences": 1
                 |     }
                 |   }
                 | ]
                 |}""".stripMargin

    val runQueryRequest: Request[IO] = Request(method = Method.POST,
      uri = Uri.unsafeFromString(s"/runQuery"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow2")))).withEntity(body)

    val response: Response[IO] = extractResponse(runQueryRequest)

    assertResult(Unauthorized)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("Username or password does not match credentials.")(entityText)
  }

  "QepService" should " validate value type of lab values when running a query" in {
    val body = """{
                 | "name": "Query_Rest",
                 | "projectId": "SHRINE",
                 | "topic": {
                 |  "id": 1,
                 |  "name": "Test Topic"
                 | },
                 | "queryTiming": "ANY",
                 | "resultOutputs": [
                 |	{
                 |		"name": "patient_count_xml"
                 |	}
                 | ],
                 | "conceptGroups": [
                 |   {
                 |     "concepts": [
                 |       {
                 |         "displayName": "Female",
                 |         "path": "\\\\SHRINE\\SHIRNE\\Demographics\\Sex\\Female\\",
                 |         "valueConstraint": {
                 |        	"valueType": "BADVALUETYPE",
                 |        	"unit": "mmol/L",
                 |        	"operator" : "LT",
                 |        	"value": "10"
                 |         }
                 |       }
                 |     ],
                 |     "isExcluded": false,
                 |     "options": {
                 |       "occurrences": 1
                 |     }
                 |   }
                 | ]
                 |}""".stripMargin

    val runQueryRequest: Request[IO] = Request(method = Method.POST,
      uri = Uri.unsafeFromString(s"/runQuery"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow")))).withEntity(body)

    val response: Response[IO] = extractResponse(runQueryRequest)

    assertResult(BadRequest)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("Error running query: Unrecognized value type BADVALUETYPE. Must be either FLAG or NUMBER.")(entityText)
  }

  before  {
    QepQueryDb.db.createTables()
  }


  after  {
    QepQueryDb.db.dropTables()
  }


  private def getQueryDataIO(user: User,
                             limitOption : Option[Int] = None,
                             skipOption : Option[Int] = None,
                             sortByOption: Option[String] = None): IO[QueriesAndCount] ={
    val queriesAndCountIO: IO[(Int, Seq[QueryWithResultStatus])] = QepQueryDb.db.selectPreviousQueriesWithResultStatusIO(
      userName = user.username,
      domain = user.domain,
      skip = skipOption,
      limit = limitOption,
      sortBy = sortByOption
    )

    queriesAndCountIO.flatMap(queriesAndCount => {
      val queryCells: Seq[QueryCell] = queriesAndCount._2.map(q =>  {
        QueryCell(q, q.queryFlag)
      })

      IO(QueriesAndCount(queriesAndCount._1, queryCells))
    })
  }


  private def getQueryResultIO(user: User,
                               networkQueryId: NetworkQueryId,
                               limitOption : Option[Int] = None,
                               skipOption : Option[Int] = None,
                               sortByOption: Option[String] = None): IO[Json] = {
    import io.circe.generic.auto.exportEncoder
    import io.circe.syntax.EncoderOps

    val queryResultsIO: IO[QueryAndResults] = QepQueryDb.db.selectResultsRowIO(networkQueryId, user)
    for{ queryResults <- queryResultsIO
         queriesAndCount <- getQueryDataIO(user, limitOption, skipOption, sortByOption)
    }
      yield{
        val resultsRow: ResultsRow = ResultsRow(queryResults)
        val queriesAndResult = PreviousQueriesAndResult(queriesAndCount.queryRowCount, skipOption.getOrElse(0), queriesAndCount.queryCells, Some(resultsRow))
        val resultJson = queriesAndResult.asJson//.toString()//.filter(_ != '\n')

        //val t = queriesAndResult.selectedQuery.get.isComplete
        //var finalResult: String = resultJson
        //if(queriesAndResult.selectedQuery.get.isComplete){
         // finalResult = finalResult.concat("\n")
        //}

       // finalResult
        resultJson
      }
  }

  val researcherUserName: String = "ben"
  val researcherFullName: String  = researcherUserName

  val researcherUser: User = User(
    fullName = researcherUserName,
    username = researcherFullName,
    domain = "domain",
    credential = new Credential("researcher's password",false),
    params = Map(),
    rolesByProject = Map()
  )

  val queryReceiveTime: Long = 1556572887669L

  val qepQuery = QepQuery(
    networkId = 1L,
    userName = researcherUser.username,
    userDomain = researcherUser.domain,
    queryName = "TestQuery",
    expression = Some("testExpression"),
    dateCreated = queryReceiveTime,
    deleted = false,
    queryXml = "testXML",
    changeDate = queryReceiveTime,
    status = QueryStatuses.SentToAdapters.statusName
  )

  val qep2Query = QepQuery(
    networkId = 2L,
    userName = researcherUser.username,
    userDomain = researcherUser.domain,
    queryName = "Another Query",
    expression = Some("testExpression"),
    dateCreated = queryReceiveTime - 200 ,
    deleted = false,
    queryXml = "testXML",
    changeDate = queryReceiveTime - 200,
    status = QueryStatuses.ReadyForAdapters.statusName
  )

  val qep3Query = QepQuery(
    networkId = 3L,
    userName = researcherUser.username,
    userDomain = researcherUser.domain,
    queryName = "AnewQuery",
    expression = Some("testExpression"),
    dateCreated = queryReceiveTime + 100,
    deleted = false,
    queryXml = "testXML",
    changeDate = queryReceiveTime,
    status = QueryStatuses.ReceivedAtHub.statusName
  )

  val qepResultRowFromMgh = QueryResultRow(
    resultId = 10L,
    networkQueryId = 1L,
    instanceId = 100L,
    adapterNode = "MGH",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 30L,
    startDate = Some(queryReceiveTime + 10),
    endDate = Some(queryReceiveTime + 20),
    status = QueryResult.StatusType.Finished.name,
    statusMessage = None,
    changeDate = queryReceiveTime + 20
  )

  val qepResultRowFromPartners = QueryResultRow(
    resultId = 10L,
    networkQueryId = 1L,
    instanceId = 100L,
    adapterNode = "Partners",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 300L,
    startDate = Some(queryReceiveTime + 10),
    endDate = Some(queryReceiveTime + 20),
    status = QueryResult.StatusType.Finished.name,
    statusMessage = None,
    changeDate = queryReceiveTime + 30
  )

  val incompleteResultRowFromBch = QueryResultRow(
    resultId = 10L,
    networkQueryId = 1L,
    instanceId = 100L,
    adapterNode = "BCH",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = -1L,
    startDate = Some(queryReceiveTime + 10),
    endDate = None,
    status = ResultStatuses.ReceivedByAdapter.statusName,
    statusMessage = None,
    changeDate = queryReceiveTime + 20
  )

  val qepResultRowFromBch = QueryResultRow(
    resultId = 10L,
    networkQueryId = 1L,
    instanceId = 100L,
    adapterNode = "BCH",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 3000L,
    startDate = Some(queryReceiveTime + 10),
    endDate = Some(queryReceiveTime + 20),
    status = QueryResult.StatusType.Finished.name,
    statusMessage = None,
    changeDate = queryReceiveTime + 40
  )

  val qepResultRowFromDfci = QueryResultRow(
    resultId = 10L,
    networkQueryId = 1L,
    instanceId = 100L,
    adapterNode = "DFCI",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 30000L,
    startDate = Some(queryReceiveTime + 10),
    endDate = Some(queryReceiveTime + 20),
    status = QueryResult.StatusType.Finished.name,
    statusMessage = None,
    changeDate = queryReceiveTime + 50
  )

  val qepResultRowFromHarvard = QueryResultRow(
    resultId = 10L,
    networkQueryId = 2L,
    instanceId = 100L,
    adapterNode = "HARVARD",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 30000L,
    startDate = Some(queryReceiveTime - 12),
    endDate = Some(queryReceiveTime - 10),
    status = QueryResult.StatusType.Finished.name,
    statusMessage = None,
    changeDate = queryReceiveTime - 5
  )

  val qepResultRowFromNU = QueryResultRow(
    resultId = 17L,
    networkQueryId = 2L,
    instanceId = 120L,
    adapterNode = "NU",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 80000L,
    startDate = Some(queryReceiveTime - 5),
    endDate = Some(queryReceiveTime - 9),
    status = QueryResult.StatusType.Finished.name,
    statusMessage = None,
    changeDate = queryReceiveTime - 7
  )

  val qepQueryWithResultErrors = QepQuery(
    networkId = 50L,
    userName = researcherUser.username,
    userDomain = researcherUser.domain,
    queryName = "TestQuery",
    expression = Some("testExpression"),
    dateCreated = queryReceiveTime,
    deleted = false,
    queryXml = "testXML",
    changeDate = queryReceiveTime,
    status = QueryStatuses.SentToAdapters.statusName
  )

  val qepQueryWithErrors = QepQuery(
    networkId = 50L,
    userName = researcherUser.username,
    userDomain = researcherUser.domain,
    queryName = "TestQuery",
    expression = Some("testExpression"),
    dateCreated = queryReceiveTime,
    deleted = false,
    queryXml = "testXML",
    changeDate = queryReceiveTime,
    status = QueryStatuses.HubError.toString
  )

  val qepResultRowFromNUError = QueryResultRow(
    resultId = 501L,
    networkQueryId = 50L,
    instanceId = 120L,
    adapterNode = "NU",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 80000L,
    startDate = Some(queryReceiveTime - 5),
    endDate = Some(queryReceiveTime - 9),
    status = QueryResult.StatusType.Error.name,
    statusMessage = None,
    changeDate = queryReceiveTime - 7
  )

  val qepResultRowFromHarvardError = QueryResultRow(
    resultId = 502L,
    networkQueryId = 50L,
    instanceId = 100L,
    adapterNode = "HARVARD",
    resultType = Some(ResultOutputType.PATIENT_COUNT_XML),
    size = 30000L,
    startDate = Some(queryReceiveTime - 12),
    endDate = Some(queryReceiveTime - 10),
    status = ResultStatuses.ErrorInShrine.statusName,
    statusMessage = None,
    changeDate = queryReceiveTime - 5
  )

}
