package net.shrine.receiver

import java.util.concurrent.atomic.AtomicReference

import cats.effect.{CancelToken, IO}
import net.shrine.adapter.RunQueryAdapter
import net.shrine.hub.data.client.HubClient
import net.shrine.hub.mom.ShrineMomClient
import net.shrine.log.Log
import net.shrine.protocol.version.v1.{Result, RunQueryForResult, UpdateQueryAtQep, UpdateQueryAtAdapter, UpdateQueryAtAdapterWithFlagging, UpdateQueryAtAdapterWithName}
import net.shrine.protocol.version.{Envelope, JsonText}
import net.shrine.qep.querydb.QepQueryDb
import net.shrine.source.ConfigSource

/**
  * Receives messages no matter what, and passes them to the Dispatcher to process
  */

object Receiver {

  private val atomicCancelToken = new AtomicReference[CancelToken[IO]](null)

  def restart(): IO[Unit] = IO { synchronized {
    stop()
    atomicCancelToken.set(ShrineMomClient.receiveUntilStop(
      HubClient.getLocalNodeIO.map(_.momQueueName),
      dispatch
    ))
  }}

  def stop(): IO[Unit] = IO { synchronized {
    Option(atomicCancelToken.getAndSet(null)).foreach { ct: CancelToken[IO] =>
      Log.debug(s"Stopping Dispatcher")
      ct.unsafeRunSync()
      Log.debug(s"Stopped Dispatcher")
    }
  }}

  val runQueryAdapter:Option[RunQueryAdapter] = {
    if (ConfigSource.config.getBoolean("shrine.adapter.create")) Some(RunQueryAdapter(ConfigSource.config))
    else None
  }

  //true to complete, false to not complete
  private def dispatch(envelope: Envelope,remainingAttempts:Int):IO[Boolean] = {

    Log.debug(s"envelope holds a ${envelope.contentsType}")

    envelope match {
        //Adapter commands
      case Envelope(contentsType,_,contents,_) if contentsType == RunQueryForResult.envelopeType =>
        val runQueryForResult:RunQueryForResult = RunQueryForResult.tryRead(new JsonText(contents)).get
        Log.info(s"Received command to run query ${runQueryForResult.query.id}")
        //if there's an adapter run the query. Otherwise complain.
        runQueryAdapter.fold{
          //todo send back an error result immediately
          Log.warn(s"Received command to run query ${runQueryForResult.query.id} but this node has no shrine adapter")
          IO(true)
        }{ rqa =>
          rqa.startRunQueryForExpectedResult(runQueryForResult).
            flatMap(_ => IO(true))
        }
      case Envelope(contentsType, _, contents, _) if contentsType == UpdateQueryAtAdapter.envelopeType =>
        val updateQuery:UpdateQueryAtAdapter = UpdateQueryAtAdapter.tryRead(new JsonText(contents)).get
        Log.info(s"Received command to update query $updateQuery at the adapter")
        val updateSomething: IO[Unit] = updateQuery match {
          case updateQueryName:UpdateQueryAtAdapterWithName => IO(renameQuery(updateQueryName))
          case updateQueryFlagging:UpdateQueryAtAdapterWithFlagging => IO(setQueryFlag(updateQueryFlagging))
          case _ => throw new IllegalStateException(s"Unexpected UpdateQueryAtAdapter command type ${updateQuery.getClass.getSimpleName} $updateQuery")
        }
        updateSomething.flatMap(_ => IO(true))
        //QEP commands
      case Envelope(contentsType, _, contents, _) if contentsType == UpdateQueryAtQep.envelopeType =>
        IO(UpdateQueryAtQep.tryRead(new JsonText(contents)).get).
          flatMap(updateQueryAtQep => QepQueryDb.db.updateQepQueryIO(updateQueryAtQep)).
          flatMap(_ => IO(true))
      case Envelope(contentsType, _, contents,_) if contentsType == Result.envelopeType =>
        IO(Result.tryRead(new JsonText(contents)).get).
          flatMap(result => QepQueryDb.db.insertQueryResultIO(result)).
          flatMap(_ => IO(true))
    }
  }

  private def renameQuery(updateQueryName:UpdateQueryAtAdapterWithName): Unit = runQueryAdapter.foreach(_.dao.renameQuery(
    updateQueryName.queryId.underlying,
    updateQueryName.queryName)
  )

  private def setQueryFlag(updateQueryFlagging:UpdateQueryAtAdapterWithFlagging): Unit = {
    if(updateQueryFlagging.flagged) runQueryAdapter.foreach(_.dao.flagQuery(
        updateQueryFlagging.queryId.underlying,
        updateQueryFlagging.flaggedMessage)
      )
    else runQueryAdapter.foreach(_.dao.unFlagQuery(
      updateQueryFlagging.queryId.underlying)
    )
  }

}
