package net.shrine.api.static.data

import cats.effect.IO
import com.typesafe.config.{ConfigException, ConfigRenderOptions}
import net.shrine.log.Log
import net.shrine.source.ConfigSource
import org.http4s.{HttpRoutes, Request, Response}
import org.http4s.dsl.impl.{->, /}
import org.http4s.dsl.io.{GET, InternalServerError, NotFound, Ok}
import org.http4s.dsl.io.{Root, http4sNotFoundSyntax, http4sOkSyntax, http4sInternalServerErrorSyntax}

import scala.util.control.NonFatal
import scala.util.{Failure, Success, Try}

case class StaticDataService() {

  val staticDataInfo: String  =
    """
      |The SHRINE static data service.
      |
      |This API retrieves static data from SHRINE.
      |
    """.stripMargin

  val service: HttpRoutes[IO] = HttpRoutes.of[IO] {
    case GET -> Root / "ping" => Ok("pong")

    case req @ GET -> Root / "config" / key => handleKey(key) match {
      case Success(e) => Ok(e)
      case Failure(x) =>  handleGeneralFailure(req, x)
    }

    case GET -> Root / "webClientConfig" => extractWebClientConfig

    case x => NotFound(s"The static data service does not respond to $x")
  }

  def extractWebClientConfig :  IO[Response[IO]] = {
    import io.circe.config.syntax.CirceConfigOps
    import io.circe.generic.auto.exportDecoder
    import io.circe.generic.auto.exportEncoder
    import io.circe.syntax.EncoderOps
    import org.http4s.circe.jsonEncoder

    val clientConfig = ConfigSource.config.getConfig("shrine.webclient")

    clientConfig.as[WebClientConfig] match {
      case Right(conf) => Ok(conf.asJson)

      case Left(e) => {
        val msg = s"An error occurred while retrieving the webclient config $e"
        Log.error(msg)
        InternalServerError(msg)
      }
    }
  }

  def handleKey(key: String): Try[String] = {
    val staticDataConfig = ConfigSource.config.getConfig("shrine.static")
    Try(staticDataConfig.getValue(key).render(ConfigRenderOptions.concise()))
  }

  private def handleGeneralFailure(req: Request[IO], x: Throwable): IO[Response[IO]] = x match {
    case cfe : ConfigException => NotFound(cfe.getMessage)
    case NonFatal(e) => InternalServerError(e.getMessage)
    case _ => throw x
  }

}

case class WebClientConfig(domain: String,
                           name: String,
                           shrineUrl: Option[String] = None,
                           siteAdminEmail: Option[String] = None,
                           usernameLabel: Option[String]= None,
                           passwordLabel: Option[String]= None,
                           defaultNumberOfOntologyChildren: Option[Int] = None,
                           queryFlaggingInstructions: Option[String]= None,
                           flaggingIconInstructions: Option[String]= None,
                           flaggingPlaceholderText: Option[String] = None)

