package net.shrine.api.static.data

  import cats.data.OptionT
  import cats.effect.{ConcurrentEffect, ContextShift, IO}
  import cats.implicits.{catsSyntaxEither => _}
  import org.http4s.dsl.io.Ok
  import org.http4s.headers.Accept
  import org.http4s.{Headers, MediaType, Method, Request, Response, Uri}
  import org.junit.runner.RunWith
  import org.scalatest.FlatSpec
  import org.scalatestplus.junit.JUnitRunner

  import scala.concurrent.ExecutionContext.Implicits.global

@RunWith(classOf[JUnitRunner])
class StaticDataServiceTest extends FlatSpec  {

  def extractResponse(request: Request[IO]): Response[IO] = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val responseOptionIo: OptionT[IO, Response[IO]] = StaticDataService().service.run(request)

    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  "StaticDataService" should "reply to ping with pong" in {
    val pingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString("/ping"))

    val response: Response[IO]= extractResponse(pingRequest)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("pong")(entityText)
  }

  "StaticDataService" should "return timeout config" in {

    val request: Request[IO] = Request(method = Method.GET,
      uri = Uri.fromString(s"/config/timeout").right.get,
      headers = Headers.of(Accept(MediaType.application.json)))

    val response: Response[IO] = extractResponse(request)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("\"45 seconds\"")(entityText)
  }

  "StaticDataService" should "return webclient config when there is no shrineUrl specified" in {
    import io.circe.Json
    import io.circe.parser.parse

    val request: Request[IO] = Request(method = Method.GET,
      uri = Uri.fromString(s"/webClientConfig").right.get,
      headers = Headers.of(Accept(MediaType.application.json)))

    val expectedResponse = """{
                              |    "domain": "foo",
                              |    "name": "SHRINE",
                              |    "shrineUrl": null,
                              |    "siteAdminEmail" : "baz@foo",
                              |    "usernameLabel" : "User Name",
                              |    "passwordLabel" : "User Password",
                              |    "queryFlaggingInstructions" : "Flag this query at all adapters",
                              |    "flaggingIconInstructions" : "Flag a query",
                              |    "flaggingPlaceholderText" : null,
                              |     "defaultNumberOfOntologyChildren" : 10000
                              |}""".stripMargin
    val expectedJson: Json = parse(expectedResponse).right.get

    val expectedClientConfig  = WebClientConfig("foo", "SHRINE")
    val response: Response[IO] = extractResponse(request)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    val resultJson: Json = parse(entityText).right.get

    assertResult(expectedJson)(resultJson)

  }
}
