package net.shrine.api.steward

import cats.effect.IO
import net.shrine.api.steward.db.{ApprovedTopicCanNotBeChanged, DetectedAttemptByWrongUserToChangeTopic, QueryParameters, SortOrder, StewardDatabase, TopicDoesNotExist}
import net.shrine.http4s.auth.BasicOrBearerAuthMiddleware
import net.shrine.i2b2.protocol.pm.User
import org.http4s.{AuthedRequest, AuthedRoutes, EntityDecoder, HttpRoutes, Response, Status}
import org.http4s.dsl.impl.{+&, ->, /, :?, Auth, IntVar, OptionalQueryParamDecoderMatcher}
import org.http4s.dsl.io.{Accepted, Forbidden, GET, NotFound, Ok, POST, UnprocessableEntity}
import org.http4s.dsl.io.{Root, http4sAcceptedSyntax, http4sForbiddenSyntax, http4sNotFoundSyntax, http4sOkSyntax, http4sUnprocessableEntitySyntax}
import io.circe.{Encoder, Json}
import io.circe.generic.auto.exportEncoder
import io.circe.syntax.EncoderOps
import net.shrine.api.steward.pmauth.Authorizer
import net.shrine.circe.CirceXml

import scala.util.{Failure, Success}
import scala.xml.NodeSeq

//noinspection ConvertExpressionToSAM
case class StewardService() extends Auth {

  val stewardServiceInfo: String  =
    """
      |The SHRINE Steward service.
      |
      |This API retrieves data from the SHRINE steward.
      |
    """.stripMargin

  object OptionalStateParam extends OptionalQueryParamDecoderMatcher[String]("state")
  object OptionalLimitParam extends OptionalQueryParamDecoderMatcher[Int]("limit")
  object OptionalSkipParam extends OptionalQueryParamDecoderMatcher[Int]("skip")
  object OptionalSortByParam extends OptionalQueryParamDecoderMatcher[String]("sortBy")
  object OptionalSortDirectionParam extends OptionalQueryParamDecoderMatcher[String]("sortDirection")
  object OptionalMinDateParam extends OptionalQueryParamDecoderMatcher[Date]("minDate")
  object OptionalMaxDateParam extends OptionalQueryParamDecoderMatcher[Date]("maxDate")
  object OptionalAsJsonParam extends OptionalQueryParamDecoderMatcher[Boolean]("asJson")

  val notAuthedService: HttpRoutes[IO] = HttpRoutes.of[IO] {
    case _@GET -> Root => Ok(stewardServiceInfo)
    case _@GET -> Root / "ping" => Ok("pong")
    case _@GET -> Root / "about" / "createTopicsMode" => getCreateTopicsMode
  }

  private val authenticatedUserService = BasicOrBearerAuthMiddleware{
    AuthedRoutes.of[User, IO] {
      case _@GET -> Root / "user" / "ping" as user => Ok(s"pong from ${user.username}")

      case _@GET -> Root / "user" / "whoami" as user => whoami(user)
    }
  }

  private val authorizedStewardService = BasicOrBearerAuthMiddleware {
    AuthedRoutes.of[User, IO] {
      case _@GET -> Root / "ping" as user => Ok(s"pong from steward user $user")

      case _@POST -> Root / "steward" / "approveTopic" / "topic" / IntVar(topicId)
        :? OptionalStateParam(_)
        +& OptionalSkipParam(_)
        +& OptionalLimitParam(_)
        +& OptionalSortByParam(_)
        +& OptionalSortDirectionParam(_)
        +& OptionalMinDateParam(_)
        +& OptionalMaxDateParam(_) as user => changeStateForTopic(TopicState.approved, user, topicId)

      case _@POST -> Root / "steward" / "rejectTopic" / "topic" / IntVar(topicId)
        :? OptionalStateParam(_)
        +& OptionalSkipParam(_)
        +& OptionalLimitParam(_)
        +& OptionalSortByParam(_)
        +& OptionalSortDirectionParam(_)
        +& OptionalMinDateParam(_)
        +& OptionalMaxDateParam(_) as user => changeStateForTopic(TopicState.rejected, user, topicId)

      case _@GET -> Root / "steward" / "topics"
        :? OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getTopicsForSteward(
        user,
        topicUser = None,
        stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case _@GET -> Root / "steward" / "topics" / "user" / topicUser
        :? OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getTopicsForSteward(
        user,
        Some(topicUser),
        stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case _@GET -> Root / "steward" / "statistics" / "queriesPerUser"
        :? OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getQueriesPerUser(
        user,
        stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case _@GET -> Root / "steward" / "statistics" / "topicsPerState"
        :? OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getTopicsPerState(
        user,
        stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case _@GET -> Root / "steward" / "queryHistory"
        :?  OptionalAsJsonParam(asJsonOption)
        +& OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getStewardQueryHistoryForUserByTopic(
        user,
        userIdOption = None,
        topicIdOption = None,
        asJsonOption,
        stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case _@GET -> Root / "steward" / "queryHistory" / "topic" / IntVar(topicId)
        :?  OptionalAsJsonParam(asJsonOption)
        +& OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getStewardQueryHistoryForUserByTopic(
        user,
        userIdOption = None,
        topicIdOption = Some(topicId),
        asJsonOption, stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case _@GET -> Root / "steward" / "queryHistory" / "user" / userId
        :?  OptionalAsJsonParam(asJsonOption)
        +& OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getStewardQueryHistoryForUserByTopic(
        user,
        userIdOption = Some(userId),
        topicIdOption = None,
        asJsonOption, stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case _@GET -> Root / "steward" / "queryHistory" / "user" / userId / "topic" / IntVar(topicId)
        :?  OptionalAsJsonParam(asJsonOption)
        +& OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getStewardQueryHistoryForUserByTopic(
        user,
        userIdOption = Some(userId),
        topicIdOption = Some(topicId),
        asJsonOption,
        stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)
    }
  }

  private val authorizedResearcherService = BasicOrBearerAuthMiddleware {
    AuthedRoutes.of[User, IO] {
      case _@GET -> Root / "researcher" / "topics"
        :? OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getUserTopics(
        user,
        stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case _@GET -> Root / "researcher" / "queryHistory"
        :? OptionalAsJsonParam(asJsonOption)
        +& OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getQueryHistoryForUserByTopic(
        Some(user.username),
        topicIdOption = None,
        asJsonOption,
        stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case _@GET -> Root / "researcher" / "queryHistory" / "topic" / IntVar(topicId)
        :? OptionalAsJsonParam(asJsonOption)
        +& OptionalStateParam(stateOption)
        +& OptionalSkipParam(skipOption)
        +& OptionalLimitParam(limitOption)
        +& OptionalSortByParam(sortByOption)
        +& OptionalSortDirectionParam(sortDirectionOption)
        +& OptionalMinDateParam(minDateOption)
        +& OptionalMaxDateParam(maxDateOption) as user => getQueryHistoryForUserByTopic(
        Some(user.username),
        topicIdOption = Some(topicId),
        asJsonOption,
        stateOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case req@POST -> Root / "researcher" / "requestTopicAccess" as _ => requestTopicAccess(req)

      case req@POST -> Root / "researcher" / "editTopicRequest" / IntVar(topicId)  as _ => editTopicRequest(req, topicId)

    }
  }

  val notFoundService : HttpRoutes[IO] = HttpRoutes.of[IO] {
    case x => NotFound(s"The QEP service does not respond to $x")
  }


  val service: HttpRoutes[IO] = {
    import cats.implicits._ // SHRINE-3272
    //IDE does not like <+> , but it compiles just fine
    notAuthedService <+> AuthenticationFailureMiddleWare(authenticatedUserService) <+> authorizedResearcherService <+> authorizedStewardService  <+> notFoundService
  }

  def getCreateTopicsMode : IO[Response[IO]] = {
    Ok(CreateTopicsMode.createTopicsInState.name)
  }

  def whoami(user: User): IO[Response[IO]] =  {
    StewardDatabase.db.upsertUserIO(user).flatMap { _ =>
      val outboundUser: OutboundUser = OutboundUser.createFromUser(user)
      val str = outboundUser.asJson.toString()
      Ok(str)
    }
  }

  def changeStateForTopic(state:TopicState,user:User, topicId: TopicId): IO[Response[IO]] = {
    validateStewardAuthorization(user) match {
      case Right(_) => StewardDatabase.db.changeTopicStateIO(topicId, state, user.username).flatMap(topicRecord => topicRecord.fold(UnprocessableEntity(s"No topic found for $topicId")) ( _ => Ok()))
      case Left(response) => response
    }
  }

  def getTopicsForSteward(user:User, topicUser: Option[UserName],
                          stateStringOption:Option[String],
                          skipOption: Option[Int],
                          limitOption: Option[Int],
                          sortByOption: Option[String],
                          sortDirectionOption: Option[String],
                          minDateOption: Option[Date],
                          maxDateOption: Option[Date]): IO[Response[IO]] ={

    validateStewardAuthorization(user) match {
      case Right(_) =>  getQueryParameters(topicUser, stateStringOption,  skipOption, limitOption, sortByOption, sortDirectionOption, minDateOption, maxDateOption){queryParameters: QueryParameters =>
        val stewardsTopicsIO: IO[StewardsTopics] = StewardDatabase.db.selectTopicsForStewardIO(queryParameters)

        stewardsTopicsIO.flatMap(stewardTopics => Ok(stewardTopics.asJson.toString()))
      }
      case Left(response) => response
    }
  }

  def validateStewardAuthorization(user:User):  Either[IO[Response[IO]], User]  = {
    StewardDatabase.db.upsertUser(user)

    if (!Authorizer.authorizeSteward(user)) {
      val stewardForbiddenResponse: IO[Response[IO]] =  Forbidden(s"The supplied user ${user.username} is not a steward and therefore not authorized to access this resource")

      Left(stewardForbiddenResponse)
    }else{
      Right(user)
    }
  }

  def getQueryParameters(userName: Option[UserName],
                         stateStringOption:Option[String],
                         skipOption: Option[Int],
                         limitOption: Option[Int],
                         sortByOption: Option[String],
                         sortDirectionOption: Option[String],
                         minDateOption: Option[Date],
                         maxDateOption: Option[Date])(processRoute:QueryParameters => IO[Response[IO]]): IO[Response[IO]] ={

    val stateTry = TopicState.stateForStringOption(stateStringOption)
    stateTry match {
      case Success(stateOption) =>
        val queryParameters = QueryParameters(userName,
          stateOption,
          skipOption,
          limitOption,
          sortByOption,
          SortOrder.sortOrderForStringOption(sortDirectionOption),
          minDateOption,
          maxDateOption
        )
        processRoute(queryParameters)
      case Failure(_) => UnprocessableEntity(s"Topic state ${stateStringOption.getOrElse(s"$stateStringOption (stateStringOption should never be None at this point)")} unknown." +
        s" Please specify one of ${TopicState.namesToStates.keySet}")
    }
  }

  def getQueriesPerUser(user:User,
                        stateStringOption:Option[String],
                        skipOption: Option[Int],
                        limitOption: Option[Int],
                        sortByOption: Option[String],
                        sortDirectionOption: Option[String],
                        minDateOption: Option[Date],
                        maxDateOption: Option[Date]): IO[Response[IO]] = {
    validateStewardAuthorization(user) match {
      case Right(_) => getQueryParameters(None, stateStringOption, skipOption, limitOption, sortByOption, sortDirectionOption, minDateOption, maxDateOption) { queryParameters: QueryParameters =>
        implicit val encodeQueriesPerUser: Encoder[QueriesPerUser] = new Encoder[QueriesPerUser] {
          final def apply(a: QueriesPerUser): Json = Json.obj(
            ("total", Json.fromInt(a.total)),
            ("queriesPerUser", Json.fromValues(a.queriesPerUser.map(t => Json.obj(
              ("_1", t._1.asJson),
              ("_2", Json.fromInt(t._2))
            ))))
          )
        }
        val queriesPerUserIO: IO[QueriesPerUser] = StewardDatabase.db.selectShrineQueryCountsPerUserIO(queryParameters)
        queriesPerUserIO.flatMap(queriesPerUser => Ok(queriesPerUser.asJson.toString()))
      }
      case Left(response) => response
    }
  }

  def getTopicsPerState(user:User,
                        stateStringOption:Option[String],
                        skipOption: Option[Int],
                        limitOption: Option[Int],
                        sortByOption: Option[String],
                        sortDirectionOption: Option[String],
                        minDateOption: Option[Date],
                        maxDateOption: Option[Date]): IO[Response[IO]] = {
    validateStewardAuthorization(user) match {
      case Right(_) => getQueryParameters(None, stateStringOption, skipOption, limitOption, sortByOption, sortDirectionOption, minDateOption, maxDateOption) { queryParameters: QueryParameters =>
        implicit val encodeTopicsPerState: Encoder[TopicsPerState] = new Encoder[TopicsPerState] {
          final def apply(a: TopicsPerState): Json = Json.obj(
            ("total", Json.fromInt(a.total)),
            ("topicsPerState", Json.fromValues(a.topicsPerState.map(t => Json.obj((t._1, Json.fromInt(t._2))))))
          )
        }
        StewardDatabase.db.selectTopicCountsPerStateIO(queryParameters).flatMap(topicsPerState => Ok(topicsPerState.asJson.toString()) )
      }
      case Left(response) => response
    }
  }

  def getStewardQueryHistoryForUserByTopic(user:User,
                                    userIdOption: Option[UserName],
                                    topicIdOption: Option[TopicId],
                                    asJsonOption: Option[Boolean],
                                    stateStringOption:Option[String],
                                    skipOption: Option[Int],
                                    limitOption: Option[Int],
                                    sortByOption: Option[String],
                                    sortDirectionOption: Option[String],
                                    minDateOption: Option[Date],
                                    maxDateOption: Option[Date]): IO[Response[IO]] = {
    validateStewardAuthorization(user) match {
      case Right(_) => getQueryHistoryForUserByTopic(
        userIdOption = userIdOption,
        topicIdOption = topicIdOption,
        asJsonOption,
        stateStringOption,
        skipOption,
        limitOption,
        sortByOption,
        sortDirectionOption,
        minDateOption,
        maxDateOption)

      case Left(response) => response
    }
  }

  def getQueryHistoryForUserByTopic(userIdOption: Option[UserName],
                                    topicIdOption: Option[TopicId],
                                    asJsonOption: Option[Boolean],
                                    stateStringOption:Option[String],
                                    skipOption: Option[Int],
                                    limitOption: Option[Int],
                                    sortByOption: Option[String],
                                    sortDirectionOption: Option[String],
                                    minDateOption: Option[Date],
                                    maxDateOption: Option[Date]): IO[Response[IO]] = {
      getQueryParameters(userIdOption, stateStringOption, skipOption, limitOption, sortByOption, sortDirectionOption, minDateOption, maxDateOption) { queryParameters: QueryParameters =>
      StewardDatabase.db.selectQueryHistoryIO(queryParameters, topicIdOption).flatMap { queryHistory =>
       if (asJsonOption.getOrElse(false)) {

         implicit val encodeOutboundShrineQueryWithJson: Encoder[Seq[OutboundShrineQueryWithJson]] = new Encoder[Seq[OutboundShrineQueryWithJson]] {
           final def apply(seq: Seq[OutboundShrineQueryWithJson]): Json = {
             import io.circe.Json
             import io.circe.syntax.EncoderOps

             val res: Seq[Json] = for (xml: OutboundShrineQueryWithJson <- seq) yield {
               val detailsJson = CirceXml.toJson(xml.queryContents) //todo pull out this json val and get the format right from somewhere else

               Json.obj(
                 ("stewardId", Json.fromLong(xml.stewardId)),
                 ("externalId", Json.fromLong(xml.externalId)),
                 ("name", Json.fromString(xml.name)),
                 ("user", xml.user.asJson),
                 ("topic", xml.topic.asJson),
                 ("queryContents", detailsJson),
                 ("stewardResponse", Json.fromString(xml.stewardResponse)),
                 ("date", Json.fromString(xml.date.toString))
               )
             }
             Json.fromValues(res)
           }
         } //end of encodeOutboundShrineQueryWithJson

         val queryRecordsWithJson: Seq[OutboundShrineQueryWithJson] = queryHistory.queryRecords.map(queryRecord => {
           OutboundShrineQueryWithJson(queryRecord.stewardId, queryRecord.externalId, queryRecord.name,
             queryRecord.user, queryRecord.topic, scala.xml.XML.loadString(queryRecord.queryContents), queryRecord.stewardResponse, queryRecord.date)
         })

         val queryHistoryWithJson = QueryHistoryToJson(queryHistory.totalCount, queryHistory.skipped, queryRecordsWithJson)
         Ok(queryHistoryWithJson.asJson.toString())



       }
       else {
         Ok(queryHistory.asJson.toString())
       }
      }
    }
  }

  def validateResearcherAuthorization(user:User):  Either[IO[Response[IO]], User]  = {
    StewardDatabase.db.upsertUser(user)

    if (!Authorizer.authorizeResearcher(user)) {
      val researcherForbiddenResponse: IO[Response[IO]] =  Forbidden(s"The supplied user ${user.username} is not a researcher and therefore not authorized to access this resource")
      Left(researcherForbiddenResponse)
    }else{
      Right(user)
    }
  }

  def getUserTopics(user:User,
                    stateStringOption:Option[String],
                    skipOption: Option[Int],
                    limitOption: Option[Int],
                    sortByOption: Option[String],
                    sortDirectionOption: Option[String],
                    minDateOption: Option[Date],
                    maxDateOption: Option[Date]): IO[Response[IO]] =  {

    validateResearcherAuthorization(user) match {
      case Right(_) => getQueryParameters(Some(user.username), stateStringOption, skipOption, limitOption, sortByOption, sortDirectionOption, minDateOption, maxDateOption) { queryParameters: QueryParameters =>
        StewardDatabase.db.selectTopicsForResearcherIO(queryParameters).flatMap(researchersTopics => Ok(researchersTopics.asJson.toString()))
      }
      case Left(response) => response
    }
  }

  def requestTopicAccess(authedRequest: AuthedRequest[IO, User]): IO[Response[IO]] =  {
    import io.circe.generic.auto.exportDecoder
    import org.http4s.circe.jsonOf
    val user = authedRequest.authInfo

    validateResearcherAuthorization(user) match {
      case Right(_) =>
        implicit val inboundTopicRequestDecoder: EntityDecoder[IO, InboundTopicRequest] = jsonOf[IO, InboundTopicRequest]
        val topicRequestIO: IO[InboundTopicRequest] = authedRequest.req.as[InboundTopicRequest]
        topicRequestIO.flatMap{ topicRequest =>
          StewardDatabase.db.createRequestForTopicAccessIO(user, topicRequest).flatMap(_ => Accepted())
        }
      case Left(response) => response
    }
  }

  def editTopicRequest(authedRequest: AuthedRequest[IO, User], topicId: TopicId): IO[Response[IO]] =  {
    import io.circe.generic.auto.exportDecoder
    import org.http4s.circe.jsonOf
    val user = authedRequest.authInfo

    validateResearcherAuthorization(user) match {
      case Right(_) =>
        implicit val inboundTopicRequestDecoder: EntityDecoder[IO, InboundTopicRequest] = jsonOf[IO, InboundTopicRequest]

        val topicRequest: InboundTopicRequest = authedRequest.req.as[InboundTopicRequest].unsafeRunSync()

        val updatedTopicIO: IO[OutboundTopic] = StewardDatabase.db.updateRequestForTopicAccessIO(user, topicId, topicRequest)

        updatedTopicIO.flatMap(updatedTopic => Accepted(updatedTopic.asJson.toString())).handleErrorWith {
          case x: TopicDoesNotExist => NotFound(x.getMessage)
          case x: ApprovedTopicCanNotBeChanged => Forbidden(x.getMessage)
          case x: DetectedAttemptByWrongUserToChangeTopic => Forbidden(x.getMessage)
          case x => throw x
        }
      case Left(response) => response
    }
  }
}

case class QueryHistoryToJson(totalCount:Int,skipped:Int,queryRecords:Seq[OutboundShrineQueryWithJson])  {
  def convertToXml = QueryHistory(totalCount, skipped, queryRecords.map(_.convertToXml))
}

case class OutboundShrineQueryToJson(
                                        stewardId:StewardQueryId,
                                        externalId:ExternalQueryId,
                                        name:String,
                                        user:OutboundUser,
                                        topic:Option[OutboundTopic],
                                        queryContents: NodeSeq,
                                        stewardResponse:TopicStateName,
                                        date:Date)
{
  def convertToXml:OutboundShrineQuery = {
    OutboundShrineQuery(stewardId, externalId, name, user, topic, queryContents.toString, stewardResponse, date)
  }
}

object AuthenticationFailureMiddleWare {

  private[steward] def apply(service: HttpRoutes[IO]) = {
    //convert 401 unauthorized to 200
    //This is to work around issue with steward app displaying a browser dialog
    // prompting again for username and password
    service.map {
      case Status.Unauthorized(resp) => resp.withStatus(Status.Ok).withEntity("AuthenticationFailed")
      case resp => resp
    }
  }
}


