package net.shrine.api.steward.email

import java.util.Date

import cats.effect.{IO, Timer}
import com.typesafe.config.Config
import courier.{Envelope, Mailer, Text}
import javax.mail.internet.InternetAddress
import net.shrine.api.steward.ResearcherToAudit
import net.shrine.api.steward.db.StewardDatabase
import net.shrine.config.ConfigExtensions
import net.shrine.email.ConfiguredMailer
import net.shrine.http4s.catsio.{ExecutionContexts, IOScheduler}
import net.shrine.log.Log
import net.shrine.problem.{AbstractProblem, ProblemSources}
import net.shrine.source.ConfigSource

import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.{Await, blocking}
import scala.concurrent.duration.{Duration, FiniteDuration, DurationInt, DurationLong}
import scala.language.postfixOps
import scala.util.control.NonFatal

/**
  * @since 1.22
  */
case class AuditEmailer(maxQueryCountBetweenAudits:Int,
                        minTimeBetweenAudits:FiniteDuration,
                        researcherLineTemplate:String,
                        emailTemplate:String,
                        emailSubject:String,
                        from:InternetAddress,
                        to:InternetAddress,
                        stewardBaseUrl:String,
                        mailer:Mailer
                       ) {
  def audit():Unit = {
    //gather a list of users to audit
    val now = System.currentTimeMillis()
    val researchersToAudit: Seq[ResearcherToAudit] = StewardDatabase.db.selectResearchersToAudit(maxQueryCountBetweenAudits,
      minTimeBetweenAudits,
      now)
    Log.info(s"Auditing users ${researchersToAudit.map(_.researcher.userName).mkString(", ")}")
    if (researchersToAudit.nonEmpty){

      val auditLines = researchersToAudit.sortBy(_.count).reverse.map { researcher =>
        researcherLineTemplate.replaceAll("FULLNAME",researcher.researcher.fullName)
          .replaceAll("USERNAME",researcher.researcher.userName)
          .replaceAll("COUNT",researcher.count.toString)
          .replaceAll("LAST_AUDIT_DATE",new Date(researcher.leastRecentQueryDate).toString)
      }.mkString("\n")

      //build up the email body
      val emailBody = Text(emailTemplate.replaceAll("AUDIT_LINES",auditLines)
        .replaceAll("STEWARD_BASE_URL",stewardBaseUrl))

      val envelope:Envelope = Envelope.from(from).to(to).subject(emailSubject).content(emailBody)

      Log.debug(s"About to send $envelope .")

      //send the email
      val future = mailer(envelope)

      try {
        blocking {
          Await.result(future, 60 seconds)
        }
        StewardDatabase.db.logAuditRequests(researchersToAudit, now)
        Log.info(s"Sent and logged $envelope .")
      } catch {
        case NonFatal(x) => CouldNotSendAuditEmail(envelope,x)
      }
    }
  }
}

object AuditEmailer {

  def audit(): IO[Unit] = IO(AuditEmailer(ConfigSource.config).audit())

  val interval: FiniteDuration = ConfigSource.config.getFiniteDuration("shrine.steward.emailDataSteward.interval")

  val initialDelay: FiniteDuration = initialDelayToSendEmail(ConfigSource.config.getFiniteDuration("shrine.steward.emailDataSteward.timeAfterMidnight"),interval)

  implicit val timer: Timer[IO] = IO.timer(ExecutionContexts.crcExecutionContext)

  private def scheduler: IOScheduler = IOScheduler.scheduleIndefinitely(
    initialDelay = initialDelay,
    interval = interval,
    task = () => audit,
    name = "AuditEmailer"
  )

  def restart(): IO[Unit] = scheduler.restart()

  def stop(): IO[Unit] = scheduler.stop()

  /**
    *
    * @param config All of shrine.conf
    */
  def apply(config:Config):AuditEmailer = {
    val config = ConfigSource.config
    val emailConfig = config.getConfig("shrine.steward.emailDataSteward")

    AuditEmailer(
      maxQueryCountBetweenAudits = emailConfig.getInt("maxQueryCountBetweenAudits"),
      minTimeBetweenAudits = emailConfig.getFiniteDuration("minTimeBetweenAudits"),
      researcherLineTemplate = emailConfig.getString("researcherLine"),
      emailTemplate = emailConfig.getString("emailBody"),
      emailSubject = emailConfig.getString("subject"),
      from = emailConfig.get("from", new InternetAddress(_)),
      to = emailConfig.get("to", new InternetAddress(_)),
      stewardBaseUrl = emailConfig.getString("externalStewardBaseUrl"),
      mailer = ConfiguredMailer.createMailerFromConfig(config.getConfig("shrine.email")))
  }

  /**
    * Check the emailer's config, log any problems
    *
    * @param config All of shrine.conf
    */
  def configCheck(config:Config):Boolean = try {
    val autoEmailer = apply(config)
    Log.info(s"DSA will request audits from ${autoEmailer.to}")
    true
  } catch {
    case NonFatal(x) =>
      CannotConfigureAuditEmailer(x)
      false
  }

  def initialDelayToSendEmail(timeFromMidnight:Duration,interval:Duration): FiniteDuration = {
    if(interval >= (1 day)) {

      import java.util.Calendar
      val c = Calendar.getInstance()
      val now = c.getTimeInMillis

      val previousMidnight = {
        c.set(Calendar.HOUR_OF_DAY, 0)
        c.set(Calendar.MINUTE, 0)
        c.set(Calendar.SECOND, 0)
        c.set(Calendar.MILLISECOND, 0)
        c.getTimeInMillis
      }
      val timeToSendToday: Long = previousMidnight + timeFromMidnight.toMillis

      val timeToSend: Long = if (timeToSendToday > now) timeToSendToday
      else timeToSendToday + (1 day).toMillis
      (timeToSend - now) milliseconds
    }
    else 0 milliseconds //if we're testing then don't delay that first send
  }
}


case class CannotConfigureAuditEmailer(ex:Throwable) extends AbstractProblem(ProblemSources.Dsa) {
  override def summary: String = "The DSA will not email audit requests due to a misconfiguration."

  override def description: String = s"The DSA will not email audit requests due to ${throwable.get}"

  override def throwable = Some(ex)
}

case class CouldNotSendAuditEmail(envelope:Envelope,ex:Throwable) extends AbstractProblem(ProblemSources.Dsa) {
  override def summary: String = "The DSA was not able to send an audit email."

  override def description: String = s"The DSA was not able to send an audit request to ${envelope.to} due to ${throwable.get}"

  override def throwable = Some(ex)

  override def detailsText: Option[String] = Some(s"Could not send $envelope")
}