package net.shrine.api.steward

import cats.data.OptionT
import cats.effect.{ConcurrentEffect, ContextShift, IO}
import cats.implicits.{catsSyntaxEither => _}
import io.circe.generic.auto.exportDecoder
import io.circe.generic.auto.exportEncoder
import io.circe.syntax.EncoderOps
import net.shrine.api.steward.db.{QueryParameters, StewardDatabase, UserRecord}
import net.shrine.i2b2.protocol.pm.User
import net.shrine.protocol.Credential
import net.shrine.util.ShouldMatchersForJUnit
import org.http4s.dsl.io.{Accepted, Forbidden, NotFound, Ok, Unauthorized, UnprocessableEntity}
import org.http4s.headers.Authorization
import org.http4s.{BasicCredentials, EntityDecoder, Headers, Method, Request, Response, Uri}
import org.junit.runner.RunWith
import org.scalatest.{BeforeAndAfter, FlatSpec}
import org.scalatestplus.junit.JUnitRunner
import net.shrine.source.ConfigSource
import org.http4s.circe.jsonOf
import net.shrine.api.steward

import scala.concurrent.ExecutionContext.Implicits.global


@RunWith(classOf[JUnitRunner])
class StewardServiceTest extends FlatSpec with BeforeAndAfter with ShouldMatchersForJUnit {


  def extractResponse(request: Request[IO]): Response[IO] = {
    implicit val contextShift: ContextShift[IO] = IO.contextShift(global)
    implicit val concurrentEffect: ConcurrentEffect[IO] = IO.ioConcurrentEffect(contextShift)

    val responseOptionIo: OptionT[IO, Response[IO]] = StewardService().service.run(request)

    responseOptionIo.map { r: Response[IO] => r }.fold {
      fail("No response from steward service")
    } { r: Response[IO] => r
    }.unsafeRunSync()
  }

  "StewardService" should " reply to ping with pong" in {
    val pingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/ping"))

    val response: Response[IO] = extractResponse(pingRequest)

    assertResult(Ok)(response.status)
    val entityText = response.as[String].unsafeRunSync
    assertResult("pong")(entityText)
  }

  "StewardService" should  "return an OK and the correct createTopicsMode name" in {

    ConfigSource.configForBlock(CreateTopicsMode.createTopicsModeConfigKey,CreateTopicsMode.Pending.name,s"${CreateTopicsMode.Pending.name} test") {
      {
        val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/about/createTopicsMode"))

        val response: Response[IO] = extractResponse(request)

        assertResult(Ok)(response.status)

        val createTopicsModeName = response.as[String].unsafeRunSync()

        assertResult(s"""${CreateTopicsMode.createTopicsInState.name}""")(createTopicsModeName)
        assertResult(s"""${CreateTopicsMode.Pending.name}""")(createTopicsModeName)
      }
    }
    ConfigSource.configForBlock(CreateTopicsMode.createTopicsModeConfigKey,CreateTopicsMode.Approved.name,s"${CreateTopicsMode.Approved.name} test") {

      val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/about/createTopicsMode"))

      val response: Response[IO] = extractResponse(request)

      assertResult(Ok)(response.status)

      val createTopicsModeName = response.as[String].unsafeRunSync()

      assertResult( s"""${CreateTopicsMode.createTopicsInState.name}""")(createTopicsModeName)
      assertResult( s"""${CreateTopicsMode.Approved.name}""")(createTopicsModeName)
    }

    ConfigSource.configForBlock(CreateTopicsMode.createTopicsModeConfigKey,CreateTopicsMode.TopicsIgnoredJustLog.name,s"${CreateTopicsMode.TopicsIgnoredJustLog.name} test") {

      val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/about/createTopicsMode"))

      val response: Response[IO] = extractResponse(request)

      assertResult(Ok)(response.status)

      val createTopicsModeName = response.as[String].unsafeRunSync()

      assertResult( s"""${CreateTopicsMode.createTopicsInState.name}""")(createTopicsModeName)
      assertResult( s"""${CreateTopicsMode.TopicsIgnoredJustLog.name}""")(createTopicsModeName)
    }
  }

  "StewardService" should  "return an OK and a valid outbound user for a user/whoami request" in {

    implicit val decoder: EntityDecoder[IO, OutboundUser] = jsonOf[IO, OutboundUser]

    val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/user/whoami"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))

    val response: Response[IO] = extractResponse(request)

    assertResult(Ok)(response.status)
    val outboundUser = response.as[OutboundUser].unsafeRunSync()
    assertResult(researcherOutboundUser)(outboundUser)
  }

  "StewardService" should  "return a 200 and AuthenticationFailed message for a user/whoami request with bad credentials" in {

    val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/user/whoami"),
      headers = Headers.of(Authorization(BasicCredentials("qep", "wrongPassword"))))
    val response: Response[IO] = extractResponse(request)

    val responseText = response.as[String].unsafeRunSync()

    assertResult("""AuthenticationFailed""")(responseText)
    assertResult(Ok)(response.status)
  }

  "StewardService" should "return the list of a researcher's Topics as Json" in {

    implicit val decoder: EntityDecoder[IO, ResearchersTopics] = jsonOf[IO, ResearchersTopics]

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()

    val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
      val response: Response[IO] = extractResponse(request)

    assertResult(Ok)(response.status)

    val topics = response.as[ResearchersTopics].unsafeRunSync()

    assertResult(true)(ResearchersTopics(researcherUserName,1,0,Seq(uncontroversialTopic)).sameExceptForTimes(topics))
  }

  "StewardService" should "return the list of a researcher's Topics as Json for various states" in {

    implicit val decoder: EntityDecoder[IO, ResearchersTopics] = jsonOf[IO, ResearchersTopics]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()

    val pendingRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?state=Pending"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val pendingResponse: Response[IO] = extractResponse(pendingRequest)

    val pendingTopics = pendingResponse.as[ResearchersTopics].unsafeRunSync()

    assertResult(true)(ResearchersTopics(researcherUserName,1,0,Seq(uncontroversialTopic)).sameExceptForTimes(pendingTopics))
    assertResult(Ok)(pendingResponse.status)

    StewardDatabase.db.changeTopicStateIO(1,TopicState.approved,stewardUserName).unsafeRunSync()

    val approvedRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?state=Approved"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val approvedResponse: Response[IO] = extractResponse(approvedRequest)

    val approvedTopics = approvedResponse.as[ResearchersTopics].unsafeRunSync()

    assertResult(true)(ResearchersTopics(researcherUserName,1,0,Seq(
      OutboundTopic(1,uncontroversialTopic.name,uncontroversialTopic.description,researcherOutboundUser,0,TopicState.approved.name,stewardOutboundUser,0))).sameExceptForTimes(approvedTopics))

    assertResult(Ok)(approvedResponse.status)


    StewardDatabase.db.changeTopicStateIO(1,TopicState.rejected,stewardUserName).unsafeRunSync()

    val rejectedRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?state=Rejected"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val rejectedResponse: Response[IO] = extractResponse(rejectedRequest)

    val rejectedTopics = rejectedResponse.as[ResearchersTopics].unsafeRunSync()

    assertResult(true)(ResearchersTopics(researcherUserName,1,0,Seq(
        OutboundTopic(1,uncontroversialTopic.name,uncontroversialTopic.description,researcherOutboundUser,0,TopicState.rejected.name,stewardOutboundUser,0))).sameExceptForTimes(rejectedTopics))

    assertResult(Ok)(rejectedResponse.status)
  }

  "StewardService" should "reject nonsense topic request states" in {
    val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?state=nonsense"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val response: Response[IO] = extractResponse(request)

    assertResult(UnprocessableEntity)(response.status)
  }

  "StewardService" should "return the list of a researcher's Topics as Json with skip and limit set" in {

    implicit val decoder: EntityDecoder[IO, ResearchersTopics] = jsonOf[IO, ResearchersTopics]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createFiveTopics()

    val topicsNoSkipRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?skip=0&limit=2"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val topicsNoSkipResponse: Response[IO] = extractResponse(topicsNoSkipRequest)

    val topicsNoSkip = topicsNoSkipResponse.as[ResearchersTopics].unsafeRunSync()

    assertResult(2)(topicsNoSkip.topics.size)
    assertResult(Ok)(topicsNoSkipResponse.status)


    val topicsSkip2Request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?skip=2&limit=2"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val topicsSkip2Response: Response[IO] = extractResponse(topicsSkip2Request)

    val topicsSkip2 = topicsSkip2Response.as[ResearchersTopics].unsafeRunSync()

    assertResult(2)(topicsSkip2.topics.size)
    assertResult(Ok)(topicsSkip2Response.status)


    val topicsSkip4Request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?skip=4&limit=2"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val topicsSkip4Response: Response[IO] = extractResponse(topicsSkip4Request)

    val topicsSkip4 = topicsSkip4Response.as[ResearchersTopics].unsafeRunSync()

    assertResult(1)(topicsSkip4.topics.size)
    assertResult(Ok)(topicsSkip4Response.status)
  }

  "StewardService" should "return the list of a researcher's Topics as Json with different sorting and ordering options" in {

    implicit val decoder: EntityDecoder[IO, ResearchersTopics] = jsonOf[IO, ResearchersTopics]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createFiveTopics()

    val topicsSortByIdRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?sortBy=id"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val topicsSortByIdResponse: Response[IO] = extractResponse(topicsSortByIdRequest)

    val topicsSortById = topicsSortByIdResponse.as[ResearchersTopics].unsafeRunSync()
    assertResult(5)(topicsSortById.topics.size)
    assertResult(Ok)(topicsSortByIdResponse.status)

    assertResult(topicsSortById.topics.sortBy(_.id))(topicsSortById.topics)

    val topicsSortByNameRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?sortBy=name"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val topicsSortByNameResponse: Response[IO] = extractResponse(topicsSortByNameRequest)

    val topicsSortByName = topicsSortByNameResponse.as[ResearchersTopics].unsafeRunSync()
    assertResult(5)(topicsSortByName.topics.size)
    assertResult(Ok)(topicsSortByNameResponse.status)
    assertResult(topicsSortByName.topics.sortBy(_.createdBy.userName))(topicsSortByName.topics)


    val topicsSortByNameDescRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?sortBy=name&sortDirection=descending"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val topicsSortByNameDescResponse: Response[IO] = extractResponse(topicsSortByNameDescRequest)

    val topicsSortByNameDesc = topicsSortByNameDescResponse.as[ResearchersTopics].unsafeRunSync()
    assertResult(5)(topicsSortByNameDesc.topics.size)
    assertResult(Ok)(topicsSortByNameDescResponse.status)
    assertResult(topicsSortByNameDesc.topics.sortBy(_.name).reverse)(topicsSortByNameDesc.topics)


    val topicsSortByNameAscRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/topics?sortBy=name&sortDirection=ascending"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val topicsSortByNameAscResponse: Response[IO] = extractResponse(topicsSortByNameAscRequest)

    val topicsSortByNameAsc = topicsSortByNameAscResponse.as[ResearchersTopics].unsafeRunSync()
    assertResult(5)(topicsSortByNameAsc.topics.size)
    assertResult(Ok)(topicsSortByNameAscResponse.status)
    assertResult(topicsSortByNameAsc.topics.sortBy(_.createdBy.userName))(topicsSortByNameAsc.topics)
  }

  "StewardService" should "return the list of a researcher's query history as Json" in {

    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()
    StewardDatabase.db.changeTopicStateIO(1,TopicState.approved,stewardUserName).unsafeRunSync()
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(0,"test query",queryContent))

    val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?asJson=false"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val response: Response[IO] = extractResponse(request)

    val queries = response.as[QueryHistory].unsafeRunSync()

    val history: QueryHistory = QueryHistory(1, 0, List(
      OutboundShrineQuery(1, 0, "test query", researcherOutboundUser, Some(
        OutboundTopic(1, uncontroversialTopic.name, uncontroversialTopic.description, researcherOutboundUser, 0, TopicState.approved.name, stewardOutboundUser, 0)
      ), getItBack, TopicState.approved.name, 0)
    ))
    assertResult(diffs)(history.differencesExceptTimes(queries))
  }

  "StewardService" should "return the list of a researcher's query history as Json, with query bodies as Json" in {

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()
    StewardDatabase.db.changeTopicStateIO(topicId = 1,TopicState.approved,stewardUserName).unsafeRunSync()
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(0,"test query",queryContent))

    val request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?asJson=true"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val response: Response[IO] = extractResponse(request)

    val queriesString: String = response.as[String].unsafeRunSync()

    val expectedQueryContents = "\"queryContents\" : {\n        \"queryDefinition\" : {\n          \"name\" : \"18-34 years old@18:31:51\",\n          \"expr\" : {\n            \"term\" : \"\\\\\\\\SHRINE\\\\SHRINE\\\\Demographics\\\\Age\\\\18-34 years old\\\\\"\n          }\n        }\n      }"

    assert(queriesString.contains(expectedQueryContents))
  }

  "StewardService" should "return the list of a researcher's query history as Json, with skip and limit parameters" in {

    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createSixQueries()

    val queriesNoLimitRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesNoLimitResponse: Response[IO] = extractResponse(queriesNoLimitRequest)

    val queriesNoLimit = queriesNoLimitResponse.as[QueryHistory].unsafeRunSync()

    assertResult(Ok)(queriesNoLimitResponse.status)
    assertResult(6)(queriesNoLimit.queryRecords.size)

    val queriesLimit5Request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?skip=0&limit=5"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesLimit5Response: Response[IO] = extractResponse(queriesLimit5Request)

    val queriesLimit5 = queriesLimit5Response.as[QueryHistory].unsafeRunSync()

    assertResult(Ok)(queriesLimit5Response.status)
    assertResult(5)(queriesLimit5.queryRecords.size)


    val queriesSkip1Limit4Request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?skip=1&limit=4"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesSkip1Limit4Response: Response[IO] = extractResponse(queriesSkip1Limit4Request)

    val queriesSkip1Limit4 = queriesSkip1Limit4Response.as[QueryHistory].unsafeRunSync()

    assertResult(Ok)(queriesSkip1Limit4Response.status)
    assertResult(4)(queriesSkip1Limit4.queryRecords.size)


    val queriesSkip4Limit5Request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?skip=4&limit=5"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesSkip4Limit5Response: Response[IO] = extractResponse(queriesSkip4Limit5Request)

    val queriesSkip4Limit5 = queriesSkip4Limit5Response.as[QueryHistory].unsafeRunSync()

    assertResult(Ok)(queriesSkip4Limit5Response.status)
    assertResult(2)(queriesSkip4Limit5.queryRecords.size)
  }

  "StewardService" should "return the list of a researcher's query history as Json, using parameters for sorting" in {

    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createSixQueries()

    val queriesSoryByExternalIdRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?sortBy=externalId"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesSoryByExternalIdResponse: Response[IO] = extractResponse(queriesSoryByExternalIdRequest)

    val queriesSortByExternalId = queriesSoryByExternalIdResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queriesSortByExternalId.queryRecords.size)
    assertResult(queriesSortByExternalId.queryRecords.sortBy(_.externalId))(queriesSortByExternalId.queryRecords)

    val queriesSoryByExternalIdAscRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?sortBy=externalId&sortDirection=ascending"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesSoryByExternalIdAscResponse: Response[IO] = extractResponse(queriesSoryByExternalIdAscRequest)

    val queriesSoryByExternalIdAsc = queriesSoryByExternalIdAscResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queriesSoryByExternalIdAsc.queryRecords.size)
    assertResult(queriesSoryByExternalIdAsc.queryRecords.sortBy(_.externalId))(queriesSoryByExternalIdAsc.queryRecords)


    val queriesSortByExternalIdDescRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?sortBy=externalId&sortDirection=descending"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesSortByExternalIdDescResponse: Response[IO] = extractResponse(queriesSortByExternalIdDescRequest)

    val queriesSortByExternalIdDesc = queriesSortByExternalIdDescResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queriesSortByExternalIdDesc.queryRecords.size)
    assertResult(queriesSortByExternalIdDesc.queryRecords.sortBy(_.externalId).reverse)(queriesSortByExternalIdDesc.queryRecords)


    val queriesSortByStewardResponseRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?sortBy=stewardResponse"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesSortByStewardResponseResponse: Response[IO] = extractResponse(queriesSortByStewardResponseRequest)

    val queriesSortByStewardResponse = queriesSortByStewardResponseResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queriesSortByStewardResponse.queryRecords.size)
    assertResult(queriesSortByStewardResponse.queryRecords.sortBy(_.stewardResponse))(queriesSortByStewardResponse.queryRecords)


    val queriesSortByNameRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?sortBy=name"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesSortByNameResponse: Response[IO] = extractResponse(queriesSortByNameRequest)

    val queriesSortByName = queriesSortByNameResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queriesSortByName.queryRecords.size)
    assertResult(queriesSortByName.queryRecords.sortBy(_.name))(queriesSortByName.queryRecords)
  }


  "StewardService" should "return the list of a researcher's query history, using parameters for sorting, skip, and limit" in {

    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createSixQueries()

    val queriesRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesResponse: Response[IO] = extractResponse(queriesRequest)

    assertResult(Ok)(queriesResponse.status)

    val queries = queriesResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queries.queryRecords.size)
    assertResult(queries.queryRecords.sortBy(_.externalId))(queries.queryRecords)

    val sixQueries:Seq[OutboundShrineQuery] = queries.queryRecords


    val queriesLimit3SortByNameRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?skip=0&limit=3&sortBy=name"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesLimit3SortByNameResponse: Response[IO] = extractResponse(queriesLimit3SortByNameRequest)

    assertResult(Ok)(queriesLimit3SortByNameResponse.status)

    val queriesLimit3SortByName = queriesLimit3SortByNameResponse.as[QueryHistory].unsafeRunSync()
    assertResult(3)(queriesLimit3SortByName.queryRecords.size)
    assertResult(queriesLimit3SortByName.queryRecords)(sixQueries.sortBy(_.name).take(3))


    val queriesSkip2Limit3SortByNameRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?skip=2&limit=3&sortBy=name"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesSkip2Limit3SortByNameResponse: Response[IO] = extractResponse(queriesSkip2Limit3SortByNameRequest)

    assertResult(Ok)(queriesSkip2Limit3SortByNameResponse.status)

    val queriesSkip2Limit3SortByName = queriesSkip2Limit3SortByNameResponse.as[QueryHistory].unsafeRunSync()
    assertResult(3)(queriesSkip2Limit3SortByName.queryRecords.size)
    assertResult(queriesSkip2Limit3SortByName.queryRecords)(sixQueries.sortBy(_.name).slice(2, 5))
  }

  "StewardService" should "return the list of a researcher's query history as Json, with filtered by topic" in {

    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createSixQueries()

    val queriesRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesResponse: Response[IO] = extractResponse(queriesRequest)

    assertResult(Ok)(queriesResponse.status)

    val queries = queriesResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queries.queryRecords.size)


    val queriesTopic1Request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory/topic/1"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesTopic1Response: Response[IO] = extractResponse(queriesTopic1Request)

    val queriesTopic1 = queriesTopic1Response.as[QueryHistory].unsafeRunSync()
    assertResult(Ok)(queriesTopic1Response.status)

    val queriesTopic1Skip1Limit2Request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory/topic/1?skip=1&limit=2"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesTopic1Skip1Limit2Response: Response[IO] = extractResponse(queriesTopic1Skip1Limit2Request)

    assertResult(Ok)(queriesTopic1Skip1Limit2Response.status)

    val queriesTopic1Skip1Limit2 = queriesTopic1Skip1Limit2Response.as[QueryHistory].unsafeRunSync()

    assertResult(2)(queriesTopic1Skip1Limit2.queryRecords.size)


    val queriesTopic2Request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory/topic/2"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesTopic2Response: Response[IO] = extractResponse(queriesTopic2Request)

    assertResult(Ok)(queriesTopic2Response.status)

    val queriesTopic2 = queriesTopic2Response.as[QueryHistory].unsafeRunSync()
    assertResult(3)(queriesTopic2.queryRecords.size)
  }

  "StewardService" should "return the list of a researcher's query history as Json, filtered by date" in {

    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)

    val startTime = System.currentTimeMillis()
    Thread.sleep(10)

    createSixQueries()

    Thread.sleep(10)
    val finishTime = System.currentTimeMillis()

    val queriesRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesResponse: Response[IO] = extractResponse(queriesRequest)
    assertResult(Ok)(queriesResponse.status)

    val queries = queriesResponse.as[QueryHistory].unsafeRunSync()

    assertResult(6)(queries.queryRecords.size)


    val queriesMinDateMaxDateRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?minDate=$startTime&maxDate=$finishTime"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesMinDateMaxDateResponse: Response[IO] = extractResponse(queriesMinDateMaxDateRequest)

    assertResult(Ok)(queriesMinDateMaxDateResponse.status)

    val queriesMinDateMaxDate = queriesMinDateMaxDateResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queriesMinDateMaxDate.queryRecords.size)


    val queriesStartTimeRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?minDate=$startTime&maxDate=$startTime"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesStartTimeResponse: Response[IO] = extractResponse(queriesStartTimeRequest)

    assertResult(Ok)(queriesStartTimeResponse.status)

    val queriesStartTime = queriesStartTimeResponse.as[QueryHistory].unsafeRunSync()
    assertResult(0)(queriesStartTime.queryRecords.size)


    val queriesFinishTimeRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?minDate=$finishTime&maxDate=$finishTime"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesFinishTimeResponse: Response[IO] = extractResponse(queriesFinishTimeRequest)

    assertResult(Ok)(queriesFinishTimeResponse.status)

    val queriesFinishTime = queriesFinishTimeResponse.as[QueryHistory].unsafeRunSync()

    assertResult(0)(queriesFinishTime.queryRecords.size)
  }

  "StewardService" should "return the list of a researcher's query history as Json, even including an unknown topic id" in {
    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(0,"test query for unknown topic",queryContent))

    val queriesRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesResponse: Response[IO] = extractResponse(queriesRequest)

    assertResult(Ok)(queriesResponse.status)

    val queries = queriesResponse.as[QueryHistory].unsafeRunSync()

    assertResult(diffs)(QueryHistory(1,0,List(OutboundShrineQuery(1,0,"test query for unknown topic",researcherOutboundUser,None,getItBack,TopicState.unknownForUser.name,0))).differencesExceptTimes(queries))
  }


  "StewardService" should "return the list of a researcher's query history as Json, sorted by topic name" in {
    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createSixQueries()

    val queriesRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/researcher/queryHistory?sortBy=topicName"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow"))))
    val queriesResponse: Response[IO] = extractResponse(queriesRequest)

    assertResult(Ok)(queriesResponse.status)

    val queries = queriesResponse.as[QueryHistory].unsafeRunSync()

    assertResult(6)(queries.queryRecords.size)

  }


  "StewardService" should "return counts of topics, total and by state" in {
    implicit val decoder: EntityDecoder[IO, TopicsPerState] = jsonOf[IO, TopicsPerState]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createFiveTopics()

    val statisticsRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/steward/statistics/topicsPerState"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val statisticsResponse: Response[IO] = extractResponse(statisticsRequest)

    assertResult(Ok)(statisticsResponse.status)

    val expectedJson =  """{
                        |  "total" : 5,
                        |  "topicsPerState" : [
                        |    {
                        |      "Pending" : 5
                        |    }
                        |  ]
                        |}""".stripMargin
    val statisticsJson = statisticsResponse.as[String].unsafeRunSync()
    assertResult(expectedJson)(statisticsJson)
  }

  "StewardService" should "injest a researcher's request to study a topic" in {

    var inboundTopicRequest = InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)

    val topicsRequest: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/researcher/requestTopicAccess"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow")))).withEntity(inboundTopicRequest.asJson.toString)

    val topicsResponse: Response[IO] = extractResponse(topicsRequest)

    assertResult(Accepted)(topicsResponse.status)
    assertResult(1)(StewardDatabase.db.selectUsers.size)
    assertResult(UserRecord(researcherUserName,researcherUserName,false))(StewardDatabase.db.selectUsers.head)

    val topics = StewardDatabase.db.selectTopics(QueryParameters())
    assertResult(1)(topics.size)

    val topic = topics.head
    assertResult(uncontroversialTopic.name)(topic.name)
    assertResult(uncontroversialTopic.description)(topic.description)
    assertResult(researcherUserName)(topic.createdBy)
    assertResult(TopicState.pending)(topic.state)
  }

  "StewardService" should "injest a second request from a researcher request to study a new topic" in {

    val secondTopicName = "Liver"
    val secondTopicDescription = "Liver Study"

    StewardDatabase.db.upsertUser(researcherUser)

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()

    var inboundTopicRequest = InboundTopicRequest(secondTopicName,secondTopicDescription)
    val topicsRequest: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/researcher/requestTopicAccess"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow")))).withEntity(inboundTopicRequest.asJson.toString)

    val topicsResponse: Response[IO] = extractResponse(topicsRequest)
    assertResult(Accepted)(topicsResponse.status)
    assertResult(1)(StewardDatabase.db.selectUsers.size)
    assertResult(UserRecord(researcherUserName,researcherUserName,false))(StewardDatabase.db.selectUsers.head)

    val topics = StewardDatabase.db.selectTopics(QueryParameters())
    assertResult(2)(topics.size)

    val firstTopic = topics(0)
    assertResult(uncontroversialTopic.name)(firstTopic.name)
    assertResult(uncontroversialTopic.description)(firstTopic.description)
    assertResult(researcherUserName)(firstTopic.createdBy)
    assertResult(TopicState.pending)(firstTopic.state)

    val secondTopic = topics(1)
    assertResult(secondTopicName)(secondTopic.name)
    assertResult(secondTopicDescription)(secondTopic.description)
    assertResult(researcherUserName)(secondTopic.createdBy)
    assertResult(TopicState.pending)(secondTopic.state)
  }

  "StewardService" should " place new topics in the Approved state in auto-approve mode" in {

    ConfigSource.configForBlock(CreateTopicsMode.createTopicsModeConfigKey,CreateTopicsMode.TopicsIgnoredJustLog.name,s"${CreateTopicsMode.TopicsIgnoredJustLog.name} test") {

      var inboundTopicRequest = InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)
      val topicsRequest: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/researcher/requestTopicAccess"),
        headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow")))).withEntity(inboundTopicRequest.asJson.toString)

      val topicsResponse: Response[IO] = extractResponse(topicsRequest)

      assertResult(Accepted)(topicsResponse.status)
      assertResult(1)(StewardDatabase.db.selectUsers.size)
      assertResult(UserRecord(researcherUserName, researcherUserName, false))(StewardDatabase.db.selectUsers.head)

      val topics = StewardDatabase.db.selectTopics(QueryParameters())
      assertResult(1)(topics.size)

      val topic = topics.head
      assertResult(uncontroversialTopic.name)(topic.name)
      assertResult(uncontroversialTopic.description)(topic.description)
      assertResult(researcherUserName)(topic.createdBy)
      assertResult(TopicState.approved)(topic.state)
    }
  }

  "StewardService" should "injest a researcher's request to edit a topic" in {

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()

    val updatedDescription = "Really you should accept this"

    var inboundTopicRequest = InboundTopicRequest(uncontroversialTopic.name,updatedDescription)
    val topicsRequest: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/researcher/editTopicRequest/1"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow")))).withEntity(inboundTopicRequest.asJson.toString)

    val topicsResponse: Response[IO] = extractResponse(topicsRequest)

    assertResult(Accepted)(topicsResponse.status)

    val topics = StewardDatabase.db.selectTopics(QueryParameters())
    assertResult(1)(topics.size)

    val topic = topics.head
    assertResult(uncontroversialTopic.name)(topic.name)
    assertResult(updatedDescription)(topic.description)
    assertResult(researcherUserName)(topic.createdBy)
    assertResult(researcherUserName)(topic.changedBy)
    assertResult(TopicState.pending)(topic.state)
  }

  "StewardService" should "reject a researcher's request to edit a topic that does not exist" in {

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()

    val updatedDescription = "Really you should accept this"

    var inboundTopicRequest = InboundTopicRequest(uncontroversialTopic.name,updatedDescription)
    val topicsRequest: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/researcher/editTopicRequest/2"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow")))).withEntity(inboundTopicRequest.asJson.toString)

    val topicsResponse: Response[IO] = extractResponse(topicsRequest)

    assertResult(NotFound)(topicsResponse.status)

    val topics = StewardDatabase.db.selectTopics(QueryParameters())
    assertResult(1)(topics.size)

    val topic = topics.head
    assertResult(uncontroversialTopic.name)(topic.name)
    assertResult(uncontroversialTopic.description)(topic.description)
    assertResult(researcherUserName)(topic.createdBy)
    assertResult(researcherUserName)(topic.changedBy)
    assertResult(TopicState.pending)(topic.state)
  }

  "StewardService" should "reject a researcher's request to edit an approved topic" in {

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()
    StewardDatabase.db.changeTopicStateIO(1,TopicState.approved,stewardUserName).unsafeRunSync()

    val updatedDescription = "Really you should accept this"

    var inboundTopicRequest = InboundTopicRequest(uncontroversialTopic.name,updatedDescription)
    val topicsRequest: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/researcher/editTopicRequest/1"),
      headers = Headers.of(Authorization(BasicCredentials(researcherUser.username, "kapow")))).withEntity(inboundTopicRequest.asJson.toString)

    val topicsResponse: Response[IO] = extractResponse(topicsRequest)

    assertResult(Forbidden)(topicsResponse.status)

    val topics = StewardDatabase.db.selectTopics(QueryParameters())
    assertResult(1)(topics.size)

    val topic = topics.head
    assertResult(uncontroversialTopic.name)(topic.name)
    assertResult(uncontroversialTopic.description)(topic.description)
    assertResult(researcherUserName)(topic.createdBy)
    assertResult(stewardUserName)(topic.changedBy)
    assertResult(TopicState.approved)(topic.state)
  }

  "StewardService" should "reject an attempt to edit a topic owned by a different user" in {

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()

    val updatedDescription = "Really you should accept this"

    var inboundTopicRequest = InboundTopicRequest(uncontroversialTopic.name,updatedDescription)
    val topicsRequest: Request[IO] = Request(method = Method.POST, uri = Uri.unsafeFromString(s"/researcher/editTopicRequest/1"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam")))).withEntity(inboundTopicRequest.asJson.toString)

    val topicsResponse: Response[IO] = extractResponse(topicsRequest)

    assertResult(Forbidden)(topicsResponse.status)

    val topics = StewardDatabase.db.selectTopics(QueryParameters())
    assertResult(1)(topics.size)

    val topic = topics.head
    assertResult(uncontroversialTopic.name)(topic.name)
    assertResult(uncontroversialTopic.description)(topic.description)
    assertResult(researcherUserName)(topic.createdBy)
    assertResult(TopicState.pending)(topic.state)
  }

  "StewardService" should "return the full history of queries as Json" in {

    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()
    StewardDatabase.db.changeTopicStateIO(1,TopicState.approved,stewardUserName).unsafeRunSync()
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(0,"test query",queryContent))

    val queriesRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/steward/queryHistory"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val queriesResponse: Response[IO] = extractResponse(queriesRequest)

    assertResult(Ok)(queriesResponse.status)

    val queries = queriesResponse.as[QueryHistory].unsafeRunSync()

    assertResult(diffs) (
      QueryHistory(1,0,List(OutboundShrineQuery(1,0,"test query",researcherOutboundUser,Some(
        OutboundTopic(1,uncontroversialTopic.name,uncontroversialTopic.description,researcherOutboundUser,0,
          TopicState.approved.name,stewardOutboundUser,0)),getItBack,TopicState.approved.name,0)))
        .differencesExceptTimes(queries))
  }

  "StewardService" should "return the history of queries for a specific user as Json" in {

    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()
    StewardDatabase.db.changeTopicStateIO(1,TopicState.approved,stewardUserName).unsafeRunSync()
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(0,"test query",queryContent))

    val queriesRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/steward/queryHistory/user/${researcherUserName}"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val queriesResponse: Response[IO] = extractResponse(queriesRequest)

    assertResult(Ok)(queriesResponse.status)

    val queries = queriesResponse.as[QueryHistory].unsafeRunSync()

    assertResult(diffs)(
      QueryHistory(1,0,List(OutboundShrineQuery(1,0,"test query",researcherOutboundUser,Some(OutboundTopic(
        1,uncontroversialTopic.name,uncontroversialTopic.description,researcherOutboundUser,0,
        TopicState.approved.name,stewardOutboundUser,0)),getItBack,TopicState.approved.name,0)))
        .differencesExceptTimes(queries))
  }

  "StewardService" should "return the query history as Json, filtered by topic" in {

    implicit val decoder: EntityDecoder[IO, QueryHistory] = jsonOf[IO, QueryHistory]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createSixQueries()

    val queriesRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/steward/queryHistory"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val queriesResponse: Response[IO] = extractResponse(queriesRequest)

    assertResult(Ok)(queriesResponse.status)

    val queries = queriesResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queries.queryRecords.size)


    val queriesForBenRequest: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/steward/queryHistory/user/ben"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val queriesForBenResponse: Response[IO] = extractResponse(queriesForBenRequest)

    assertResult(Ok)(queriesForBenResponse.status)

    val queriesForBen = queriesForBenResponse.as[QueryHistory].unsafeRunSync()
    assertResult(6)(queriesForBen.queryRecords.size)


    val queriesForTopic1Request: Request[IO] = Request(method = Method.GET, uri = Uri.unsafeFromString(s"/steward/queryHistory/topic/1"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val queriesForTopic1Response: Response[IO] = extractResponse(queriesForTopic1Request)

    assertResult(Ok)(queriesForTopic1Response.status)

    val queriesForTopic1 = queriesForTopic1Response.as[QueryHistory].unsafeRunSync()
    assertResult(3)(queriesForTopic1.queryRecords.size)


    val queriesForBenAndTopic1Request: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/queryHistory/user/ben/topic/1"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val queriesForBenAndTopic1Response: Response[IO] = extractResponse(queriesForBenAndTopic1Request)

    assertResult(Ok)(queriesForBenAndTopic1Response.status)

    val queriesForBenAndTopic1 = queriesForBenAndTopic1Response.as[QueryHistory].unsafeRunSync()
    assertResult(3)(queriesForBenAndTopic1.queryRecords.size)


    val queriesForTopic1Skip1Limit2Request: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/queryHistory/topic/1?skip=1&limit=2"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val queriesForTopic1Skip1Limit2Response: Response[IO] = extractResponse(queriesForTopic1Skip1Limit2Request)

    assertResult(Ok)(queriesForTopic1Skip1Limit2Response.status)

    val queriesForTopic1Skip1Limit2 = queriesForTopic1Skip1Limit2Response.as[QueryHistory].unsafeRunSync()
    assertResult(2)(queriesForTopic1Skip1Limit2.queryRecords.size)


    val queriesForTopic2Request: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/queryHistory/topic/2"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val queriesForTopic2Response: Response[IO] = extractResponse(queriesForTopic2Request)

    assertResult(Ok)(queriesForTopic2Response.status)

    val queriesForTopic2 = queriesForTopic2Response.as[QueryHistory].unsafeRunSync()
    assertResult(3)(queriesForTopic2.queryRecords.size)
  }

  "StewardService" should "return counts of queries, total and by user" in {

    implicit val decoder: EntityDecoder[IO, QueriesPerUser] = jsonOf[IO, QueriesPerUser]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createSixQueries()

    val queriesPerUserRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/statistics/queriesPerUser"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val queriesPerUserResponse: Response[IO] = extractResponse(queriesPerUserRequest)
    assertResult(Ok)(queriesPerUserResponse.status)

    val expectedJson = """{
                         |  "total" : 6,
                         |  "queriesPerUser" : [
                         |    {
                         |      "_1" : {
                         |        "userName" : "ben",
                         |        "fullName" : "ben",
                         |        "roles" : [
                         |          "Researcher"
                         |        ]
                         |      },
                         |      "_2" : 6
                         |    }
                         |  ]
                         |}""".stripMargin
    val statisticsJson = queriesPerUserResponse.as[String].unsafeRunSync()

    assertResult(expectedJson)(statisticsJson)
  }

  "StewardService" should "return the topics for a specific user as Json" in {

    implicit val decoder: EntityDecoder[IO, StewardsTopics] = jsonOf[IO, StewardsTopics]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()


    val topicsRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/topics/user/${researcherUserName}"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val topicsResponse: Response[IO] = extractResponse(topicsRequest)
    assertResult(Ok)(topicsResponse.status)

    val topics = topicsResponse.as[StewardsTopics].unsafeRunSync()

    assertResult(true)(StewardsTopics(1,0,Seq(uncontroversialTopic)).sameExceptForTimes(topics))
  }


  "StewardService" should "return the topics for a specific user as Json, given skip and limit parameters" in {

    implicit val decoder: EntityDecoder[IO, StewardsTopics] = jsonOf[IO, StewardsTopics]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    createFiveTopics()

    val topicsRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/topics/user/${researcherUserName}"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val topicsResponse: Response[IO] = extractResponse(topicsRequest)
    assertResult(Ok)(topicsResponse.status)

    val topics = topicsResponse.as[StewardsTopics].unsafeRunSync()

    assertResult(5)(topics.topics.size)


    val topicsLimit3Request: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/topics/user/${researcherUserName}?skip=0&limit=3"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val topicsLimit3Response: Response[IO] = extractResponse(topicsLimit3Request)
    assertResult(Ok)(topicsLimit3Response.status)

    val topicsLimit3 = topicsLimit3Response.as[StewardsTopics].unsafeRunSync()
    assertResult(3)(topicsLimit3.topics.size)


    val topicsSkip2Limit3Request: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/topics/user/${researcherUserName}?skip=2&limit=3"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val topicsSkip2Limit3Response: Response[IO] = extractResponse(topicsSkip2Limit3Request)
    assertResult(Ok)(topicsSkip2Limit3Response.status)

    val topicsSkip2Limit3 = topicsSkip2Limit3Response.as[StewardsTopics].unsafeRunSync()
    assertResult(3)(topicsSkip2Limit3.topics.size)


    val topicsSkip3Limit4Request: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/topics/user/${researcherUserName}?skip=3&limit=4"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val topicsSkip3Limit4Response: Response[IO] = extractResponse(topicsSkip3Limit4Request)
    assertResult(Ok)(topicsSkip3Limit4Response.status)

    val topicsSkip3Limit4 = topicsSkip3Limit4Response.as[StewardsTopics].unsafeRunSync()
    assertResult(2)(topicsSkip3Limit4 .topics.size)
  }

  "StewardService" should "return all of the topics" in {

    implicit val decoder: EntityDecoder[IO, StewardsTopics] = jsonOf[IO, StewardsTopics]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()

    val topicsRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/topics"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val topicsResponse: Response[IO] = extractResponse(topicsRequest)
    assertResult(Ok)(topicsResponse.status)

    val topics = topicsResponse.as[StewardsTopics].unsafeRunSync()
    assertResult(true)(StewardsTopics(1,0,Seq(uncontroversialTopic)).sameExceptForTimes(topics))
  }


  "StewardService" should "return all of the pending topics" in {

    implicit val decoder: EntityDecoder[IO, StewardsTopics] = jsonOf[IO, StewardsTopics]

    StewardDatabase.db.upsertUser(researcherUser)
    StewardDatabase.db.upsertUser(stewardUser)
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()


    val topicsRequest: Request[IO] = Request(method = Method.GET,
      uri = Uri.unsafeFromString(s"/steward/topics?state=Pending"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val topicsResponse: Response[IO] = extractResponse(topicsRequest)
    assertResult(Ok)(topicsResponse.status)

    val topics = topicsResponse.as[StewardsTopics].unsafeRunSync()
    assertResult(true)(StewardsTopics(1,0,Seq(uncontroversialTopic)).sameExceptForTimes(topics))
  }

  "StewardService" should "approve a researcher's request to study a topic" in {

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()

    val topicsRequest: Request[IO] = Request(method = Method.POST,
      uri = Uri.unsafeFromString(s"/steward/approveTopic/topic/${uncontroversialTopic.id}"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val topicsResponse: Response[IO] = extractResponse(topicsRequest)
    assertResult(Ok)(topicsResponse.status)
  }

  "StewardService" should "reject a researcher's request to study a topic" in {
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()

    val topicsRequest: Request[IO] = Request(method = Method.POST,
      uri = Uri.unsafeFromString(s"/steward/rejectTopic/topic/${uncontroversialTopic.id}"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val topicsResponse: Response[IO] = extractResponse(topicsRequest)
    assertResult(Ok)(topicsResponse.status)
  }

  "A steward's attempt to approve or reject a topic that doesn't exist" should "report an error" in {

    val approveTopicRequest: Request[IO] = Request(method = Method.POST,
      uri = Uri.unsafeFromString(s"/steward/approveTopic/topic/${uncontroversialTopic.id}"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val approveTopicResponse: Response[IO] = extractResponse(approveTopicRequest)
    assertResult(UnprocessableEntity)(approveTopicResponse.status)


    val rejectTopicRequest: Request[IO] = Request(method = Method.POST,
      uri = Uri.unsafeFromString(s"/steward/rejectTopic/topic/${uncontroversialTopic.id}"),
      headers = Headers.of(Authorization(BasicCredentials(stewardUser.username, "kablam"))))
    val rejectTopicResponse: Response[IO] = extractResponse(rejectTopicRequest)
    assertResult(UnprocessableEntity)(rejectTopicResponse.status)
  }

  before  {
    StewardDatabase.db.createTables()
  }

  after  {
    StewardDatabase.db.nextTopicId.set(1)
    StewardDatabase.db.dropTables()
  }

  val researcherUserName = "ben"
  val researcherFullName = researcherUserName

  val stewardUserName = "dave"
  val stewardFullName = stewardUserName
  /**
    * to run these tests with I2B2
    * add a user named qep, to be the qep
    * add a Boolean parameter for qep, qep, true
    * add a user named ben, to be a researcher
    * add a user named dave, to be the data steward
    * add a Boolean parameter for dave, DataSteward, true
    * add all three users to the i2b2 project
    */

  val researcherUser = User(
    fullName = researcherFullName,
    username = researcherUserName,
    domain = "domain",
    credential = new Credential("ben's password",false),
    params = Map(),
    rolesByProject = Map()
  )

  val stewardUser = User(
    fullName = stewardFullName,
    username = stewardUserName,
    domain = "domain",
    credential = new Credential("dave's password",false),
    params = Map(stewardRole -> "true"),
    rolesByProject = Map()
  )

  val researcherOutboundUser = OutboundUser.createFromUser(researcherUser)
  val stewardOutboundUser = OutboundUser.createFromUser(stewardUser)

  val uncontroversialTopic = OutboundTopic(1,"UncontroversialKidneys","Study kidneys without controversy",researcherOutboundUser,0L,TopicState.pending.name,researcherOutboundUser,0L)
  val forbiddenTopicId = 0
  private val queryContent: QueryContents = "<queryDefinition><name>18-34 years old@18:31:51</name><expr><term>\\\\SHRINE\\SHRINE\\Demographics\\Age\\18-34 years old\\</term></expr></queryDefinition>"
  val diffs = List.empty
  val getItBack = scala.xml.XML.loadString(queryContent).toString

  def createFiveTopics(): Unit ={
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("slightlyControversial","who cares?")).unsafeRunSync()
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("controversial","controversial")).unsafeRunSync()
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("moderatelyControversial","more controversial than that")).unsafeRunSync()
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("veryControversial","Just say no")).unsafeRunSync()
  }

  def createSixQueries() = {
    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest(uncontroversialTopic.name,uncontroversialTopic.description)).unsafeRunSync()
    StewardDatabase.db.changeTopicStateIO(1,TopicState.approved,stewardUserName).unsafeRunSync()

    StewardDatabase.db.createRequestForTopicAccessIO(researcherUser,InboundTopicRequest("Forbidden topic","No way is the data steward going for this")).unsafeRunSync()
    StewardDatabase.db.changeTopicStateIO(2,TopicState.rejected,stewardUserName).unsafeRunSync()

    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(0,"A test query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(2),InboundShrineQuery(1,"B forbidden query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(2," C test query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(2),InboundShrineQuery(3,"4 forbidden query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(1),InboundShrineQuery(4,"7 test query",queryContent))
    StewardDatabase.db.logAndCheckQuery(researcherUserName,Some(2),InboundShrineQuery(5,"% forbidden query",queryContent))
  }
}
