package net.shrine.utilities.http4sclienttest

import cats.effect.IO
import io.netty.handler.ssl.{SslContext, SslContextBuilder}
import org.asynchttpclient.Dsl
import io.netty.handler.ssl.util.InsecureTrustManagerFactory
import org.asynchttpclient.proxy.ProxyServer
import org.http4s.client.Client
import org.http4s.client.asynchttpclient.AsyncHttpClient
import org.http4s.headers.Accept
import org.http4s.{Headers, Method, ParseResult, Request, Uri}

import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.duration.DurationInt
import scala.language.postfixOps

/**
 * @author dwalend
 * @since 1.25
 */

object Http4sClientTest {
  def main(args: Array[String]): Unit = {
    try {
      if (args.length != 2) throw WrongNumberOfArguments("Requires two arguments: the http verb (PUT or GET) and the URL to call.")

      val verbString = args(0)
      val verb: ParseResult[Method] = Method.fromString(verbString)

      val urlString = args(1)
      val uri: ParseResult[Uri] = Uri.fromString(urlString)

      (verb,uri) match {
        case (Right(v),Right(u)) => {
          val requestResponseIo: IO[String] = testHttpRequest(v,u)
          requestResponseIo.unsafeRunTimed(30 seconds).fold{println("No response after 30 seconds")}{println(_)}
//          println(requestResponseIo.unsafeRunSync())
          System.exit(0)
        }
        case sad => println(s"something went wrong with $sad") //todo do better - and look at error messages
      }
    } catch {
      case wnoa: WrongNumberOfArguments => {
        printUsage()
        System.exit(1)
      }
      case x:Throwable => {
        x.printStackTrace()
        System.exit(2)
      }
    }
  }

  def printUsage(): Unit = {
    println(
      """Usage: ./http4sClientTest VERB URL
        |
        |Try the URL using the http VERB via the http4s client. This is primarily meant to test the http4s client in
        |environments with https client-side proxies, and to detect other potential obsticals between a downstream node
        |and a hub in a SHRINE network.
        |
        |Exit codes: 0 success
        |            1 known error
        |            2 unknown error (with an accompanying stack trace).
        |
        |Examples:
        |./http4sClientTest GET https://shrine-dev1.catalyst:6443/shrine-api/mom/getQueues
        |./http4sClientTest PUT https://shrine-dev1.catalyst:6443/shrine-api/mom/createQueue/HarmlessTestQueue
      """.stripMargin)
  }

  def testHttpRequest(verb:Method,uri:Uri): IO[String] = {
    val sslContext: SslContext = SslContextBuilder.forClient().trustManager(InsecureTrustManagerFactory.INSTANCE).build()

    val proxyHost = Option(System.getProperty("http.proxyHost"))
    val proxyPort = Option(System.getProperty("http.proxyPort")).map(_.toInt)

    val config = (proxyHost,proxyPort) match {
      case (Some(host),Some(port)) => {
        val proxyServer = new ProxyServer.Builder(host,port).build()
        Dsl.config().setSslContext(sslContext).setProxyServer(proxyServer).build
      }
      case (None,None) => Dsl.config().setSslContext(sslContext).build
      case x => throw new IllegalArgumentException(s"Only one of http.proxyHost and http.proxyPort set $x")
    }
    implicit val timer = IO.timer(global)
    implicit val concurrentEffect = IO.contextShift(global)
    val httpClient: Client[IO] = AsyncHttpClient.resource[IO](config).use(client => IO(client)).unsafeRunSync()

    val acceptHeader = Accept(org.http4s.MediaType.application.json)

    val request = Request[IO](
      method = verb,
      uri = uri,
      headers = Headers.of(acceptHeader)
    )
    httpClient.expect[String](request)
  }

  case class WrongNumberOfArguments(message:String) extends Exception(message)
}
